#include "os.h"

#include "bitmap.h"
#include "resman.h"
#include "window.h"
#include "app.h"
#include "palette.h"
#include "scigame.h"
#include "sciroom.h"


int bGetPaletteEntries = 1;
extern void DoGammaCorrection(PALETTEENTRY far * pPes, int iNumColors, double fGamma);
extern void DoGammaCorrection(RGBQUAD * quad);

typedef struct {
	WORD bfType;
	DWORD bfSize;
	WORD bfReserved1;
	WORD bfReserved2;
	DWORD bfOffBits;

	DWORD biSize;
	DWORD biWidth;
	DWORD biHeight;
	WORD biPlanes;
	WORD biBitCount;
	DWORD biCompression;
	DWORD biSizeImage;
	DWORD biXPelsPerMeter;
	DWORD biYPelsPerMeter;
	DWORD biClrUsed;
	DWORD biClrImportant;
	RGBQUAD bmiColors[256];
} BTMP_RES;

BTMP_RES btmpResGlobalHeader;
WORD *intColors;
RGBQUAD palettecolors[256];

#pragma optimize( "agt", on )

Bitmap::Bitmap(int newResourceID, ResourceManager * newResourceManager, BOOL activatePalette)
:VBitmap(newResourceID, newResourceManager)
{
	int i;
	// j;
	BYTE remapping[256];
	DWORD colorsUsed;
	int size;

	mnBits = 8;

	if (!data) {
		// Bitmap wasn't found by resource loader.
		wWidth = 0;
		wHeight = 0;
		wWidthInBytes = 0;
		return;
	}
	// wWidth = (WORD)(pBTMPRes->biWidth);
	wWidth = *(WORD *) (data + 18);
	// wHeight = (WORD)(pBTMPRes->biHeight);
	wHeight = *(WORD *) (data + 22);
	wWidthInBytes = wWidth + 3 - ((wWidth - 1) % 4);	// Something like this.

	nOriginX = nGetOriginX();
	nOriginY = nGetOriginY();


	// Allocation memory for the pixels.
	size = (DWORD) (wWidthInBytes) * (DWORD) (wHeight);
	mpPixels = new BYTE[size];
	if (!mpPixels) {
		// Error allocating memory for bitmap.
		if (data)
			delete[] data;
		return;
	}
	memset(mpPixels, 0, size);

	// Get the palette from the resource.
	colorsUsed = *(DWORD *) (data + 46);
	if (colorsUsed) {
		for (i = 0; i < (int) colorsUsed; ++i) {
			palettecolors[i] = *(RGBQUAD *) (data + 54 + i * 4);
			//DoGammaCorrection(&(palettecolors[i]));
		}
		for (i = (int) colorsUsed; i < 256; ++i) {
			palettecolors[i] = *(RGBQUAD *) (data + 54);
			//DoGammaCorrection(&(palettecolors[i]));
		}
	} else {
		for (i = 0; i < 256; ++i) {
			palettecolors[i] = *(RGBQUAD *) (data + 54 + i * 4);
			//DoGammaCorrection(&(palettecolors[i]));
		}
	}

	//if (activatePalette) {
	//	ActivatePalette();
	//}
	// Remap the bitmap to the current palette.
	// Generate a remapping table;
	
	// let's generate a 16-bit palette
	if (pApp->poFrameBuf() && pApp->poFrameBuf()->mnBits == 16) {
		mpoPal16 = new Palette16(palettecolors);
	}

	remapping[0] = 0;
	for (i = 1; i < 256; ++i) {
		remapping[i] = i;
		/*
		if (pApp->pCurrentPalette) {
			for (j = 1; j < 256; ++j) {
				if ((palettecolors[i].rgbRed == pApp->pCurrentPalette->colors[j].rgbRed) &&
					(palettecolors[i].rgbGreen == pApp->pCurrentPalette->colors[j].rgbGreen) &&
					(palettecolors[i].rgbBlue == pApp->pCurrentPalette->colors[j].rgbBlue)
					) {
					remapping[i] = j;
					break;
				}
			}
		}
		*/
	}

	// Check for RLE compression
	switch (*(WORD *) (data + 30)) {
	case BI_RLE8:
		// Compressed
		// VBitmap::CopyRLEPixelsWithRemap(wWidthInBytes, mpPixels, data +
		// *(DWORD*)(data + 10), remapping);	// bottom-up
		VBitmap::CopyRLEPixelsWithRemap(wWidthInBytes, mpPixels + (DWORD) (wWidthInBytes) * (DWORD) (wHeight - 1), data + *(DWORD *) (data + 10), remapping);	// top-down
		break;
	case BI_RGB:
		// Uncompressed
		// VBitmap::CopyPixelsWithRemap(wWidthInBytes, mpPixels, data +
		// *(DWORD*)(data + 10), remapping);	// bottom-up
		VBitmap::CopyPixelsWithRemap(wWidthInBytes, mpPixels + (DWORD) (wWidthInBytes) * (DWORD) (wHeight - 1), data + *(DWORD *) (data + 10), remapping);	// top-down
		break;
	default:
		// Unknown format, don't use it.
		if (mpPixels) {
			delete[] mpPixels;
			mpPixels = NULL;
		}
		break;
	}

	if (data) {
		delete[] data;
		data = NULL;
	}
}

#pragma optimize( "agt", on )

Bitmap::Bitmap(WORD wNewWidth, WORD wNewHeight, int nBits)
:VBitmap(wNewWidth, wNewHeight, nBits)
{
	//int i;

	/*
	if (btmpResGlobalHeader.biXPelsPerMeter != 2835) {
		for (i = 0; i < 256; ++i) {
			btmpResGlobalHeader.bmiColors[i].rgbRed = 0;
			btmpResGlobalHeader.bmiColors[i].rgbGreen = 0;
			btmpResGlobalHeader.bmiColors[i].rgbBlue = 0;
			btmpResGlobalHeader.bmiColors[i].rgbReserved = 0;
		}
		btmpResGlobalHeader.biSize = sizeof(BITMAPINFOHEADER);
		btmpResGlobalHeader.biWidth = Width();
		btmpResGlobalHeader.biHeight = Height();
		btmpResGlobalHeader.biPlanes = 1;
		btmpResGlobalHeader.biBitCount = 8;
		btmpResGlobalHeader.biCompression = BI_RGB;
		btmpResGlobalHeader.biSizeImage = 0;
		btmpResGlobalHeader.biXPelsPerMeter = 2835;
		btmpResGlobalHeader.biYPelsPerMeter = 2835;
		btmpResGlobalHeader.biClrUsed = 0;
		btmpResGlobalHeader.biClrImportant = 0;
	}
	*/
	nOriginX = nGetOriginX();
	nOriginY = nGetOriginY();

}

#pragma optimize( "agt", on )

Bitmap::Bitmap(RESOURCE_TYPES theType, char mask2, int iResourceID, ResourceManager * newResourceManager)
:VBitmap(theType, mask2, iResourceID, newResourceManager)
{
	nOriginX = nGetOriginX();
	nOriginY = nGetOriginY();
}

#pragma optimize( "agt", on )


Bitmap::~Bitmap()
{
}

#pragma optimize( "agt", on )

//void Bitmap::CopyPixels(Bitmap *pDestBitmap, RECT rcDest, RECT rcSource, BOOL bTransparent)
//	{
//	VBitmap::CopyPixels(pDestBitmap, rcDest, rcSource, bTransparent);
//	}

// PROBLEM WITH COMPILER SPEED OPTIMIZATION IN THIS FUNCTION
#pragma optimize("", off)

void Bitmap::CopyPixels(Window * pDestWindow, RECT rcDest, RECT rcSource)
{
	RECT rcClipSource,
	 rcClipDest;

	if (mpPixels == 0 || pDestWindow->hDC == NULL)
		return;

	// Offset rcDest to match rcInner.
	SetRect(&rcClipSource, 0, 0, Width(), Height());
	SetRect(&rcClipDest, 0, 0, pDestWindow->Width(), pDestWindow->Height());
	OffsetRect(&rcDest, pDestWindow->rcInner.left, pDestWindow->rcInner.top);
	OffsetRect(&rcClipDest, pDestWindow->rcInner.left, pDestWindow->rcInner.top);
	ClipRects(rcDest, rcClipDest, rcSource, rcClipSource);

#ifdef FLIP_DIB
	// Flip top and bottom because DIBitmaps are supposed to be bottom to
	// top.
	RECT rcTemp;

	rcTemp.top = Height() - rcSource.bottom;
	rcTemp.bottom = Height() - rcSource.top;
	rcTemp.right = rcSource.right;
	rcTemp.left = rcSource.left;
	rcSource = rcTemp;
#endif

	if (rcSource.bottom == rcSource.top)
		return;

	BYTE *pSourcePixels;

	pSourcePixels = mpPixels;
	pSourcePixels += (DWORD) rcSource.top * wWidthInBytes;

	int iOldHeight = (int) (btmpResGlobalHeader.biHeight);
	int iOldWidth = (int) (btmpResGlobalHeader.biWidth);

	//*** BEFORE BLTING, match the DIB color table to the
	//*** current palette to match the animated palette
	int i;
	HPALETTE hOldPalette = 0;

	intColors = (WORD *) (btmpResGlobalHeader.bmiColors);
	for (i = 0; i < 256; ++i)
		intColors[i] = i;

	btmpResGlobalHeader.biHeight = -(rcSource.bottom - rcSource.top);
	btmpResGlobalHeader.biWidth = Width();
	SetDIBitsToDevice(pDestWindow->hDC, rcDest.left, rcDest.top,
				rcDest.right - rcDest.left, (rcDest.bottom - rcDest.top),
					  rcSource.left, 0,
					  0, Height(),
					  pSourcePixels, (LPBITMAPINFO) (((BYTE *) & btmpResGlobalHeader) + 16), DIB_PAL_COLORS);
	btmpResGlobalHeader.biHeight = iOldHeight;
	btmpResGlobalHeader.biWidth = iOldWidth;
}

#pragma optimize("", on)

#pragma optimize( "agt", on )

void Bitmap::CopyPixels(HDC hdc, RECT rcDest, RECT rcSource)
{
	RECT rcClipSource,
	 rcClipDest;

	if (mpPixels == 0 || hdc == NULL)
		return;

	// Offset rcDest to match rcInner.
	SetRect(&rcClipSource, 0, 0, Width(), Height());
	SetRect(&rcClipDest, 0, 0, 640, 480);
	OffsetRect(&rcDest, 0, 0);
	OffsetRect(&rcClipDest, 0, 0);
	ClipRects(rcDest, rcClipDest, rcSource, rcClipSource);

#ifdef FLIP_DIB
	// Flip top and bottom because DIBitmaps are supposed to be bottom to
	// top.
	RECT rcTemp;

	rcTemp.top = Height() - rcSource.bottom;
	rcTemp.bottom = Height() - rcSource.top;
	rcTemp.right = rcSource.right;
	rcTemp.left = rcSource.left;
	rcSource = rcTemp;
#endif

	if (rcSource.bottom == rcSource.top)
		return;

	BYTE *pSourcePixels;

	pSourcePixels = mpPixels;
	pSourcePixels += (DWORD) rcSource.top * wWidthInBytes;

	int iOldHeight = (int) (btmpResGlobalHeader.biHeight);
	int iOldWidth = (int) (btmpResGlobalHeader.biWidth);

	//*** BEFORE BLTING, match the DIB color table to the
	//*** current palette to match the animated palette
	int i;
	HPALETTE hOldPalette = 0;

	intColors = (WORD *) (btmpResGlobalHeader.bmiColors);
	for (i = 0; i < 256; ++i)
		intColors[i] = i;
	btmpResGlobalHeader.biHeight = rcSource.bottom - rcSource.top;
	btmpResGlobalHeader.biWidth = Width();
	SetDIBitsToDevice(hdc, rcDest.left, rcDest.top,
				rcDest.right - rcDest.left, (rcDest.bottom - rcDest.top),
					  rcSource.left, 0,
					  0, Height(),
					  pSourcePixels, (LPBITMAPINFO) (((BYTE *) & btmpResGlobalHeader) + 16), DIB_PAL_COLORS);
	btmpResGlobalHeader.biHeight = iOldHeight;
	btmpResGlobalHeader.biWidth = iOldWidth;
}

#pragma optimize( "agt", on )

BYTE *Bitmap::GetData()
{
	return (BYTE *) mpPixels;
}

#pragma optimize( "agt", on )
void Bitmap::ActivatePalette()
{
	int i;

	if (pApp->pCurrentPalette)
		pApp->pResourceManager->PurgeResource(pApp->pCurrentPalette);
	pApp->pCurrentPalette = (Palette *) pApp->pResourceManager->LoadResource(RES_PALETTE, 0);
	for (i = 0; i < 256; ++i) {
		pApp->pCurrentPalette->colors[i] = palettecolors[i];
	}
	pApp->pCurrentPalette->ActivatePalette();
}

#pragma optimize( "agt", on )
void Bitmap::SetCel(int theCel)
{
}

void Bitmap::SetLoop(int theLoop)
{
}

int Bitmap::GetNumLoops()
{
	return 1;
}

int Bitmap::GetNumCels()
{
	return 1;
}

int Bitmap::nGetCel()
{
	return 0;
}

int Bitmap::nGetLoop()
{
	return 0;
}

void Bitmap::vIncrementCel()
{
}

void Bitmap::vDecrementCel()
{
}
void Bitmap::vIncrementLoop()
{
}

void Bitmap::vDecrementLoop()
{
}


#pragma optimize( "agt", on )
void Bitmap::DrawRect(COLORREF color, RECT rCurrentPos, sciPlane * poPlane, int nThickness)
{
	WORD wBeginX = 0;
	WORD wBeginY = 0;

	WORD i,
	 wWidth = Width();
	WORD wHeight = Height();
	int nThCnt;					// thickness loop counter

	for (nThCnt = 0; nThCnt < nThickness; nThCnt++) {

		// memset(mpPixels,color, wWidth);
		memset(mpPixels + wWidthInBytes * nThCnt, color, wWidth);

		for (i = 1; i < wHeight; i++) {
			// Left vertical
			*(mpPixels + (i * wWidthInBytes) + nThCnt) = (char) color;
			// Right vertical
			*(mpPixels + (i * wWidthInBytes) + wWidth - nThCnt - 1) = (char) color;
		}

		// Bottom horizontal
		memset(mpPixels + (wHeight - nThCnt - 1) * wWidthInBytes, color, wWidth);

	}



	// RECT rCurrent;
	RECT rMine;
	RECT rScreen;
	RECT rGlobal;

	RECT *prClip;
	prClip = poPlane->prGetClip();
	RECT rClip = {prClip->left, prClip->top, prClip->right, prClip->bottom};

	assert(poPlane);

	// convert to the global rect
	rGlobal = rCurrentPos;
	poPlane->vLocalToGlobal(&rGlobal);

	// clip to the given area
	// if we're outside the clip area, return

	if (!IntersectRect(&rScreen, &rGlobal, &rClip)) {
		return;
	}
	// make the source & dest rects the same size
	SetRect(&rMine,
			rScreen.left - rGlobal.left,
			rScreen.top - rGlobal.top,
			Width() - (rGlobal.right - rScreen.right),
			Height() - (rGlobal.bottom - rScreen.bottom));

	POINT tScreen;
	tScreen.x = rScreen.left;
	tScreen.y = rScreen.top;

	if (pApp && pApp->poFrameBuf()) {
		vDrawToFrame(pApp->poFrameBuf(), tScreen, rMine);
	}
}

#pragma optimize( "agt", on )

void Bitmap::vDraw(int nX, int nY, sciPlane * poPlane, BYTE jAlpha, COLOR16 wRemapColor, BYTE jRemapAmount, RECT *prLocal, WORD *pwPalette)
{
	RECT rMine;
	RECT rScreen;
	RECT rGlobal;
	//RECT rCurrent;

	//int nWidth = Width();
	//int nHeight = Height();

	if (bFlip) {
		nX -= (Width() - nOriginX) - 1;
	} else {
		nX -= nOriginX;
	}
	nY -= nOriginY;

	assert(poPlane);

	// start with the unclipped (local) rect
	SetRect(&rGlobal, nX, nY, nX + Width(), nY + Height());

	if (prLocal) {
		*prLocal = rGlobal; // still local, not global at this point
	}

	// now convert to the global rect
	poPlane->vLocalToGlobal(&rGlobal);

	// clip to the given area
	// if we're outside the clip area, return
	if (!IntersectRect(&rScreen, &rGlobal, poPlane->prGetClip())) {
		return;
	}
	// I'm drawing, so add myself to the drawn list
//	mpoPlane->vAddToDrawnList(this);

	// make the source & dest rects the same size
	if (bFlip) {
		SetRect(&rMine,
				rGlobal.right - rScreen.right,
				rScreen.top - rGlobal.top,
				Width() - (rScreen.left - rGlobal.left),
				Height() - (rGlobal.bottom - rScreen.bottom));
	} else {
		SetRect(&rMine,
				rScreen.left - rGlobal.left,
				rScreen.top - rGlobal.top,
				Width() - (rGlobal.right - rScreen.right),
				Height() - (rGlobal.bottom - rScreen.bottom));
	}

	POINT tScreen;

	// if flipped, for speed's sake hand off the upper-right corner rather than the
	// upper-left corner, since drawing will start at that pixel
	if (bFlip) {
		tScreen.x = rScreen.right - 1;
	} else {
		tScreen.x = rScreen.left;
	}
	tScreen.y = rScreen.top;

	if (pApp && pApp->poFrameBuf()) {
		vDrawToFrame(pApp->poFrameBuf(), tScreen, rMine, jAlpha, wRemapColor, jRemapAmount, pwPalette);
	}
}

void Bitmap::vSetOriginPoint(short nNewX, short nNewY, int nLoop, int nCel)
{

}
