// BUILDING.CPP	 Things you can go inside
//
// DJM

#include "scimain.h"
#include "building.h"
#include "mapplane.h"
#include "gamemap.h"
#include "objectdb.h"
#include "procs.h"
#include "colordef.h"
#include "unit.h"
#include "player.h"
#include "debug.h"
#include "mapedit.h"

Building::Building(MapPlane *poPlane, int nTemplateID)
:DrawObj(poPlane, nTemplateID)
{
	meClass = OC_BUILDING;

	mpoOutsideBitmap = mpoBitmap;
	mpoInsideBitmap = NULL;
	vSetOpen(FALSE);
	mbForceInterior = FALSE;
	mpoPlane->mpoMap->AddBuilding(this);
	vSetInsideBitmap(mpoTemplate->lGetValue(TAG_Bint, 0));
}

void Building::vSetInsideBitmap(int nBitmap) {
	if (nBitmap == nInsideBitmapID()) {
		return;
	}
	if (mpoInsideBitmap) {
		pApp->pResourceManager->FreeResource(mpoInsideBitmap);
		mpoInsideBitmap = NULL;
	}
	
	mpoInsideBitmap = (Bitmap*)pApp->pResourceManager->LoadResource(RES_BITMAP, nBitmap);
}

Building::~Building()
{
	mpoPlane->mpoMap->RemoveBuilding(this);
	if (mpoOutsideBitmap) {
		pApp->pResourceManager->FreeResource(mpoOutsideBitmap);
		mpoOutsideBitmap = NULL;
	}
	if (mpoInsideBitmap) {
		pApp->pResourceManager->FreeResource(mpoInsideBitmap);
		mpoInsideBitmap = NULL;
	}
	mpoBitmap = NULL;
}


void Building::vChangeToInside() {
	if (mpoInsideBitmap) {
		//swap bitmaps
		mpoBitmap = mpoInsideBitmap;
		vSetOpen(TRUE);
		vSetLayer(DP_FLOOR);
		vMarkUpdate();
	}
}

void Building::vChangeToOutside() {
	if (mpoOutsideBitmap) {
		//swap bitmaps
		mpoBitmap = mpoOutsideBitmap;
		vSetOpen(FALSE);
		vSetLayer(DP_OBSTACLES);
		vMarkUpdate();
	}

}

void Building::vCheckState() {
	// If local player is present then pop the top it open and change layers
	// Note - gbToggleInteriors is a Debug variable that allows CTRL-v to 'cheat' and allow us
	//		  to see in all interiors. We may want to remove it later on - TBD, TW
	if (mfHealth == 0) return;
	if (bLocalPlayerInside() || mbForceInterior || gbToggleInteriors) {	
		if (!mbOpen) {
			vChangeToInside();
		}
	}
	else {
		if (mbOpen) {
			vChangeToOutside();
		}
	}

}

Boolean Building::HandleEvent(sciEvent & oEvent)
{
	if (gbMapEditing) {
		DrawObj::HandleEvent(oEvent);
		return oEvent.claimed;
	}

	//if the event is on us and we're a child that's suppose to bubble, then pass to parent
	if (mpoParent && mbEventBubbling && bOnMe(oEvent))
		return (mpoParent->HandleEvent(oEvent));

	//See if we're a parent that got this event as it bubbled up to us
	if ((mpoChildren && bOnChildren(oEvent)))
	{
		//don't do anything
	} else {
		//finally do standard check
		if	(!bOnMe(oEvent)) 
			return oEvent.claimed;
	}

	if (!oEvent.claimed && oEvent.type == MOUSE_DOWN) 
	{
		if (oEvent.modifiers & RIGHT_MOUSE) // RIGHT MOUSE
		{
			// TBD - do not route these to npc's
			if (gpoMap->nSelectionSize() > 0)  // // oEvent.modifiers & RIGHT_MOUSE) 
			{
				//MapNode *poNode = gpoMap->poNodeFromEvent(oEvent);
				MapNode *poNode = poNearestInterior(gpoMap->poNodeFromEvent(oEvent));
				if (poNode) 
				{
					Command *poCommand = new Command;
					poCommand->mnX = poNode->mnXPos;
					poCommand->mnY = poNode->mnYPos;
					poCommand->mnUnit = ID_NONE - 1;
					if (mpoPlane->mpoMap->GetMode() == UIMODE_PATROL)
					{
						poCommand->meType = CM_PATROL_WAYPOINT;
					}
					else
					{
						poCommand->meType = CM_MOVE;
					}

					game->poLocalPlayer()->vPostCommand(poCommand);
					//gpoMap->vAudioConfirmCommand();
					oEvent.claimed = TRUE;
				}
			} 
		}
	}
			
	return oEvent.claimed;
}

// TBD - PB - THIS CURRENTLY JUST SCANS FOR RANGE AND RETURNS A FIXED POSITION
// CHANGE TO RETURN VALUE JUST OUTSIDE OF ENTRANCE -- THIS SHOULD WORK FOR NOW
MapNode * Building::GetPositionOutside()
{
	int nWidth = gpoMap->nGetWidth();
	int nHeight = gpoMap->nGetHeight();

	int nMinX = nWidth - 1;
	int nMinY = nHeight - 1;
	int nMaxX = 0;
	int nMaxY = 0;

	long plListSize;
	long *plIntList = mpoTemplate->plGetList(TAG_Libs, &plListSize);
	if (plListSize) 
	{
		// we have data in Libs tag
		if (plListSize % 2 == 0) 
		{	//make sure we have even multiple of coords
			int nXPos = mpoCurNode->mnXPos;
			int nYPos = mpoCurNode->mnYPos;
			int nNewNodeX, nNewNodeY;
			int i;
			for (i=0; i < plListSize; i+=2) 
			{
				nNewNodeX = nXPos + plIntList[i];
				nNewNodeY = nYPos + plIntList[i+1];
				if ( (nNewNodeX >= 0) && (nNewNodeX < nWidth) &&
					  (nNewNodeY >= 0) && (nNewNodeY < nHeight) ) 
				{
					nMinX = min(nMinX, nNewNodeX);
					nMinY = min(nMinY, nNewNodeY);
					nMaxX = max(nMaxX, nNewNodeX);
					nMaxY = max(nMaxY, nNewNodeY);
				}				
			}
		}
	}

	plIntList = mpoTemplate->plGetList(TAG_Lobs, &plListSize);
	if (plListSize) 
	{
		// we have data in Lobs tag
		if (plListSize % 2 == 0) 
		{	//make sure we have even multiple of coords
			int nXPos = mpoCurNode->mnXPos;
			int nYPos = mpoCurNode->mnYPos;
			int nNewNodeX, nNewNodeY;
			int i;
			for (i=0; i < plListSize; i+=2) 
			{
				nNewNodeX = nXPos + plIntList[i];
				nNewNodeY = nYPos + plIntList[i+1];
				if ( (nNewNodeX >= 0) && (nNewNodeX < nWidth) &&
					  (nNewNodeY >= 0) && (nNewNodeY < nHeight) ) 
				{
					nMinX = min(nMinX, nNewNodeX);
					nMinY = min(nMinY, nNewNodeY);
					nMaxX = max(nMaxX, nNewNodeX);
					nMaxY = max(nMaxY, nNewNodeY);
				}				
			}
		}
	}
	int nX = min(nMaxX + 1, nWidth - 1);
	int nY = min(nMaxY + 1, nHeight - 1);

	assert(nX >= 0 && nY >= 0 && nX < nWidth && nY < nHeight);
	MapNode * pReturn = gpoMap->poGetNode(nX, nY);
	return pReturn;
}

MapNode * Building::GetPositionInside()
{
	//go through the interior list and check for a local player
	int nWidth = gpoMap->nGetWidth();
	int nHeight = gpoMap->nGetHeight();
	long plListSize;
	long *plIntList = mpoTemplate->plGetList(TAG_Libs, &plListSize);
	if (plListSize) {
		// we have data in Libs tag
		if (plListSize % 2 == 0) 
		{	//make sure we have even multiple of coords
			int nXPos = mpoCurNode->mnXPos;
			int nYPos = mpoCurNode->mnYPos;
			int nNewNodeX, nNewNodeY;
			int i;
			for (i=0; i < plListSize; i+=2) 
			{
				nNewNodeX = nXPos + plIntList[i];
				nNewNodeY = nYPos + plIntList[i+1];
				if ( (nNewNodeX >= 0) && (nNewNodeX < nWidth) &&
					  (nNewNodeY >= 0) && (nNewNodeY < nHeight) ) 
				{
					MapNode *poNode = gpoMap->poGetNode(nNewNodeX, nNewNodeY);
					assert (poNode);
					return poNode;
				}				
			}
		}
	}	
	return NULL;
}

Boolean Building::bLocalPlayerInside() {
	//go through the interior list and check for a local player
	int nWidth = gpoMap->nGetWidth();
	int nHeight = gpoMap->nGetHeight();
	long plListSize;
	long *plIntList = mpoTemplate->plGetList(TAG_Libs, &plListSize);
	if (plListSize) {
		// we have data in Libs tag
		if (plListSize % 2 == 0) {	//make sure we have even multiple of coords
			int nXPos = mpoCurNode->mnXPos;
			int nYPos = mpoCurNode->mnYPos;
			int nNewNodeX, nNewNodeY;
			int i;
			for (i=0; i < plListSize; i+=2) {
				nNewNodeX = nXPos + plIntList[i];
				nNewNodeY = nYPos + plIntList[i+1];
				if ( (nNewNodeX >= 0) && (nNewNodeX < nWidth) &&
					  (nNewNodeY >= 0) && (nNewNodeY < nHeight) ) {
					MapNode *poNode = gpoMap->poGetNode(nNewNodeX, nNewNodeY);
					assert (poNode);
					if (poNode->mpoUnit) {
						//check for a local human player
						if (game->poLocalPlayer() == poNode->mpoUnit->poOwner() 
							 && poNode->mpoUnit->bAlive() ) {	
							return TRUE;
						}
					}
				}
			}
		}
	}
	return FALSE;	//no local human player found
}

void Building::vDrag(sciEvent &oEvent) {
	if (gpoMapEdit->bLegalMove(this, oEvent)) {
		DrawObj::vDrag(oEvent);	// call the super
//		if (mbOpen) {
		gpoMapEdit->vUpdateUnitPositions(this);	
//		}
	}
}

// For adding to the map.
// The building object being dragged around makes a copy of itself and inserts it in into the
// node nearest the event
void Building::vCopySelf(sciEvent &oEvent) {
	MapNode *poNode = mpoPlane->mpoMap->poNodeFromEvent(oEvent);
	if (poNode) {

		// If self is on clipboard and a building interior is in spot where we want a copy to go,
		// then we can't allow it as only 1 interior node is allowed. This condition can and will
		// occur when you add a bldg with the right mouse button, don't move it, and click the 
		// button again.
		//		OR
		// we right click on a node that has an interior node from another building
		if ( (gpoMapEdit->bOnClipBoard(this) && !gpoMapEdit->bLegalMove(this, oEvent)) ||
			 !gpoMapEdit->bLegalMove(this, oEvent) ) {
			vDebugOut("can't add building there\n");
			pApp->pWindow->vRemoveEventHog(this);
			delete this;
			gpoMapEdit->vSetMode(EM_NONE);
			gpoMapEdit->vSetObject(NULL);
		} else {

			// clear self out of node the copy is suppose to go into and put back onto clipboard
			gpoMapEdit->mpoMapNode->vAddObj(this); //this lines removes self from previous node and puts onto clipboard

			// now create copy and place into node
			Building *poCopy = new Building(mpoPlane, nTemplateID());	
			poCopy->vSetLayer(mnLayer);
			poNode->vAddObj(poCopy);
		}
	}
}

// For adding to the map.
// The building object being dragged around inserts ITSELF in the node nearest its event
void Building::vInsertSelf(sciEvent &oEvent) {
	MapNode *poNode = mpoPlane->mpoMap->poNodeFromEvent(oEvent);
	if (poNode) {

		// if we're on the clipboard
		if (gpoMapEdit->bOnClipBoard(this)) {
			if (gpoMapEdit->bLegalMove(this, oEvent)) {
				//most likely vDrag will have already inserted us into the node but if it didn't
				//we need to catch it here.
//				vSetLayer(DP_OBSTACLES);
				poNode->vAddObj(this); //this lines also removes self from previous node

			} else {
				vDebugOut("can't add building there\n"); //TBD - add some real error message here - TW
				delete this;
			}
		}
		
	} else {
		// we clicked somewhere not on the game map so get rid of self
		delete this;
	}

}


void Building::vSaveBitmap() {
	RESOURCE_TYPES oldType;

	if (mpoBitmap == mpoOutsideBitmap)
	{
			oldType = mpoOutsideBitmap->type;
			mpoOutsideBitmap->type = RES_VIEWBITMAP; 
			mpoOutsideBitmap->vWriteFile();
			//restore to original just in case
			mpoOutsideBitmap->type = oldType;			
	}

	else if (mpoBitmap == mpoInsideBitmap)
		{
			oldType = mpoInsideBitmap->type;
			mpoInsideBitmap->type = RES_VIEWBITMAP; 
			mpoInsideBitmap->vWriteFile();
			//restore to original just in case
			mpoInsideBitmap->type = oldType;
		}

 ///////////////////////////
/// TEMP for milestone ///
/////////////////////////
		
/*	//write exterior view first
	
	//need to 'fake' the resource into thinking it's a ViewBitmap
	//so it'll attach TEX as an extension instead of BMP
	oldType = mpoOutsideBitmap->type;
	mpoOutsideBitmap->type = RES_VIEWBITMAP; 

	if (mbOpen) {
		mpoOutsideBitmap->vSetOriginPoint(mpoInsideBitmap->nOriginX, mpoInsideBitmap->nOriginY);
	}
	mpoOutsideBitmap->vWriteFile();
	
	//restore to original just in case
	mpoOutsideBitmap->type = oldType;			

	//then write interior if present
	if (mpoInsideBitmap) {
		if (!mbOpen) {
			mpoInsideBitmap->vSetOriginPoint(mpoOutsideBitmap->nOriginX, mpoOutsideBitmap->nOriginY);
		}
		
		//need to 'fake' the resource into thinking it's a ViewBitmap
		//so it'll attach TEX as an extension instead of BMP
		oldType = mpoInsideBitmap->type;
		mpoInsideBitmap->type = RES_VIEWBITMAP; 
															 
		mpoInsideBitmap->vWriteFile();
		
		//restore to original just in case
		mpoInsideBitmap->type = oldType;
	}
*/	
}


void Building::vSetOpen (Boolean bOpen)
{
	mbOpen = bOpen;
	vSetParentOpen(bOpen);
}

//Returns an interior node closest to poNode
MapNode *Building::poNearestInterior(MapNode *poNode)
{
	if (!poNode)
		return NULL;

	MapNode *poClosestNode = NULL;
	int nWidth = gpoMap->nGetWidth();
	int nHeight = gpoMap->nGetHeight();
	long plListSize;
	long *plIntList = mpoTemplate->plGetList(TAG_Libs, &plListSize);
	if (plListSize) {
		// we have data in Libs tag
		if (plListSize % 2 == 0) 
		{	//make sure we have even multiple of coords
			int nXPos = mpoCurNode->mnXPos;
			int nYPos = mpoCurNode->mnYPos;
			int nNewNodeX, nNewNodeY;
			int i;
			for (i=0; i < plListSize; i+=2) 
			{
				nNewNodeX = nXPos + plIntList[i];
				nNewNodeY = nYPos + plIntList[i+1];
				if ( (nNewNodeX >= 0) && (nNewNodeX < nWidth) &&
					  (nNewNodeY >= 0) && (nNewNodeY < nHeight) ) 
				{
					MapNode *poNewNode = gpoMap->poGetNode(nNewNodeX, nNewNodeY);
					assert (poNewNode);
					if (poNode == poNewNode)
						return poNode;		//actually clicked on an interior node so just return it
					if (poClosestNode)
					{
						int nOldDiff = abs(poNode->mnXPos - poClosestNode->mnXPos) + abs(poNode->mnYPos - poClosestNode->mnYPos);
						int nNewDiff = abs(poNewNode->mnXPos - poNode->mnXPos) + abs(poNewNode->mnYPos - poNode->mnYPos);
						if (nNewDiff < nOldDiff)
							poClosestNode = poNewNode;
					} else {
						poClosestNode = poNewNode;
					}
			
				}				
			}
			return poClosestNode;
		}
	}	
	return NULL;

}