// COMMAND.H -- command class
//
// DJM

#ifndef _COMMAND_
#define _COMMAND_

class DrawObj;
class MapNode;

#define MAX_TURN_QUEUE 20

typedef enum _ATTR_CODE_ 
{
	ATTR_AGGRESSION,
	ATTR_TENACITY,
	ATTR_HEALTH,
	ATTR_CONVERSION,
	ATTR_NEWUPGRADE_00,
	ATTR_NEWUPGRADE_01,
	ATTR_NEWUPGRADE_02,
	ATTR_NEWUPGRADE_03,
	ATTR_NEWUPGRADE_04,
	ATTR_NEWUPGRADE_05,
	ATTR_NEWUPGRADE_06,
	ATTR_NEWUPGRADE_07,
	ATTR_NEWUPGRADE_TOPPLUSONE,
	ATTR_ANXIETY,
	ATTR_READYSPELL,
	ATTR_DEFAULT_ATTACK
} ATTR_CODE;

#include "game.h"

typedef enum _COMMAND_TYPE 
{
	CM_NONE,							// do nothing
	CM_SELECT,							// change the selection
	CM_ATTACK,							// attack someone
	CM_MOVE,							// move somewhere
	CM_ATTR,							// change attribute(s)
	CM_CONVERT,							// convert a vampire
	CM_SPELL,							// cast a spell
	CM_FEED,							// feed on a human
	CM_GUARD,							// guard a unit
	CM_FORMATION,						// formation relative to a unit
	CM_GUARD_MOVEINTOPOSITION,			// sub command of guard
	CM_GUARD_MOVEINTERNAL,				// sub command of guard
	CM_GUARD_MOVETOVERIFY,				// sub command of guard	
	CM_PATROL_WAYPOINT,					// accumulate list of waypoints
	CM_WAIT_500MS,						//
	CM_CLOAK,
	CM_TAKE,
	CM_GIVE,
	CM_DROP,
	CM_CAST,
	CM_MOVE_INVITE,
	CM_MOVE_FOLLOW,
	CM_RETREAT
} COMMAND_TYPE;

#define STATE_COMMAND_NONE				0

#define STATE_RETREAT_RUNAWAYFROMENEMY	1
#define STATE_RETREAT_TOUNIT			2

#define MAX_SELECTION_SIZE  9

class Command : public Object {
public:
	Command() {
		mnPlayer = 0;
		meType = CM_NONE;
		mbTargetUnit = FALSE;
		mnX = 0;
		mnY = 0;
		mnUnit = ID_NONE;
		mnAttributeID = ID_NONE;
		mnNewValue = 0;
		mnSelectSize = 0;
		mlChecksum = 0;
		mbInternal = FALSE;
		mnState = 0;
	}

	Command(BYTE *pBinaryCommand);

	Command(Command &o) {
		mnPlayer = o.mnPlayer;
		meType = o.meType;
		mbTargetUnit = o.mbTargetUnit;
		mnX = o.mnX;
		mnY = o.mnY;
		mnUnit = o.mnUnit;
		mnAttributeID = o.mnAttributeID;
		mnNewValue = o.mnNewValue;
		mnSelectSize = o.mnSelectSize;
		mlChecksum = o.mlChecksum;
		mbInternal = o.mbInternal;
		mnState = o.mnState;
		int i;
		for (i = 0; i < mnSelectSize; ++i) {
			manSelection[i] = o.manSelection[i];
		}

	}

	int nCreateBinaryCommand(BYTE *pBinaryCommand);

	void vGetTargetPos(int *pnX, int *pnY);
	MapNode *poGetTargetNode(void);
	DrawObj *poGetTarget(void);


	COMMAND_TYPE	meType;

	// not all of these are used for every turn type
	int			mnPlayer;		// which player made this command?
	BOOL		mbTargetUnit; // targeting a unit or a location?
	int			mnX;
	int			mnY;
	int			mnUnit;
	int			mnAttributeID;
	int			mnNewValue;
	int			mnSelectSize;
	int			manSelection[MAX_SELECTION_SIZE];
	long		mlChecksum;
	BOOL		mbInternal;
	int			mnState;
};

class CommandStack : public Object {
protected:
	Command*	maoCommands[MAX_TURN_QUEUE];
	int		mnCurrentPos; // current pos in the stack
	int		mnCurrentTurn; // TBD -- decide should this be here?

	int		nTurnFromPos(int nPos) {
		int nOff = nPos - mnCurrentPos;
		if (nOff < 0) {
			nOff += MAX_TURN_QUEUE;
		}
		return mnCurrentTurn + nOff;
	}
	int		nPosFromTurn(int nTurn) {
		return (((nTurn - mnCurrentTurn) + mnCurrentPos) % MAX_TURN_QUEUE);
	}

	int nCurrentTurn(void) {return game->nNetTurn();}

public:
	CommandStack() {
		memset(maoCommands, 0, MAX_TURN_QUEUE * sizeof(Command*));
		mnCurrentPos = 0;
		mnCurrentTurn = 0;
	}
	~CommandStack() {
		int i;
		for (i = 0; i < MAX_TURN_QUEUE; ++ i) {
			if (maoCommands[i]) {
				delete maoCommands[i];
				maoCommands[i] = NULL;
			}
		}
	}

	void	vPushTurn(int nTurn, Command *poTurn) {
		maoCommands[nPosFromTurn(nTurn)] = poTurn;
	}
	void	vPopTurn(int nTurn) {
		int nPos = nPosFromTurn(nTurn);
		if (maoCommands[nPos]) {
			delete maoCommands[nPos];
			maoCommands[nPos] = NULL;
		}
	}
	Command*	poPeekTurn(int nTurn) {
		return maoCommands[nPosFromTurn(nTurn)];
	}
	BOOL	bTurnAvailable(int nTurn) {
		return (poPeekTurn(nTurn) != NULL);
	}
	long	lTurnChecksum(int nTurn){
		assert (poPeekTurn(nTurn));  // pab - turn must be available
		return (poPeekTurn(nTurn)->mlChecksum);
	}
};

#if 0
/*
typedef enum _COMMAND {
	CM_NONE,
	CM_MOVE_TO,
	CM_ATTACK,
	CM_CONVERT,
	CM_FEED,
} COMMAND;

class Command {
public:
	COMMAND		meCommand;
	int			mnX;
	int			mnY;
	DrawObj		*mpoWho; // who to attack, feed off, etc.

	Command() {
		meCommand = CM_NONE;
		mnX = 0;
		mnY = 0;
		mpoWho = NULL;
	}
	Command(Command &oCommand) {
		meCommand = oCommand.meCommand;
		mnX = oCommand.mnX;
		mnY = oCommand.mnY;
		mpoWho = oCommand.mpoWho;
	}
};
*/
#endif

#endif