/*==========================================================================

  PB - THIS CODE ORIGINALLY APPEARED IN DEMO IN DIRECTX 5 SDK.  USED HERE ACCORDING TO
  DIRECTX5 SDK LICENSE AGREEMENT.

 *
 *  File:   dsstream.c
 *  Copyright (C) 1995-1997 Microsoft Corporation. All Rights Reserved.
 *
 ***************************************************************************/
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <dsound.h>
#include <commctrl.h>
#include <commdlg.h>
#include <memory.h>
#include <cderr.h>

#include "dsstream.h"

char szAppTitle[64];
char szAppCaption[64];
char szPan[32];
char szVolume[32];
char szFrequency[32];
char szProgress[32];
char szOpenFilter[128];
char szOpenDLGTitle[64];


char szTemp[256];
char szDebug[128];
char szFileBuffer[MAX_PATH];
char szFileTitle[MAX_PATH];
char szCDStartPath[MAX_PATH];	        // The path to start the Open dialog in
 
// Registry Key and Value names that allow us to retrive a path something like
// "C:\DXSDK\SDK\MEDIA", but matching the current install.
static const TCHAR gszRegKeyDirect3D[] = TEXT("Software\\Microsoft\\Direct3D");
static const TCHAR gszRegValueD3DPath[] = TEXT("D3D Path");

// Size of longest token below
#define MAX_TOKEN_LEN	7


static TCHAR gszPlayToken[] = TEXT("PLAY");
static TCHAR gszLoopToken[] = TEXT("LOOP");
static TCHAR gszCloseToken[] = TEXT("CLOSE");
static TCHAR gszStickyToken[] = TEXT("STICKY");
static TCHAR gszGlobalToken[] = TEXT("GLOBAL");

LPDIRECTSOUND		lpDS = NULL;
LPDIRECTSOUNDBUFFER	lpDSBStreamBuffer = NULL;
LPDIRECTSOUNDNOTIFY lpDirectSoundNotify = NULL;

WAVEINFOCA		wiWave;

HWND	hWndMain, hWndPan, hWndPanText, hWndVol, hWndVolText, hWndFreqText;
HWND	hWndBar, hWndPlay, hWndStop, hWndLoopCheck, hWndFreq, hWndProg;
HWND	hWndProgText;

#ifdef DEBUG
HWND			hWndList;
#endif

HINSTANCE	hInst;

static BOOL	bFileOpen = FALSE, bPlaying = FALSE;
static BOOL     bEnumDrivers = FALSE, gfCloseOnDone = FALSE;
static UINT     uTimerID = 0, uLastPercent = 100;
static GUID     guID;
static DWORD	gdwFocus = 0;

static BOOL InitApp( HINSTANCE );
static BOOL InitInstance( HINSTANCE, int );
static void BuildTitleBarText( void );
// Call this in initialization to set the szCDStartPath[] global variable
static void GetMediaStartPath( void );
static void FillDataBuffer( void );
static void LoadFromCommandLine( LPSTR lpszCmd );
extern void UpdateProgressBar(void);


/*****************************************************************************/
/* StreamBufferSetup()                                                       */
/*                                                                           */
/* This function uses the filename stored in the global character array to*/
/* open a WAVE file. Then it creates a secondary DirectSoundBuffer object    */
/* which will later be used to stream that file from disk during playback.   */
/*                                                                           */
/*****************************************************************************/
int StreamBufferSetup( void )
    {
    DSBUFFERDESC dsbd;
    HRESULT      dsRetVal;

    int nChkErr;
	int nRem;

    /* This portion of the WAVE I/O is patterned after what's in DSTRWAVE, which
     * was in turn adopted from WAVE.C which is part of the DSSHOW sample.
     */

    if(( nChkErr = WaveOpenFile( szFileBuffer, &wiWave.hmmio, &wiWave.pwfx, &wiWave.mmckInRIFF )) != 0 )
        {
        //ErrorMessageBox( IDS_ERROR_WAVEFILEOPEN, MB_ICONEXCLAMATION );
        return( ERR_WAVE_OPEN_FAILED );
        }

    if( wiWave.pwfx->wFormatTag != WAVE_FORMAT_PCM )
        {
        //ErrorMessageBox( IDS_ERROR_WAVENOTPCM, MB_ICONEXCLAMATION );
        WaveCloseReadFile( &wiWave.hmmio, &wiWave.pwfx );
        return( ERR_WAVE_INVALID_FORMAT );
		}
    // Seek to the data chunk. mmck.ckSize will be the size of all the data in the file.
    if(( nChkErr = WaveStartDataRead( &wiWave.hmmio, &wiWave.mmck, &wiWave.mmckInRIFF )) != 0 )
        {
        //ErrorMessageBox( IDS_ERROR_WAVESEEKFAILED, MB_ICONEXCLAMATION );
        WaveCloseReadFile( &wiWave.hmmio, &wiWave.pwfx );
        return( ERR_WAVE_CORRUPTED_FILE );
        }
	
    // Calculate a buffer length 3 sec. long. This should be an integral number of the
	// number of bytes in one notification period. 
	wiWave.dwNotifySize = wiWave.pwfx->nSamplesPerSec * 3 * (DWORD)wiWave.pwfx->nBlockAlign;
	wiWave.dwNotifySize = wiWave.dwNotifySize/NUM_PLAY_NOTIFICATIONS;
	// the notify size should be an intergral multiple of the nBlockAlignvalue.
	if ((nRem = wiWave.dwNotifySize%(DWORD)wiWave.pwfx->nBlockAlign) != 0)
	{
		wiWave.dwNotifySize += (wiWave.pwfx->nBlockAlign - nRem);
	}
	wiWave.dwBufferSize = wiWave.dwNotifySize * NUM_PLAY_NOTIFICATIONS;	
//    DPF(0, "BufferSize = %lu ", wiWave.dwBufferSize);

#ifdef DEBUG
    wsprintf( szDebug, "BufferSize = %lu", wiWave.dwBufferSize );
    ListBox_AddString( hWndList, szDebug );
#endif

     //Create the secondary DirectSoundBuffer object to receive our sound data.
    memset( &dsbd, 0, sizeof( DSBUFFERDESC ));
    dsbd.dwSize = sizeof( DSBUFFERDESC );
    // Use new GetCurrentPosition() accuracy (DirectX 2 feature)
    dsbd.dwFlags = DSBCAPS_CTRLPOSITIONNOTIFY| DSBCAPS_CTRLDEFAULT | DSBCAPS_GETCURRENTPOSITION2 | gdwFocus;
    dsbd.dwBufferBytes = wiWave.dwBufferSize;

    //Set Format properties according to the WAVE file we just opened
    dsbd.lpwfxFormat = wiWave.pwfx;
    
	dsRetVal = lpDS->lpVtbl->CreateSoundBuffer( lpDS, &dsbd,&lpDSBStreamBuffer,NULL );						
    if( dsRetVal != DS_OK )
        {
        //ErrorMessageBox( IDS_ERROR_DSBCREATE, MB_ICONEXCLAMATION );
        return( ERR_CREATEDSB_FAILED );
        }

    wiWave.lpDSBStreamBuffer = lpDSBStreamBuffer;
    wiWave.bFoundEnd = FALSE;
	wiWave.dwNextWriteOffset = 0;
    wiWave.bLoopFile = Button_GetCheck( hWndLoopCheck );

	// now get the pointer to the notification interface.
	dsRetVal = IDirectSoundNotify_QueryInterface(lpDSBStreamBuffer, &IID_IDirectSoundNotify, &((LPVOID)lpDirectSoundNotify));
	if (dsRetVal != DS_OK)
	{	
		//ErrorMessageBox(IDS_ERROR_QINOTIFY_FAILED, MB_ICONEXCLAMATION);
	    return (dsRetVal);
    }

	// Fill data in the buffer.
    FillDataBuffer();	

	// we're set to play now.
    wiWave.bDonePlaying = FALSE;

#ifdef DEBUG
    wsprintf( szDebug, "wiWave.dwBufferSize = %lu", wiWave.dwBufferSize );
  //  DPF( 3, "StreamBufferSetup Debug" );
  //  DPF( 3, szDebug );
#endif

   // SendMessage( hWndVol, TBM_SETPOS, TRUE, VOL_MAX );
   // SendMessage( hWndPan, TBM_SETPOS, TRUE, PAN_CENTER );
    //SendMessage( hWndFreq, TBM_SETPOS, TRUE,(LPARAM)wiWave.pwfx->nSamplesPerSec / FREQ_MULTIPLIER );
    //PostMessage( hWndMain, WM_DSSTREAM_PROGRESS, 0L, 0L );
  //  UpdateFromControls();
    return( 0 );
    }


/*****************************************************************************/
/* ResetWavePlayer()                                                         */
/*                                                                           */
/*  Performs a subset of the above operations (in StreamBufferSetup). Things */
/* not done include creating a DSB and opening the file (it's already open). */
/*                                                                           */
/*****************************************************************************/
void ResetWavePlayer( void )
    {
	int nChkErr = DS_OK;
//	DPF(1, "Resetting wave player");

    /* Seek to the data chunk */
    if(( nChkErr = WaveStartDataRead( &wiWave.hmmio, &wiWave.mmck, &wiWave.mmckInRIFF )) != 0 )
        {
		// This shouldn't happen, since we've done this before.
      //  DPF(0, "Error seeking to file in ResetWavePlayer");
		return;	
        }

    wiWave.bFoundEnd = FALSE;
	wiWave.dwNextWriteOffset = 0;
	
//	DPF(0, "Reset: Filldatabuffer");
    FillDataBuffer();

    wiWave.bDonePlaying = FALSE;
    PostMessage( hWndMain, WM_DSSTREAM_PROGRESS, 0L, 0L );
    }


/*****************************************************************************/
/* FillDataBuffer()                                                          */
/*                                                                           */
/*   This function fills the sound buffer with a block of data, starting at  */
/* the current read position.  The entire buffer is filled.                  */
/*                                                                           */
/*****************************************************************************/
 void FillDataBuffer( void )
 {
    LPBYTE  lpWrite1, lpWrite2;
    DWORD	dwLen1, dwLen2;
    UINT	uActualBytesWritten;
    int		nChkErr;
    HRESULT	dsRetVal;
	DWORD dwBytes = wiWave.dwBufferSize; 

	// This is the initial read. So we fill the entire buffer.
	// This will not wrap around so the 2nd pointers will be NULL.
    dsRetVal = lpDSBStreamBuffer->lpVtbl->Lock( lpDSBStreamBuffer, 0, dwBytes,
												&((LPVOID)lpWrite1), &dwLen1,
												&(LPVOID)lpWrite2, &dwLen2, 0 );

	if (dwLen1 != dwBytes)
	{
	//	DPF(0, "FillDataBuffer: dwBytes != dwLen1");
	}


    if( dsRetVal != DS_OK )
	{
		return; 
	}


    nChkErr = WaveReadFile( wiWave.hmmio, (UINT)dwLen1, lpWrite1,
                            &wiWave.mmck, &uActualBytesWritten );
	// if the number of bytes written is less than the 
	// amount we requested, we have a short file.
	if (uActualBytesWritten < dwLen1)
	{
	//	DPF(0, "FillDataBuffer: Actual written is less than dwlen1");
		if (!wiWave.bLoopFile)
		{
			// we set the bFoundEnd flag if the length is less than
			// one notify period long which is when the first notification comes in.
			// The next notification will then call send a message to process a stop. 
			if (uActualBytesWritten < wiWave.dwNotifySize)
				wiWave.bFoundEnd = TRUE;
			// Fill in silence for the rest of the buffer.
			//	DPF(0, "Filling in silence");
				FillMemory(lpWrite1+uActualBytesWritten, dwLen1-uActualBytesWritten, 
						(BYTE)(wiWave.pwfx->wBitsPerSample == 8 ? 128 : 0));
		}
		else
		{
			// we are looping.
			UINT uWritten = uActualBytesWritten;	// from previous call above.
			while (uWritten < dwLen1)
			{	// this will keep reading in until the buffer is full. For very short files.
				nChkErr = WaveStartDataRead( &wiWave.hmmio, &wiWave.mmck, &wiWave.mmckInRIFF );
				nChkErr = WaveReadFile(wiWave.hmmio, (UINT)dwLen1-uWritten, 
								lpWrite1 + uWritten, &wiWave.mmck, &uActualBytesWritten);
				uWritten += uActualBytesWritten;
			} // while
		} // else
	}

		// now unlock the buffer.
	dsRetVal = lpDSBStreamBuffer->lpVtbl->Unlock( lpDSBStreamBuffer, (LPVOID)lpWrite1, dwLen1,		NULL, 0 );
                                                                             
	wiWave.dwNextWriteOffset += dwLen1;	
	// this is a circular buffer. Do mod buffersize.
	if (wiWave.dwNextWriteOffset >= wiWave.dwBufferSize)
		wiWave.dwNextWriteOffset -= wiWave.dwBufferSize; 

//	DPF(3, "Setting dwProgress = 0");
	wiWave.dwProgress = 0;
	wiWave.dwLastPos = 0;
}






/********************************************************************************/
/* HandlePanScroll()                                                            */
/*                                                                              */
/*   Handles the pan trackbar scroll when a WM_HSCROLL is received.             */
/*                                                                              */
/********************************************************************************/
void HandlePanScroll( int nCode, int nPos )
    {
    long  lPan, lDelta;

    lPan = (LONG)SendMessage( hWndPan, TBM_GETPOS, (WPARAM)0, (LPARAM)0 );

    switch( nCode )
        {
        case TB_LINEUP:
            if( lPan >= PAN_MIN - 1 )
                lDelta = -1;
            break;
        case TB_LINEDOWN:
            if( lPan <= PAN_MAX + 1 )
                lDelta = 1;
            break;
        case TB_PAGEUP:
            if( lPan >= PAN_MIN - PAN_PAGESIZE )
                lDelta = -16;
            break;
        case TB_PAGEDOWN:
            if( lPan <= PAN_MAX + PAN_PAGESIZE )
                lDelta = 16;
            break;
        case TB_ENDTRACK:
            return;
        default:
            lDelta = 0;
        }

    if( lDelta )
        SendMessage( hWndPan, TBM_SETPOS, TRUE, lPan + lDelta );
    else
        SendMessage( hWndPan, TBM_SETPOS, TRUE, (long)nPos );
    //UpdateFromControls();
    }


/********************************************************************************/
/* HandleVolScroll()                                                            */
/*                                                                              */
/*   Handles the volume trackbar scrolling when a WM_HSCROLL is received.       */
/*                                                                              */
/********************************************************************************/
void HandleVolScroll( int nCode, int nPos )
    {
    long  lVol, lDelta;

    lVol = (LONG)SendMessage( hWndVol, TBM_GETPOS, (WPARAM)0, (LPARAM)0 );

    switch( nCode )
        {
        case TB_LINEDOWN:
            if( lVol <= VOL_MAX - 1 )
                lDelta = 1;
            break;
        case TB_LINEUP:
            if( lVol >= VOL_MIN + 1 )
                lDelta = -1;
            break;
        case TB_PAGEDOWN:
            if( lVol <= VOL_MAX - VOL_PAGESIZE )
                lDelta = 10;
            break;
        case TB_PAGEUP:
            if( lVol >= VOL_MIN + VOL_PAGESIZE )
                lDelta = -10;
            break;
        case TB_ENDTRACK:
            return;
        default:
            lDelta = 0;
        }

    if( lDelta )
        SendMessage( hWndVol, TBM_SETPOS, TRUE, (lVol + lDelta));
    else
        SendMessage( hWndVol, TBM_SETPOS, TRUE, (long)nPos );
    //UpdateFromControls();
    }


/********************************************************************************/
/* HandleFreqScroll()                                                           */
/*                                                                              */
/*   Handles the volume trackbar scrolling when a WM_HSCROLL is received.       */
/*                                                                              */
/********************************************************************************/
void HandleFreqScroll( int nCode, int nPos )
    {
    long  lFreq, lDelta;

    lFreq = (LONG)SendMessage( hWndFreq, TBM_GETPOS, (WPARAM)0, (LPARAM)0 );

    switch( nCode )
        {
        case TB_LINEDOWN:
            if( lFreq <= FREQ_MAX-1 )
                lDelta = 1;
            break;
        case TB_LINEUP:
            if( lFreq >= FREQ_MIN+1 )
                lDelta = -1;
            break;
        case TB_PAGEDOWN:
            if( lFreq <= FREQ_MAX - FREQ_PAGESIZE )
                lDelta = 10;
            break;
        case TB_PAGEUP:
            if( lFreq >= FREQ_MIN + FREQ_PAGESIZE )
                lDelta = -10;
            break;
        case TB_ENDTRACK:
            return;
        default:
            lDelta = 0;
        }

    if( lDelta )
        SendMessage( hWndFreq, TBM_SETPOS, TRUE, (lFreq + lDelta));
    else
        SendMessage( hWndFreq, TBM_SETPOS, TRUE, (long)nPos );
    //UpdateFromControls();
    }


/****************************************************************************/
/* ErrorMessageBox()                                                        */
/*                                                                          */
/*   A little routine to load error messages from the string resource table */
/* and pop them up in a MessageBox() for the world to see. The dwMBFlags    */
/* parameter allows the caller to specify the type of icon to use.          */
/*                                                                          */
/****************************************************************************/
void ErrorMessageBox( UINT uID, DWORD dwMBFlags )
    {
    LoadString( hInst, uID, szTemp, sizeof(szTemp));
    MessageBox( GetActiveWindow(), szTemp, szAppTitle, MB_OK | dwMBFlags );
    }


/****************************************************************************/
/* BuildTitleBar()                                                          */
/*                                                                          */
/*   Small routine to build and set the title bar text depending on whether */
/* or not a file is open.                                                   */
/****************************************************************************/
void BuildTitleBarText( void )
    {
    char szTitle[ sizeof(szAppCaption) + MAX_PATH + sizeof(" - ")];

    lstrcpy( szTitle, szAppCaption );
    if( bFileOpen )
	{
	lstrcat( szTitle, " - " );
	lstrcat( szTitle, szFileTitle );
	}
    SetWindowText( hWndMain, szTitle );
    }


/****************************************************************************/
/* GetMediaStartPath()                                                      */
/*                                                                          */
/*   This helper function attempts to get the media directory for Direct3D, */
/* which is where all the installed DX wave files go. If it can't find that */
/* it settles for the media sub-directory of the Windows directory.         */
/****************************************************************************/
void GetMediaStartPath( void )
    {
    HKEY    hReg;
    DWORD   cbStartPathLen;

    if( ERROR_SUCCESS != RegOpenKeyEx( HKEY_LOCAL_MACHINE,
			gszRegKeyDirect3D,
			0, KEY_READ, &hReg ))
	{
        goto REG_OPEN_FAILED;
	}
    else
	{
	// Query the Registry for the path to the media directory
	cbStartPathLen = sizeof( szCDStartPath );
	if( ERROR_SUCCESS != RegQueryValueEx( hReg, gszRegValueD3DPath,
				NULL, NULL,
				szCDStartPath, &cbStartPathLen ))
	    {
	    goto REG_OPEN_FAILED;
	    }
	RegCloseKey( hReg );
	hReg = NULL;
	}

    return;

REG_OPEN_FAILED:
    // Start off by getting the Windows directory -- we're trying to build a
    // file path like "C:\WINDOWS\MEDIA", but the WINDOWS directory could be
    // named anything, so we must ask.
    GetWindowsDirectory( szCDStartPath, sizeof(szCDStartPath));
    // If there's no trailing backslash, append one
    if( lstrcmp( &szCDStartPath[lstrlen(szCDStartPath)], TEXT("\\") ))
	lstrcat( szCDStartPath, TEXT("\\"));
    // Now add on the MEDIA part of the path
    lstrcat( szCDStartPath, TEXT("MEDIA"));
    }



