#ifndef _game
#define _game

#include "scigame.h"
#include "list.h"
#include "stdio.h"
#include "GameSync.h"

class MapPlane;
class InterfacePlane;
class MissionGoal;

//#include "player.h"

class Player;
class Template;
class Clan;
class Spell;
class NonPlayerClass;
class Bitmap;

// Maximum number of players at a single time. Don't go above 32!!!
#define MAX_PLAYERS 8

typedef enum TAG_ENUM_NPC
{
	NPC_PEASANT,
	NPC_VAMPIREHUNTER,
	NPC_SOLDIER,
	NPC_GYPSY,
	NPC_MERCHANT,
	NPC_OTHER,
	ENUM_NPC_COUNT
} ENUM_NPC;

typedef struct _NET_TURN_HEAD {
	BYTE	jKind;
	long	lTurn;
	long	lChecksum;
} NET_TURN_HEAD, *PNET_TURN_HEAD;

typedef struct _NET_INTERNAL_HEAD {
	BYTE	jKind;
	long	lTurn;
	long	lChecksum;
} NET_INTERNAL_HEAD, *PNET_INTERNAL_HEAD;

class Game : public sciGame
{
public:
	Game();
	~Game();

	int nGetPlayerCount() { return mnPlayers; }
	// debugging stuff -- checksum and text dumping
	long lGetChecksum(void);
	void vDumpSyncText(FILE *fileDump);
	
	void vGameTick(void);
	void vGameCycle(void);

	int nNetTurn(void) {return mnCurrentNetTurn;}
	int nPostingTurn(void) {return mnPostingNetTurn;}
	int nPostingDelay(void) {return mnTurnPostingDelay;}

	int nGameCycle(void) {return mnGameCycles;}
	int nTicksPerGameCycle(void) {return mnTicksPerGameCycle;}

	Player *poLocalPlayer(void) {return mpoLocalPlayer;}
	List<Player> *poPlayers(void) {return &moPlayers;}
	void vClearPlayers(void);
	void vAddPlayer(Player *poPlayer);

	void vSetupGame(void);

	void vStartGame(void);
	void vKillGame(void);

	BOOL bNetGame(void) {return mbNetGame;}

	int nDrawsSkipped(void) {return mnDrawsSkipped;}

	Template *mpoTemplate; // the global game template

	List<Clan> moClans;
	List<Spell> moSpells;
	RefList<MissionGoal> moGoals;

	NonPlayerClass * GetNonPlayerClass(ENUM_NPC eNpc);
	Spell * GetSpell(int nIndex);
	Spell * GetTrueSpell(int nIndex);	//uses true indexing
	BYTE ColorHighlight() { return mclrHighlight; }
	
	void ToggleClearGameMap();
	void SetClearGameMapInterval(int nVal);
	void SetNetworkGameSynchronizationInterval(int nVal);

	void AddGoal(MissionGoal * poMissionGoal);
	int GoalCount();
	int GoalCountCompleted();

	Bitmap * GetHighlightBitmap();

	WORD wGetMiniMapColor(int nColorIndex);

protected:
	BOOL mbClearGameMap;
	int mnClearGameMapCyclesInterval;
	int mnNetworkGameSynchronizationInterval;

	List<NonPlayerClass> moNpcList;
	
	virtual void LoadTheRoom(int roomNum);
	BOOL bHandleNetTurn();
	void vCheckNet(void);

	// players
	Player			*mpoLocalPlayer;
	List<Player>	moPlayers;
	int				mnPlayers;

	//
	// GAME CYCLES INFO
	//
	BOOL	mbGameStarted;
	int		mnGameCycles;
	int		mnLastGameCycleTime;
	BOOL	mbGameCycling;
	int		mnTicksPerGameCycle;
	int		mnCycleOverage; // how much past the real cycle time are we
	int		mnDrawsSkipped;
	int		mnLastDrawCycle;

	//
	// NET TURNS INFO
	//
	int		mnCurrentNetTurn;
	int		mnPostingNetTurn;
	int		mnCyclesPerNetTurn;
	int		mnLastNetTurnCycle;
	int		mnLastNetStopTime;
	BOOL	mbNetPause;
	int		mnTurnPostingDelay;
	int		mnLastNetCheckTime;
	BOOL	mbNetGame;

	GameSync moGameSync;
	BOOL	mbBadSynch;
	BYTE	mclrHighlight;

	MapPlane * mpoMapPlane;
	InterfacePlane * mpoIntPlane;
};	

extern Game *game;

#ifndef _VWDB_
#include "mapnode.h"
#endif


#endif // _game
