#include "os.h"
#include "procs.h"
#include "GameSync.h"

#define DEBUG_GAMESYNC 0

void GameSync::init()
{
	mnTurnMaxNew = INT_MIN+1;
	mnTurnMax = INT_MIN+1;

	for (int i=0; i<mnPlayers; i++)
	{
		mnTurn[i] = INT_MIN;
	}
}

void GameSync::SetPlayerCount(int nPlayers)
{ 
	// TBD - if we need more than 32 players then dynamically allocate max slots
	assert(nPlayers <= PLAYERS_MAX);
	assert(nPlayers >= 0);
	mnPlayers = nPlayers;

	init();
}

BOOL GameSync::IsValid(UINT unPlayer, int nTurn, int lChecksum)
{
	// validate parms
	assert(unPlayer >= 0);
	assert(unPlayer < PLAYERS_MAX);
	
	// if highest turn so far then record
	if (nTurn > mnTurnMaxNew)
	{
		mnTurnMaxNew = nTurn;
	}
	
	// too far ahead ... problem?
#if DEBUG_GAMESYNC
	assert((mnTurnMaxNew < 5*SENDSYNCONNTHTURN + mnTurnMax)||(mnTurnMax==(INT_MIN+1)));
#endif

	// if this player already submitted turn 
	//    for this sync check
	if (mnTurn[unPlayer] == mnTurnMax)
	{
		// throw out this turn
#if DEBUG_GAMESYNC
		vDebugOut("Already have turn %d for player %d - discard turn %d\n", mnTurnMax, unPlayer, nTurn);
#endif
		return TRUE; 
	}

	// need a later turn
	if (nTurn < mnTurnMax) 
	{
#if DEBUG_GAMESYNC
		vDebugOut("Waiting for turn %d for player %d - discard turn %d\n", mnTurnMax, unPlayer, nTurn);
#endif
		return TRUE;
	}

	// if past the player's last recorded turn
	if (nTurn > mnTurn[unPlayer])
	{
		// NEW TURN MAX
		if (nTurn > mnTurnMax)
		{
			mnTurnMax = nTurn;
		}

		// RECORD IT FOR THE PLAYER
		mnTurn[unPlayer] = nTurn;
		mlChecksum[unPlayer] = lChecksum;

#if DEBUG_GAMESYNC
		vDebugOut("Turn Arrived:: Turn %d :: Player %d :: Checksum %d\n", nTurn, (int)unPlayer, (int)lChecksum);
#endif
	}

	int i=0;
	
	BOOL bResetTurnMax = TRUE;

	while(i < mnPlayers)
	{
		if (mnTurn[i] == mnTurnMax)
		{
			if (lChecksum != mlChecksum[i])
			{
#if DEBUG_GAMESYNC
				vDebugOut("BADCHECKSUM\n");
#endif
				return FALSE;
			}
		}
		else   //if (mnTurn[i] < mnTurnMax)
		{
			bResetTurnMax = FALSE;
		}
		i++;
	}


	if (bResetTurnMax)  // ALL PLAYERS HAVE CHECKED IN FOR SYNCHRONIZATION SO
						// ACCEPT THE NEXT ONE
	{
#if DEBUG_GAMESYNC
		vDebugOut("All Checksums valid for turn %d.\n", mnTurnMax);
#endif
		mnTurnMax = mnTurnMaxNew + 1;
#if DEBUG_GAMESYNC
		vDebugOut("Next Synch Check at turn %d.\n", mnTurnMax);
#endif
	}

	return TRUE;
}
