#include "unit.h"
#include "human.h"
#include "NonPlayerClass.h"
#include "mob.h"
#include "guard.h"
#include "movement.h"
#include "item.h"
#include "objectdb.h"
#include "procs.h"

#define CYCLES_500MS	10


Human::Human(Player *poOwner, MapPlane *poPlane, MapNode *poNode, int nTemplateID):Unit(poOwner,poPlane,poNode,nTemplateID)
{
	// SETUP THE NPC
	mnCyclesBeforeJoiningNewMob = 0;

	int nVal = mpoTemplate->lGetValue(TAG_clan, NPC_PEASANT);
	ENUM_NPC eVal = NPC_PEASANT;
	if (nVal>=0 && nVal < ENUM_NPC_COUNT)
	{
		eVal = (ENUM_NPC)nVal;
	}
	else
	{
		assert(0); // this should only happen if
					// the code gets modified incorrectly elsewhere
	}
	mpoNpc = game->GetNonPlayerClass(eVal);

	vSetLevel(mnLevel);

	SetSpeed(SPEED_NORMAL);

	// SETUP HUMAN SPECIFICS
	mnVisibleRange = vGetVisibleRange();
	mnVampireRange[0] = mnVisibleRange;
	mnVampireRange[1] = mnVisibleRange;
	mnVampireRange[2] = mnVisibleRange;

	mfAnxiety = 0;
}

Template * Human::vGetClanTemplate()
{
	if (!mpoNpc) return NULL;
	return mpoNpc->GetTemplate();
}

vClan * Human::vGetClan()
{
	return mpoNpc;
}

// GAME CYCLE SHOULD BE CALLED BY DERIVED CLASSES
void Human::vGameCycle()
{
	// we'll assume we're not moving until otherwise contradicted (in vMoveOneCycle)
	mbMoving = FALSE;

	// IF MOB LEADER THEN MAKE THE MOB CYCLE TOO
	if (bIsMobLeader())
	{
		mpoMobCoordinator->GameCycle();
	}

	// DECREMENT UNDER ATTACK
	if (mnUnderAttackCycles)
	{
		mnUnderAttackCycles--;
	}

	// NOT ALIVE
	if (!mbAlive)
	{
		if (moCurrentCommand.meType != CM_NONE)
		{
			vCommandNone();
		}
		return;
	}

	// UNDER THRALL
	if (mpoThrallOwner)
	{
		mnThrallCyclesRemaining--;
		if (mnThrallCyclesRemaining)
		{
			// IF NOT A VALID THRALL COMMAND
			if (!(moCurrentCommand.meType == CM_MOVE && moCurrentCommand.mbTargetUnit))
			{
				// NO NEED TO CLEAR THE COMMAND EVERY TIME
				return;
			}
			// ELSE PROCESS THIS COMMAND
		}
		else
		{
			// CLEAR THE COMMAND
			vCommandNone();
			mpoThrallOwner = NULL;
			return;
		}
	}
	else
	{
		// NOT UNDER THRALL
		// SO CHECK THIS
		if (mnCyclesBeforeJoiningNewMob)
		{
			mnCyclesBeforeJoiningNewMob--;
		}
		
		// AND NOT MOVING?
		if (!mbBetweenPoints)
		{
			vScan();
		}
		vThink();
		vRegeneratePower();
	}

	// IF MOVING THEN MOVE -- NO THINKING?
	if (mbBetweenPoints)
	{
		vMoveOneCycle();
		return;
	}

	// ANXIETY FALLOFF
	if (mfAnxiety > 0)
		mfAnxiety = max(0, mfAnxiety - mpoNpc->fGetAnxietyFalloffRate());

	vFormAMobIfNecessary();

	// FOR PEASANTS ONLY
	if (moCurrentCommand.meType == CM_NONE)
	{
		// IGNORE EVERY
		if (!bIsMobMember() || bIsMobLeader())
		{
			if (!mpoSequencer)
			{
				if (!bIsMobLeader())
				{
					SelectNewRandomBehavior();
				}
				else
				{
					Sequence_Clear();
					Sequence_WanderMethodically(0);
				}
			}
			assert(mpoSequencer);
			Command * poCommand;

			mpoSequencer->Get(&poCommand);
			mpoSequencer->ProceedToNext();
			if (poCommand) 
			{
				vCommand(*poCommand, TRUE);
			}
			
			if (mpoSequencer->bAtBeginning())
			{
				delete mpoSequencer;
				mpoSequencer = NULL;
			}
		}
	}

	// we're at a square -- move, stay or fight?
	switch (moCurrentCommand.meType) {
	case CM_NONE:
		{
			if (IsGuarded())
			{
				mpoGuardCoordinator->Idle();
			}

			// if we're controlled by a player, bail
			if (mpoOwner) 
			{
				return;
			}
		}
		break;
	case CM_GUARD_MOVEINTOPOSITION:

		// are we at the destination? Then kill this move stuff
		if (mpoCurNode->mnXPos == moCurrentCommand.mnX && mpoCurNode->mnYPos == moCurrentCommand.mnY) 
		{
			// POST A GUARD_READY MESSAGE TO GUARDED UNIT
			vCommandNone();
			if (mpoGuardedUnit)
			{
				mpoGuardedUnit->Message(this, GUARD_READY);
			}
			
			return;
		}

		// move
		vMoveToNextNode();

		// TBD - check bad start position / blocked / send busy

/*			if (pNode == NULL)
			{
				vCommandNone();
				mpoGuardedUnit->Message(this, GUARD_BUSY);
			}*/
		
		break;
	case CM_RETREAT:
		SetSpeed(SPEED_MOBBING);

		/*
		//	1) REMAIN UNCLOAKED DURING RETREAT
		if (!mnAdrenalineCycleTicker)
		{
			// POP ADRENALINE IF NOT ALREADY
			// THIS WILL UNCLOAK IF NEEDED	
			mnAdrenalineCycleTicker = mpoClan->GetAdrenalineDuration();

			SetSpeedAdrenalineBoost();
		}

		// MAKE SURE WE ARE UNCLOAKED
		if (mbCloaked)
		{
			vUncloak();
			// TBD -- SHOULD WE DELAY ONE CYCLE
		}
		*/
		// DETERMINE WHERE TO GO
		BOOL bDone;
		bDone = FALSE;
		int nX;
		int nY;

		if (moCurrentCommand.mnState == STATE_COMMAND_NONE)
		while (!bDone)
		{
			switch (moCurrentCommand.mnState)
			{
			case STATE_COMMAND_NONE:
				{
					moCurrentCommand.mnState = STATE_RETREAT_RUNAWAYFROMENEMY;
					PlayClassSound(TAG_Swei, 80, 1, 3);
				}			
				break;
			case STATE_RETREAT_RUNAWAYFROMENEMY:
				// TBD - FIND MOST UNPOPULATED DIRECTION FROM ENEMY -- THEN RUN!!!!
				{
					nX = 1;
					nY = 1;
					FindRetreatLocation(1, &nX, &nY);
					//nX = nX * 4;
					//nY = nY * 4;
					
					moCurrentCommand.mnX = nGetX() + nX;
					moCurrentCommand.mnY = nGetY() + nY;
					moCurrentCommand.mnX = max(0, moCurrentCommand.mnX);
					moCurrentCommand.mnX = min(gpoMap->nGetWidth()-1, moCurrentCommand.mnX);

					moCurrentCommand.mnY = max(0, moCurrentCommand.mnY);
					moCurrentCommand.mnY = min(gpoMap->nGetHeight()-1, moCurrentCommand.mnY);
				}
				bDone = TRUE;
				break;	
			case STATE_RETREAT_TOUNIT:
				bDone = TRUE;
				break;	
			}
		}

		BOOL bCancel;
		bCancel = FALSE;

		// CONTINUE TOWARDS DESTINATION UNTIL THERE ARE NO ENEMY IN VISIBLE RANGE
		//  OR REGAINED ENOUGH HEALTH TO JOIN FIGHT
		{
			int nCount = 0;
			int nNumPlayers = game->nGetPlayerCount();
			//int nThisPlayer = mpoOwner->mnPlayer;
			int i;

			// COUNT ENEMY UNITS
			for (i=0;i<nNumPlayers;i++)
			{
				nCount += mnVisibleTotal[i].count;
			}
			if (nCount == 0)
			{
		int aOffset[] = {	 1,  0, 
							 1, -1,
							 0, -1,
							-1, -1, 
							-1,  0,
							-1,  1,
							 0,  1,
							 1,  1 };

				//SetSpeed(SPEED_NORMAL);
				
				vCommandNone();
				Sequence_Clear();
				Sequence_WanderMethodically(0,aOffset[mnLastRetreatDirection*2],aOffset[mnLastRetreatDirection*2+1]);
				mnLastRetreatDirection = -1;
				Sequence_WaitShort();
				return;
			}
		}
/*
		// STATE_RETREAT_TOUNIT
		if (moCurrentCommand.mbTargetUnit)
		{
			int nTargetX, nTargetY;
			Unit *poTarget;

			poTarget = (Unit *)goObjectDB.poGetObject(moCurrentCommand.mnUnit);
			
			if (!poTarget || !poTarget->bAlive())
			{
				// OUR TARGET HAS BEEN DELETED
				moCurrentCommand.mnState = STATE_COMMAND_NONE;
				return;
			}
			
			moCurrentCommand.vGetTargetPos(&nTargetX, &nTargetY);

			int nTargetDist = lDistToSpace(nTargetX, nTargetY);

			if (nTargetDist <= 2)
			{
				// CLOSE ENOUGH
				vCommandNone();
				return;
			} 
			else 
			{
				// MOVE CLOSER
				if (nTargetX != mnLastTargetX || nTargetY != mnLastTargetY) 
				{
					vCalcRoute();
				}
				vMoveToNextNode();
			}
		}
		else	// STATE_RETREAT_RUNAWAYFROMENEMY
		*/
		{
			if (mpoCurNode->mnXPos == moCurrentCommand.mnX && mpoCurNode->mnYPos == moCurrentCommand.mnY) 
			{
				moCurrentCommand.mnState = STATE_COMMAND_NONE;
				//vCommandNone();
				return;
			}
			vMoveToNextNode();
			break;
		}

		//	7) continue retreating until no visible enemy or enough health to continue attack

		break;

	case CM_TAKE:
		{
			int nTargetX, nTargetY;
			Item *poTarget;

			if (moCurrentCommand.mbTargetUnit) 
			{
				poTarget = (Item *)goObjectDB.poGetObject(moCurrentCommand.mnUnit);
				
				if (!poTarget) 
				{
					vCommandNone();
					return;
				}
			}
			
			moCurrentCommand.vGetTargetPos(&nTargetX, &nTargetY);

			int nTargetDist = lDistToSpace(nTargetX, nTargetY);

			if (nTargetDist <= 1) 
			{
				Take(poTarget->poNode(), poTarget);
				vCommandNone();
				return;
			} 
			else 
			{
				// MOVE CLOSER
				if (nTargetX != mnLastTargetX || nTargetY != mnLastTargetY) 
				{
					// STOP THE TAKE
					vCommandNone();
					return;
				}
				vMoveToNextNode();
			}
			break;
		}
		break;
	case CM_FEED:
		assert(0);  // UNIT SPECIFIC BEHAVIOR
		break;
	case CM_CONVERT:
		assert(0);  // UNIT SPECIFIC BEHAVIOR
		break;
	case CM_CAST:
		assert(0);
		break;
	case CM_ATTACK:
		{
			int nTargetX, nTargetY;
			Unit *poTarget;

			// check for bad target
			if (moCurrentCommand.mbTargetUnit) 
			{
				poTarget = (Unit *)goObjectDB.poGetObject(moCurrentCommand.mnUnit);

				// if the target is gone or dead, we're done
				if (!poTarget)
				{
					vCommandNone();
					return;
				}
				if (!poTarget->bAlive()) 
				{
					vCommandNone();
					if (bIsMobLeader())
					{
						mpoMobCoordinator->Regroup();
						vCommandNone();
						Sequence_Clear();
						Sequence_WanderMethodically(0);
					}
					return;
				}
			}
			
			moCurrentCommand.vGetTargetPos(&nTargetX, &nTargetY);

			// are we close enough to uncloak and dash? Then do so.
			int nTargetDist = lDistToSpace(nTargetX, nTargetY);

			/*
			if (nTargetDist <= GetRangeToDecloak()) 
			{
				vUncloak();
			}
			*/

			if (moCurrentCommand.mbInternal)		// INTERNAL COMMAND
			{
				if (!mnUnderAttackCycles && !bAttackEnemyUnit(poTarget))	// CONTINUE THE ATTACK?
				{
					vCommandNone();					// STOP THE ATTACK
					return;
				}
			}

			// are we close enough to attack? Then attack, don't move
			if (nTargetDist <= mnAttackRange) 
			{
				Unit * poUnit = poTarget;
				if ((poUnit->mpoThrallOwner == this)||(poUnit->mpoThrallOwner == NULL))
				{
					poUnit->SetThrallOwner(this, 50);
				}
				vChangeFacing(mpoCurNode->mnXPos, mpoCurNode->mnYPos, nTargetX, nTargetY);
				vSetAction(ACT_ATTACK);
				vAttack();
			} 
			else 
			{
				// move closer

				// if the target has moved, recalculate the path
				if (nTargetX != mnLastTargetX || nTargetY != mnLastTargetY) 
				{
					vCalcRoute();
				}

				// so now we move.
				vMoveToNextNode();
			}
			break;
		}
		
	case CM_GUARD_MOVETOVERIFY:
	case CM_GUARD_MOVEINTERNAL:
		if (mpoCurNode->mnXPos == moCurrentCommand.mnX && mpoCurNode->mnYPos == moCurrentCommand.mnY) 
		{
			if (IsPatrolling())
			{
				mpoMovement->ProceedToNext();
				Patrol();
			}
			else
			{
				if (CM_GUARD_MOVETOVERIFY == moCurrentCommand.meType)
				{
					assert(mpoGuardedUnit);

					vCommandNone();

					if (mpoGuardedUnit)
					{
						mpoGuardedUnit->Message(this, GUARD_VERIFIED);
					}
				}
				else
				{
					vCommandNone();
				}
			}
			return;
		}
		// move
		vMoveToNextNode();

		break;

	case CM_MOVE_INVITE:
		if (moCurrentCommand.mbTargetUnit)
		{
			int nTargetX, nTargetY;
			Unit *poTarget;

			poTarget = (Unit *)goObjectDB.poGetObject(moCurrentCommand.mnUnit);
			
			if (!poTarget) 
			{
				vCommandNone();
				return;
			}
			
			moCurrentCommand.vGetTargetPos(&nTargetX, &nTargetY);

			int nTargetDist = lDistToSpace(nTargetX, nTargetY);

			if (nTargetDist < 4)
			{
				if (poTarget->bMoving())
				{
					poTarget->vCommandNone();
				}
			}
			if (nTargetDist < 2)
			{
				poTarget->vSetMobbing(TRUE, mpoMobCoordinator);
				mpoMobCoordinator->AddUnitAsFollower(poTarget);
				// LET'S GO
				mpoMobCoordinator->vPlayRandomSound(TAG_Satk);
				SetThrallOwner(this, 25);
				vCommandNone();
			}
			else 
			{
				// MOVE CLOSER
				if (nTargetX != mnLastTargetX || nTargetY != mnLastTargetY) 
				{
					vCalcRoute();
				}
				vMoveToNextNode();
			}
		}
		else
		{
			vCommandNone();
		}
		break;

	case CM_MOVE_FOLLOW:
		if (moCurrentCommand.mbTargetUnit)
		{
			Command oCommand = moCurrentCommand;
			int nTargetX, nTargetY;
			Unit *poTarget;

			poTarget = (Unit *)goObjectDB.poGetObject(moCurrentCommand.mnUnit);
			
			if (!poTarget) 
			{
				vCommandNone();
				return;
			}
			
			moCurrentCommand.vGetTargetPos(&nTargetX, &nTargetY);

			int nTargetDist = lDistToSpace(nTargetX, nTargetY);

			if (poTarget->bMoving())
			{
				if (nTargetDist < 2 * moCurrentCommand.mnNewValue)
				{
					return;
				}
			}
			else 
			{
				if (nTargetDist < moCurrentCommand.mnNewValue)
				{
					return;
				}
			}

			// MOVE CLOSER
			if (nTargetX != mnLastTargetX || nTargetY != mnLastTargetY) 
			{
				vCalcRoute();
			}
			vMoveToNextNode();
			if (moCurrentCommand.meType == CM_NONE)
			{
				moCurrentCommand = oCommand;
			}
		}
		else
		{
			vCommandNone();
		}
		break;

	case CM_MOVE:
		{
			BOOL bCancel = FALSE;

		// are we at the destination? Then kill this move stuff
		if (moCurrentCommand.mbTargetUnit) 
		{
			int nTargetX, nTargetY;
			Unit *poTarget;

			poTarget = (Unit *)goObjectDB.poGetObject(moCurrentCommand.mnUnit);
			
			if (!poTarget) 
			{
				vCommandNone();
				return;
			}
			
			moCurrentCommand.vGetTargetPos(&nTargetX, &nTargetY);

			int nTargetDist = lDistToSpace(nTargetX, nTargetY);

			if (nTargetDist <= 2)
			{
				return;
			} 
			else 
			{
				// MOVE CLOSER
				if (nTargetX != mnLastTargetX || nTargetY != mnLastTargetY) 
				{
					vCalcRoute();
				}
				vMoveToNextNode();
			}
		}
		else
		{
			bCancel = TRUE;
			if (mpoCurNode->mnXPos == moCurrentCommand.mnX && mpoCurNode->mnYPos == moCurrentCommand.mnY) {
				if (IsPatrolling())
				{
					mpoMovement->ProceedToNext();
					Patrol();
				}
				else
				{
					if (bIsMobLeader())
					{
						mpoMobCoordinator->mState = 0;
					}
					vCommandNone();
				}
			}
			else if (IsGuarded())
			{
				assert(mpoGuardCoordinator);
				{
					BOOL bInterior = (moCurrentCommand.mnUnit == (ID_NONE - 1));
					mpoGuardCoordinator->Move(gpoMap->poGetNode(moCurrentCommand.mnX, moCurrentCommand.mnY), bInterior);
				}
				vCommandNone();
			}
			else
			{
				// move
				vMoveToNextNode();
				bCancel = TRUE;
			}


		}
		if (moCurrentCommand.meType == CM_NONE)
		{
			// IT WILL AUTOMATICALLY REBUILD
			if (bIsMobLeader() && bCancel)
			{
				int score = 0;
				int sx = max(0, nGetX()-1);
				int sy = max(0, nGetY()-1);
				int ex = min(gpoMap->nGetWidth()-1, nGetX()+1);
				int ey = min(gpoMap->nGetHeight()-1, nGetY()+1);
				int count = 0;
				for (int i=sx;i<ex;i++)
					for(int j=sy;j<ey;j++)
					{
						count++;
						if (gpoMap->poGetNode(i,j)->bBlocked())
						{
							score++;
						}
					}
				if (count == score)
				{
					delete mpoSequencer;
					mpoSequencer = NULL;
				}
				if (!mpoSequencer)
				{
					mpoMobCoordinator->ChooseNewLeader();
					mpoMobCoordinator->Regroup();
				}
				return;
			}
		}
		}
		break;
	case CM_WAIT_500MS:
		{
			if (!(mdwFlags & UNIT_FLAG_WAITING))
			{
//				vDebugOut("Started Waiting...\n");
				mdwFlags |= UNIT_FLAG_WAITING;
				mnStartWaitCycle = game->nGameCycle();
			}

			int nCurCycle = game->nGameCycle();
			if ((nCurCycle - mnStartWaitCycle) >= CYCLES_500MS) 
			{
//				vDebugOut("Finished Waiting...\n");
				mdwFlags = mdwFlags & (~(DWORD)UNIT_FLAG_WAITING );
				vCommandNone();
				return;
			}
		}
		break;

	default:
		break;
	} //switch

}

BOOL Human::vCanBeEnthralled()
{
	if (mpoNpc->GetTemplate()->lGetValue(TAG_eThr, 1) != 0)
		return TRUE;
	else
		return FALSE;
}

int Human::GetUpgradeSight()
{
	return mpoNpc->GetUpgradeSight();
}

int Human::GetUpgradeAttack()
{
	return mpoNpc->GetUpgradeAttack();
}

int Human::GetUpgradeDefend()
{
	return mpoNpc->GetUpgradeDefend();
}

int Human::GetUpgradeSpeed()
{
	return mpoNpc->GetUpgradeSpeed();
}

int Human::GetUpgradeStealth()
{
	return mpoNpc->GetUpgradeStealth();
}

int Human::GetUpgradeDetection()
{
	return mpoNpc->GetUpgradeDetection();
}

int Human::GetUpgradeSpells()
{
	return mpoNpc->GetUpgradeCasting();
}

int Human::GetUpgradeLevel()
{
	return mnLevel;
}

short Human::vGetRelativeHealth(Player * poPlayer)
{
	vUpdateRelativeHealth(poPlayer);

	return mnRelativeHealth;
}

void Human::vUpdateRelativeHealth(Player * poPlayer)
{
	float fValue;

	// IF ON THE SAME TEAM MAKE IT ALL VISIBLE
	float fHealth = (mfHealth / mfMaxHealth) * 100;
	float fPower = (mfPower / mfMaxPower) * 100;

	//only local vampires get the relative health formula
	if (fHealth < fPower) 
	{
		// Health is lower so add it to 1/x of Power
		fValue = ((float)mnRelHealthPercent / 100) * fPower + fHealth;
	} 
	else 
	{
		// Power is lower(or equal) so add it to 1/x of Health
		fValue = ((float)mnRelHealthPercent / 100) * fHealth + fPower;
	}
	mnRelativeHealth = min(100, (short)fValue);
}

int Human::GetAnxiety()
{
	return (int)mfAnxiety;
}

void Human::vIncreaseAnxiety(float nAmount)
{
	mfAnxiety = min(mfAnxiety + nAmount, mpoNpc->nGetMaxAnxiety());
	vDebugOut("Human increasing anxiety by %d to %f\n", nAmount, mfAnxiety);
}

void Human::vSetAnxiety(int nAmount)
{
	mfAnxiety = (float)max(0, min(nAmount, mpoNpc->nGetMaxAnxiety()));
}

void Human::vLeaveMobIfNecessary()
{
	if (mpoMobCoordinator)
	{
		if (mfAnxiety <= mpoNpc->nGetAnxietyLeaveMob())
		{
			BOOL bOut;
			mpoMobCoordinator->RemoveUnit(&bOut, this);

			mnCyclesBeforeJoiningNewMob = mpoNpc->GetTemplate()->lGetValue(TAG_mTLJ, 5) * 20;
			// TBD - THIS STUFF SHOULD REALLY BE IN REMOVEUNIT
			if (!mpoMobCoordinator->GetSize())
			{
				delete mpoMobCoordinator;
			}
			mpoMobCoordinator = NULL;
		}
	}
}

void Human::vFormAMobIfNecessary()
{
	// SHOULD THIS UNIT FORM A MOB?
	if (!mpoMobCoordinator)
	{
		if (!vIsReadyToEnterMob()) return;
		
		if (!gnMobCount)
		{
			if (mfAnxiety >= mpoNpc->nGetAnxietyEnterMob())
			{
				mpoMobCoordinator = new MobCoordinator(*this);
				Sequence_Clear();
				Sequence_WanderMethodically(0);
				vSetMobbing(TRUE, mpoMobCoordinator);
			}
		}
	}
}

BOOL Human::vIsReadyToEnterMob()
{
	return (mnCyclesBeforeJoiningNewMob == 0);
}

