// INTBUTTN.CPP -- the various interface buttons
//
// TW

#include "intbuttn.h"
#include "statusln.h"
#include "intplane.h"
#include "unit.h"
#include "util.h" //for vDebugPurposes

#define DEFAULT_TEXT_UPGRADE_BG	3

#define ICON_FIRST_SPACING		14		//the spacing between the  Icon and it's first button
#define BTN_SPACING				6		//the spacing between each button

UpgradeButton *UpgradeButton::mpoCurrentHilight = NULL;

class UpgradeButtonRepeatScript : public sciScript 	{
	public:
		Boolean mbRightClicking;

		UpgradeButtonRepeatScript(Boolean bRightClicking) : sciScript(NULL) {
			mbRightClicking = bRightClicking;
		}

		void ChangeState(int newState) {
			sciScript::ChangeState(newState);
			switch (state) 				{
			case 0: //Setup
				SetTicks(((UpgradeButton*)client)->mnStartDelay);
				break;
			case 1:
				if (((UpgradeButton*)client)->bEnabled() &&
						!((UpgradeButton*)client)->bHidden())
				{
					if (mbRightClicking)
						((UpgradeButton*)client)->vDoRightClick();
					else
						((UpgradeButton*)client)->vDoLeftClick();
				}
				SetTicks(((UpgradeButton*)client)->mnRepeatDelay);
				--state;
				break;
			}
		}
	};



UpgradeButton::UpgradeButton(int nButtonID, Boolean bSelected) {
	mnButtonID = nButtonID;
	
	mbEnabled = TRUE;
	mbSelected = bSelected;
	mbPending = FALSE;
	mbMouseDown = FALSE;
	
	mnState = BTN_STATE_UNSELECTED;
	mpoSelectedView = NULL;
	mpoPendingView = NULL;
	mpoText = NULL;
	mpoSelectionList = NULL;
	mnJustification = JUST_LEFT | JUST_TOP;

	mcHotkey = 0;
	mcHotkey2 = 0;
	mcRightHotkey = 0;
	mnStartDelay = -1;
	mnRepeatDelay = -1;

	mnUnselectedCel = 0;
	mnSelectedCel = 0;
	mnHilightCel = 0;
	mnDepressedCel = 0;
	mnDisabledCel = 0;

	//some basic text defaults
	mnPendingBG = 223;
	mnNonPendingBG = DEFAULT_TEXT_UPGRADE_BG;
	mnUnselectedFont = 2;	
	mnSelectedFont = 3;
	mnHilightFont = 4;	
	mnDepressedFont = 5;
	mnDisabledFont = 5;	

}

UpgradeButton::~UpgradeButton() {
	if (mpoText) {
		delete mpoText;
		mpoText = NULL;
	}
	if (mpoSelectedView) {
		delete mpoSelectedView;
		mpoSelectedView = NULL;
	}
	if (mpoPendingView) {
		delete mpoPendingView;
		mpoPendingView = NULL;
	}
}	
 
Boolean UpgradeButton::HandleEvent(sciEvent &oEvent) {
//	sciProp::HandleEvent(oEvent);
//	return oEvent.claimed;
	
	if (!mbEnabled || hidden) {
		return FALSE;
	}

	switch (oEvent.type)
		{
		case MOUSE_DOWN:
			if ((!OnMe(oEvent.x, oEvent.y) && !mbMouseDown) || oEvent.claimed)
				break;
			if ((oEvent.modifiers & RIGHT_MOUSE) && !mbMouseDown)
			{
				mbMouseDown = TRUE;
				pApp->pWindow->vAddEventHog(this); //make a (modal)hog
//				plane->CaptureEvents(this);
				vSetState(BTN_STATE_DEPRESSED);
				oEvent.claimed = TRUE;
				if (mnStartDelay != -1)
					SetScript(new UpgradeButtonRepeatScript(TRUE));
			}
			if ((oEvent.modifiers & LEFT_MOUSE) && !mbMouseDown)
			{
				mbMouseDown = TRUE;
				pApp->pWindow->vAddEventHog(this); //make a (modal)hog
//				plane->CaptureEvents(this);
				vSetState(BTN_STATE_DEPRESSED);
				oEvent.claimed = TRUE;
				if (mnStartDelay != -1)
					SetScript(new UpgradeButtonRepeatScript(FALSE));
			}
			break;
		case MOUSE_MOVE:
			if (mbMouseDown)
			{
				oEvent.claimed = TRUE;
				if (!OnMe(oEvent.x, oEvent.y))
				{
					//SetState(DEFAULT_BUTTON_NORMAL); - tw: i think i like the hilight 'look' better
					vSetState(BTN_STATE_HILIGHTED);
					SetScript(NULL);
				}
				else
				{
					vSetState(BTN_STATE_DEPRESSED);
				}
			}
			else
			{
//				vDebugOut ("OnMe:%d\n", OnMe(oEvent.x, oEvent.y));
				if (OnMe(oEvent.x, oEvent.y))
				{
					if (oEvent.claimed) 
						{
						vSetSelectedState();
						}
					else
						{
						vSetState(BTN_STATE_HILIGHTED);
						oEvent.claimed = TRUE;
						}
				}
				else
				{
					vSetSelectedState();
				}
			}
			break;
		case MOUSE_UP:
			if ((!OnMe(oEvent.x, oEvent.y) && !mbMouseDown) || oEvent.claimed)
				break;
			SetScript(NULL);
			if ((oEvent.modifiers & LEFT_MOUSE) && mbMouseDown)
			{
				if (OnMe(oEvent.x, oEvent.y))
				{
					vSetState(BTN_STATE_HILIGHTED);
					vDoLeftClick();
				}
				else
				{
					vSetSelectedState();
				}
				mbMouseDown = FALSE;
				oEvent.claimed = TRUE;
			}

			if ((oEvent.modifiers & RIGHT_MOUSE) && mbMouseDown)
			{
				if (OnMe(oEvent.x, oEvent.y))
				{
					vSetState(BTN_STATE_HILIGHTED);
					vDoRightClick();
				}
				else
				{
					vSetState(DEFAULT_BUTTON_NORMAL);
				}
				mbMouseDown = FALSE;
				oEvent.claimed = TRUE;
			}
			//remove from  hog list
			pApp->pWindow->vRemoveEventHog(this);
//			plane->CaptureEvents(NULL);
			break;
		case KEY_CHAR:
			if ((tolower(oEvent.c) == mcHotkey) || (tolower(oEvent.c) == mcHotkey2))
			{
				vSetState(BTN_STATE_DEPRESSED);
				plane->window->ForceUpdate();
				Sleep(100);
				vSetSelectedState();
				vDoLeftClick();
				oEvent.claimed = TRUE;
			}
			else if (tolower(oEvent.c) == mcRightHotkey)
			{
				vSetState(BTN_STATE_DEPRESSED);
				plane->window->ForceUpdate();
				Sleep(100);
				vSetSelectedState();
				vDoRightClick();
				oEvent.claimed = TRUE;
			}
			break;

		}

	//if mouse is down(and thus a hog) then we need to make us a 'modal' hog
	if (mbMouseDown == TRUE) {
		oEvent.claimed = TRUE;
	}

	return oEvent.claimed;
}

void UpgradeButton::Render(RECT &rUpdateRect) {
	sciProp::Render(rUpdateRect);
}

void UpgradeButton::vSetText(char *pstrText)
{
	if (mpoText)
		delete mpoText;
	
	mpoText = new sciText();
	if (mpoText) {
		mpoText->plane = plane;
		mpoText->SetBackgroundColor(mnNonPendingBG);
		mpoText->Init();
		mpoText->SetText(pstrText);
		mpoText->vSetSize();
		mpoText->Posn(GetX(), GetY());
		mpoText->SetWidth(GetWidth());
		mpoText->SetHeight(GetHeight());
		mpoText->SetJust(mnJustification);
	}
	
}

void UpgradeButton::vSetFont(int nUnsel, int nSel, int nHi, int nDep, int nDis) { 
	mnUnselectedFont = nUnsel;	
	mnSelectedFont = nSel;
	mnHilightFont = nHi;	
	mnDepressedFont = nDep;
	mnDisabledFont = nDis;	
}

void UpgradeButton::vSetFontBG(int nPending, int nNonPending) {
	mnPendingBG = nPending;
	mnNonPendingBG = nNonPending;
}

void UpgradeButton::vSetViews(int nUnselectView, int nSelectView, int nPendingView) {
	mpoSelectedView = new sciView();
	mpoSelectedView->vSetPlane(plane);
	mpoSelectedView->SetView(nSelectView);
	
	mpoPendingView = new sciView();
	mpoPendingView->vSetPlane(plane);
	mpoPendingView->SetView(nPendingView);
	
	sciProp::SetView(nUnselectView);
	if(mpoText) {
		mpoText->SetWidth(GetWidth());
		mpoText->SetHeight(GetHeight());
	}

}	

void UpgradeButton::vSetButtonCels(int nUnsel, int nSel, int nHi, int nDep, int nDis) {
	mnUnselectedCel = nUnsel;
	mnSelectedCel = nSel;	
	mnHilightCel = nHi;	
	mnDepressedCel = nDep;		
	mnDisabledCel = nDis;
}


void UpgradeButton::vUpdateCel() {
	if (mpoText) {
		if (mbPending)
			mpoText->SetBackgroundColor(mnPendingBG);
		else
			mpoText->SetBackgroundColor(mnNonPendingBG);
	}
	
	switch (mnState) 	{
	case BTN_STATE_UNSELECTED:
		SetCel(mnUnselectedCel);
		if(mpoText)
			mpoText->SetFont(mnUnselectedFont);
		break;
	case BTN_STATE_SELECTED:
		SetCel(mnSelectedCel);
		if(mpoText)
			mpoText->SetFont(mnSelectedFont);
		break;
	case BTN_STATE_HILIGHTED:
		if (UpgradeButton::mpoCurrentHilight != this) {
			if (UpgradeButton::mpoCurrentHilight != NULL) {
				UpgradeButton::mpoCurrentHilight->vSetSelectedState();
			}
			UpgradeButton::mpoCurrentHilight = this;
		}

		SetCel(mnHilightCel);
		if(mpoText)
			mpoText->SetFont(mnHilightFont);
		break;
	case BTN_STATE_DEPRESSED:
		SetCel(mnDepressedCel);
		if(mpoText)
			mpoText->SetFont(mnDepressedFont);
		break;
	case BTN_STATE_DISABLED:
		SetCel(mnDisabledCel);
		if(mpoText)
			mpoText->SetFont(mnDisabledFont);
		break;
	}
	pApp->pWindow->Update(rcWindow);
}


void UpgradeButton::Init(int nWidth, int nHeight) {
	sciProp::Init();

	if (mpoSelectedView) {
		mpoSelectedView->Init();
	}

	if (mpoPendingView) {
		mpoPendingView->Init();
	}

	if(mpoText) {
		if(!nWidth)
			nWidth = GetWidth();
		if(!nHeight)
			nHeight = GetHeight();
		SetWidth(nWidth);
		SetHeight(nHeight);
		mpoText->SetWidth(GetWidth());
		mpoText->SetHeight(GetHeight());
		mpoText->SetPri(GetPri() + 1);
		mpoText->Posn(GetX(), GetY());
		mpoText->Init();
	}

}

void UpgradeButton::vSetPlane(sciPlane *poPlane) {
	assert(poPlane);
	plane = poPlane;
	if (mpoSelectedView) {
		mpoSelectedView->vSetPlane(poPlane);
	}
	if (mpoPendingView) {
		mpoPendingView->vSetPlane(poPlane);
	}
}

void UpgradeButton::vSetState(int nState) {
	mnState = nState;
	vUpdateCel();
}

void UpgradeButton::vSetSelectedState() {
	if (mbSelected) {
		vSetState(BTN_STATE_SELECTED);
	} else {
		vSetState(BTN_STATE_UNSELECTED);
	}
}

void UpgradeButton::vDoLeftClick() {

}

void UpgradeButton::vDoRightClick() {

}


void UpgradeButton::vEnable() {
	if (!mbEnabled) {
		mbEnabled = TRUE;
		mbMouseDown = FALSE;
		if (OnMe(plane->GetCursorX(), plane->GetCursorY()))
			vSetState(BTN_STATE_HILIGHTED);
		else
			vSetSelectedState();
	}
}

void UpgradeButton::vDisable() {
	if (mbEnabled) {
		mbEnabled = FALSE;
		mbMouseDown = FALSE;
		mbPending = FALSE;
		SetScript(NULL);
		vSetState(BTN_STATE_DISABLED);
	}
}

void UpgradeButton::vSetRepeat(int nStartDelay, int nRepeatDelay) {
	mnStartDelay = nStartDelay;
	mnRepeatDelay = nRepeatDelay;
}

void UpgradeButton::vSetPri(int nPri) {
	sciProp::SetPri(nPri);
	
	if (mpoText) {
		mpoText->SetPri(nPri + 1);
	}
	
	if (mpoSelectedView) {
		mpoSelectedView->SetPri(nPri + 2);
	}

	if (mpoPendingView) {
		mpoPendingView->SetPri(nPri + 3);
	}
}


void UpgradeButton::Posn(int nX, int nY) {
	sciProp::Posn(nX,nY);
	
	if (mpoText) {
		mpoText->Posn(nX,nY);
	}
	
	if (mpoSelectedView) {
		mpoSelectedView->Posn(nX,nY);
	}

	if (mpoPendingView) {
		mpoPendingView->Posn(nX,nY);
	}
}

void UpgradeButton::vTextPosn(int nAddX, int nAddY) {
	if (mpoText) {
		mpoText->Posn(mpoText->GetX() + nAddX, mpoText->GetY() + nAddY);
	}
}


void UpgradeButton::vSetTextJust(int nJust) {
	mnJustification = nJust;
	if (mpoText) {
		mpoText->SetJust(nJust);
	}
}

void UpgradeButton::Show() {
	sciProp::Show();

	if (mpoText) {
		mpoText->Show();
	}
	
	if (mpoSelectedView) {
		mpoSelectedView->Show();
	}

	if (mpoPendingView) {
		mpoPendingView->Show();
	}

}

void UpgradeButton::Hide() {
	mbMouseDown = FALSE;
	SetScript(NULL);
	sciProp::Hide();
	if (mpoText) {
		mpoText->Hide();
	}
	
	if (mpoSelectedView) {
		mpoSelectedView->Hide();
	}

	if (mpoPendingView) {
		mpoPendingView->Hide();
	}
	
	mpoSelectionList = NULL;
	UpgradeButton::mpoCurrentHilight = NULL;
}


//sets self as pending and clears pending of any other button in button list
void UpgradeButton::vSetAsPending() {
	assert(mpoButtonGroupList);
	UpgradeButton *poBtn;
	for (int i = 0; i < mpoButtonGroupList->Size(); i++) {
		poBtn = ((UpgradeButton *)(mpoButtonGroupList->At(i)));
		if (poBtn == this) {
			vSetPending();
			vUpdateCel();
		} else {
			poBtn->vClearPending();
			poBtn->vUpdateCel();
		}
	}

}











UpgradeButtonGroup::UpgradeButtonGroup() {
	mpoStatusLine = NULL;
	mpoStatusText = NULL;
	mpoIconText = NULL;
	mpoIconView = NULL;
	mpoButtonList = NULL;
	SetRect(&rcWindow,0,0,0,0);
}

UpgradeButtonGroup::~UpgradeButtonGroup()
{
	if (mpoStatusText) {
		delete[] mpoStatusText;
		mpoStatusText = NULL;
	}

	if (mpoIconText) {
		delete mpoIconText;
		mpoIconText = NULL;
	}

	if (mpoIconView) {
		delete mpoIconView;
		mpoIconView = NULL;
	}

	if (mpoButtonList) {
		//mpoButtonList->Release();
		delete mpoButtonList;
		mpoButtonList = NULL;
	}

}

Boolean UpgradeButtonGroup::HandleEvent(sciEvent &oEvent) {
	if (hidden) {
		return FALSE;
	}

	switch (oEvent.type) {

	case MOUSE_MOVE:
		if (OnMe(oEvent.x, oEvent.y)) {
			//vDebugOut("On me\n");
			//vDebugOut("x: %d, y:%d\n", plane->GetCursorX(), plane->GetCursorY());
			assert(mpoStatusLine);
			if (mpoStatusText) {
				mpoStatusLine->vAddToBar((UpgradeButtonGroup*)this, mpoStatusText);			
			}
//			oEvent.claimed = TRUE;
		} else {
			assert(mpoStatusLine);
			if (mpoStatusLine->bOnBar((UpgradeButtonGroup*)this)) {
				mpoStatusLine->vClearFromBar((UpgradeButtonGroup*)this);
			}
		}
		break;
	}

	return oEvent.claimed;
}

// Make sure to call this AFTER the Show function
void UpgradeButtonGroup::vSetGroupRect() {
	sciView *poView;
	for (int i = 0; i < mpoButtonList->Size(); i++) {
		poView = ((sciView *)(mpoButtonList->At(i)));
		if (poView->IsNotHidden())
			UnionRect(&rcWindow, &rcWindow, poView->prGetClip());
	}

	if (mpoIconView) 
		UnionRect(&rcWindow, &rcWindow, mpoIconView->prGetClip());

	if (mpoIconText)
		UnionRect(&rcWindow, &rcWindow, mpoIconText->prGetClip());

	left = rcWindow.left;
	top = rcWindow.top;
	right = rcWindow.right;
	bottom = rcWindow.bottom;
}

void UpgradeButtonGroup::vSetStatusText(char *pstrNewText) {
	int nStrLen;

	if (mpoStatusText && pstrNewText && !strcmp(mpoStatusText, pstrNewText)) {
		return;
	}

	if (mpoStatusText) {
		delete[] mpoStatusText;
	}

	if (pstrNewText) {
		nStrLen = strlen(pstrNewText);
		mpoStatusText = new char[nStrLen + 1];
		strcpy(mpoStatusText, pstrNewText);
	}
}

void UpgradeButtonGroup::vSetIconText(char *pstrNewText)
{
	if (mpoIconText)
		delete mpoIconText;
	
	mpoIconText = new sciText();
	if (mpoIconText) {
		mpoIconText->plane = plane;
		mpoIconText->SetBackgroundColor(DEFAULT_TEXT_UPGRADE_BG);
		mpoIconText->Init();
		mpoIconText->SetText(pstrNewText);
		mpoIconText->vSetSize();
//		mpoIconText->Posn(GetX(), GetY());
//		mpoIconText->SetWidth(GetWidth());
//		mpoIconText->SetHeight(GetHeight());
//		mpoIconText->SetJust(justification);
	}
}

void UpgradeButtonGroup::vSetIconView(int nNewView) {
	if (mpoIconView) 
		delete mpoIconView;

	mpoIconView = new sciView();
	if (mpoIconView) {
		mpoIconView->plane = plane;
		mpoIconView->SetView(nNewView);
	}
}

void UpgradeButtonGroup::vAddToGroup(UpgradeButton *poButton) {
	if(!mpoButtonList)
		mpoButtonList = new RefList<UpgradeButton>();

	if(!mpoButtonList->Contains(poButton)) 	{
		mpoButtonList->AddToBack(poButton);
		poButton->vSetButtonGroupList(mpoButtonList);
		//poButton->vSetPlane(plane); //for now I think I'll set the button's plane when I setup each individual button
	}
}

void UpgradeButtonGroup::vUpdateButtonState(RefList<DrawObj> *poUnitList) {
	AttributeButton *poBtn;
	for (int i = 0; i < mpoButtonList->Size(); i++) {
		poBtn = ((AttributeButton *)(mpoButtonList->At(i)));
		poBtn->vClearPending();
		poBtn->vUnselect();
		for (int j = 0; j < poUnitList->GetSize(); ++j) {
			DrawObj *poUnit = poUnitList->GetItem(j);
			poBtn->vSetSelectionList(poUnitList);
			poBtn->vUpdateState((Vampire*)poUnit);
			poBtn->plane->window->Update(*(poBtn->prGetClip()));
		}
	}
}

void UpgradeButtonGroup::vUpdateButtonState(Vampire *poVampire) {
	AttributeButton *poBtn;
	for (int i = 0; i < mpoButtonList->Size(); i++) {
		poBtn = ((AttributeButton *)(mpoButtonList->At(i)));
		poBtn->vClearPending();
		poBtn->vUnselect();
		poBtn->vUpdateState(poVampire);
		poBtn->plane->window->Update(*(poBtn->prGetClip()));
	}
}

/*void UpgradeButtonGroup::Show(RefList<DrawObj> *poUnitList) {
	AttributeButton *poBtn;
	for (int i = 0; i < mpoButtonList->Size(); i++) {
		poBtn = ((AttributeButton *)(mpoButtonList->At(i)));
		poBtn->vClearPending();
		poBtn->vUnselect();
		for (int j = 0; j < poUnitList->GetSize(); ++j) {
			DrawObj *poUnit = poUnitList->GetItem(j);
			poBtn->vSetSelectionList(poUnitList);
			poBtn->vUpdateState((Unit*)poUnit);
		}
	}
	Show();
}
*/

void UpgradeButtonGroup::Show() {
	if (mpoIconText) {
		mpoIconText->Show();
	}

	if (mpoIconView) {
		mpoIconView->Show();
	}

	sciView::Show();

	UpgradeButton *poBtn;
	for (int i = 0; i < mpoButtonList->Size(); i++) {
		poBtn = ((UpgradeButton *)(mpoButtonList->At(i)));
		if (poBtn->nGetID() == 6 && (!game->poLocalPlayer()->bCanBeLevelSix()) )	//only Level can be a 6 so its ok to do this check here
			poBtn->Hide();
		else
			poBtn->Show();
	}
	vSetGroupRect();
}

void UpgradeButtonGroup::Hide() {
	if (mpoIconText) {
		mpoIconText->Hide();
	}

	if (mpoIconView) {
		mpoIconView->Hide();
	}

	sciView::Hide();

	UpgradeButton *poBtn;
	for (int i = 0; i < mpoButtonList->Size(); i++) {
		poBtn = ((UpgradeButton *)(mpoButtonList->At(i)));
		poBtn->Hide();
	}

}

void UpgradeButtonGroup::vEnableAll() {
	for(int i = 0; i < mpoButtonList->Size(); i++)
		((UpgradeButton *)mpoButtonList->At(i))->vEnable();
}

void UpgradeButtonGroup::vDisableAll() {
	for(int i = 0; i < mpoButtonList->Size(); i++)
		((UpgradeButton *)mpoButtonList->At(i))->vDisable();
}

void UpgradeButtonGroup::Posn(int nX, int nY)
{
	sciView::Posn(nX,nY);

	int nIconWidth = 0; //width of icon(or icon text)

	if (mpoIconView) {
		mpoIconView->Posn(nX,nY);
		nIconWidth = mpoIconView->GetWidth();
	}

	if (mpoIconText) {
		mpoIconText->Posn(nX,nY);
		nIconWidth += mpoIconText->GetWidth();
	}
	

	int nBtnWidth = 0;	//the width of a button in the group
	UpgradeButton *poBtn=(UpgradeButton *)mpoButtonList->At(0);
		if (poBtn)
			nBtnWidth = poBtn->GetWidth();
	
	for (int i = 0; i < mpoButtonList->Size(); i++)
		((UpgradeButton *)mpoButtonList->At(i))->Posn(nX + nIconWidth + ICON_FIRST_SPACING + (i * (nBtnWidth + BTN_SPACING)),nY);

}






AttributeButton::AttributeButton(BUTTON_ATTR eAttribute, int nButtonID, Boolean bSelected)
 : UpgradeButton(nButtonID, bSelected) {

	meAttribute = eAttribute;
}

AttributeButton::~AttributeButton() {

}

void AttributeButton::vUpdateState(Vampire *poUnit) {
	int nCurrentSetting = -1;
	int nDesiredSetting = -1;

	if (!poUnit->bIsVampire()) return;

	switch (meAttribute) {
	case BTN_ATTR_AGGRESSION:
		if (mnButtonID == poUnit->GetAggression()) 
			vSelect();
		break;
	case BTN_ATTR_TENACITY:
		if (mnButtonID == poUnit->GetTenacity()) 
			vSelect();
		break;
	case BTN_ATTR_SIGHT:
		nCurrentSetting = poUnit->GetUpgradeSight();
		nDesiredSetting = poUnit->GetDesiredSight();
		break;
	case BTN_ATTR_ATTACK:
		nCurrentSetting = poUnit->GetUpgradeAttack();
		nDesiredSetting = poUnit->GetDesiredAttack();
		break;
	case BTN_ATTR_DEFEND:
		nCurrentSetting = poUnit->GetUpgradeDefend();
		nDesiredSetting = poUnit->GetDesiredDefend();
		break;
	case BTN_ATTR_SPEED:
		nCurrentSetting = poUnit->GetUpgradeSpeed();
		nDesiredSetting = poUnit->GetDesiredSpeed();
		break;
	case BTN_ATTR_STEALTH:
		nCurrentSetting = poUnit->GetUpgradeStealth();
		nDesiredSetting = poUnit->GetDesiredStealth();
		break;
	case BTN_ATTR_DETECTION:
		nCurrentSetting = poUnit->GetUpgradeDetection();
		nDesiredSetting = poUnit->GetDesiredDetection();
		break;
	case BTN_ATTR_SPELLS:
		nCurrentSetting = poUnit->GetUpgradeSpells();
		nDesiredSetting = poUnit->GetDesiredSpells();
		break;
	case BTN_ATTR_LEVEL: 
		//vDebugOut("ID:%d, le: %d, dl:%d\n", mnButtonID, poUnit->GetUpgradeLevel(), poUnit->GetDesiredLevel());
		nCurrentSetting = poUnit->GetUpgradeLevel();
		//Example : limit of 2 level 2's and 1 level 3 and you have 1 level 2 and 3 level 1s
		//You try to upgrade all to level three. The level 2 will goto 3 and one level 1
		//will goto level 2. This if will get rid of the pending state for the remaining
		//level 1s
		if (bCanUpgrade()) 
			nDesiredSetting = poUnit->GetDesiredLevel();
		break;
	}

	if (nCurrentSetting == mnButtonID) {
		mbSelected = TRUE;
		mnState = BTN_STATE_SELECTED;
	} 

	if ( (nDesiredSetting == mnButtonID) && (nDesiredSetting != nCurrentSetting) ) {
		mbPending = TRUE;
	}
//	else {
//		mbPending = FALSE;
//	}
	vUpdateCel();

}

void AttributeButton::vDoLeftClick() {
//	vDebugOut ("Level:%d Attr:%d\n", buttonID, meAttribute);
	//vSetAsPending();
	//vSetSelectedState();
	//plane->vMarkUpdate();

	switch (meAttribute) {
	case BTN_ATTR_AGGRESSION:
		vSetAsPending();
		Vampire::SetAggression((SETTING_LMH)mnButtonID);
		break;
	case BTN_ATTR_TENACITY:
		vSetAsPending();
		Vampire::SetTenacity((SETTING_LMH)mnButtonID);
		break;
	case BTN_ATTR_SIGHT:
		if (bCanUpgrade()) {
			Vampire::GroupUpgradeSight(mnButtonID);
			vSetAsPending();
		}
		break;
	case BTN_ATTR_ATTACK:
		if (bCanUpgrade()) {
			Vampire::GroupUpgradeAttack(mnButtonID);
			vSetAsPending();
		}
		break;
	case BTN_ATTR_DEFEND:
		if (bCanUpgrade()) {
			Vampire::GroupUpgradeDefend(mnButtonID);
			vSetAsPending();
		}
		break;
	case BTN_ATTR_SPEED:
		if (bCanUpgrade()) {
			Vampire::GroupUpgradeSpeed(mnButtonID);
			vSetAsPending();
		}
		break;
	case BTN_ATTR_STEALTH:
		if (bCanUpgrade()) {
			Vampire::GroupUpgradeStealth(mnButtonID);
			vSetAsPending();
		}
		break;
	case BTN_ATTR_DETECTION:
		if (bCanUpgrade()) {
			Vampire::GroupUpgradeDetection(mnButtonID);
			vSetAsPending();
		}
		break;
	case BTN_ATTR_SPELLS:
		if (bCanUpgrade()) {
			Vampire::GroupUpgradeSpells(mnButtonID);
			vSetAsPending();
		}
		break;
	case BTN_ATTR_LEVEL: 
		if (bCanUpgrade()) {
			Vampire::GroupUpgradeLevel(mnButtonID);
			vSetAsPending();
		}
		break;
	}
	
}

Boolean AttributeButton::bCanUpgrade() {
	if (hidden)
		return FALSE;

	assert(mpoSelectionList);
	for (int i = 0; i < mpoSelectionList->GetSize(); ++i) {
		Vampire * poUnit = (Vampire *)mpoSelectionList->GetItem(i);
		assert(poUnit);
		assert(poUnit->bIsVampire());
		switch (meAttribute) {
		case BTN_ATTR_SIGHT:
			if (poUnit->bCanUpgradeSight(mnButtonID))
				return TRUE;
			break;
		case BTN_ATTR_ATTACK:
			if (poUnit->bCanUpgradeAttack(mnButtonID))
				return TRUE;
			break;
		case BTN_ATTR_DEFEND:
			if (poUnit->bCanUpgradeDefend(mnButtonID))
				return TRUE;
			break;
		case BTN_ATTR_SPEED:
			if (poUnit->bCanUpgradeSpeed(mnButtonID))
				return TRUE;
			break;
		case BTN_ATTR_STEALTH:
			if (poUnit->bCanUpgradeStealth(mnButtonID))
				return TRUE;
			break;
		case BTN_ATTR_DETECTION:
			if (poUnit->bCanUpgradeDetection(mnButtonID))
				return TRUE;
			break;
		case BTN_ATTR_SPELLS:
			if (poUnit->bCanUpgradeSpells(mnButtonID))
				return TRUE;
			break;
		case BTN_ATTR_LEVEL: 
//			if (poBtn->nGetID() == 6 && (!game->poLocalPlayer()->bCanBeLevelSix()) )
//				return FALSE;

			if (poUnit->bCanUpgradeLevel(mnButtonID))
				return TRUE;
			break;
		}
	
	}
	
	return FALSE;
}
	
void AttributeButton::Show() 
{
	Player *poPlayer = game->poLocalPlayer();

	switch (meAttribute) {
	case BTN_ATTR_LEVEL:
		if (mnButtonID > poPlayer->manUpgradeLevel[UPGRADE_Level]) {
			Hide();
			return;
		}
		break;
	case BTN_ATTR_ATTACK:
		if (mnButtonID > poPlayer->manUpgradeLevel[UPGRADE_Attack]) {
			Hide();
			return;
		}
		break;
	case BTN_ATTR_DEFEND:
		if (mnButtonID > poPlayer->manUpgradeLevel[UPGRADE_Defend]) {
			Hide();
			return;
		}
		break;
	case BTN_ATTR_DETECTION:
		if (mnButtonID > poPlayer->manUpgradeLevel[UPGRADE_Detection]) {
			Hide();
			return;
		}
		break;
	case BTN_ATTR_SIGHT:
		if (mnButtonID > poPlayer->manUpgradeLevel[UPGRADE_Sight]) {
			Hide();
			return;
		}
		break;
	case BTN_ATTR_SPEED:
		if (mnButtonID > poPlayer->manUpgradeLevel[UPGRADE_Speed]) {
			Hide();
			return;
		}
		break;
	case BTN_ATTR_SPELLS:
		if (mnButtonID > poPlayer->manUpgradeLevel[UPGRADE_Spells]) {
			Hide();
			return;
		}
		break;
	case BTN_ATTR_STEALTH:
		if (mnButtonID > poPlayer->manUpgradeLevel[UPGRADE_Stealth]) {
			Hide();
			return;
		}
		break;
	default:
		break;
	}

	UpgradeButton::Show();
}





CommandButton::CommandButton(BTN_MODE eMode) {
	meMode = eMode;
	mpoStatusLine = NULL;
	mpoStatusText = NULL;

}

CommandButton::~CommandButton() {
	if (mpoStatusText) {
		delete[] mpoStatusText;
		mpoStatusText = NULL;
	}

}

void CommandButton::vSetStatusText(char *pstrNewText) {
	int nStrLen;

	if (mpoStatusText && pstrNewText && !strcmp(mpoStatusText, pstrNewText)) {
		return;
	}

	if (mpoStatusText) {
		delete[] mpoStatusText;
	}

	if (pstrNewText) {
		nStrLen = strlen(pstrNewText);
		mpoStatusText = new char[nStrLen + 1];
		strcpy(mpoStatusText, pstrNewText);
	}
}

Boolean CommandButton::HandleEvent(sciEvent &oEvent) {
	if (hidden) {
		return FALSE;
	}

	switch (oEvent.type) {

	case MOUSE_MOVE:
		if (OnMe(oEvent.x, oEvent.y)) {
			//vDebugOut("On me\n");
			//vDebugOut("x: %d, y:%d\n", plane->GetCursorX(), plane->GetCursorY());
			assert(mpoStatusLine);
			if (mpoStatusText) {
				mpoStatusLine->vAddToBar((CommandButton*)this, mpoStatusText);			
			}
//			oEvent.claimed = TRUE;
		} else {
			assert(mpoStatusLine);
			if (mpoStatusLine->bOnBar((CommandButton*)this)) {
				mpoStatusLine->vClearFromBar((CommandButton*)this);
			}
		}
		break;
	}

	return sciButton::HandleEvent(oEvent);
}


void CommandButton::DoLeftClick() {
	switch (meMode) {
	case BTN_MOVE:
		gpoMap->SetMode(UIMODE_MOVE);
		break;
	case BTN_ATTACK:
		gpoMap->SetMode(UIMODE_ATTACK);
		break;
	case BTN_FEED:
		gpoMap->SetMode(UIMODE_FEED);
		break;
	case BTN_CONVERT:
		gpoMap->SetMode(UIMODE_CONVERT);
		break;
	case BTN_GUARD:
		gpoMap->SetMode(UIMODE_GUARD);
		break;
	case BTN_CASTSPELL:
		//gpoMap->SetMode(UIMODE_CAST);
		gpoInterface->vShowPicker(FALSE);
		break;
	case BTN_SCOUT:
		gpoMap->SetMode(UIMODE_SCOUT);
		break;
	case BTN_PATROL:
		if (gpoMap->GetMode() == UIMODE_PATROL) {
			gpoMap->SetMode(UIMODE_SELECT);
		} else {
			gpoMap->SetMode(UIMODE_PATROL);
		}
		break;
	case BTN_TAKE:
		gpoMap->SetMode(UIMODE_TAKE);
		break;
	case BTN_CLOAK:
		{
			Command * poCommand = new Command;
			poCommand->meType = CM_CLOAK;
			game->poLocalPlayer()->vPostCommand(poCommand, FALSE);
		}
		break;
	case BTN_CNTR_PANEL:
		break;
	case BTN_MAP_ZOOM:
		gpoInterface->vZoomMap();
		break;
	case BTN_DEF_ATTACK:
		gpoInterface->vShowPicker(TRUE);
		break;
	case BTN_UNDEFINED:
		break;
	}
}








RadioDisplayButton::RadioDisplayButton(int nID, Boolean bSelected)
 : sciRadioButton(nID, bSelected) {
	mpoStatusLine = NULL;
	mpoStatusText = NULL;
}

RadioDisplayButton::~RadioDisplayButton() {
	if (mpoStatusText) {
		delete[] mpoStatusText;
		mpoStatusText = NULL;
	}

}

void RadioDisplayButton::vSetStatusText(char *pstrNewText) {
	int nStrLen;

	if (mpoStatusText && pstrNewText && !strcmp(mpoStatusText, pstrNewText)) {
		return;
	}

	if (mpoStatusText) {
		delete[] mpoStatusText;
	}

	if (pstrNewText) {
		nStrLen = strlen(pstrNewText);
		mpoStatusText = new char[nStrLen + 1];
		strcpy(mpoStatusText, pstrNewText);
	}
}

Boolean RadioDisplayButton::HandleEvent(sciEvent &oEvent) {
	if (hidden) {
		return FALSE;
	}

	switch (oEvent.type) {

	case MOUSE_MOVE:
		if (OnMe(oEvent.x, oEvent.y)) {
			//vDebugOut("On me\n");
			//vDebugOut("x: %d, y:%d\n", plane->GetCursorX(), plane->GetCursorY());
			assert(mpoStatusLine);
			if (mpoStatusText) {
				mpoStatusLine->vAddToBar((RadioDisplayButton*)this, mpoStatusText);			
			}
//			oEvent.claimed = TRUE;
		} else {
			assert(mpoStatusLine);
			if (mpoStatusLine->bOnBar((RadioDisplayButton*)this)) {
				mpoStatusLine->vClearFromBar((RadioDisplayButton*)this);
			}
		}
		break;
	}

	return sciRadioButton::HandleEvent(oEvent);
}


void RadioDisplayButton::DoLeftClick() {
	switch (buttonID) {
	case DISP_BTN_UPGRADES:
		if (selected)
			gpoInterface->vSetDisplay(DISPLAY_UPGRADES);
		else 
			mpoButtonGroup->SelectButtonByID(DISPLAY_UPGRADES);		//we clicked on it again so go to default
		break;
	case DISP_BTN_SPELLS:
		if (selected)
			gpoInterface->vSetDisplay(DISPLAY_SPELLS);
		else
			mpoButtonGroup->SelectButtonByID(DISPLAY_SPELLS);		//we clicked on it again so go to default
		break;
	}
}

