// MapNode.H -- the main vampire wars database includes
//
// DJM

#ifndef _VWDB_
#define _VWDB_
#include "scimain.h"
#include "command.h"
//#include "drawobj.h"
//#include "mapplane.h"
//#include "player.h"
#include "gamemap.h"

class MapPlane;

//typedef WORD	BITMAPID;
typedef WORD	OBJID;
typedef WORD    UNITID;
//typedef unsigned int	BOOL;

#define ADD_TO_FRONT	0
#define ADD_TO_END		((WORD)-1)

extern Boolean gbToggleObstacleGrid;

typedef enum _TERRAINTYPE {
	TT_CLEAR,
	TT_OBSTACLE,
	TT_ENTRANCE,
	TT_ROUGH,
	TT_INTERIOR,
} TERRAINTYPE;

typedef enum  _MINI_MAP_TERRAIN_TYPE {
	MMTT_NONE,		//0
	MMTT_GRASS,		//1
	MMTT_SAND,		//2
	MMTT_SNOW,		//3
	MMTT_OBSTACLE,	//4
	MMTT_BUILDING,	//5
	MMTT_TREES,		//6
	MMTT_STONE,		//7
	MMTT_ROADS,		//8
	MMTT_WATER,		//9
	MMTT_BRIDGE,	//10
	MMTT_GATE,		//11
	MMTT_DIRT,		//12
} MINI_MAP_TERRAIN_TYPE	;

typedef struct _TERRAININFO {
	WORD		wXDelta;
	WORD		wYDelta;
	TERRAINTYPE	eType;
} TERRAININFO, *P_TERRAININFO;

typedef enum _DRAW_PASS {
	DP_GROUND,
	DP_FLOOR,
	DP_ITEMS,
	DP_OBSTACLES,
	DP_DEBUG,
	DP_NUM_LAYERS
} DRAW_PASS;

typedef enum _OBJ_TYPE {
	OT_GROUND,
	OT_FLOOR,
	OT_OBSTACLE,
	OT_BUILDING,
	OT_STONE,
	OT_ARTIFACT
} OBJ_TYPE;

// forward references
class MapNode;
//class GameMap;
class Unit;

class PlayScript {
};

#define MAX_PLAYER_NAME 50
#define MAX_CHAT_STRING	256
#define MAX_CLAN_NAME	50

typedef enum TAG_ENUM_SEX {
	SEX_MALE,
	SEX_FEMALE
} ENUM_SEX;

class vClan : public Object {
public:
	vClan(Template & oClanTemplate);
	Template * GetTemplate();
	FLOAT GetXMinDamage() { return mfMinDamage; }
	FLOAT GetXMaxDamageAttack()	 { return mfMaxDamageAttack; }
	FLOAT GetXMaxDamageDefend()	 { return mfMaxDamageDefend; }
	int GetDistanceIncludeFriendlies() { return mnDistanceIncludeFriendlies; }
	virtual int GetBaseSight();

	// THESE RETURN A PERCENT VALUE TO MODIFY HEALTH / POWER SETTINGS USING Vampire::GetXXXXForLevel
	double GetPowerCoefficient();
	double GetHealthCoefficient();
	float GetSpellCostCoefficient1();
	float GetSpellCostCoefficient2();
	int GetAdrenalineDuration();
	int GetAdrenalineResetTime();
	double GetAdrenalineSpeedBoost();

protected:
	Template * mpoTemplate;
	FLOAT		mfMinDamage;
	FLOAT		mfMaxDamageAttack;
	FLOAT		mfMaxDamageDefend;
	int			mnDistanceIncludeFriendlies;
};

class Clan : public vClan {
public:
	Clan(Template & oClanTemplate);
	int GetBaseSight();
	int GetRangeCallForHelp();
	int GetBaseAttackSpeed() { return mnBaseAttackSpeed; }
	int GetTimeToUpLevel() { return mnTimeUpgradeLevel; }
	int GetTimeToUpSkillFrom(int nCurrentLevel) { return mnTimeUpgradeSkillFrom[nCurrentLevel]; }
	FLOAT GetPercentPowerToUpLevel() { return mfPowerUpgradeLevel; }
	FLOAT GetPercentPowerToUpSkillFrom(int nCurrentLevel) { return mfPowerUpgradeSkillFrom[nCurrentLevel]; }
	FLOAT GetPowerSaveForEmergency() { return mfPowerSaveForEmergency; }
	int GetUnitsPerLevel(int nLevel, int nTopLevel);
	FLOAT GetPercentMinimumFeedingHealth();
	FLOAT GetPercentHealthConversionFloor();
	FLOAT GetPercentPowerToHealth() { return mfMaxPercentPowerToHealth; }
	FLOAT GetHealthPerPower() { return mfHealthPerPower; } 
	int GetConversionSpeed() { return mnConvertSpeed; }
	FLOAT GetPowerRegenPercent(); 
	int GetPowerCostToConvert();
	int GetNewConvertHealth();
	int GetNewConvertPower();
	int GetRangeToDecloak() { return mnRangeToDecloak; }
	int GetIdleTimeLimit() { return mnIdleTimeLimit; }
	int GetDistanceToLookForAnother() { return mnDistanceToLookForAnother; }
	int nInterfaceTemplateNumber() {return  mnInterfaceTemplateNumber;}
	int GetTemplateIdForSexAndLevel(ENUM_SEX eSex, int nLevel);
	Template * GetTemplate() { return mpoTemplate; }
	int GetAnxietyBlastForFeeding();
	int GetAnxietyBlastForConverting();
	int GetAnxietyBlastForAttacking();
	int GetAnxietyBlastForKilling();
	int GetAnxietyCausedByFinding();
	float GetAnxietyCausedByUncloaking();
	int GetAlertCycleLimit();

	WORD *pwGetClanPalette(int nColorIndex) {
		assert(nColorIndex >= 0 && nColorIndex < MAX_PLAYERS);

		return maoPalettes[nColorIndex].pwGetPalette();
	}



protected:
	int			mnAlertCycleLimit;
	int			mnDistanceToLookForAnother;
	int			mnIdleTimeLimit;
	int			mnRangeToDecloak;
	char		mstrName[MAX_CLAN_NAME];
	UINT		mnCountFeedingOnOnePeasant;
	FLOAT		mfPowerSaveForEmergency;
	FLOAT		mfPercentMinimumFeedingHealth;
	FLOAT		mfPercentHealthConversionFloor;
	FLOAT		mfMaxPercentPowerToHealth;
	FLOAT		mfHealthPerPower;
	FLOAT		mfPowerRegenPercent;

	int			mnCostToConvert;
	int			mnNewConvertHealth;
	int			mnNewConvertPower;
	int			mnConvertSpeed;
	int			mnBaseAttackSpeed;
	int			mnTimeUpgradeLevel;
	int			mnTimeUpgradeSkillFrom[2];

	int			mnInterfaceTemplateNumber;

	FLOAT		mfPowerUpgradeLevel;
	FLOAT		mfPowerUpgradeSkillFrom[2];

	int			mnCountUnitsPerLevel[6];

	RefList<Template> mpoUnitTemplates;
	Palette16	maoPalettes[MAX_PLAYERS];
};

class House : public Object {
};

// class that stores template information for an object
class ObjectInfo : public Resource {
};


// forward reference
class Unit;

#define FAST_DRAW

class MapNode : public Object {
	friend class DebugHandler;
public:
	// debugging stuff -- checksum and text dumping
	long lGetChecksum(void);
	void vDumpSyncText(FILE *fileDump);
	
	~MapNode();
	void vReset(void); 
	void vSaveMap(FILE *file, WORD wVersion);	
	void vLoadMap(FILE *file, WORD wVersion, MapPlane *poPlane);
	void vHighlight(void);	//for overlaying obstacle grids/debugging
	
	void vAddCover(DrawObj *poObj) {mpoCovering = poObj;}
	void vClearCover(void) {mpoCovering = NULL;}
	DrawObj *poGetCover(void) {return mpoCovering;}
	
	void vSetTerrainType(TERRAINTYPE eType) {meType = eType;}
	TERRAINTYPE eTerrainType() {return meType;}
	MINI_MAP_TERRAIN_TYPE	eMiniMapTerrainType () {return meMMType;}
	
	void vAddObj(DrawObj *poObj);
	void vAddObjToFront(DrawObj * poObj);
	void vRemoveObj(DrawObj *poObj);

	//void vClearVisibility(void);
	//void vAddVisibility(WORD wVisibility);
	BYTE jGetVisibility(void) {
		return gpoMap->jGetVisibility(mnXPos, mnYPos);
	}
	void vSetVisibility(BYTE jVisibility) {
		gpoMap->vSetVisibility(mnXPos, mnYPos, jVisibility);
	}

#ifdef FAST_DRAW
	void vDraw(DRAW_PASS ePass, int nX, int nY) {
		if (!mpoUnit && !moDrawn.GetSize() && !gbToggleObstacleGrid) {
			return;
		} else {
			vReallyDraw(ePass, nX, nY);
		}
	}
#else
	void vDraw(DRAW_PASS ePass, int nX, int nY) {
		vReallyDraw(ePass, nX, nY);
	}
#endif

	void vReallyDraw(DRAW_PASS ePass, int nX, int nY);

	int nPixelX(void) {return mnXPos * mpoMap->nGetSpacingX();}
	int nPixelY(void) {return mnYPos * mpoMap->nGetSpacingY();}
	int nDrawnSize(void) {return moDrawn.GetSize();}

	BOOL bBlocked(BOOL *pbTemp = NULL);

	MapNode(GameMap *poMap, int nX, int nY);
	void vUpdateObstacles(DrawObj *poObj); //paramter usagae described in CPP file
	void vResetChildren(DrawObj *poParent);	//a child was modified therefore modify all children of other parents of same ID
	
	GameMap		*mpoMap;	// my map
	int			mnXPos;		// my position
	int			mnYPos;		

	TERRAINTYPE				meType;
	MINI_MAP_TERRAIN_TYPE	meMMType;	//terrain type for the mini-map
	
	// These 2 functions return TRUE if the node contains the given parameter
	Boolean bContains(DrawObj *poObj) {return (moDrawn.Find(poObj) != ID_NONE);}
	Boolean bContains(Unit *poUnit) {return (mpoUnit ==poUnit);}

	// first to draw -- ground
	//List<DrawObj>	moGround;
	
	// next floors
	//List<DrawObj>	moFloor;

	// next to draw -- items & movable stuff
	//List<DrawObj>	moItems;

	// last to draw -- buildings, trees
	//List<DrawObj>	moObstacles;

	//WORD		mwVisibility;	// from 0 (none) to 5 (full)

	Unit		*mpoUnit;		// the unit occupying this space (if any)

	////
	//// USED FOR OBSTACLE CALCULATION
	////
	int		mnHDist; // heuristic distance from goal
	int		mnGDist; // generated distance from goal
	int		mnFDist; // above two added -- purely for convenience
	int		mbOpen;	 // is this one of the nodes we're working on?
	//int		mbClosed; // has this node had its shortest distance calculated?
	MapNode	*mpoParent; // the node along the path back to start
	MapNode *mpoTouchedChain; // used to make a chain of touched nodes for clearing

protected:
	// all non-unit draw objects
	List<DrawObj>	moDrawn;
 
	// not drawn, but listed in another square's terrain list
//	RefList<DrawObj>	moCovering;
	//tw - I'm going to try this with a single ptr instead of a list
	DrawObj *mpoCovering;

};

#endif