// MAPSCROLLER.H -- the map scroller plane 
//
// TW

#ifndef _invplane
#define _invplane

#include "sciplane.h"
#include "sciprop.h"

class StatusLine;
class Bitmap;
class Template;
class MiniMap;
class MapFrame;

// The inner box you drag around the map scroller area
class MapScrollerBox : public sciFeature
{
public:
	
	MapScrollerBox(BYTE jRed, BYTE jGreen, BYTE jBlue, int nThickness, int nZoom);
	~MapScrollerBox();
	void vCenter(int nX, int nY);		//centers self around points
	void Render (RECT &rUpdateRect);
	void Doit();
	Boolean HandleEvent(sciEvent & oEvent);
	void vCheckPoints(int &nNewX, int &nNewY);	
	void Posn(int posnX, int posnY);
	void vSetupBitmap();
	void vScrollGameMap();			//given position of box within plane, calculate appropiate map pos and scroll it
protected:
	void vGameMapScrolled();		//checks to see if gamemap has scrolled, if it has it repositions self within plane and scrolls minimap if neccessary

protected:
	int		nXScroll(void);		//returns x/y position of scroll box based on current and 
	int		nYScroll(void);		//previsous map locations and scrolls MapScroller

	Bitmap	*mpoBitmap;			// the box's bitmap
	WORD	mwColor;			// color of box
	int		mnZoom;				//1=single pixels, 2 = pixel doubling, etc
	int		mnThickness;		// thickess of box
	RECT	mrRect;				// the box's rectangle coordinates
	Boolean	mbMouseDown;		//TRUE when dragging box around
	int		mnDragDiffX;		//for dragging around
	int		mnDragDiffY;		//for dragging around
	int		mnNewX;				//for dragging around - the new X position of box when buttn released
	int		mnNewY;				//for dragging around - the new X position of box when buttn released

	int		mnPrevMapPosX;		//previous map location
	int		mnPrevMapPosY;

	int		mnPrevMousePosX;	//previous location of mouse within plane
	int		mnPrevMousePosY;

	int		mnMinX;				//Min and max values the box can move within its plane
	int		mnMinY;				//before it would start to be 'clipped'
	int		mnMaxX;				//	Note - these values are local to the plane
	int		mnMaxY;				//

	int		mnBitmapW;			//width of bitmap in pixels
	int		mnBitmapH;			//height of bitmap in pixels
	
	int		mnMouseScrollSpeed;	//scroll speed from global game template
	
	//store these to speed up various lookups
	int		mnMapW;				//number of horizontal map grids
	int		mnMapH;				//number of vertical map grids
	int		mnPlaneL;			//planes left coordinate (globally)
	int		mnPlaneT;			//planes top coordinate (globally)
	int		mnPlaneR;			//planes right coordinate (globally)
	int		mnPlaneB;			//planes bottom coordinate (globally)
	int		mnPlaneW;			//width of our plane in pixels 
	int		mnPlaneH;			//height of our plane in pixels 


};

class MapScroller : public sciPlane {
	
public:
	MapScroller(int nPri, Template *poInterfaceTemplate);
	~MapScroller();
	void Doit();
	void Render (RECT &rUpdateRect);
	Boolean HandleEvent(sciEvent & oEvent);
	void vSetStatusLine(StatusLine *poStatusLine) {mpoStatusLine = poStatusLine;}
	void vSetStatusText(char *pstrNewText);
	void vSetSizeAndPosition();
	void vReAdjust();
	void vScrollRight(int nHowMuch);
	void vScrollLeft(int nHowMuch);
	void vScrollUp(int nHowMuch);
	void vScrollDown(int nHowMuch);
	void vGameCycle();
	void vForceUpdate();
	int	nPosX() {return mnPosX;}
	int	nPosY() {return mnPosY;}
	void vFrameEvent(sciEvent oEvent);	//a handle event that the frame received
	void vZoom();						//zooms map in/out
	
protected:
	void			vInitScrollBox();
	void			vRightClick(sciEvent &oEvent);	//sends a right click event to the map plane
	MiniMap			*mpoMiniMap;
	MapScrollerBox	*mpoScrollBox;
	Template		*mpoInterfaceTemplate;
	StatusLine		*mpoStatusLine;
	char			*mpoStatusText;	//the text that will be display on the status line when mouse is over it
	int				mnPosX;			//position of the box in the scrollable plane(much like GameMap)
	int				mnPosY;
	int				mnZoom;				//1=single pixels, 2 = pixel doubling, etc
	MapFrame		*mpoMapFrame;

};

class MapFrame : public sciPlane {
public:
	MapFrame(int nPri, Template *poInterfaceTemplate, MapScroller *poParent);
	~MapFrame();
	void vErase(Template *poInterfaceTemplate);
	Boolean HandleEvent(sciEvent & oEvent);
//	void vSetStatusLine(StatusLine *poStatusLine) {mpoStatusLine = poStatusLine;}
//	void vSetStatusText(char *pstrNewText);
protected:
	MapScroller		*mpoParent;
	//	StatusLine		*mpoStatusLine;
//	char			*mpoStatusText;	//the text that will be display on the status line when mouse is over it
};


#endif
