#include "scimain.h"
#include "drawobj.h"
#include "mapplane.h"
#include "gamemap.h"
#include "objectdb.h"
#include "procs.h"
#include "colordef.h"
#include "unit.h"
#include "player.h"
#include "debug.h"
#include "mapedit.h"
#include "vampire.h"
#include "missile.h"
#include "soundlist.h"
#include "cycler.h"
#include "path.h"
#include "math.h"
#include "spell.h"

#define MAX_STOPPED_CYCLES 40
#define MAX_BLOCKED_CYCLES 40
#define FADERESOLUTION 20


void Missile::vSetPixelPos(int nX, int nY)
{
	mdX = nX;
	mdY = nY;

	vMarkUpdate();
}

void Missile::vMoveOneCycle()
{
	Missile * poMissile;
	poMissile = new Missile(MISSILE_TRACERANIMATED, gpoMap->mpoPlane, (Unit*)this, NULL, NULL, nGetID(), 10, 0, 0, NULL, mdX, mdY, 2, 0);

	assert(mnSpeed > 0);
	int nBaseX = mpoLastNode->nPixelX();
	int nBaseY = mpoLastNode->nPixelY();
	int nDeltaX = mpoCurNode->nPixelX() - nBaseX;
	int nDeltaY = mpoCurNode->nPixelY() - nBaseY;

	if (mnPartialMove < mnSpeed) 
	{
		++mnPartialMove;
	}
	else
	{
		mbBetweenPoints = FALSE;
	}

	int nNewX;
	int nNewY;

	nNewX = nBaseX + ((mnPartialMove * nDeltaX) / mnSpeed);
	nNewY = nBaseY + ((mnPartialMove * nDeltaY) / mnSpeed);

	//meDirection =  eDirection(nBaseX,nBaseY,nNewX,nNewY);
	
	mdX = nNewX;
	mdY = nNewY;

	mbMoving = TRUE;

	if (mpoUnit && mbFlashEmbeddedUnitVisibility)
	{
		mpoUnit->mpoCurNode = mpoCurNode;
		mpoUnit->vScan();
	}

}


MapNode * Missile::poPopOneNode()
{
	MapNode *poNode;
	poNode = moCurrentRoute.PopLast();
	return poNode;
}

MapNode * Missile::poGetNextMoveNode()
{
	MapNode *poTempNode = poPopOneNode();

	// if there's no current node to move to, try to make a route
	if (!poTempNode) 
	{
		mnState = 0;
		return NULL;
	}

/*
	if (!poTempNode) 
	{
		// vCalcRoute();
		// poTempNode = poPopOneNode();

		// if the route fails, return, and possibly give up
		//if (!poTempNode) 
		{
			//if (++mnStoppedCycles > MAX_STOPPED_CYCLES) 
			{
				// we're giving up
				// TBD -- IF MISSILE IS BLOCKED MAYBE BLOW UP WITH AREA DAMAGE?>
				mnState = 0;
			}

			// no moving in any case
			return NULL;
		}
	}

	BOOL bTemp;

	// if we're blocked, how badly?

	if (0) {  //poTempNode->bBlocked(&bTemp)

		// is it just temporary? Then wait a bit
		if (bTemp) {
			if (++mnTempBlockedCycles > MAX_BLOCKED_CYCLES) {
				// it's been too long; recalc the route
				vCalcRoute();
				poTempNode = poPopOneNode();
			} 
			else 
			{
				// no, just wait some more

				// push back the node
				moCurrentRoute.AddToEnd(poTempNode);
				return NULL;
			}
		} else {

			// we're permanently blocked -- find a new route
			vCalcRoute();
			poTempNode = poPopOneNode();
		}

		// if after all this we don't have a node, then give up and bail
		if (!poTempNode) 
		{
			// TBD -- WHAT TO DO
			mnState = 0;
			return NULL;
		}
	}
*/
	return poTempNode;
}



void Missile::vMoveToNextNode(void)
{
	MapNode *poTempNode = poGetNextMoveNode();

	// NO NODE SO DONE
	if (!poTempNode) 
	{
		mnState = 0;
		return;
	}
	vMoveToNode(poTempNode);
	if (mbSoundEmitter)
		pApp->pSoundManager->SetSoundEmitterPos(this, (int)mdX, (int)mdY);
}

void Missile::vMoveToNode(MapNode *poNode)
{
	mnStoppedCycles = 0;
	mnTempBlockedCycles = 0;

	mpoLastNode = mpoCurNode;
	DrawObj::vMoveToNode(poNode);

	mbBetweenPoints = TRUE;
	mnPartialMove = 0;
	vMoveOneCycle();
}

MapNode * Missile::poGetTargetNode(void)
{
	if (mpoTargetUnit) 
	{
		return mpoTargetUnit->poNode();
	} 
	else 
	{
		return gpoMap->poGetNode((int)mdTargetX / 17, (int)mdTargetY / 10);
	}
}

void Missile::vCalcRoute()
{
	GameMap *poMap = gpoMap;
	MapNode *poGoal = NULL;
	MapNode *poDest = NULL;

	moCurrentRoute.RemoveAll();
	//
	// get the goal either from the unit we're targeting or the
	// locationi
	//
	{
		int nTargetX, nTargetY;
		nTargetX = poGetTargetNode()->mnXPos;
		nTargetY = poGetTargetNode()->mnYPos;
		mnLastTargetX = nTargetX;
		mnLastTargetY = nTargetY;
	}

	poGoal = poGetTargetNode();
	if (poGoal) 
	{
		bFindPath(poMap, mpoCurNode, poGoal, 
			  &moCurrentRoute, &poDest);
	}
}

void Missile::SetCycleCountdown(BOOL bCycleCountdown)
{
	mbCycleCountdown = (bCycleCountdown != FALSE);
}

void Missile::SetCyclesRemaining(UINT nCyclesRemaining)
{
	mnCyclesRemaining = nCyclesRemaining;
	mbCycleCountdown = TRUE;
}

void Missile::SetUnitWasSelected(BOOL bValue)
{
	mbUnitWasSelected = (bValue != FALSE);
}

void Missile::SetObstructView(BOOL bValue)
{
	mbObstructView = (bValue != FALSE);
}

Missile::Missile(MISSILE_TYPE eType, MapPlane *poPlane, Unit * poFromUnit,  
				 MapNode * poSourceNode, MapNode * poTargetNode, int nTemplateID, 
				 int nWait, BOOL bSolid, float fPercent, Unit * poTargetUnit, 
				 double X, double Y, int nFallOff, int nDirection):
	DrawObj(poPlane, nTemplateID),mdX(0.0),mdY(0.0), mfPower(0)
{
	// PERFORM ALL GENERAL INITIALIZATION
	mbDrawInFogOfWar = mpoTemplate->lGetValue(TAG_dfow, 0);
	mpoFoggedUnit = NULL;
	meDamageType = DAMAGE_AREASQUARE;
	mbObstructView = FALSE;
	mpoSpell = NULL;
	mnFalloffCycle = 0;
	mbFlashEmbeddedUnitVisibility = FALSE;
	mbUnitWasSelected = TRUE;
	mbCycleCountdown = FALSE;
	mpoUnit = NULL;
	mnCyclesRemaining = 0;
	meType = eType;
	meImpact = IMPACT_NONE;
	meClass = OC_MISSILE;
	meDamageType = DAMAGE_AREASQUARE;
	mnDirection = nDirection;
	mnFallOff = nFallOff;
	mbCheck = 0;
	mbSolid = bSolid;
	mbSoundEmitter = FALSE;
	mbCastShadow = bSolid;
	mbImpactObject = bSolid;
	mnState = 0;

	if (mbSolid)
	{
		if (eType == MISSILE_DEATHFROMABOVE)
			meSmokeTrail = SMOKETRAIL_FADECYCLE;
		else 
			meSmokeTrail = SMOKETRAIL_FIREBALL;
		if (eType == MISSILE_SMOKE)
		{
			mbCastShadow = FALSE;
		}
	}
	else 
	{
		meSmokeTrail = SMOKETRAIL_RESIDUAL;
	}
	mpoSourceUnit = poFromUnit;
	mpoTargetUnit = poTargetUnit;

	// FOR PATH MISSILES
	mnPartialMove = 0;
	mbMoving = FALSE;
	mbBetweenPoints = FALSE;
	mnSpeed = 3;
	mpoLastNode = NULL;

	// SHADOW 
	// TBD -- PB THIS SHOULD BE TRACKED AS A LINKED SEPARATE OBJECT FOR PROPER ZORDER
	mpoShadowBitmap = (Bitmap*)pApp->pResourceManager->LoadResource(RES_BITMAP, mpoTemplate->lGetValue(TAG_Bpor, 398));

	// SETUP SOURCE AND TARGET NODES
	if (poTargetUnit)
	{
		poTargetNode = poTargetUnit->poNode();
	}
	if ((poFromUnit) && (meType != MISSILE_TRACER) && (meType != MISSILE_TRACERANIMATED))
	{
		poSourceNode = poFromUnit->poNode();
	}
	
	// PERFORM TYPE SPECIFIC INITIALIZATION
	switch(meType)
	{
		case MISSILE_GROUNDHOMING:
			mnState = 2;
			mdX = poSourceNode->mnXPos * 17.0;
			mdY = poSourceNode->mnYPos * 10.0;
			mdZ = 0.0;

			mnCel = nWait;	
			assert(mnCel >= 0);
			mnWait = 0;		
			mpoCurNode = poSourceNode;
			mdTargetX = poTargetNode->mnXPos * 17.0;
			mdTargetY = poTargetNode->mnYPos * 10.0;
			break;
		case MISSILE_INSTANT:
		case MISSILE_FOG:
			mjAlpha = 255;
			mnCel = 0;
			mpoCurNode = poTargetNode;
			mdX = mpoCurNode->mnXPos * 17.0;
			mdY = mpoCurNode->mnYPos * 10.0;
			mdZ = 0.0;
			mnWait = nWait;
			break;
		case MISSILE_DRIFTINGSMOKE:
			mdX = X;
			mdY = Y;
			
			// THESE ARE HACKS REUSING PARAMETERS FOR VARIOUS PURPOSES
			mdZ = fPercent;
			fPercent = 0;
	
			mnCelCount = mpoBitmap->GetNumCels();
			mnCel = 0;
			//assert(mnCel >= 0);
			mnWait = 0;		

			// UNUSED
			mdTargetX = mdX; 
			mdTargetY = mdY;
			
			/*mddX = 0;
			mddY = 0;
			mddZ = 0;
			mjAlpha = (BYTE)(0.65 * 255);*/

			mddX = -1.7;
			mddY = -1.0;
			mddZ = -0.5;

			break;
		case MISSILE_LINGERINGSMOKE:
			mdX = X;
			mdY = Y;

			mnShadeCount = nWait;
			if (mnShadeCount>10)
			{
				 mnShadeCount -= GlobalRand(8);
			}
			mnShade = mnShadeCount;
			
			mnCelCount = mpoBitmap->GetNumCels();
			mnCel = mnCelCount-1;
			// THESE ARE HACKS REUSING PARAMETERS FOR VARIOUS PURPOSES
			mdZ = fPercent;
			fPercent = 0;
			
			//mnCel = nWait;	
			//assert(mnCel >= 0);
			mnWait = 0;		

			// UNUSED
			mdTargetX = mdX; 
			mdTargetY = mdY;
			
			mddX = 0;
			mddY = 0;
			mddZ = 0;
			mjAlpha = (BYTE)(0.65 * 255);

			/*mddX = -1.7;
			mddY = -1.0;
			mddZ = -0.5;*/

			break;
		case MISSILE_ANIMATEANDFADEDRAWOBJ:
			mnFallOff = 4;
		case MISSILE_TRACERANIMATED:
			mdX = X;
			mdY = Y;
			mdZ = fPercent;
			fPercent = 0;
			mpoSourceUnit = NULL;
			//poSourceNode = poFromUnit->poNode();
			poSourceNode = GetNodeForCurrentPosition();
			assert(poFromUnit);

			if (eType != MISSILE_ANIMATEANDFADEDRAWOBJ)
			{
				// BITMAP
				mpoBitmap = (Bitmap*)pApp->pResourceManager->LoadResource(RES_BITMAP, poFromUnit->nBitmapID());
				// CELL
				mnCel = poFromUnit->nGetCel();
				assert(mnCel >= 0);
			}
			mnShadeCount = nWait;
			mnShade = mnShadeCount;
			if (!mpoCycler)
			{
				mpoCycler = new ReverseCycler(this);
			}
			break;

		case MISSILE_TRACER:
			mdX = X;
			mdY = Y;
			mdZ = fPercent;
			fPercent = 0;
			mpoSourceUnit = NULL;
			//poSourceNode = poFromUnit->poNode();
			poSourceNode = GetNodeForCurrentPosition();
			assert(poFromUnit);

			// BITMAP
			mpoBitmap = (Bitmap*)pApp->pResourceManager->LoadResource(RES_BITMAP, poFromUnit->nBitmapID());

			mpwSpecialPalette = poFromUnit->mpwSpecialPalette;

			// CELL
			mnCel = poFromUnit->nGetCel();
			mnLoop = poFromUnit->nGetLoop();
			mbFlip = poFromUnit->bFlip();
			assert(mnCel >= 0);
			mnShadeCount = nWait*2;
			mnShade = mnShadeCount*2/3;
			break;
		case MISSILE_DROPBOMB:
			mnShade = 0;
			if (nTemplateID == 6020)
			{
				mnShadeCount = 0; 
				mjAlpha = 0;
			}
			else
			{
				mnShadeCount = 4;  // FINISHED
				mjAlpha = 255;
			}
		default:
			mdX = poSourceNode->mnXPos * 17.0;
			mdY = poSourceNode->mnYPos * 10.0;
			assert(poTargetNode);
			mdTargetX = poTargetNode->mnXPos * 17.0;
			mdTargetY = poTargetNode->mnYPos * 10.0;
			mnCelCount = 1;
			mnCel = 0;
			mdZ = 10;
			mnWait = nWait;
			if (fPercent == 0.0)
			{
				mddX = (mdTargetX - mdX) / (20.0 * 2.0);
				mddY = (mdTargetY - mdY) / (20.0 * 2.0);
				mddZ = 10;
				mdddZ = - 0.5;
			}
			else
			{
				mddX = (mdTargetX - mdX) / (20.0 * 1.0);
				mddY = (mdTargetY - mdY) / (20.0 * 1.0);
				mddZ = 10 * fPercent;
				mdddZ = -0.4;
			}
			if (mddY > 0)
			{
				mnDirection = 0;
			}
			else 
			{
				mnDirection = 1;
			}
			break;
	}
	

	// TBD - SOME HACKS UNTIL WE HAVE PROPER CYCLING IN PLACE
	if (nTemplateID == 6000)
	{
		mnCelCount = mpoBitmap->GetNumCels();

		for (int i=0;i<mnCelCount; i++)
		{
			mpoBitmap->vSetOriginPoint(14, 17, 0, i);
		}

		mnCel = GlobalRand(mnCelCount);
		//if (meType == MISSILE_DEATHFROMABOVE)		
		//	pApp->pSoundManager->EmitFromObject(this, SOUND_FireballLaunch);
	}
	else if ((nTemplateID >= 6002)&&(nTemplateID <= 6004))
	{
		mnCelCount = mpoBitmap->GetNumCels();	
		//mnCel = max(0, mnCelCount - mnCel);
	//	assert(mnCel >= 0);
		for (int i=0;i<mnCelCount; i++)
		{
			mpoBitmap->vSetOriginPoint(20, 20, 0, i);
		}
		mjAlpha = (mnCelCount - mnCel) * 96 / mnCelCount;
	}
	mpoBitmap->SetCel(mnCel);

	// ADD OBJECT TO MAP
	mpoCurNode = poSourceNode;
	if (mnDirection == 1)
	{
		mpoCurNode->vAddObj(this);
	}
	else
	{
		mpoCurNode->vAddObjToFront(this);
	}
	UpdateNode();
	vSetLayer(DP_OBSTACLES);
	
	vMarkUpdate();

	//mpoCycler = new ForwardCycler(this);
	//pApp->pWindow->vAddEventHog(this);
	if (nTemplateID == 6007) mdZ = 23.0;
	PlaySoundCreate();
}

void Missile::PlaySoundCreate()
{
	int nSound = mpoTemplate->lGetValue(TAG_sn01, 0);
	if (nSound)
	{
		mbSoundEmitter = TRUE;
		pApp->pSoundManager->EmitFromObject(this, nSound);
	}
}

void Missile::PlaySoundDestroy()
{
	int nSound = mpoTemplate->lGetValue(TAG_sn02, 0);
	if (nSound)
	{
		pApp->pSoundManager->EmitFromPosition((int)mdX, (int)mdY, nSound);
	}
}

void Missile::Destroy()
{
	PlaySoundDestroy();
	delete this;
}

Missile::~Missile()
{
	mpoCurNode->vRemoveObj(this);
	mpoCurNode = NULL;
	mpoPlane->vRemoveFromDrawnList(this);
}

// returns calculated node position, not actual node position
MapNode * Missile::GetNodeForCurrentPosition()
{
	int nX,nY;

	GetPixelPosition(&nX, &nY);
	nX = max(0, nX);
	nY = max(0, nY);
	return gpoMap->poGetNode(nX/17, nY/10);
}

// updates current position of layer for draw order
void Missile::UpdateNode()
{
	MapNode * poNewNode = GetNodeForCurrentPosition();
	
	if (poNewNode != mpoCurNode)
	{
		mpoCurNode->vRemoveObj(this);
		mpoCurNode = poNewNode;
		if (mnDirection == 1)
		{
			mpoCurNode->vAddObj(this);
		}
		else
		{
			mpoCurNode->vAddObjToFront(this);
		}
		if (mbSoundEmitter)
			pApp->pSoundManager->SetSoundEmitterPos(this, (int)mdX, (int)mdY);
	}
	
}

void Missile::vDraw()
{
	if (mbVisible) 
	{
		int nX, nY, nZ;
		GetPixelPosition(&nX, &nY);
		nX = nX % 17;
		nY = nY % 10;
		nZ = (int)mdZ;

		int nXofs = (int)(mdZ * 0.4);
		int nYofs = (int)(-mdZ * 0.1);

		// ONLY SOLID MISSILES CAST SHADOWS
		if (mbCastShadow)
		{
			mpoShadowBitmap->vDraw(nXofs + nX + nPixelX() - mpoPlane->nMapOrgPixelX(), nYofs + nY + nPixelY() - mpoPlane->nMapOrgPixelY(), mpoPlane);
		}

		// TBD -- THIS IS CYCLING FOR NOW
		if (meType != MISSILE_TRACER)
		{
			assert(mnCel >= 0);
			vSetCel(mnCel);
		}
		if (meType == MISSILE_GROUNDHOMING)
		{
			GetPixelPosition(&nX, &nY);
			DrawObj::vDraw(nX - mpoPlane->nMapOrgPixelX(), nY - mpoPlane->nMapOrgPixelY() - nZ);
		}
		else
			DrawObj::vDraw(nX + nPixelX() - mpoPlane->nMapOrgPixelX(), nY + nPixelY() - mpoPlane->nMapOrgPixelY() - nZ);
	}
}

// returns actual position in x,y map coordinates
void Missile::GetPixelPosition(int * pnX, int * pnY)
{
	*pnX = (int)mdX;
	*pnY = (int)mdY;
}

void Missile::GetPlayPosition(unsigned int * unX, unsigned int *unY)
{
	*unX = (UINT)max(0, mdX);
	*unY = (UINT)max(0, mdY);
}

extern Missile * CreateMissile(Unit * poFromUnit, MapNode * poSourceNode, MapNode * poTargetNode, MISSILE_TYPE missile, int nWait, BOOL bSolid, float fPercent = 0.0, double X = 10.0, double Y = 10.0, int nFallOff = 2, int nDirection = 1);

void Missile::SetDamage(TYPE_DAMAGE eDamage)
{
	meDamageType = eDamage;
}

void Missile::SetDirection(int nDegrees)
{
	mnDirectionDegrees = nDegrees;
}

void Missile::SetRotation(int nDegrees)
{
	mnRotationDegrees = nDegrees;
}

void Missile::SetVelocity(double dVelocity)
{
	mdVelocity = dVelocity;
}

void Missile::SetFalloffCycle(int nFalloffCycle)
{
	mnFalloffCycle = nFalloffCycle;
}

void Missile::vGameCycle()
{
	switch (meType)
	{
	case MISSILE_FOG:
	{
		Missile * poMissile;
		MapNode *poNode;
		
		int j=1;
#define MAKEFOG {if(poNode->eTerrainType() == TT_CLEAR) { poMissile = CreateMissile((Unit*)this, poNode, poNode, MISSILE_FOG, 10, FALSE, -18.0f, poNode->nPixelX(), poNode->nPixelY(), 3, 1); poMissile->SetFalloffCycle(max(0, mnSpeed - 10) + GlobalRand(20)); if (GlobalRand(2)==0) poMissile->vSetVisibility(TRUE); else poMissile->vSetVisibility(FALSE);} }
		static int nCount = 0;
		nCount++;
		mnState++;
#define SPEEDOFEXPANSION 3
		if (mnState == SPEEDOFEXPANSION * mnRotationDegrees)
		{
			Destroy();
			return;
		}
		else if (mnState == 1)
		{
			poNode = mpoCurNode;
			MAKEFOG;
		}
		if (mnState % SPEEDOFEXPANSION) return;
		if (mbVisible) return;
		int nState = mnState / SPEEDOFEXPANSION;
		int mnVisibleRange = mnRotationDegrees;
		// CREATE MORE FOG OUTWARD
		int nSquareRange = mnVisibleRange * mnVisibleRange;// + mnVisibleRange / 2;
		int i;
		int nRange;
		//int nVisibility;

		//
		// we do a 1/8th triangle of the full range area
		// then use rotations and reflections to check the other nodes
		//for (i = -mnState; i <= 0; ++i) 
		i = -nState;
		{
			//j=i;
			for (j = i; j <= 0; ++j) 
			{
				int nSquare = i * i + j * j;
				if (nSquare <= nSquareRange) {
					double fRange = sqrt((double)nSquare);
					nRange = (int)fRange;
/*
		#if VIS_FUZZY_BORDER
					nVisibility = (int)(((mnVisibleRange + 0.5) - fRange) * (255.0 / VIS_FUZZY_BORDER) + 0.5);
					nVisibility = min(255, max(0, nVisibility));
		#else
					nVisibility = 255;
		#endif
		*/
					// this sets non-local players' visibility to zero without
					// using an extra if -- test is above, outside the loop
		//			nVisibility &= nLocalMask;

					// do all 8 reflected combinations
					int nCenterX = mpoCurNode->mnXPos;
					int nCenterY = mpoCurNode->mnYPos;
					poNode = gpoMap->poGetNodeRC(nCenterX + i, nCenterY + j);
					if (poNode) 
					{
						MAKEFOG;
					}
					if (i != j) 
					{
						poNode = gpoMap->poGetNodeRC(nCenterX + j, nCenterY + i);
						if (poNode) 
						{
							MAKEFOG;
						}
					}

					if (j) {
						// don't bother doing -j if j == 0
						poNode = gpoMap->poGetNodeRC(nCenterX + i, nCenterY - j);
						if (poNode) 
						{
							MAKEFOG;
						}
						if (i != -j) 
						{
							poNode = gpoMap->poGetNodeRC(nCenterX - j, nCenterY + i);
							if (poNode) 
							{
								MAKEFOG;
							}
						}
					}
					if (i) {
						// don't bother doing -i if i == 0
						poNode = gpoMap->poGetNodeRC(nCenterX - i, nCenterY + j);
						if (poNode) 
						{
							MAKEFOG;
						}
						if (i != -j) 
						{
							poNode = gpoMap->poGetNodeRC(nCenterX + j, nCenterY - i);
							if (poNode) 
							{
								MAKEFOG;
							}
						}
						if (j) {
							// don't bother doing -j if j == 0
							poNode = gpoMap->poGetNodeRC(nCenterX - i, nCenterY - j);
							if (poNode) 
							{
								MAKEFOG;
							}
							if (i != j) 
							{
								poNode = gpoMap->poGetNodeRC(nCenterX - j, nCenterY - i);
								if (poNode) 
								{
									MAKEFOG;
								}
							}
						}
					}
				}
			}
		}
		}
		return;
	case MISSILE_INSTANT:
		//if (mpoCycler) mpoCycler->vAnimationCycle();
		mnState++;
		if (mnState > 2 && mnState < 12)
		{
			mjAlpha = 64 + GlobalRand(192);
		}
		switch(mnState)
		{
		case 1:
			Damage();
			if (meDamageType != DAMAGE_NONE)
			{
				Missile * poMissile;
				poMissile = new Missile(MISSILE_TIMER, gpoMap->mpoPlane, mpoSourceUnit, mpoCurNode, mpoCurNode, 6018, 0, FALSE, 1.18f, NULL, 10.0, 10.0, 2, 0);
				poMissile->vSetPixelPos((int)mdX, (int)mdY+8);
				poMissile->vSetLayer(DP_ITEMS);
				poMissile->UpdateNode();
				poMissile->SetCyclesRemaining(200);
			}
			break;
		case 12:
			Destroy();
		}
		return;
	case MISSILE_DROPBOMB:
		{
			int nDirection;
			nDirection = 0;
	
/*
			if (mnShadeCount == 0)
			{
				mnShade++;
				if (mnShade == 5)
				{
					mnShadeCount = 1;
					mnShade = 0;
				}
			}
			else if (mnShadeCount == 1)
			{
				mnShade += 255 / 15;
				if (mnShade > 255)
				{
					mnShade = 255;
					mnShadeCount = 2;
				}
				mjAlpha = mnShade;
			}
			else if (mnShadeCount == 2)
			{
				mnShade++;
				if (mnShade == 265)
				{
					mnShade = 255;
					mnShadeCount = 3;
				}
			}
			else if (mnShadeCount == 3)
			{
				mnShade -= 255 / 30;
				if (mnShade <= 0)
				{
					mnShade = 0;
					mnShadeCount = 4;
				}
				mjAlpha = mnShade;
			}
*/			
			mjAlpha = 255;
			mnWait++;
			if (mnWait == 5)
			{
				Missile * poMissile = CreateMissile(NULL, mpoCurNode, mpoCurNode, MISSILE_LINGERINGSMOKE, 10, FALSE, (float)mdZ, mdX, mdY, 3, nDirection);
				assert(mpoSpell);
				poMissile->SetSpell(mpoSpell);
				mnWait = 0;
			}

			if (mddX < 0)
			{
				if (mdX > mdTargetX)
				{
					mdX = mdX + mddX;
				}
				else 
				{
					mdX = mdTargetX;
				}
			}
			else if (mddX > 0)
			{
				if (mdX < mdTargetX)
				{
					mdX = mdX + mddX;
				}
				else 
				{
					mdX = mdTargetX;
				}
			}
			if (mddY < 0)
			{
				if (mdY > mdTargetY)
				{
					mdY = mdY + mddY;
				}
				else 
				{
					mdY = mdTargetY;
				}
			}
			else if (mddY > 0)
			{
				if (mdY < mdTargetY)
				{
					mdY = mdY + mddY;
				}
				else 
				{
					mdY = mdTargetY;
				}
			}
			mddZ = mddZ + mdddZ;
			mdZ = mdZ + mddZ;
			
			if ( mnCelCount > 1 )
			{
				mbCheck++;
				if (mbCheck == 3)
				{
					mbCheck = 0;
					mnCel--;
					if (mnCel < 0)
					{
						mnCel = mnCelCount - 1;
					}
					assert(mnCel >= 0);
					vSetCel(mnCel);
				}
			}
			UpdateNode();

			// ARRIVED AT TARGET
			if (mdZ <= 0)
			{
				if (mbImpactObject)
				{
					Impact();
				}
				Destroy();
				return;
			}
		}
		break;
	case MISSILE_GROUNDSPIRAL:
		assert(mbCycleCountdown);
		if (mnCyclesRemaining > 0)
		{
			mnCyclesRemaining--;
		}
		else
		{
			Impact();
			Destroy();
			return;
		}
		// CONTINUE SCALING OUTWARD
		if (mbCheck == 0)
		{
			Missile * poMissile = new Missile(MISSILE_LINGERINGSMOKE, gpoMap->mpoPlane, NULL, mpoCurNode, mpoCurNode, 6012, 5, FALSE, (float)mdZ, NULL, mdX, mdY, 3, 1);
			//Missile * poMissile = CreateMissile(NULL, mpoCurNode, mpoCurNode, MISSILE_LINGERINGSMOKE, 12, FALSE, (float)mdZ, mdX, mdY, 3, 1);
			assert(mpoSpell);
			poMissile->SetSpell(mpoSpell);
			mbCheck = 1;
		}
		else
		{
			mbCheck++;
			if (mbCheck == 3) mbCheck = 0;
		}
		
		mdX = mdX + (int)(mdVelocity * cos(2.0 * PI * (double)mnDirectionDegrees / 360.0));
		mdY = mdY + (int)((10.0 / 17.0) * mdVelocity * sin(2.0 * PI * (double)mnDirectionDegrees / 360.0));
		// CONTINUE ROTATING
		mnDirectionDegrees = mnDirectionDegrees + mnRotationDegrees;
		// WRAP
		if (mnDirectionDegrees >= 360)
			mnDirectionDegrees = mnDirectionDegrees % 360;
		else if (mnDirectionDegrees < 0)
			mnDirectionDegrees = 360 - (abs(mnDirectionDegrees) % 360);
		MapNode * poNode;
		poNode = mpoCurNode;
		UpdateNode();
		if (poNode != mpoCurNode)
		{
			if (mpoCurNode->mpoUnit && mpoCurNode->mpoUnit != mpoSourceUnit)
			{
				mpoCurNode->mpoUnit->vDamage(20.0f+10.0f*(float)mpoSourceUnit->nLevel(), mpoSourceUnit);
			}
		}
		return;
	case MISSILE_TIMERFADEDRAWOBJ:
		if (mbCycleCountdown)
		{
			if (mnCyclesRemaining > 0)
			{
				mnCyclesRemaining--;
			}
			else
			{
				if (mpoUnit->bIsVisible())
				{
					mnState++;
					BYTE jAlpha = 255 * (FADERESOLUTION - mnState) / FADERESOLUTION;
					if (jAlpha == 0)
					{
						mpoUnit->vSetVisibility(FALSE);
					}
					else
					{
						mpoUnit->vSetAlpha(jAlpha);
					}
				}
				else
				{
					Impact();
					//if (mpoUnit) delete mpoUnit;
					Destroy();
				}
			}
		}

		return;
	case MISSILE_ANIMATESTOPANDFADEAWAY:
		if (mnCel == nGetNumCels()-1)
		{
			if (mpoCycler) 
			{
				delete mpoCycler;
				mpoCycler = NULL;
			}
		}
		if (mbCycleCountdown)
		{
			if (mnCyclesRemaining > 0)
			{
				mnCyclesRemaining--;
			}
			else
			{
				if (mpoUnit->bIsVisible())
				{
					mnState++;
					BYTE jAlpha = 255 * (FADERESOLUTION - mnState) / FADERESOLUTION;
					if (jAlpha == 0)
					{
						mpoUnit->vSetVisibility(FALSE);
						vSetVisibility(FALSE);
					}
					else
					{
						mpoUnit->vSetAlpha(jAlpha);
						vSetAlpha(jAlpha);
					}
				}
				else
				{
					Impact();
					//if (mpoUnit) delete mpoUnit;
					Destroy();
				}
			}
		}
		return;
	case MISSILE_TIMER:
		if (mbCycleCountdown)
		{
			if (mnCyclesRemaining > 0)
			{
				mnCyclesRemaining--;
			}
			else
			{
				if (mbVisible)
				{
					mnState++;
					mjAlpha = 255 * (FADERESOLUTION - mnState) / FADERESOLUTION;
					if (mjAlpha == 0)
						vSetVisibility(FALSE);
				}
				else
				{
					Impact();
					Destroy();
				}
			}
		}
		return;
	case MISSILE_GROUNDHOMING:
		//if (mpoCycler) mpoCycler->vAnimationCycle();

		// IF MOVING THEN MOVE -- NO THINKING?
		if (mnState == 2)
		{
			vCalcRoute();
			mnState = 1;
		}
		
		//Missile * poMissile;
		//poMissile = new Missile(MISSILE_TRACER, gpoMap->mpoPlane, (Unit*)this, NULL, mpoLastNode, 6005, 10, 0, 0, NULL, mdX, mdY, 1, 1);

		if (mbBetweenPoints)
		{
			vMoveOneCycle();
		}
		else if (mnState == 1 && mpoCurNode != poGetTargetNode())
		{
			int nTargetX;
			int nTargetY;

			nTargetX = poGetTargetNode()->mnXPos;
			nTargetY = poGetTargetNode()->mnYPos;
			
			if (nTargetX != mnLastTargetX || nTargetY != mnLastTargetY) 
			{
				vCalcRoute();
			}
			vMoveToNextNode();
		}
		else
		{
			if (mbImpactObject)
			{
				if (mpoTargetUnit && mpoTargetUnit->lDistToSpace(mpoCurNode->mnXPos, mpoCurNode->mnYPos) <= 1)
				{
					Impact();
				}
				else
					Impact();
			}
			Destroy();
			return;
		}
		if (mbCycleCountdown)
		{
			if (mnCyclesRemaining > 0)
			{
				mnCyclesRemaining--;
			}
			else
				Destroy();
		}
		return;
	case MISSILE_ANIMATEANDFADEDRAWOBJ:
		int bDelete;
		bDelete = FALSE;
		if (mnCel == nGetNumCels() - 1)
		{
			bDelete = TRUE;
		}
		if (mnCel == mnFallOff)
		{
			if (mpoUnit)
			{
				mpoUnit->DestroyFromMap();
				mpoUnit = NULL;
			}
		}
		if (bDelete)
		{
			Destroy();
		}
		return;
	case MISSILE_TRACERANIMATED:
		//if (mpoCycler) mpoCycler->vAnimationCycle();
		// FALL THROUGH
	case MISSILE_TRACER:
		if (mbObstructView)
		{
			if (mpoFoggedUnit)
			{
				if (mpoCurNode->mpoUnit == NULL)
				{
					mpoFoggedUnit->SetNoFog();
					mpoFoggedUnit = NULL;
				}
			}
			else
			{
				if (mpoCurNode->mpoUnit)
				{
					mpoCurNode->mpoUnit->SetInFog();
					mpoFoggedUnit = mpoCurNode->mpoUnit;
				}
			}
			//gpoMap->vObstructVisibility((int)(mdX / 17.0), (int)(mdY / 10.0), MAP_VISIBILITY_EXPOSED_MAP);
		}
		if (mnFalloffCycle)
		{
			mnFalloffCycle--;
			int n = GlobalRand(4);
			//if (n == 0)
				mjAlpha = 127 + (BYTE)((double)n * 128.0 / 3.0);
			//else if (n == 2)
			//	mjAlpha = 64;
			//e/lse
			//	mjAlpha = 255;
			return;
		}
		mbCheck++;
		if (mbCheck == mnFallOff)
		{
			mnShade--;
			mbCheck = 0;
			mjAlpha = mnShade * 255 / mnShadeCount;
			UpdateNode();
		}
		if (mnShade == 0)
		{
			if (mpoFoggedUnit)
			{
				mpoFoggedUnit->SetNoFog();
				mpoFoggedUnit = NULL;
			}
			Destroy();
		}
		return;
		break;
	case MISSILE_DRIFTINGSMOKE:
		if ( mnCelCount > 1 )
		{
			mdX = mdX + mddX;
			mdY = mdY + mddY;
			mdZ = mdZ + mddZ;

			mbCheck++;

			// FADES OUT FROM 138 to 16 AT mnCelCount RESOLUTION
			//mjAlpha = 16 + (mnCelCount - mnCel) * 112 / mnCelCount;
			if (mbCheck == mnFallOff)
			{
				mbCheck = 0;
				mnCel++;
				if (mnCel >= mnCelCount)
				{
					Destroy();
					return;
				}
			}
			UpdateNode();
			return;
		}
		Destroy();
		return;
	case MISSILE_LINGERINGSMOKE:
		if ( mnCelCount > 1 )
		{
			mdX = mdX + mddX;
			mdY = mdY + mddY;
			mdZ = mdZ + mddZ;

			mbCheck++;

			if (mbCheck == mnFallOff)
			{
				mnShade--;
				//mjAlpha = 16 + (mnCelCount - mnCel) * 112 / mnCelCount;
				mjAlpha = 16 + mnShade * 112 / mnShadeCount;
				mbCheck = 0;
				if (mnCel > 0) mnCel--;
				if (mnShade < 0)
				{
					Destroy();
					return;
				}
			}
			UpdateNode();
			return;
		}
		Destroy();
		return;
	default:
		{
			// WAIT FOR GAME CYCLES
			//if (mpoCycler) mpoCycler->vAnimationCycle();

			int nDirection;

			nDirection = 0;
			switch (meSmokeTrail)
			{
			case SMOKETRAIL_NONE:
				break;
			case SMOKETRAIL_FIREBALL:
				if (GlobalRand(3) == 2)
					CreateMissile(NULL, mpoCurNode, mpoCurNode, MISSILE_LINGERINGSMOKE, 24, FALSE, (float)mdZ, mdX, mdY, 2, nDirection);
				break;
			case SMOKETRAIL_RESIDUAL:
				new Missile(MISSILE_LINGERINGSMOKE, gpoMap->mpoPlane, NULL, mpoCurNode, mpoCurNode, 6002+GlobalRand(3), 8, FALSE, (float)mdZ, NULL, mdX, mdY, 1, nDirection);
				//CreateMissile(NULL, mpoCurNode, mpoCurNode, MISSILE_LINGERINGSMOKE, 10, FALSE, (float)mdZ, mdX, mdY, 1, nDirection);
				break;
			case SMOKETRAIL_FADECYCLE:
				mbCheck++;
				if (mbCheck == 3)
				{
					mbCheck = 0;
					CreateMissile(NULL, mpoCurNode, mpoCurNode, MISSILE_DRIFTINGSMOKE, 0, FALSE, (float)mdZ, mdX, mdY, 2, nDirection);
				}
				break;
			}

			if (mddX < 0)
			{
				if (mdX > mdTargetX)
				{
					mdX = mdX + mddX;
				}
				else 
				{
					mdX = mdTargetX;
				}
			}
			else if (mddX > 0)
			{
				if (mdX < mdTargetX)
				{
					mdX = mdX + mddX;
				}
				else 
				{
					mdX = mdTargetX;
				}
			}
			if (mddY < 0)
			{
				if (mdY > mdTargetY)
				{
					mdY = mdY + mddY;
				}
				else 
				{
					mdY = mdTargetY;
				}
			}
			else if (mddY > 0)
			{
				if (mdY < mdTargetY)
				{
					mdY = mdY + mddY;
				}
				else 
				{
					mdY = mdTargetY;
				}
			}
			mddZ = mddZ + mdddZ;
			mdZ = mdZ + mddZ;
			
			
			/*
			if ( mnCelCount > 1 )
			{
				mbCheck++;
				if (mbCheck == 3)
				{
					mbCheck = 0;
					mnCel--;
					if (mnCel < 0)
					{
						mnCel = mnCelCount - 1;
					}
					assert(mnCel >= 0);
					vSetCel(mnCel);
				}
			}
			*/
			UpdateNode();

			// ARRIVED AT TARGET
			if ((mdY == mdTargetY)&&(mdX == mdTargetX))
			{
				if (mbImpactObject)
				{
					Impact();
				}
				Destroy();
				return;
			}
		}
		break;
	}

}

void Missile::EmbedUnit(Unit * poUnit)
{
	assert(poUnit);
	assert(mpoUnit==0);  // just once
	assert(poUnit->mpoCurNode == NULL); // not on map
    mpoUnit = poUnit;
}

void Missile::SetDrawObj(int nUnit)
{
	mpoUnit = (Unit * ) goObjectDB.poGetObject(nUnit);
}

void Missile::Impact()
{
	// IMPACT
	switch(meImpact)
	{
	case IMPACT_NONE:
		return;
	case IMPACT_DEATHRAIN:
		Impact_DeathRain();
		break;
	case IMPACT_DEATHFROMABOVE:
		Impact_DeathFromAbove();
		break;
	case IMPACT_DRAINPOWERANDRETURN:
		Impact_DrainPowerAndReturn();
		break;
	case IMPACT_INCREASEPOWER:
		Impact_IncreasePower();
		break;
	case IMPACT_UNITREAPPEARS:
		Impact_UnitReappears();
		break;
	case IMPACT_SPYOFF:
		Impact_SpyOff();
		break;
	case IMPACT_MELT:
		Impact_Melt();
		break;
	case IMPACT_CLEARINVISIBLE:
		Impact_ClearInvisible();
		break;
	case IMPACT_BOMBLET:
		Impact_Bomblet();
		break;
	default:
		assert(0); // NOT IMPLEMENTED
	}
}

void Missile::Impact_ClearInvisible()
{
	mpoTargetUnit->ClearInvisible();
}

void Missile::Impact_IncreasePower()
{
	// TRANSFER POWER FROM MISSILE TO TARGET UNIT
	mpoTargetUnit->IncreasePower(mfPower);
//	pApp->pSoundManager->EmitFromPosition((int)mdX, (int)mdY, SOUND_EnergyArrived);
}

void Missile::Impact_SpyOff()
{
	mpoTargetUnit->ClearSpy(mpoSourceUnit->poOwner());
}

void Missile::Impact_UnitReappears()
{
	assert(mpoCurNode->mpoUnit == NULL);
	mpoUnit->mpoCurNode = mpoCurNode;
	mpoUnit->vSetPixelPos(mpoCurNode->nPixelX(), mpoCurNode->nPixelY());
	mpoCurNode->mpoUnit = mpoUnit;
	mpoUnit->mbAlive = TRUE;
	if (mbUnitWasSelected)
	{
		if (0 == mpoPlane->mpoMap->nSelectionSize())
		{
			mpoPlane->mpoMap->vAddSelection(mpoUnit);
			mpoPlane->mpoMap->vCommitSelection();
		}
	}
	mpoUnit->vMarkUpdate();
	//pApp->pSoundManager->EmitFromObject(mpoUnit, SOUND_RicochetQuickSound);
	mpoUnit = NULL;
}

void Missile::Impact_DrainPowerAndReturn()
{
	// DRAIN POWER UP TO DAMAGE FROM POWER (N)
	float fPower = PowerFromUnit((float)(50 + 5 * mpoSourceUnit->nLevel()), mpoTargetUnit);
	//float fPower = mpoTargetUnit->vDamage((float)(50 + 5 * mpoSourceUnit->nLevel()), mpoSourceUnit);

	// CREATE NEW MISSILE TO POINT TO TARGET UNIT
	// CREATE A HOMING MISSILE
	Unit * poUnit = mpoSourceUnit;

	// FILL NEW MISSILE WITH N POWER
	Missile * poMissile = new Missile(MISSILE_GROUNDHOMING, gpoMap->mpoPlane, mpoTargetUnit, mpoCurNode, 0, 6015, 0, 0, 0, poUnit);
#define MISSILESPEED 3
	// HOW FAR CAN IT GO?
	poMissile->SetCycleCountdown(TRUE);
	// SET NEW MISSILE CYCLES REMAINING TO SAME AS THIS MISSILE
	poMissile->SetCyclesRemaining(mnCyclesRemaining);
	poMissile->SetImpact(IMPACT_INCREASEPOWER);
	poMissile->EmbedPower(fPower);
	// RELEASE IT
	//pApp->pSoundManager->EmitFromPosition((int)mdX, (int)mdY, SOUND_AcquiringEnergy);

}

void Missile::SetHeight(double dZ)
{
	mdZ = dZ;
}

void Missile::Impact_Bomblet()
{
	if (mpoCurNode->mpoUnit)
	{
		mpoCurNode->mpoUnit->vDamage(30.0f + 5.0f * (float)mpoSourceUnit->nLevel(), mpoSourceUnit);
	}
}

void Missile::Impact_DeathRain()
{
	// CREATE 20 DOWNWARD PROJECTILES THAT LAND IN SQUARE AREA
	for (int i=0; i<20; i++)
	{
		// FIND ITS TARGET WITHIN SQUARE AREA
		int nDimension = 7;
		int x = mpoCurNode->mnXPos;
		int y = mpoCurNode->mnYPos;
		x = min(gpoMap->nGetWidth(), max(0, x + GlobalRand(nDimension) - (nDimension/2)));
		y = min(gpoMap->nGetHeight(), max(0, y + GlobalRand(nDimension) - (nDimension/2)));
		MapNode * poTargetNode = gpoMap->poGetNode(x, y);

		// NEW MISSILE
		Missile * poMissile = new Missile(MISSILE_DROPBOMB, gpoMap->mpoPlane, 0, mpoCurNode, poTargetNode, 6020, 0, 0, 0);
		
		// BEGIN DROP FROM HERE TO THERE
		poMissile->SetHeight(mdZ);
		poMissile->mddZ = 7.0 * GlobalRand(100) / 100.0 - 1.4;
		poMissile->mdddZ = -0.2;
		poMissile->mbCastShadow = TRUE;
		poMissile->SetImpact(IMPACT_BOMBLET);
		poMissile->mpoSourceUnit = mpoSourceUnit;
	}
}

void Missile::Impact_DeathFromAbove()
{
	if (mbSolid)
	{
		//pApp->pSoundManager->EmitFromPosition((int)mdX, (int)mdY, SOUND_FireballBoom, SOUNDPRIORITY_UNITSOUNDS, FALSE, 128);

		int nCount = 5 + GlobalRand(4);
		int nAdjustX = (int)(mddX * 18.0 / 17.0);
		int nAdjustY = (int)(mddY * 18.0 / 10.0);

		for (int i=0;i<nCount;i++)
		{
			int dx = GlobalRand(15) - 7;
			int dy = GlobalRand(11) - 5;

			dx = dx + nAdjustX;
			dy = dy + nAdjustY;

			int X = mpoCurNode->mnXPos + dx;
			int Y = mpoCurNode->mnYPos + dy;

			if (X < 0 || X >= gpoMap->nGetWidth() || Y < 0 || Y >= gpoMap->nGetHeight())
			{
				continue;
			}

			MapNode * poTargetNode = gpoMap->poGetNode(mpoCurNode->mnXPos + dx, mpoCurNode->mnYPos + dy); 
			if (poTargetNode)
			{
				CreateMissile(NULL, mpoCurNode, poTargetNode, MISSILE_SMOKE, 0, TRUE, (float)0.68);
			}
			CreateMissile(NULL, mpoCurNode, mpoCurNode, MISSILE_DRIFTINGSMOKE, 8, FALSE, (float)GlobalRand(10), mpoCurNode->mnXPos * 17.0 + GlobalRand(20) - 5, mpoCurNode->mnYPos * 10.0 + GlobalRand(10) - 5, 2, 1);
		}
		Damage();
		Missile * poMissile = new Missile(MISSILE_DRIFTINGSMOKE, gpoMap->mpoPlane, (Unit*)this, NULL, NULL, 6017, 10, 0, 0, NULL, mdX, mdY, 3, 0);
		poMissile->SetDrift(DRIFT_NONE);
		poMissile = new Missile(MISSILE_TIMER, gpoMap->mpoPlane, mpoSourceUnit, mpoCurNode, mpoCurNode, 6018, 0, FALSE, 1.18f, NULL, 10.0, 10.0, 2, 1);
		poMissile->vSetPixelPos((int)mdX, (int)mdY+8);
		poMissile->vSetLayer(DP_ITEMS);
		poMissile->UpdateNode();
		poMissile->SetCyclesRemaining(200);
	}
}

void Missile::SetDrift(TYPE_DRIFT eDrift)
{
	switch (eDrift)
	{
	case DRIFT_NONE:
		mddX = 0;
		mddY = 0;
		mddZ = 0;
		break;
	}
}

void Missile::Damage()
{
	switch(meDamageType)
	{
	case DAMAGE_AREASQUARE:
		{
			// RANGE, DAMAGEBASE, DAMAGELEVELMULTIPLIER
			Damage_AreaSquare();
		}
		break;
	case DAMAGE_NONE:
		break;
	default:
		assert(0);
		break;
	}
}

void Missile::Damage_AreaSquare()
{
	assert(mpoSpell);
	assert(mpoCurNode);
	assert(mpoSourceUnit);

	UINT nDimension = mpoSpell->GetDamageRange(mpoSourceUnit);

	MapNode * poNodeEpicenter = mpoCurNode;
	float nMaxDamage = mpoSpell->GetDamage(mpoSourceUnit);
	float nAdjustDamage = nMaxDamage; 

	// ADJUST TO ODD VALUE IF NOT ZERO
	if (nDimension && (nDimension % 2 == 0)) nDimension++;

	int nMaxRange = nDimension / 2;

	int nStartX = max(0, poNodeEpicenter->mnXPos - nMaxRange);
	int nStartY = max(0, poNodeEpicenter->mnYPos - nMaxRange);
	int nEndX = min(gpoMap->nGetWidth()-1, poNodeEpicenter->mnXPos + nMaxRange);
	int nEndY = min(gpoMap->nGetHeight()-1, poNodeEpicenter->mnYPos + nMaxRange);

	for (int nX = nStartX; nX<nEndX; nX++)
	{
		for(int nY = nStartY; nY<nEndY; nY++)
		{
			MapNode * poNode = gpoMap->poGetNode(nX, nY);
			Unit * poUnit = poNode->mpoUnit;
			DrawObj * poCover = poNode->poGetCover();
			int nDistanceFromCenter;
			if (poUnit || poCover)
			{
				// THIS IS THE DISTANCE MODEL -- NOT TRUE DISTANCE
				if (nMaxRange)
				{
					nDistanceFromCenter = max( abs(nEndY - nMaxRange - nY), abs(nEndX - nMaxRange - nX));
					nAdjustDamage = nMaxDamage * (nMaxRange - nDistanceFromCenter / nMaxRange);
				}
				if (poUnit)
				{
					assert(mpoSourceUnit);
					poUnit->vDamage((float)nAdjustDamage, mpoSourceUnit);
				}						
				if (poCover)
				{
					poCover->vDamage((float)nAdjustDamage, mpoSourceUnit);
				}
			}
		}
	}
}

// possibly use this for higher resolution than game cycle
void Missile::SubCycle(int nSubCycle)
{
	// check mnLastSubCycle against nSubCycle
	// increment status by difference in subcycles
	// update on screen status
}

// TBD -- PAB -- handle blocks or countermeasures
Boolean Missile::HandleEvent(sciEvent & oEvent)
{
	return FALSE;
}

void Missile::vInsertSelf(sciEvent &oEvent) {
	MapNode *poNode = mpoPlane->mpoMap->poNodeFromEvent(oEvent);
	if (poNode) {

		//most likely vDrag will have already inserted us into the node(and into moDrawn)
		//but if it didn't we need to catch it here.
//		vSetLayer(DP_ITEMS);
		poNode->vAddObj(this); //this lines also removes self from previous node
		//gpoMapEdit->mpoMapNode->vReset(); //clear out clipboard - no need see above line

	} else {
		// we clicked somewhere not on the game map so get rid of self
		delete this;
	}
}


// For adding to the map.
// The draw object being dragged around makes a copy of itself and inserts in into the
// node nearest the event
void Missile::vCopySelf(sciEvent &oEvent) 
{
	MapNode *poNode = mpoPlane->mpoMap->poNodeFromEvent(oEvent);
	if (poNode) {
		DrawObj *poCopy = new DrawObj(mpoPlane, nTemplateID());	
		poCopy->vSetLayer(mnLayer);
		poNode->vAddObj(poCopy);
	}
}

void Missile::SetImpact(TYPE_IMPACT eImpact)
{
	if (eImpact == IMPACT_NONE)
	{
		mbImpactObject = FALSE;
	}
	else
	{
		mbImpactObject = TRUE;
	}
	meImpact = eImpact;
}

float Missile::PowerFromUnit(float fMaxPower, Unit * poFromUnit)
{
	assert(poFromUnit);
	return poFromUnit->vDrainPower(fMaxPower);
}

void Missile::EmbedPower(float fPower)
{
	mfPower = fPower;
}

void Missile::SetFlashEmbeddedUnitVisibility(BOOL bValue)
{
	mbFlashEmbeddedUnitVisibility = (bValue != FALSE);
}

void Missile::Impact_Melt()
{
	if (mpoTargetUnit)
		mpoTargetUnit->ClearFrozen();
}

void Missile::SetAlpha(BYTE jAlpha) 
{ 
	mjAlpha = jAlpha; 
}

void Missile::SetSpell(Spell * poSpell)
{
	mpoSpell = poSpell;
}