// OBJECTDB.CPP -- the map/draw objects database
//
// DJM

#include "scimain.h"
#include "mapnode.h"
#include "drawobj.h"
#include "objectdb.h"

ObjectDB goObjectDB;

ObjectDB::ObjectDB()
{
	nTotalObjects = 0;
}

void ObjectDB::vAddObject(DrawObj *poObject)
{
	int i;
	int nSize = moObjects.GetSize();

	assert(poObject);
	assert(poObject->nGetID() == ID_NONE);

	for (i = 0; i < nSize; ++i) {
		if (!moObjects.GetItem(i)) {
			break;
		}
	}

	poObject->vSetID(i);
	moObjects.SetItem(i, poObject);
	nTotalObjects++;
	assert(nTotalObjects <= moObjects.GetSize());
}

void ObjectDB::vAddObjectFast(DrawObj *poObject)
// don't search for empty slots, just add to end
{
	assert(poObject);
	assert(poObject->nGetID() == ID_NONE);

	poObject->vSetID(moObjects.GetSize());
	moObjects.SetItem(moObjects.GetSize(), poObject);
	nTotalObjects++;
	assert(nTotalObjects <= moObjects.GetSize());
}

void ObjectDB::vRemoveObject(DrawObj *poObject)
{
	int nLoc;

	assert(poObject);
	assert(poObject->nGetID() != ID_NONE);

	poObject->vSetID(ID_NONE);
	nLoc = moObjects.Find(poObject);
	if (nLoc != ID_NONE) {
		moObjects.SetItem(nLoc, NULL);
		nTotalObjects--;
		assert(nTotalObjects >= 0);
	} else {
		DebugString("Failed to find object in vRemoveObject. DJM");
	}
}
	
void ObjectDB::vRemoveObject(int nID)
{
	DrawObj *poObject = (DrawObj*)moObjects.GetItem(nID);

	assert(poObject);

	poObject->vSetID(ID_NONE);
	moObjects.SetItem(nID, NULL);
	nTotalObjects--;
	assert(nTotalObjects >= 0);
}