#include "os.h"

#include "app.h"
#include "window.h"
#include "palette.h"
#include "util.h"
#include "resman.h"
#include "inifile.h"
#include "framebuf.h"

#include <string.h>

extern Application *pApp;
extern char szIniFilename[];
extern char *szAppName;
extern int bGetPaletteEntries;

#include <math.h>

void DoGammaCorrection(PALETTEENTRY far * pPes, int iNumColors, double fGamma)
{
	// temp -- don't do gamma.
	return;

	int i;
	double fInvGamma;

	fInvGamma = double (1) / fGamma;

	for (i = 0; i < iNumColors; ++i) {
		pPes[i].peRed = (BYTE) (.5 + 256 * pow(double (pPes[i].peRed) / 256, fInvGamma));
		pPes[i].peGreen = (BYTE) (.5 + 256 * pow(double (pPes[i].peGreen) / 256, fInvGamma));
		pPes[i].peBlue = (BYTE) (.5 + 256 * pow(double (pPes[i].peBlue) / 256, fInvGamma));
	}
}

double defaultGamma = 0;

void DoGammaCorrection(RGBQUAD * quad)
{
	PALETTEENTRY pe;

	pe.peRed = quad->rgbRed;
	pe.peGreen = quad->rgbGreen;
	pe.peBlue = quad->rgbBlue;
	DoGammaCorrection(&pe, 1, defaultGamma);
	quad->rgbRed = pe.peRed;
	quad->rgbGreen = pe.peGreen;
	quad->rgbBlue = pe.peBlue;
}

Palette::Palette(int resourceID)
:VPalette(resourceID)
{
	LPLOGPALETTE lpLogPalette;
	char far *lpPalRes;
	int i;
	hpalTheOneTruePalette = NULL;
	hpalOldSave = NULL;

	if (GetData()) {
		lpPalRes = (char far *) GetData();
		lpLogPalette = (LPLOGPALETTE) (lpPalRes + 20);

		for (i = 0; i < (int) (lpLogPalette->palNumEntries); ++i) {
			colors[i].rgbRed = lpLogPalette->palPalEntry[i].peRed;
			colors[i].rgbGreen = lpLogPalette->palPalEntry[i].peGreen;
			colors[i].rgbBlue = lpLogPalette->palPalEntry[i].peBlue;
			colors[i].rgbReserved = 0;
			DoGammaCorrection(&(colors[i]));
		}
		for (; i < 256; ++i) {
			colors[i].rgbRed = 0;
			colors[i].rgbGreen = 0;
			colors[i].rgbBlue = 0;
			colors[i].rgbReserved = 0;
		}
	} else {
		for (i = 0; i < 256; ++i) {
			colors[i].rgbRed = 0;
			colors[i].rgbGreen = 0;
			colors[i].rgbBlue = 0;
			colors[i].rgbReserved = 0;
		}
	}
}

Palette::~Palette()
{
	if (hpalOldSave && pApp->pWindow)
	{
		SelectPalette(pApp->pWindow->hDC, hpalOldSave, TRUE);
	}
	if (hpalTheOneTruePalette)
	{
		DeleteObject(hpalTheOneTruePalette);
		hpalTheOneTruePalette = NULL;
	}
}

void Palette::ActivatePalette()
{
#if 0
	if (pApp->pWindow) {
		if (hpalOldSave) {
			ActivatePalette(pApp->pWindow->hDC);
		} else {
			hpalOldSave = ActivatePalette(pApp->pWindow->hDC);
		}
	}
#else
	FrameBuffer *poFrameBuf = pApp->poFrameBuf();
	if (poFrameBuf) {
		Palette16 *poPal = new Palette16(this);
		poFrameBuf->vSetPalette(poPal);
	}
#endif
}

HPALETTE Palette::ActivatePalette(HDC hDC)
{
	int i;
	HPALETTE hOldPalette = 0;
	LPLOGPALETTE lpNoCollapsePalette;

	VPalette::ActivatePalette();
	bGetPaletteEntries = 1;

	if (pApp->pWindow) {
		if (hpalTheOneTruePalette == NULL) {
			lpNoCollapsePalette = (LOGPALETTE *) new BYTE[UINT(sizeof(LOGPALETTE) + 256 * sizeof(PALETTEENTRY))];
			lpNoCollapsePalette->palVersion = 0x0300;
			lpNoCollapsePalette->palNumEntries = 256;
			lpNoCollapsePalette->palPalEntry[0].peRed = 0;
			lpNoCollapsePalette->palPalEntry[0].peGreen = 0;
			lpNoCollapsePalette->palPalEntry[0].peBlue = 0;
			lpNoCollapsePalette->palPalEntry[0].peFlags = PC_RESERVED;
			for (i = 1; i < 256; ++i) {
				lpNoCollapsePalette->palPalEntry[i].peRed = colors[i].rgbRed;
				lpNoCollapsePalette->palPalEntry[i].peGreen = colors[i].rgbGreen;
				lpNoCollapsePalette->palPalEntry[i].peBlue = colors[i].rgbBlue;
				// lpNoCollapsePalette->palPalEntry[i].peFlags =
				// PC_RESERVED;
				lpNoCollapsePalette->palPalEntry[i].peFlags = 0;
			}

			hpalTheOneTruePalette = CreatePalette(lpNoCollapsePalette);

			/* if (GetDeviceCaps(hDC, RASTERCAPS) & RC_PALETTE) {
			   hOldPalette = SelectPalette(hDC, hpalTheOneTruePalette,
			   FALSE); RealizePalette(hDC);
			
			// Figure out what the realize did.
			   GetSystemPaletteEntries(hDC, 0, 256,
			   lpNoCollapsePalette->palPalEntry); for (i=0; i<10; ++i) {
			   lpNoCollapsePalette->palPalEntry[i].peFlags = 0; } for
			   (i=10; i<236; ++i) {
			   lpNoCollapsePalette->palPalEntry[i].peFlags =
			   PC_NOCOLLAPSE; } for (i=236; i<256; ++i) {
			   lpNoCollapsePalette->palPalEntry[i].peFlags = 0; } for
			   (i=0; i<256; ++i) { colors[i].rgbRed =
			   lpNoCollapsePalette->palPalEntry[i].peRed;
			   colors[i].rgbGreen =
			   lpNoCollapsePalette->palPalEntry[i].peGreen;
			   colors[i].rgbBlue =
			   lpNoCollapsePalette->palPalEntry[i].peBlue; } if
			   (hpalTheOneTruePalette)
			   DeleteObject(hpalTheOneTruePalette); hpalTheOneTruePalette
			   = CreatePalette(lpNoCollapsePalette); } */

			delete[] lpNoCollapsePalette;
		}
		hOldPalette = SelectPalette(hDC, hpalTheOneTruePalette, FALSE);
		RealizePalette(hDC);

		return hOldPalette;
	}
	return 0;
}

Palette16::Palette16()
{
	vInit();
}

Palette16::Palette16(const VPalette *poVPal)
{
	vInit();
	assert(poVPal);
	vConvert((RGBQUAD*)poVPal->colors);
}

Palette16::Palette16(RGBQUAD *pRGBQuads)
{
	vInit();
	assert(pRGBQuads);
	vConvert(pRGBQuads);
}

Palette16::Palette16(Palette16& oP)
{
	vInit();
	vAllocate();
	oP.vGetCopy(mpwPalette);
}

Palette16::~Palette16()
{
	if (mpwPalette) {
		delete [] mpwPalette;
		mpwPalette = NULL;
	}
}

Palette16& Palette16::operator=(Palette16& oP)
{
	vAllocate();
	oP.vGetCopy(mpwPalette);
	return *this;
}

void Palette16::vInit()
{
	mpwPalette = NULL;
	assert(gpoBlender);
}

void Palette16::vConvert(const VPalette *poVPal)
{
	vConvert((RGBQUAD*)poVPal->colors);
}

void Palette16::vConvert(RGBQUAD *pRGBQuads)
{
	assert(gpoBlender);

	vAllocate();

	int i;
	for (i = 0; i < 256; ++i) {
		mpwPalette[i] = (WORD)gpoBlender->dwPack(&(pRGBQuads[i]));
	}
}

void Palette16::vAllocate(void)
{
	if (mpwPalette) {
		return;
	}
	mpwPalette = new WORD[256];

	assert(mpwPalette);
}

void Palette16::vGetCopy(WORD *pwPalette)
{
	assert(pwPalette);
	memcpy(pwPalette, mpwPalette, 256 * sizeof(WORD));
}

void Palette16::vSetColor(BYTE jRef, WORD wColor) 
{
	vAllocate();
	mpwPalette[jRef] = wColor;
}

void Palette16::vSetColor(BYTE jRef, RGBQUAD *pRGBQuad) 
{
	vAllocate();
	assert(pRGBQuad);

	mpwPalette[jRef] = (WORD)gpoBlender->dwPack(pRGBQuad);
}

void Palette16::vSetColor(BYTE jRef, BYTE jRed, BYTE jGreen, BYTE jBlue) 
{
	vAllocate();
	mpwPalette[jRef] = (WORD)gpoBlender->dwPack(jRed, jGreen, jBlue);
}

