// PLAYER.H -- the info about one of the players
//
// DJM

#ifndef _PLAYER_
#define _PLAYER_


#include "object.h"
#include "list.h"
#include "mapnode.h"
#include "stringob.h"
#include "command.h"
#include "unit.h"

class DrawObj;
class GameMap;
class MapPlane;
class Clan;
//class Unit;
class Vampire;

#define PL_HUMAN	((BOOL)TRUE)
#define PL_COMPUTER ((BOOL)FALSE)

#define PL_LOCAL ((BOOL)TRUE)
#define PL_REMOTE ((BOOL)FALSE)

class Player : public Object {
public:
	Player(BOOL bHuman, BOOL bLocal, int nPlayer);
	~Player();

	int CallForHelpFromFriendlies(Vampire * poVampire, int nRange);
	// debugging stuff -- checksum and text dumping
	long lTurnChecksum(int nTurn);
	long lGetChecksum(void);
	void vDumpSyncText(FILE *fileDump);

	void vPostCommand(Command *poCommand, BOOL bAudioConfirm = TRUE);
	void vPostRemoteCommand(Command *poCommand, int nTurn);
	BOOL bTurnAvailable(int nTurn);

	virtual void vNetTurn(int nTurn);

	void vAddUnit(Unit *poUnit);
	void vDeleteUnit(Unit *poUnit) {moUnits.Remove(poUnit);}
	int GetUnitCountOfLevel(int nLevel);

	BOOL	mbHuman;
	BOOL	mbLocal;

	Clan	*mpoClan;

	String		msoName;
	String		msoChat; // current chat string

	int		mnPlayer;	// which player am I (1 is person who started game)

	void SendChecksumInfo();
	void SetTopLevel(int nLevel);
	int GetTopLevel();
	void GetVampireLord(Unit ** poUnit);
	void SetFlags(int nFlag);
	void ClearFlags(int nFlag);
	Vampire * FindFriendlyUnitToRetreatTowards(Vampire * poThisUnit);
	int GetLiveFriendlyUnitCount();

	Boolean bCanBeLevelSix() {return FALSE;}	//should return TRUE when level 6 is finally attainable
	short manUpgradeLevel[UPGRADE_COUNT];	//the current upgrades and levels they can go up to. A 0 means player doesn't know that upgrade.
	int nKnownUpgrades();					//the number of known upgrades (Level excluded)
	int nUpgradePosition(int nUpgrade);		//returns the 'position' of the known upgrade. 

	void vInitSpellSpace(int nNumSpells);	//allocate space for mpjSpells based on the number of spells in the game
	BYTE jSpellKnown(int nNdx);				//returns the value of indexed item in mpjSpells 
	void vAddSpell(int nNdx, int nLevel);	//add a known spell

	WORD wMiniMapColor(void) {return mwMiniMapColor;}
	int nColorIndex(void) {return mnColorIndex;}

protected:
	DWORD	mdwFlags;
	int		mnTopLevel;
	int		mnColorIndex; // 0-7 for global color values
	WORD	mwMiniMapColor; // the 16-bit color value used in the minimap
	CommandStack	moTurns;

	Player() { };  // pab - protected 

	RefList<Unit>			moUnits;

	//RefList<DrawObj>		moVisibleSelection; // what's currently showing

	//RefList<DrawObj>		moCurrentSelection; // what's currently being operated on

	int			mnSelectSize;
	int			manSelection[MAX_SELECTION_SIZE];

	GameMap		*mpoMap;
	MapPlane	*mpoPlane;
	BYTE		*mpjSpells;			//the current spells(parallel with game->moSpells list)
									//I'm using a BYTE instead of a boolean to store the spell level in case spells might go up levels
};


#endif