#include "os.h"

#include "roomint.h"
#include "rooms.h"
#include "app.h"
#include "window.h"
#include "palette.h"
#include "sounder.h"
#include "util.h"
#include "inifile.h"
#include "sciplane.h"
#include "scigame.h"
#include "sciscrip.h"
#include "scicursr.h"
#include "game.h"
#include <digitalv.h>
#include <stdio.h>
#include "splash.h"

// globals
extern HINSTANCE hInstance;
extern char cdDriveLetter;
extern char *movieFilename;

Palette *savedCurrentPalette;
Boolean endMovie;
WORD wMCIDeviceID = 0;	/* MCI Device ID for the AVI file */
HWND hwndMovie;		/* window handle of the movie */
DWORD movieLength;
DWORD moviePosition;
RECT rcMovie;		/* the rect where the movie is positioned      */
LONG oldWndProc;

BOOL initAVI();
void fileOpenMovie(HWND hWnd, char *movieName);
void playMovie(HWND hWnd);
void fileCloseMovie(HWND hWnd);
void positionMovie(HWND hWnd);
LRESULT CALLBACK stopProc(
    HWND hwnd,	// handle of window
    UINT uMsg,	// message identifier
    WPARAM wParam,	// first message parameter
    LPARAM lParam 	// second message parameter
   );

#define BUFFER_LENGTH  256

RoomIntro::RoomIntro()
	{
	SetPic(9998);
	delete pApp->pSounder;
	pApp->pSounder = NULL;
	endMovie = FALSE;

	MENUITEMINFO mii;
	mii.cbSize = sizeof(mii);
	mii.fMask = MIIM_STATE;
	mii.fType = 0;
	mii.fState = MFS_GRAYED;
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 0, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 1, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 2, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 3, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 4, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 5, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 6, TRUE, &mii);
	DrawMenuBar(pApp->pWindow->hWnd);
	}

RoomIntro::~RoomIntro()
	{
	if (!pApp->pCurrentPalette)
		{
		pApp->pCurrentPalette = savedCurrentPalette;
		if (pApp->pCurrentPalette)
			pApp->pCurrentPalette->ActivatePalette();
		}
	fileCloseMovie(pApp->pWindow->hWnd);
	pApp->pSounder = new Sounder;

	MENUITEMINFO mii;
	mii.cbSize = sizeof(mii);
	mii.fMask = MIIM_STATE;
	mii.fType = 0;
	mii.fState = MFS_ENABLED;
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 0, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 1, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 2, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 3, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 4, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 5, TRUE, &mii);
	SetMenuItemInfo(GetMenu(pApp->pWindow->hWnd), 6, TRUE, &mii);
	DrawMenuBar(pApp->pWindow->hWnd);
	}

class RoomIntroScript : public sciScript
	{
	public:
		void RoomIntroScript::ChangeState(int newState)
			{
			HANDLE h;
			WIN32_FIND_DATA findFileData;
			Boolean found;
			char str[256];
			char movieName[256];
			MCI_DGV_STATUS_PARMS	mciStatus;

			sciScript::ChangeState(newState);
			switch (state)
				{
				case 0: //General Setup
					if (!initAVI())
						{
						// quit the room.
						endMovie = TRUE;
						}

					// Find the movie
					// Try to find it locally.
					h = FindFirstFile(movieFilename, &findFileData);
					if (h != INVALID_HANDLE_VALUE)
						{
						found = TRUE;
						strcpy(movieName, movieFilename);
						FindClose(h);
						}
					else
						{
						strcpy(str, "..\\");
						strcat(str, movieFilename);
						h = FindFirstFile(str, &findFileData);
						if (h != INVALID_HANDLE_VALUE)
							{
							found = TRUE;
							strcpy(movieName, str);
							FindClose(h);
							}
						else
							{
							found = FALSE;
							}
						}

					// Try to find it on the CD, if there is one.
					if ((found == FALSE) && cdDriveLetter)
						{
						str[0] = cdDriveLetter;
						str[1] = 0;
						strcat(str, ":\\");
						strcat(str, movieFilename);
						h = FindFirstFile(str, &findFileData);
						if (h != INVALID_HANDLE_VALUE)
							{
							found = TRUE;
							strcpy(movieName, str);
							FindClose(h);
							}
						}
					if (found)
						{
						// Start the movie
						fileOpenMovie(pApp->pWindow->hWnd, movieName);
						playMovie(pApp->pWindow->hWnd);

						mciStatus.dwItem = MCI_STATUS_LENGTH;
						mciSendCommand(wMCIDeviceID, 
							MCI_STATUS, MCI_STATUS_ITEM,
							(DWORD)(LPMCI_STATUS_PARMS)&mciStatus);
						movieLength = mciStatus.dwReturn;
						}
					else
						{
						endMovie = TRUE;
						}

					SetTicks(0);
					break;
				case 1:
					//game->curRoom->plane->cursor->Hide();
					// Is the movie done?
					mciStatus.dwItem = MCI_STATUS_POSITION;
					mciSendCommand(wMCIDeviceID, 
						MCI_STATUS, MCI_STATUS_ITEM,
						(DWORD)(LPMCI_STATUS_PARMS)&mciStatus);
					moviePosition = mciStatus.dwReturn;
					mciStatus.dwItem = MCI_STATUS_MODE;
					mciSendCommand(wMCIDeviceID, 
						MCI_STATUS, MCI_STATUS_ITEM,
						(DWORD)(LPMCI_STATUS_PARMS)&mciStatus);
					if ((mciStatus.dwReturn == MCI_MODE_PAUSE) ||
						(mciStatus.dwReturn == MCI_MODE_NOT_READY) ||
						(mciStatus.dwReturn == MCI_MODE_STOP)
						)
						{
						char buff[128];

						sprintf(buff, "%d\n", mciStatus.dwReturn);
						DebugString(buff);
						endMovie = TRUE;
						}
					if (endMovie || (moviePosition >= movieLength))
						{
						// Yes, cleanup.
						game->SetNextRoom(pApp->pIniFile->ReadValue("Room", ROOM_0));
						pApp->pCurrentPalette = savedCurrentPalette;
						if (pApp->pCurrentPalette)
							pApp->pCurrentPalette->ActivatePalette();
						}
					else
						{
						// No, wait and check again.
						SetTicks(6);
						SetState(GetState() - 1);
						}
					break;
				}
			}
	};

void RoomIntro::Init()
	{
	sciRoom::Init();
	SetScript(new RoomIntroScript);
	savedCurrentPalette = pApp->pCurrentPalette;
	pApp->pCurrentPalette = NULL;
	}

Boolean RoomIntro::HandleEvent(sciEvent &event)
	{
	if (event.type == MOUSE_DOWN)
		endMovie = TRUE;
	if (event.type == KEY_PRESS)
		endMovie = TRUE;
	return FALSE;
	}

/*--------------------------------------------------------------+
| initAVI - initialize avi libraries				|
|								|
+--------------------------------------------------------------*/
BOOL initAVI()
{
    MCI_DGV_OPEN_PARMS	mciOpen;
		 
    /* set up the open parameters */
    mciOpen.dwCallback = NULL;
    mciOpen.wDeviceID = 0;
    mciOpen.lpstrDeviceType = "avivideo";
    mciOpen.lpstrElementName = NULL;
    mciOpen.lpstrAlias = NULL;
    mciOpen.dwStyle = 0;
    mciOpen.hWndParent = NULL;
		 
   /* try to open the driver */
   return (mciSendCommand(0, MCI_OPEN, (DWORD)(MCI_OPEN_TYPE), 
                          (DWORD)(LPMCI_DGV_OPEN_PARMS)&mciOpen) == 0);
}

/*--------------------------------------------------------------+
| termAVI - Closes the opened AVI file and the opened device    |
|           type.                                               |
|                                                               |
+--------------------------------------------------------------*/
void termAVI(void)
{
    WORD               wID;
    MCI_GENERIC_PARMS  mciClose;

    //
    // Get the device ID for the opened device type and then close
    // the device type.
    //
    wID = mciGetDeviceID("avivideo");
    mciSendCommand(wID, MCI_CLOSE, 0L, 
                   (DWORD)(LPMCI_GENERIC_PARMS)&mciClose);
}

/*--------------------------------------------------------------+ 
| fileOpenMovie - open an AVI movie. Use CommDlg open box to	|
|	        open and then handle the initialization to	|
|		show the movie and position it properly.  Keep	|
|		the movie paused when opened.			|
|								|
|		Sets <fMovieOpened> on success.			|
+--------------------------------------------------------------*/
void fileOpenMovie(HWND hWnd, char *movieName)
{
   static char szFile [BUFFER_LENGTH];
   static char szFileTitle [BUFFER_LENGTH];
	 MCI_DGV_OPEN_PARMS	mciOpen;
	 MCI_DGV_WINDOW_PARMS	mciWindow;
	 MCI_DGV_STATUS_PARMS	mciStatus;

	 
	 /* we have a .AVI movie to open, use MCI */
	 /* set up the open parameters */
	 mciOpen.dwCallback = NULL;
	 mciOpen.wDeviceID =0;
	 mciOpen.lpstrDeviceType = NULL;
	 mciOpen.lpstrElementName = movieName;
	 mciOpen.lpstrAlias = NULL;
	 mciOpen.dwStyle = WS_CHILD;
	 mciOpen.hWndParent = hWnd;

	 /* try to open the file */
	 if (mciSendCommand(0, MCI_OPEN, 
		 (DWORD)(MCI_OPEN_ELEMENT|MCI_DGV_OPEN_PARENT|MCI_DGV_OPEN_WS), 
       (DWORD)(LPMCI_DGV_OPEN_PARMS)&mciOpen) == 0){

		 /* we opened the file o.k., now set up to */
		 /* play it.				   */
		 wMCIDeviceID = mciOpen.wDeviceID;	/* save ID */

		 /* show the playback window */
		 mciWindow.dwCallback = NULL;
		 mciWindow.hWnd = pApp->pWindow->hWnd;
		 //mciWindow.hWnd = NULL;
		 mciWindow.nCmdShow = SW_SHOW;
		 mciWindow.lpstrText = (LPSTR)NULL;
		 mciSendCommand(wMCIDeviceID, MCI_WINDOW, 
			 MCI_DGV_WINDOW_STATE, 
			 (DWORD)(LPMCI_DGV_WINDOW_PARMS)&mciWindow);

		 /* get the window handle */
		 mciStatus.dwItem = MCI_DGV_STATUS_HWND;
		 mciSendCommand(wMCIDeviceID, 
			 MCI_STATUS, MCI_STATUS_ITEM,
			 (DWORD)(LPMCI_STATUS_PARMS)&mciStatus);
		 hwndMovie = (HWND)mciStatus.dwReturn;

		 /* now get the movie centered */
		 positionMovie(hWnd);
//	 } else {
//		/* generic error for open */
//		MessageBox(hWnd, "Unable to open Movie", NULL, 
//			      MB_ICONEXCLAMATION|MB_OK);
	 }
   /* cause an update to occur */
   InvalidateRect(hWnd, NULL, FALSE);
   UpdateWindow(hWnd);
}

/*--------------------------------------------------------------+
| playMovie - play/pause the movie depending on the state	|
|		of the <fPlaying> flag.				|
|								|
| This function sets the <fPlaying> flag appropriately when done|
|								|
+--------------------------------------------------------------*/
void playMovie(HWND hWnd)
{
	   DWORD		            dwFlags;
	   MCI_DGV_PLAY_PARMS	mciPlay;
		   
	   /* init to play all */
 	   mciPlay.dwCallback = (DWORD)hWnd;
	   mciPlay.dwFrom = mciPlay.dwTo = 0;
	   dwFlags = MCI_NOTIFY;
		   
	   mciSendCommand(wMCIDeviceID, MCI_PLAY, dwFlags,
		               (DWORD)(LPMCI_DGV_PLAY_PARMS)&mciPlay);
}

/*--------------------------------------------------------------+ 
| fileCloseMovie - close the movie and anything associated	|
|		   with it.					|
|								|
| This function clears the <fPlaying> and <fMovieOpen> flags	|
|								|
+--------------------------------------------------------------*/
void fileCloseMovie(HWND hWnd)
{
  MCI_GENERIC_PARMS  mciGeneric;
    
  mciSendCommand(wMCIDeviceID, MCI_CLOSE, 0L, 
                 (DWORD)(LPMCI_GENERIC_PARMS)&mciGeneric);

  /* cause a total repaint to occur */
  InvalidateRect(hWnd, NULL, TRUE);
  UpdateWindow(hWnd);
}

/*--------------------------------------------------------------+ 
| positionMovie - sets the movie rectange <rcMovie> to be	|
|		centered within the app's window.		|
|								|
+--------------------------------------------------------------*/
VOID positionMovie(HWND hWnd)
	{
	RECT	rcClient, rcMovieBounds;
	MCI_DGV_RECT_PARMS	mciRect;

	oldWndProc = SetWindowLong(hwndMovie, GWL_WNDPROC, (LONG)stopProc);

	GetClientRect(hWnd, &rcClient);	/* get the parent windows rect */
   
	/* get the original size of the movie */
	mciSendCommand(wMCIDeviceID, MCI_WHERE, 
		(DWORD)(MCI_DGV_WHERE_SOURCE), 
		(DWORD)(LPMCI_DGV_RECT_PARMS)&mciRect);
	rcMovieBounds = mciRect.rc;	/* get it in the movie bounds rect */

	rcMovie.left = (rcClient.right/2) - (rcMovieBounds.right / 2);
	rcMovie.top = (rcClient.bottom/2) - (rcMovieBounds.bottom / 2);
	rcMovie.right = rcMovie.left + rcMovieBounds.right;
	rcMovie.bottom = rcMovie.top + rcMovieBounds.bottom;

	/* reposition the playback (child) window */
	MoveWindow(hwndMovie, rcMovie.left, rcMovie.top,
		rcMovieBounds.right, rcMovieBounds.bottom, TRUE);
	}

LRESULT CALLBACK stopProc(
    HWND hwnd,	// handle of window
    UINT uMsg,	// message identifier
    WPARAM wParam,	// first message parameter
    LPARAM lParam 	// second message parameter
   )
	{
	switch(uMsg)
		{
		case WM_LBUTTONDOWN:
			SendMessage(pApp->pWindow->hWnd, uMsg, wParam, lParam);
			break;
		}
	return CallWindowProc((WNDPROC)oldWndProc, hwnd, uMsg, wParam, lParam);
	}


RoomSelect::RoomSelect()
{
	new OpeningScreen();
}

RoomSelect::~RoomSelect()
{
	game->SetNextRoom(ROOM_0);
}

void RoomSelect::Init()
{
	sciRoom::Init();
}

Boolean RoomSelect::HandleEvent(sciEvent &event)
{
	if (event.type == MOUSE_DOWN)
		endMovie = TRUE;
	if (event.type == KEY_PRESS)
		endMovie = TRUE;
	return FALSE;
}

