#include "os.h"

#include "app.h"
#include "rc.h"
#include "scibuttn.h"
#include "util.h"
#include "sigs.h"
#include "game.h"
#include "sciroom.h"
#include "sciplane.h"
#include "scicursr.h"
#include "sciscrip.h"

extern Application *pApp;
extern Game *game;
extern SIGS *sigs;

sciButton *sciButton::mpoCurrentHilight = NULL;

class sciButtonRepeatScript : public sciScript
	{
	public:
		Boolean rightClicking;

		sciButtonRepeatScript(Boolean newRightClicking) : sciScript(NULL)
			{
			rightClicking = newRightClicking;
			}

		void ChangeState(int newState)
			{
			sciScript::ChangeState(newState);
			switch (state)
				{
				case 0: //Setup
					SetTicks(((sciButton*)client)->startDelay);
					break;
				case 1:
					if (((sciButton*)client)->IsEnabled() &&
							((sciButton*)client)->IsNotHidden())
					{
						if (rightClicking)
							((sciButton*)client)->DoRightClick();
						else
							((sciButton*)client)->DoLeftClick();
					}
					SetTicks(((sciButton*)client)->repeatDelay);
					--state;
					break;
				}
			}
	};

sciButton::sciButton()
{
	state		= DEFAULT_BUTTON_NORMAL;
	text		= NULL;
	SetButtonDefaultCels();
	mouseDown	= FALSE;
	enabled		= TRUE;
	hotkey		= 0;
	hotkey2		= 0;
	rightHotkey = 0;
	//justification = JUST_HORIZONTAL | JUST_VERTICAL; //tw- I dont like this default
	justification = JUST_LEFT | JUST_TOP;
	startDelay	= -1;
	repeatDelay = -1;	
}

sciButton::~sciButton()
{
	if(text) {delete text; text = NULL;}
}

void sciButton::Init(int newWidth, int newHeight)
{
	sciProp::Init();
	if(text)
	{
		if(!newWidth)
			newWidth = GetWidth();
		if(!newHeight)
			newHeight = GetHeight();
		SetWidth(newWidth);
		SetHeight(newHeight);
		text->SetWidth(GetWidth());
		text->SetHeight(GetHeight());
		text->SetPri(GetPri() + 1);
		text->Posn(GetX(), GetY());
		text->Init();
	}
}

void sciButton::Render(RECT &updateRect) {
	if (text) {
		text->vDraw();
	} else {
		sciProp::Render(updateRect);
	}
}

void sciButton::SetButtonCels(int newNormal, int newDepressed, int newDisabled, int newHilighted)
{
	normal = newNormal;
	depressed = newDepressed;
	disabled = newDisabled;
	hilighted = newHilighted;
	UpdateCel();
}

void sciButton::SetButtonDefaultCels()
{
	normal = DEFAULT_BUTTON_NORMAL;
	depressed = DEFAULT_BUTTON_DEPRESSED;
	disabled = DEFAULT_BUTTON_DISABLED;
	hilighted = DEFAULT_BUTTON_HILIGHTED;
	UpdateCel();
}

Boolean sciButton::HandleEvent(sciEvent &event)
{
	if (!enabled || hidden)
		return FALSE;
	switch (event.type)
		{
		case MOUSE_DOWN:
			if ((!OnMe(event.x, event.y) && !mouseDown) || event.claimed)
				break;
			if ((event.modifiers & RIGHT_MOUSE) && !mouseDown)
			{
				mouseDown = TRUE;
				pApp->pWindow->vAddEventHog(this); //make a (modal)hog
//				plane->CaptureEvents(this);
				SetState(DEFAULT_BUTTON_DEPRESSED);
				event.claimed = TRUE;
				if (startDelay != -1)
					SetScript(new sciButtonRepeatScript(TRUE));
			}
			if ((event.modifiers & LEFT_MOUSE) && !mouseDown)
			{
				mouseDown = TRUE;
				pApp->pWindow->vAddEventHog(this); //make a (modal)hog
//				plane->CaptureEvents(this);
				SetState(DEFAULT_BUTTON_DEPRESSED);
				event.claimed = TRUE;
				if (startDelay != -1)
					SetScript(new sciButtonRepeatScript(FALSE));
			}
			break;
		case MOUSE_MOVE:
			if (mouseDown)
			{
				event.claimed = TRUE;
				if (!OnMe(event.x, event.y))
				{
					//SetState(DEFAULT_BUTTON_NORMAL); - tw: i think i like the hilight 'look' better
					SetState(DEFAULT_BUTTON_HILIGHTED);
					SetScript(NULL);
				}
				else
				{
					SetState(DEFAULT_BUTTON_DEPRESSED);
				}
			}
			else
			{
//				vDebugOut ("OnMe:%d\n", OnMe(event.x, event.y));
				if (OnMe(event.x, event.y))
				{
					if (event.claimed)
						{
						SetState(DEFAULT_BUTTON_NORMAL);
						}
					else
						{
						SetState(DEFAULT_BUTTON_HILIGHTED);
						event.claimed = TRUE;
						}
				}
				else
				{
					SetState(DEFAULT_BUTTON_NORMAL);
				}
			}
			break;
		case MOUSE_UP:
			if ((!OnMe(event.x, event.y) && !mouseDown) || event.claimed)
				break;
			SetScript(NULL);
			if ((event.modifiers & LEFT_MOUSE) && mouseDown)
			{
				if (OnMe(event.x, event.y))
				{
					SetState(DEFAULT_BUTTON_HILIGHTED);
					DoLeftClick();
				}
				else
				{
					SetState(DEFAULT_BUTTON_NORMAL);
				}
				mouseDown = FALSE;
				event.claimed = TRUE;
			}

			if ((event.modifiers & RIGHT_MOUSE) && mouseDown)
			{
				if (OnMe(event.x, event.y))
				{
					SetState(DEFAULT_BUTTON_HILIGHTED);
					DoRightClick();
				}
				else
				{
					SetState(DEFAULT_BUTTON_NORMAL);
				}
				mouseDown = FALSE;
				event.claimed = TRUE;
			}
			//remove from  hog list
			pApp->pWindow->vRemoveEventHog(this);
//			plane->CaptureEvents(NULL);
			break;
		case KEY_CHAR:
			if ((tolower(event.c) == hotkey) || (tolower(event.c) == hotkey2))
			{
				SetState(DEFAULT_BUTTON_DEPRESSED);
				plane->window->ForceUpdate();
				Sleep(100);
				SetState(DEFAULT_BUTTON_NORMAL);
				DoLeftClick();
				event.claimed = TRUE;
			}
			else if (tolower(event.c) == rightHotkey)
			{
				SetState(DEFAULT_BUTTON_DEPRESSED);
				plane->window->ForceUpdate();
				Sleep(100);
				SetState(DEFAULT_BUTTON_NORMAL);
				DoRightClick();
				event.claimed = TRUE;
			}
			break;

		}

	//if mouse is down(and thus a hog) then we need to make us a 'modal' hog
	if (mouseDown == TRUE) {
		event.claimed = TRUE;
	}

	return event.claimed;
}

void sciButton::Enable()
{
	if (!enabled)
		{
		enabled = TRUE;
		mouseDown = FALSE;
		if (OnMe(plane->GetCursorX(), plane->GetCursorY()))
			SetState(DEFAULT_BUTTON_HILIGHTED);
		else
			SetState(DEFAULT_BUTTON_NORMAL);
		}
}

void sciButton::Disable()
{
	if (enabled)
		{
		enabled = FALSE;
		mouseDown = FALSE;
		SetScript(NULL);
		SetState(DEFAULT_BUTTON_DISABLED);
		}
}

Boolean sciButton::IsEnabled()
{
	return enabled;
}

void sciButton::DoLeftClick()
{
//	if (sigs)
//		sigs->Send(MENU_MESSAGE, (char *)&message, sizeof(message));
}

void sciButton::SetHotkey(char c)
{
	hotkey = tolower(c);
}

void sciButton::SetHotkey2(char c)
{
	hotkey2 = tolower(c);
}

void sciButton::SetRightHotkey(char c)
{
	rightHotkey = c;
}

void sciButton::SetState(int newState)
{
	state = newState;
	UpdateCel();
}

void sciButton::UpdateCel()
{
	switch (state)
	{
		case DEFAULT_BUTTON_NORMAL:
			if(text)
				text->SetFont(normalFont);
			SetCel(normal);
			break;
		case DEFAULT_BUTTON_DEPRESSED:
			if(text)
				text->SetFont(depressedFont);
			SetCel(depressed);
			break;
		case DEFAULT_BUTTON_DISABLED:
			SetCel(disabled);
			if(text)
				text->SetFont(disabledFont);
			break;
		case DEFAULT_BUTTON_HILIGHTED:
			if (sciButton::mpoCurrentHilight != this) {
				if (sciButton::mpoCurrentHilight != NULL) {
					sciButton::mpoCurrentHilight->SetState(DEFAULT_BUTTON_NORMAL);
				}
				sciButton::mpoCurrentHilight = this;
			}

			if(text)
				text->SetFont(hilightedFont);
				//text->SetFont(disabledFont);
			SetCel(hilighted);
			break;
	}
	pApp->pWindow->Update(rcWindow);
}

void sciButton::DoRightClick()
{
}

void sciButton::SetRepeat(int newStartDelay, int newRepeatDelay)
{
	startDelay = newStartDelay;
	repeatDelay = newRepeatDelay;
}

void sciButton::SetText(char *newText)
{
	if (text)
		delete text;
	
	text = new sciText();
	if (text) {
		text->plane = plane;
//		text->SetBackgroundColor(35);
		text->Init();
		text->SetText(newText);
		text->vSetSize();
		text->Posn(GetX(), GetY());
		text->SetWidth(GetWidth());
		text->SetHeight(GetHeight());
		text->SetJust(justification);
	}

	// TODO:  More stuff, e.g. resizing button rect.
}

void sciButton::SetFont(int newNormalFont, int newDepressedFont, int newHiFont, int newDisabledFont)
{
	normalFont = newNormalFont;

	// set up highlighted font when mouse moves over it
	if(newHiFont != -1)
		hilightedFont = newHiFont;
	else
		hilightedFont = newNormalFont;

	//set up font when button/key is pressed on it
	if(newDepressedFont != -1)
		depressedFont = newDepressedFont;
	else
		depressedFont = newNormalFont;

	// set up font when it is diabled
	if(newDisabledFont != -1)
		disabledFont = newDisabledFont;
	else
		disabledFont = newNormalFont;

	if (text)
		text->SetFont(newNormalFont);
}

void sciButton::ToggleFont(Boolean bHilighted)
{
	if(bHilighted)
		text->SetFont(hilightedFont);
	else
		text->SetFont(normalFont);
}

void sciButton::SetView(int newView)
{
	sciProp::SetView(newView);
	if(text) {
		text->SetWidth(GetWidth());
		text->SetHeight(GetHeight());
	}
}
void sciButton::SetPri(int newPri)
{
	sciProp::SetPri(newPri);
	if (text)
		text->SetPri(newPri + 1);
}

void sciButton::Posn(int x, int y)
{
	sciProp::Posn(x, y);
	if(text)
		text->Posn(x, y);
}

void sciButton::TextPosn(int addX, int addY)
{
	text->Posn(text->GetX() + addX, text->GetY() + addY);
}

void sciButton::SetTextJust(int newJust)
{
	justification = newJust;
	if (text)
		text->SetJust(newJust);
}

void sciButton::Show()
{
	sciProp::Show();
	if (text)
		text->Show();
		//text->vDraw();
}

void sciButton::Hide()
{
	mouseDown = FALSE;
	SetScript(NULL);
	sciProp::Hide();
	if (text)
		text->Hide();
	sciButton::mpoCurrentHilight = NULL;
}
