#include "os.h"
#include "scicycle.h"
#include "sciprop.h"
#include "scigame.h"
#include "sciscrip.h"
#include "game.h"

sciCycle::sciCycle(sciScript *newCaller)
	{
	client = NULL;
	caller = newCaller;
	clientLastCel = 0;
	cycleDir = 0;
	cycleCnt = game->GetTime();
	inMotion = FALSE;
	}

sciCycle::~sciCycle()
	{
	}

void sciCycle::Init(sciProp *newClient)
	{
	client = newClient;
	clientLastCel = client->LastCel();
	inMotion = TRUE;
	}

int sciCycle::NextCel()
	{
	if (abs(game->GetTime() - cycleCnt) < client->cycleSpeed)
		return client->GetCel();
	else
		{
		cycleCnt = game->GetTime();
		return client->GetCel() + cycleDir;
		}
	}

void sciCycle::Doit()
	{
	}

void sciCycle::CycleDone()
	{
	if (caller)
		caller->SetTicks(0);
	inMotion = FALSE;
	client->SetCycle(NULL);	// Detach and dispose of "this".
	}

sciForward::sciForward()
	: sciCycle(NULL)
	{
	cycleDir = CYCLEDIR_FORWARD;
	}

sciForward::~sciForward()
	{
	}

void sciForward::Doit()
	{
	int newCel;

	if (!inMotion)
		return;

	newCel = NextCel();
	if (client->GetCel() != newCel)
		{
		if (newCel > clientLastCel)
			{
			CycleDone();
			}
		else
			{
			client->SetCel(newCel);
			}
		}
		
	}

void sciForward::CycleDone()
	{
	client->SetCel(0);
	}

sciReverse::sciReverse()
	: sciCycle(NULL)
	{
	cycleDir = CYCLEDIR_REVERSE;
	}

sciReverse::~sciReverse()
	{
	}

void sciReverse::Doit()
	{
	int newCel;

	if (!inMotion)
		return;

	newCel = NextCel();
	if (client->GetCel() != newCel)
		{
		if (newCel < 0)
			{
			CycleDone();
			}
		else
			{
			client->SetCel(newCel);
			}
		}
		
	}

void sciReverse::CycleDone()
	{
	client->SetCel(clientLastCel);
	}

sciCycleTo::sciCycleTo(sciScript *newCaller)
	: sciCycle(newCaller)
	{
	endCel = 0;
	}

sciCycleTo::~sciCycleTo()
	{
	}

void sciCycleTo::Doit()
	{
	int newCel;

	if (!inMotion)
		return;

	newCel = NextCel();
	if (client->GetCel() != newCel)
		{
		if (newCel == endCel)
			{
			client->SetCel(endCel);
			CycleDone();
			}
		else
			{
			client->SetCel(newCel);
			}
		}
		
	}

sciBegLoop::sciBegLoop(sciScript *newCaller)
	: sciCycleTo(newCaller)
	{
	}

void sciBegLoop::Init(sciProp *newClient)
	{
	sciCycle::Init(newClient);
	endCel = 0;
	cycleDir = CYCLEDIR_REVERSE;
	}

sciEndLoop::sciEndLoop(sciScript *newCaller)
	: sciCycleTo(newCaller)
	{
	}

void sciEndLoop::Init(sciProp *newClient)
	{
	sciCycle::Init(newClient);
	endCel = clientLastCel;
	cycleDir = CYCLEDIR_FORWARD;
	}

sciOscillate::sciOscillate()
	: sciCycle(NULL)
	{
	cycleDir = CYCLEDIR_FORWARD;
	}

sciOscillate::~sciOscillate()
	{
	}

void sciOscillate::Doit()
	{
	int newCel;

	if (!inMotion)
		return;

	newCel = NextCel();
	if (client->GetCel() != newCel)
		{
		if ((newCel > clientLastCel) || (newCel < 0))
			{
			CycleDone();
			}
		else
			{
			client->SetCel(newCel);
			}
		}
		
	}

void sciOscillate::CycleDone()
	{
	if (cycleDir == CYCLEDIR_FORWARD)
		cycleDir = CYCLEDIR_REVERSE;
	else
		cycleDir = CYCLEDIR_FORWARD;
	}

