#ifndef _sciplane
#define _sciplane

#include "sciobj.h"
//#include "scilist.h"
#include "list.h"
#include "window.h"

class sciFeature;
class sciEvent;
class sciCursor;
class sciProp;

class sciPlane : public sciObject
	{
	public:
		sciPlane(Window *parentWindow, Boolean bModal = FALSE, int nPriority = 0);
		~sciPlane();
		sciPlane *poGetPlane() {return this;}
		void AddToCast(sciFeature *feature);
		void DeleteFromCast(sciFeature *feature);
		void AddToProps(sciProp *prop);
		void DeleteFromProps(sciProp *prop);
		virtual Boolean HandleEvent(sciEvent &event);
		virtual void Doit();
		virtual void Render(RECT &updateRect);
		virtual void vMarkUpdate() {window->Update(mrClip);}
		void Show();
		void Hide();
		void SetPic(int picNumber);
		void vSetBackgroundColor(BYTE jColor);
		void vSetBackgroundColor(BYTE jRed, BYTE jGreen, BYTE jBlue);
		void Posn(int x, int y, BOOL updateWindow = TRUE);
		int GetX();
		int GetY();
		int GetRight() {return right;}
		int GetBottom() {return bottom;}
		int GetCursorX();
		int GetCursorY();
		Boolean IsModal();
		void vSetPri(int nPriority);
		int nGetPri(void) {return mnPriority;};

		Boolean OnMe(int onMeX, int onMeY);	//doesn't care if the points are localized to the plane
		Boolean OnMeLocalized(int onMeX, int onMeY); //localizes the points first, then does an OnMe check
		Boolean OnMeGlobalized(int onMeX, int onMeY); //globalizes the points first, then does an OnMe check

		void vSetRect(int nLeft, int nTop, int nRight, int nBottom);
		void vSetRect(RECT &r) {
			vSetRect(r.left, r.top, r.right, r.bottom);
		};

		RECT *prGetClip() {return &mrClip;};

		// we're inlining these for speed
		void vLocalToGlobal(RECT *prThis) {
			assert(prThis);
			prThis->top += top;
			prThis->bottom += top;
			prThis->left += left;
			prThis->right += left;
		};
		void vLocalToGlobal(int *pnX, int *pnY) {
			assert(pnX && pnY);
			*pnX += left;
			*pnY += top;
		};
		void vGlobalToLocal(RECT *prThis) {
			assert(prThis);
			prThis->top -= top;
			prThis->bottom -= top;
			prThis->left -= left;
			prThis->right -= left;
		};
		void vGlobalToLocal(int *pnX, int *pnY) {
			assert(pnX && pnY);
			*pnX -= left;
			*pnY -= top;
		};

		int nWidth(void) {return mnWidth;}
		int nHeight(void) {return mnHeight;}

		sciCursor *cursor;
		Window	*window;

	protected:
		void LoadPic();
		void vCalcClip();

		int		top;
		int		left;
		int		bottom;
		int		right;
		int		mnWidth;
		int		mnHeight;

		RefList<sciFeature>	cast;	// Used to pass HandleEvents()s along.
		RefList<sciProp> props;	// Used to pass Doit()s along.
		int		isNotHidden;
		int		picNumber;
		int		mnPriority; // used for plane sorting
		Boolean	itsMyWindow;
		Boolean mbModal;

		RECT	mrClip;	// clipping area (combo of self and screen)

		Bitmap *pBackgroundBitmap;
		Bitmap *pBackgroundColor;


	};

#endif // _sciplane
