#include "os.h"
#include "scitext.h"
#include "app.h"
#include "resman.h"
#include "viewbmp.h"
#include "sciplane.h"

extern Application *pApp;

sciText::sciText()
	{
	fontNumber = -1;
	text = NULL;
	fontBitmap = NULL;
	backgroundColor = 0;
	caretX = 0;
	caretY = 0;
	SetWidth(128);
	SetHeight(64);
	SetJust(JUST_LEFT | JUST_TOP);
	SetFont(2);
	}

sciText::~sciText()
	{
	if (text)
		delete[] text;

	// If the font has already been set, undo the stuff.
	if (fontBitmap)
		{
		resourceManager->FreeResource(fontBitmap);
		fontBitmap = NULL;
		}
	if (bitmap)
		{
		delete bitmap;
		bitmap = NULL;
		}
	}

void sciText::SetWidth(int newWidth)
	{
	if (bitmap)
		{
		delete bitmap;
		bitmap = NULL;
		}
	width = newWidth;
	sciView::SetWidth(width);
	RenderText();
	}

void sciText::SetHeight(int newHeight)
	{
	if (bitmap)
		{
		delete bitmap;
		bitmap = NULL;
		}
	height = newHeight;
	sciView::SetHeight(height);
	RenderText();
	}

void sciText::SetFont(int newFont)
	{
	Resource *resource;
	FONT *pFont;
	int i;

	if (fontNumber == newFont)
		return;

	// Get the font properties.
	resource = resourceManager->LoadResource(RES_FONT, newFont);
	if ((!resource) || (!resource->GetData()))
		{
		newFont = 2;
		resource = resourceManager->LoadResource(RES_FONT, newFont);
		}
	fontNumber = newFont;
	pFont = (FONT *)(resource->GetData());
	fontResource.bitmapID = pFont->bitmapID;
	fontResource.shadowWidth = pFont->shadowWidth;
	fontResource.shadowHeight = pFont->shadowHeight;
	fontResource.characterSpacing = pFont->characterSpacing;
	fontResource.lineSpacing = pFont->lineSpacing;
	fontResource.numChars = pFont->numChars;
	for (i = 0; i < fontResource.numChars; ++i)
		{
		fontResource.theChars[i] = pFont->theChars[i];
		}
	resourceManager->FreeResource(resource);

	// If the font has already been set, undo the stuff.
	if (fontBitmap)
		{
		resourceManager->FreeResource(fontBitmap);
		fontBitmap = NULL;
		}

	// Load the bitmap containing the font.
	fontBitmap = (Bitmap*)resourceManager->LoadResource(RES_BITMAP, fontResource.bitmapID);

	RenderText();
	}

void sciText::SetText(char *newText)
	{
	int strLen;

	if (text && newText && !strcmp(text, newText))
		return;
	if (text)
		delete[] text;
	if (newText)
		{
		strLen = strlen(newText);
		text = new char[strLen + 1];
		strcpy(text, newText);
		}
	RenderText();
	}

char* sciText::GetText()
	{

	if (text)
		return text;
	else
		return "";
	}

void sciText::SetBackgroundColor(BYTE newColor)
	{
	backgroundColor = newColor;
	RenderText();
	}

void sciText::RenderText()
	{
	int strLen;
	RECT sourceRect, destRect;
	int leftPosition, topPosition;
	int i;
	int sizeWidth, sizeHeight;

	if (!(text && fontBitmap))
		return;

	// Create a bitmap to hold the rendered text.
	if (!bitmap)
		bitmap = new Bitmap(width, height);

	// Justify
	CalculateSize(text, &sizeWidth, &sizeHeight);
	if (just & JUST_LEFT)
		leftPosition = 0;
	if (just & JUST_RIGHT)
		leftPosition = width - sizeWidth;
	if (just & JUST_HORIZONTAL)
		leftPosition = (width - sizeWidth) / 2;
	if (just & JUST_TOP)
		topPosition = 0;
	if (just & JUST_BOTTOM)
		topPosition = height - sizeHeight;
	if (just & JUST_VERTICAL)
		topPosition = (height - sizeHeight) / 2;

	destRect.left = 0;
	destRect.top = 0;
	destRect.right = bitmap->Width();
	destRect.bottom = bitmap->Height();
	bitmap->vErase8(destRect, backgroundColor);
	strLen = strlen(text);
	for (i = 0; i < strLen; ++i)
		{
		sourceRect.top = fontResource.theChars[text[i]].top;
		sourceRect.left = fontResource.theChars[text[i]].left;
		sourceRect.bottom = fontResource.theChars[text[i]].bottom + 1 +
			fontResource.shadowHeight;
		sourceRect.right = fontResource.theChars[text[i]].right + 1 + 
			fontResource.shadowWidth;
		destRect.left = leftPosition;
		destRect.top = topPosition;
		destRect.right = leftPosition + (sourceRect.right - sourceRect.left);
		destRect.bottom = topPosition + (sourceRect.bottom - sourceRect.top);

		fontBitmap->CopyPixels(bitmap, destRect, sourceRect, 1);
		leftPosition += (sourceRect.right - sourceRect.left) -
			fontResource.shadowWidth + fontResource.characterSpacing;
		if ((text[i+1] == '\n') || (leftPosition +
				(fontResource.theChars[text[i+1]].right - 
				fontResource.theChars[text[i+1]].left -
				fontResource.shadowWidth + 
				fontResource.characterSpacing + 1) >
				bitmap->Width()))
			{
			leftPosition = 0;
			topPosition += fontResource.lineSpacing;
			CalculateSize(text + i + 1, &sizeWidth, &sizeHeight);
			if (just & JUST_LEFT)
				leftPosition = 0;
			if (just & JUST_RIGHT)
				leftPosition = width - sizeWidth;
			if (just & JUST_HORIZONTAL)
				leftPosition = (width - sizeWidth) / 2;
			}
		}
	sourceRect.top = GetY();
	sourceRect.left = GetX();
	sourceRect.bottom = sourceRect.top + bitmap->Height();
	sourceRect.right = sourceRect.left + bitmap->Width();

	top = sourceRect.top;
	left = sourceRect.left;
	bottom = sourceRect.bottom;
	right = sourceRect.right;

	SetRect(&rcWindow, left, top, right, bottom);

	if (inited && IsNotHidden()) {
		//plane->window->Update(sourceRect);
		bitmap->vDraw(left, top, plane);
	}

	caretX = leftPosition;
	caretY = topPosition;
	}

void sciText::CalculateSize(char *theText, int *theWidth, int *theHeight)
	{
	int leftPosition, topPosition;
	int strLen;
	int i;
	int characterWidth;

	*theWidth = 0;
	*theHeight = 0;
	leftPosition = 0;
	topPosition = 0;
	strLen = strlen(theText);
	for (i = 0; i < strLen; ++i)
		{
		characterWidth = fontResource.theChars[theText[i]].right + 1 + 
			fontResource.shadowWidth -
			fontResource.theChars[theText[i]].left;
		leftPosition += characterWidth -
			fontResource.shadowWidth + fontResource.characterSpacing;
		if (leftPosition  > *theWidth)
			*theWidth = leftPosition;
		if ((theText[i+1] == '\n') || (leftPosition +
				(fontResource.theChars[theText[i+1]].right - 
				fontResource.theChars[theText[i+1]].left -
				fontResource.shadowWidth + 
				fontResource.characterSpacing + 1) >
				width))
			{
			leftPosition = 0;
			topPosition += fontResource.lineSpacing;
			}
		}
	*theHeight = topPosition + fontResource.lineSpacing;
	*theWidth -= fontResource.shadowWidth;
	}

void sciText::SetJust(int newJust)
	{
	just = newJust;
	RenderText();
	}

int sciText::GetCaretX()
	{
	return caretX;
	}

int sciText::GetCaretY()
	{
	return caretY;
	}

void sciText::Init()
	{
	Bitmap *savedBitmap;
	savedBitmap = bitmap;
	bitmap = NULL;
	sciView::Init();
	hidden = FALSE; //assume it's not hidden at the beginning	
	bitmap = savedBitmap;
	RenderText();

	}

void sciText::vRefresh() { 
	if (bitmap && IsNotHidden()) {
		bitmap->vDraw(left, top, plane);
	}
}

//Sets the size of its fields(width,height) based upon the length of the longest line within 
//the 'text' string field. 
void sciText::vSetSize() {
	if (text) {
		int sizeWidth, sizeHeight;
		//temporarily expand width/height
		width = 640;
		height= 480;
		CalculateSize(text, &sizeWidth, &sizeHeight);
		SetWidth(sizeWidth);
		int nTop = max(fontResource.theChars['y'].top, fontResource.theChars['Y'].top);
		int nBot = max(fontResource.theChars['y'].bottom, fontResource.theChars['Y'].bottom);
		sizeHeight += (int)((nBot - nTop) * .3);	//add a little padding
		SetHeight(sizeHeight);
	}
}

// if nNumCharacters is passed, an 'input line' is being referenced else a standard 'text'
// line is being referenced. Calculate in the oPoint structure the sizes.
void sciText::vGetSize(POINT &oPoint, int nNumCharacters) {
	oPoint.x = 0;
	oPoint.y = 0;

	if (nNumCharacters) {
		//we're checking for an 'input line' 
		int nRight = fontResource.theChars['W'].right;
		int nLeft = fontResource.theChars['W'].left;
		int nTop = fontResource.theChars['y'].top;
		int nBot = fontResource.theChars['y'].bottom;
		oPoint.x = (int)(((nRight - nLeft) * nNumCharacters) * .75);
		oPoint.y = (nBot - nTop) + (int)((nBot - nTop) * .3);
	}
	else {
		//we're checking a set 'text' field
		if (text) {
//			int sizeWidth, sizeHeight;
//			CalculateSize(text, &sizeWidth, &sizeHeight);
//			oPoint.x = sizeWidth;
//			oPoint.y = sizeHeight;
			oPoint.x = width;
			oPoint.y = height;
		}
	}
}

void sciText::vDraw() {
	if (bitmap && IsNotHidden()) {
		bitmap->vDraw(left,top,plane);
	}
}

//returns the height of a given character
int sciText::nCharHeight(char cChar)
{
	int nTop = fontResource.theChars[cChar].top;
	int nBot = fontResource.theChars[cChar].bottom;
	return nBot - nTop + fontResource.shadowHeight;
}

//returns the width of a given character
int sciText::nCharWidth(char cChar)
{
	int nRight = fontResource.theChars[cChar].right;
	int nLeft = fontResource.theChars[cChar].left;
	return nRight - nLeft;
}
