#include "os.h"
#include "sciview.h"
#include "sciplane.h"
#include "app.h"
#include "resman.h"
#include "viewbmp.h"

extern Application *pApp;

sciView::sciView()
	{
	view = 0;
	loop = 0;
	cel = 0;
	bitmap = NULL;

	// Someone else has to delete resourceManager;
	resourceManager = pApp->pResourceManager;
	}

sciView::~sciView()
	{
	if (bitmap)
		{
		resourceManager->FreeResource(bitmap);
		bitmap = NULL;
		}

	// Someone else has to delete resourceManager;
	}

void sciView::Init()
	{
	sciFeature::Init();
	if (bitmap)
		resourceManager->FreeResource(bitmap);
	bitmap = NULL;
	if (view)
		{
		bitmap = (Bitmap*)resourceManager->LoadResource(RES_BITMAP, view);
		if (bitmap)
			{
			SetRect(&rcWindow, left, top, left + bitmap->Width(), top + bitmap->Height());
			}
		top = rcWindow.top;
		left = rcWindow.left;
		bottom = rcWindow.bottom;
		right = rcWindow.right;
		}
	if (cel > LastCel())
		cel = LastCel();
	if (cel < 0)
		cel = 0;
	}

void sciView::Posn(int posnX, int posnY)
	{
	if (inited && IsNotHidden())
		{
		plane->window->Update(rcWindow);
		}
	sciFeature::Posn(posnX, posnY);
	SetRect(&rcWindow, left, top, right, bottom);
	if (inited && IsNotHidden())
		{
		plane->window->Update(rcWindow);
		}
	}

Boolean sciView::OnMe(int onMeX, int onMeY)
{
	if (!bitmap)
		return sciFeature::OnMe(onMeX, onMeY);

	onMeX += (bitmap->nOriginX);
	onMeY += (bitmap->nOriginY);

	if (IsNotHidden() &&
		((onMeX >= left) && (onMeX <= right)) &&
		((onMeY >= top) && (onMeY <= bottom))
		)
		return (TRUE);
	else
		return (FALSE);
}

void sciView::SetView(int newView)
	{
	if (view != newView)
		{
		view = newView;
		if (inited)
			{
			plane->window->Update(rcWindow);
			Init();
			if (IsNotHidden())
				{
				plane->window->Update(rcWindow);
				}
			}
		}
	else
		{
		view = newView;
		}
	}

void sciView::SetLoop(int newLoop)
	{
	loop = newLoop;
	if (bitmap)
		{
		bitmap->SetLoop(loop);
		}
	}

void sciView::SetCel(int newCel)
	{
	if (cel != newCel)
		{
		if (inited)
			{
			if (newCel > LastCel())
				newCel = LastCel();
			if (newCel < 0)
				newCel = 0;
			plane->window->Update(rcWindow);
			}
		}
	cel = newCel;
	if (bitmap)
		{
		bitmap->SetCel(cel);
		}
	}

int sciView::GetCel()
	{
	return cel;
	}

int sciView::GetLoop()
	{
	return loop;
	}

int sciView::GetView()
	{
	return view;
	}

void sciView::Hide()
	{
	if (!hidden)
		{
		sciFeature::Hide();
		if (inited)
			{
			plane->window->Update(rcWindow);
			}
		}
	}

void sciView::Show()
	{
	if (hidden)
		{
		sciFeature::Show();
		if (inited)
			{
			plane->window->Update(rcWindow);
			}
		}
	}

int sciView::LastCel()
	{
	if ((bitmap == NULL) || (bitmap->Height() == 0))
		return 0;
	else
		return bitmap->GetNumCels() - 1;
	return 0;
	}

int sciView::LastLoop()
	{
	return bitmap->GetNumLoops() - 1;
	}

void sciView::Render(RECT &updateRect)
{
#if 0
	RECT my;
	RECT clippedUpdateRect;

	clippedUpdateRect = rcWindow;
	plane->vLocalToGlobal(&clippedUpdateRect);

	IntersectRect(&clippedUpdateRect, &clippedUpdateRect, &updateRect);
	if (bitmap && plane && IsNotHidden())
		{
		my.left = clippedUpdateRect.left - rcWindow.left;
		my.top = clippedUpdateRect.top - rcWindow.top;
		my.right = my.left + (clippedUpdateRect.right - clippedUpdateRect.left);
		my.bottom = my.top + (clippedUpdateRect.bottom - clippedUpdateRect.top);
		bitmap->CopyPixels(plane->window->pWorkBitmap, clippedUpdateRect, my, TRUE);
		}
#endif
	
	//new way of drawing
	if (bitmap && plane && IsNotHidden())
		bitmap->vDraw(left,top,plane);

}
void sciView::vSetFlip(Boolean bFlip)
{
	if (bitmap)
		bitmap->vSetFlip(bFlip);
}

int sciView::nBitmapWidth(void)
{
	return bitmap ? bitmap->Width()  : 0;
}

int sciView::nBitmapHeight(void)
{
	return bitmap ? bitmap->Height() : 0;
}
