#include "os.h"

#include "app.h"
#include "window.h"
#include "sigs.h"
#include "scigame.h"
#include "scisound.h"
#include "rc.h"
#include "util.h"
#include "chat.h"
#include <time.h>
#include "game.h"

extern Application *pApp;
extern char *szAppName;
extern SendChatButton *sendChatButton;
//extern BOOL netDemo;
extern BOOL startNet;
extern char *szAppTitle;
extern char versionText[32];

#define HEADER_DATATYPE unsigned short

typedef int (*SNWValidateFunc)(char *x, char *y, long *xx, long *yy, char *z1, char *z2);
extern SNWValidateFunc SNWValidate;
typedef int (*InitializeDLLFunc)(char *GPServerIPAddr, long inPort, long outPort);
extern InitializeDLLFunc InitializeDLL;
typedef int (*CloseDLLFunc)(void);
extern CloseDLLFunc CloseDLL;
typedef int (*SendTCPMessageFunc)(BYTE *data, HEADER_DATATYPE datalengthinbytes);
extern SendTCPMessageFunc SendTCPMessage;
typedef int (*CheckForAnyTCPMessageFunc)();
extern CheckForAnyTCPMessageFunc CheckForAnyTCPMessage;
typedef int (*PeekForTCPMessageFunc)(BYTE *data, HEADER_DATATYPE buflen, int playerindex);
extern PeekForTCPMessageFunc PeekForTCPMessage;
typedef int (*RecvTCPMessageFunc)(BYTE *data, HEADER_DATATYPE buflen, int playerindex);
extern RecvTCPMessageFunc RecvTCPMessage;
typedef int (*GetNumberOfPlayersFunc)(LPVOID data);
extern GetNumberOfPlayersFunc GetNumberOfPlayers;
typedef int (*GetMyPlayerNumberFunc)();
extern GetMyPlayerNumberFunc GetMyPlayerNumber;
typedef int (*GetPlayerXNameFunc)(char *data, HEADER_DATATYPE buflen);
extern GetPlayerXNameFunc GetPlayerXName;
typedef int (*CheckGameConnectStatusFunc)();
extern CheckGameConnectStatusFunc CheckGameConnectStatus;
typedef int (*GameCloseFunc)();
extern GameCloseFunc GameClose;
typedef int (*GetGameOptionFunc)(char *option, char *value);
extern GetGameOptionFunc GetGameOption;
typedef int (*CheckPlayerConnectStatusFunc)(int playerIndex);
extern CheckPlayerConnectStatusFunc CheckPlayerConnectStatus;

SNWValidateFunc SNWValidate = NULL;
InitializeDLLFunc InitializeDLL = NULL;
CloseDLLFunc CloseDLL = NULL;
SendTCPMessageFunc SendTCPMessage = NULL;
CheckForAnyTCPMessageFunc CheckForAnyTCPMessage = NULL;
PeekForTCPMessageFunc PeekForTCPMessage = NULL;
RecvTCPMessageFunc RecvTCPMessage = NULL;
GetNumberOfPlayersFunc GetNumberOfPlayers = NULL;
GetMyPlayerNumberFunc GetMyPlayerNumber = NULL;
GetPlayerXNameFunc GetPlayerXName = NULL;
CheckGameConnectStatusFunc CheckGameConnectStatus = NULL;
GameCloseFunc GameClose = NULL;
GetGameOptionFunc GetGameOption = NULL;
CheckPlayerConnectStatusFunc CheckPlayerConnectStatus = NULL;

SIGS *sigs;

SIGS::SIGS(char *gpName)
	{
	int i;
	char GPServerIP[18];
	long GPInPort, GPOutPort;
	DWORD dw;
	char exePath[MAX_PATH];
	char relaunchCommand[MAX_PATH];
	int nIndex;
	char windowText[64];

	sigs = this;

	// Null out all the pointers.
	validateDLL = 0;
	sigsDLL = 0;
	FreeSIGS();

	playerToListenTo = -1;
	sprintf(windowText, "%s %s", szAppTitle, versionText);
	//SetWindowText(pApp->pWindow->hWnd, windowText);
	sendMessageNumber = 0;
	receiveMessageNumber = 0;
	numPlayers = 0;
	playerNum = 0;
	HoldMessages(TRUE);
	for (i = 0; i < MAX_NUM_PLAYERS; ++i)
		{
		livePlayers[i] = FALSE;
		networkPlayerInfo[i].name[0] = 0;
		networkPlayerInfo[i].version[0] = 0;
		}
	numLivePlayers = 0;

	// Load the Validate DLL
	validateDLL = LoadLibrary("SNWValid.DLL");
	if (validateDLL == 0)
		{
		// Error
		MessageBox(NULL, "Internet features can't be found. Re-run Setup.", szAppName, MB_OK | MB_TASKMODAL);
		FreeSIGS();
		return;
		}
	SNWValidate			=
		(SNWValidateFunc)GetProcAddress(validateDLL, "SNWValidate");
	nIndex = ::GetModuleFileName(NULL, exePath, MAX_PATH);
	if (nIndex > 0)
		{
		// Make relaunch command.
		strcpy(relaunchCommand, exePath);
//		if (netDemo)
//			strcat(relaunchCommand, " -netDemo");
//		else
		if (!stricmp(gpName, "Blackjack"))
			strcat(relaunchCommand, " -startNetBlackjack");
		else if (!stricmp(gpName, "Poker"))
			strcat(relaunchCommand, " -startNetPoker");

		// Remove executable from the path.
		while ((nIndex > 0) && (exePath[nIndex] != '\\'))
			--nIndex;
		exePath[nIndex] = '\\';
		exePath[nIndex+1] = '\0';
		}
	else
		{
		relaunchCommand[0] = 0;
		exePath[0] = 0;
		}
	i = SNWValidate(gpName, GPServerIP, &GPInPort, &GPOutPort, exePath, relaunchCommand);

	// Free the Validate DLL, we're finished with it.
	if (validateDLL)
		{
		FreeLibrary(validateDLL);
		validateDLL = 0;
		}
	SNWValidate = NULL;

	switch (i)
		{
		case 0:
			break;
		case -1:
			// Error
			FreeSIGS();
			return;
		case -2:
			// Product needs patch, exit now!
			FreeSIGS();
			game->QuitGame(TRUE);
			return;
		}

	// Load the SIGS DLL
	sigsDLL = LoadLibrary("SierraNW.DLL");
	if (sigsDLL == 0)
		{
		// Error
		dw = GetLastError();
		switch ((dw & 0xFFFF))
			{
			case ERROR_DLL_NOT_FOUND:
				MessageBox(NULL, "SierraNW.DLL could not be found.", "Network Error", MB_OK | MB_TASKMODAL);
				break;
			default:
				MessageBox(NULL, "Unknown error loading SierraNW.DLL", "Network Error", MB_OK | MB_TASKMODAL);
				break;
			}
		FreeSIGS();
		return;
		}

	// Get pointers to all the used functions.
	InitializeDLL =
		(InitializeDLLFunc)GetProcAddress(sigsDLL, "InitializeDLL");
	CloseDLL =
		(CloseDLLFunc)GetProcAddress(sigsDLL, "CloseDLL");
	SendTCPMessage =
		(SendTCPMessageFunc)GetProcAddress(sigsDLL, "SendTCPMessage");
	CheckForAnyTCPMessage =
		(CheckForAnyTCPMessageFunc)GetProcAddress(sigsDLL, "CheckForAnyTCPMessage");
	PeekForTCPMessage =
		(PeekForTCPMessageFunc)GetProcAddress(sigsDLL, "PeekForTCPMessage");
	RecvTCPMessage =
		(RecvTCPMessageFunc)GetProcAddress(sigsDLL, "RecvTCPMessage");
	GetNumberOfPlayers =
		(GetNumberOfPlayersFunc)GetProcAddress(sigsDLL, "GetNumberOfPlayers");
	GetMyPlayerNumber =
		(GetMyPlayerNumberFunc)GetProcAddress(sigsDLL, "GetMyPlayerNumber");
	GetPlayerXName =
		(GetPlayerXNameFunc)GetProcAddress(sigsDLL, "GetPlayerXName");
	CheckGameConnectStatus =
		(CheckGameConnectStatusFunc)GetProcAddress(sigsDLL, "CheckGameConnectStatus");
	GameClose =
		(GameCloseFunc)GetProcAddress(sigsDLL, "GameClose");
	GetGameOption =
		(GetGameOptionFunc)GetProcAddress(sigsDLL, "GetGameOption");
	CheckPlayerConnectStatus =
		(CheckPlayerConnectStatusFunc)GetProcAddress(sigsDLL, "CheckPlayerConnectStatus");

	// Start the SIGS DLL.
	i = InitializeDLL(GPServerIP, GPInPort, GPOutPort);
	if (i != 1)
		{
		// Error
		FreeSIGS();
		return;
		}
	networkOK = TRUE;
	}

SIGS::~SIGS()
	{
	sigs = NULL;
	//SetWindowText(pApp->pWindow->hWnd, szAppTitle);
	if (NetworkOK())
		{
		// Free the DLL.
		if (CloseDLL)
			CloseDLL();
		FreeSIGS();
		}
//	if (netDemo)
//		game->QuitGame();
	}

void SIGS::Send(MESSAGE_TYPE messageType, char *buffer, int bufferLen)
	{
	int i;
	BYTE *sendBuff;

	if (!NetworkOK())
		{
		// Error! This shouldn't happen.
		return;
		}

	sendBuff = new BYTE[bufferLen + 5];
	*(int*)(sendBuff) = bufferLen;
	sendBuff[4] = (BYTE)messageType;
	if (bufferLen)
		memcpy(sendBuff+5, buffer, bufferLen);
	i = SendTCPMessage(sendBuff, bufferLen + 5);
	delete[] sendBuff;
	++sendMessageNumber;
	}

void SIGS::Pump()
	{
	int i, j;
	BYTE *buff;
	BYTE *messageBuff;
	int messageType;
	int numBytes;
	static int lastPeekTime = 0;
	Message *message;
	char whosIn[256];

	if (!NetworkOK())
		return;

	// Check to see if the Network connection is still valid.
	switch (Status())
		{
		case IN_GAME:
			// Game in progress.
			if (firstTimeInGame)
				{
				numPlayers = GetNumberOfPlayers(whosIn);
				playerNum = GetMyPlayerNumber();
				firstTimeInGame = FALSE;
				for (i = 0; i < numPlayers; ++i)
					{
					livePlayers[i] = TRUE;
					GetPlayerXName(networkPlayerInfo[i].name, i);
					}
				numLivePlayers = numPlayers;
				}
			// Check if player that is playing is still alive.
			// Only once in a while
			if ((game->GetTime() > lastPeekTime + 60) && !holdMessages)
				{
				lastPeekTime = game->GetTime();
				if (playerToListenTo != -1)
					{
					CheckPlayerStatusNow(playerToListenTo);
					}
				else
					{
					for (i = 0; i < numPlayers; ++i)
						{
						if (livePlayers[i])
							CheckPlayerStatusNow(i);
						}
					}
				}
			break;
		case IN_GP:
			firstTimeInGame = TRUE;
			return;
		case DISCONNECTED:
			// GP shut down.
			return;
		}

	switch (CheckForAnyTCPMessage())
		{
		case -2:
			// No Data
			break;
		case -1:
			// Error - Unload other network players.
			break;
		case 0:
			// Data
			// There's data waiting.
			for (i = 0; i < MAX_NUM_PLAYERS; ++i)
				{
				// Only peek on valid characters.
				if (livePlayers[i] && (i != playerNum))
					{
					j = PeekForTCPMessage((BYTE*)&numBytes, 4, i);
					}
				else
					j = -2;
				switch (j)
					{
					case -1:
						// Error
						//RemovePlayer(i);
						break;
					case -2:
						// No data.
						break;
					default:
						// Data
						++receiveMessageNumber;

						buff = new BYTE[numBytes+5];
						RecvTCPMessage(buff, numBytes+5, i);
						messageType = *(BYTE*)(buff+4);
						messageBuff = buff+5;
						message = new Message;
						message->from = i;
						message->type = messageType;
#ifdef _DEBUG
						char debugBuff1[64];
						wsprintf(debugBuff1, "%d: Receiving Message(%d) #%d\n", i, message->type, receiveMessageNumber);
						DebugString(debugBuff1);
#endif // _DEBUG
						if (numBytes)
							{
							message->data = new BYTE[numBytes];
							memcpy(message->data, messageBuff, numBytes);
							}
						else
							{
							message->data = NULL;
							}
						listOfMessages.AddToEnd(message);
						delete buff;
						break;
					}
				}
			break;
		}

	CheckForAMessage();
	}

void SIGS::HoldMessages(Boolean hold)
	{
#ifdef _DEBUG
	if (holdMessages && !hold)
		DebugString("Not Holding Messages.\n");
	if (!holdMessages && hold)
		DebugString("Holding Messages.\n");
#endif // _DEBUG
	holdMessages = hold;
	}

Boolean SIGS::NetworkOK()
	{
	return networkOK;
	}

void SIGS::FreeSIGS()
	{
	if (validateDLL)
		FreeLibrary(validateDLL);
	if (sigsDLL)
		FreeLibrary(sigsDLL);
	SNWValidate = NULL;
	InitializeDLL = NULL;
	CloseDLL = NULL;
	SendTCPMessage = NULL;
	CheckForAnyTCPMessage = NULL;
	PeekForTCPMessage = NULL;
	RecvTCPMessage = NULL;
	GetNumberOfPlayers = NULL;
	GetMyPlayerNumber = NULL;
	GetPlayerXName = NULL;
	CheckGameConnectStatus = NULL;
	GameClose = NULL;
	GetGameOption = NULL;
	CheckPlayerConnectStatus = NULL;
	networkOK = FALSE;
	firstTimeInGame = TRUE;
	}

void SIGS::CheckPlayerStatusNow(int i)
	{
	int j;

	// Only peek on valid characters.
	if ((i < MAX_NUM_PLAYERS) && livePlayers[i] && (i != playerNum))
		j = CheckPlayerConnectStatus(i);
	else
		j = -2;
	switch (j)
		{
		case 0:
			// I don't know what 0 stands for.
			break;
		case -1:
			// Error
			RemovePlayer(i);
			break;
		case -2:
			break;
		}
	}

void SIGS::CheckForAMessage()
	{
	Message *message;
	int i;
	int anotherMessage;

	i = 0;
	anotherMessage = TRUE;
	message = NULL;
	while ((i < listOfMessages.Size()) && anotherMessage)
		{
		message = (Message*)(listOfMessages.At(i));
		if (message)
			{
			if ((message->type >= FIRST_BLACKJACK_MESSAGE) &&
				(message->type <= LAST_BLACKJACK_MESSAGE))
				{
				}
				
			else if ((message->type >= FIRST_QUEUED_MESSAGE) &&
					(message->type < FIRST_NON_QUEUED_MESSAGE))
				{
				// Its a queued message. Check for holdMessages.
				if (holdMessages || (numLivePlayers == 0))
					{
					// Leave this message queued.
					message = NULL;
					++i;
					}
				else
					{
					if ((playerToListenTo == -1) ||
						(playerToListenTo == message->from))
						{
						// Process this message.
						anotherMessage = FALSE;
						}
					else
						{
						// Leave this message queued.
						message = NULL;
						++i;
						}
					}
				}
			else if ((message->type >= FIRST_NON_QUEUED_MESSAGE))
				{
				// Its a non-queued message.
				// Process any other message.
				anotherMessage = FALSE;
				}
			}
		else
			{
			anotherMessage = FALSE;
			}
		}

	if (!message)
		return;

	listOfMessages.Delete(message);

	switch (message->type)
		{
		case MENU_MESSAGE:
		case CHAT_MESSAGE:
		case MYVERSION_MESSAGE:
		case DECK_MESSAGE:
		case INSURANCE_MESSAGE:
		case PLAYSOUND_MESSAGE:
			ProcessAMessage(message);
			break;
		default:
			ProcessTheMessage(message);
			break;
		}

	if (message->data)
		delete[] message->data;
	delete message;
	}

void SIGS::ProcessAMessage(Message *message)
	{
	//char str[128];

	switch (message->type)
		{
		case MENU_MESSAGE:
			break;
		case CHAT_MESSAGE:
			if (sendChatButton)
				sendChatButton->AddAChatLine((char*)(message->data));
			break;
		case MYVERSION_MESSAGE:
			HandleMyVersion(message->from, (MYVERSION_STRUCT*)message->data);
			break;
		case DECK_MESSAGE:
			break;
		case PLAYSOUND_MESSAGE:
			game->sound1->Play(*(int*)message->data);
			break;
		}
	}

GP_CONNECT_STATUS SIGS::Status()
	{
	if (CheckGameConnectStatus)
		{
		switch (CheckGameConnectStatus())
			{
			case 1:
				return IN_GAME;
			case 0:
				return IN_GP;
			case -1:
			default:
				return DISCONNECTED;
			}
		}
	else
		{
		return DISCONNECTED;
		}
	}

void SIGS::HandleMyVersion(int playerNum, MYVERSION_STRUCT *myVersion)
	{
	}

void SIGS::RemovePlayer(int i)
	{
	if (livePlayers[i])
		{
		--numLivePlayers;
		livePlayers[i] = FALSE;
		}
	}
