#include "os.h"
#include "app.h"
#include "resman.h"
#include "unihand.h"
#include "util.h"
#include "SoundManagerDirect.h"
#include "dsstream.h"
#include "sound.h"

extern Application *pApp;
extern BYTE btSoundBitSupport;
LPDIRECTSOUNDBUFFER	lpDSBStreamBuffer = NULL;
LPDIRECTSOUNDNOTIFY lpDirectSoundNotify = NULL;


Sound::Sound(int iResourceID, ResourceManager *resourceManager)
	: VSound(iResourceID, resourceManager)
{
	HMMIO			hmmio;
	MMCKINFO		mmckinfoParent;
	MMCKINFO		mmckinfoSubchunk;
	MMIOINFO *		pmmioInfo;
	HPSTR			hpWaveData;
	DWORD			dwWaveDataSize;
	PCMWAVEFORMAT	wfWaveFormat;

	mpDSB = NULL;

	if (!data)
	{
		return;
	}

	// IS THIS A STREAMER?
	if (!strcmp((const char *)data, "STREAMER"))
	{
		// PREPARE FOR STREAMING
		mbStreaming = TRUE;
		return;
	}

	mbStreaming = FALSE;
	
	/* 
		Open the resource for reading using buffered I/O.
	 */
	pmmioInfo = new MMIOINFO;
	pmmioInfo->pIOProc = NULL;
	pmmioInfo->fccIOProc = FOURCC_MEM;
	pmmioInfo->pchBuffer = (char *)data;
	pmmioInfo->cchBuffer = numOfBytes;
	pmmioInfo->adwInfo[0] = NULL;
	pmmioInfo->dwFlags = NULL;
	pmmioInfo->wErrorRet = NULL;
	pmmioInfo->htask = NULL;
	pmmioInfo->pchNext = NULL;
	pmmioInfo->pchEndRead = NULL;
	pmmioInfo->pchEndWrite = NULL;
	pmmioInfo->lBufOffset = NULL;
	pmmioInfo->lDiskOffset = NULL;
	pmmioInfo->dwReserved1 = NULL;
	pmmioInfo->dwReserved2 = NULL;
	pmmioInfo->hmmio = NULL;

	hmmio = mmioOpen(NULL, pmmioInfo, MMIO_READWRITE);

	// Locate a 'RIFF' chunk with a 'WAVE' form type 
	// to make sure it's a WAVE file.
	mmckinfoParent.fccType = mmioFOURCC('W', 'A', 'V', 'E');
	mmioDescend(hmmio, (LPMMCKINFO) &mmckinfoParent, NULL, MMIO_FINDRIFF);

	// Now, find the format chunk (form type 'fmt '). It should be
	// a subchunk of the 'RIFF' parent chunk.
	mmckinfoSubchunk.ckid = mmioFOURCC('f', 'm', 't', ' ');
	mmioDescend(hmmio, &mmckinfoSubchunk, &mmckinfoParent, MMIO_FINDCHUNK);

	// Read the format chunk.
	mmioRead(hmmio, (HPSTR) &wfWaveFormat, sizeof(wfWaveFormat));

	memcpy(&wfFormatEx, &(wfWaveFormat.wf), sizeof(WAVEFORMAT));
	wfFormatEx.wBitsPerSample = wfWaveFormat.wBitsPerSample;
	wfFormatEx.cbSize = 0;
	
	switch (wfWaveFormat.wf.nChannels) {
		case 1:
			switch (wfWaveFormat.wf.nBlockAlign) {
				case 1:
					wfFormat = WF_8BITMONO;
					break;
				case 2:
					wfFormat = WF_16BITMONO;
					break;
				}
			break;
		case 2:
			switch (wfWaveFormat.wf.nBlockAlign) {
				case 2:
					wfFormat = WF_8BITSTEREO;
					break;
				case 4:
					wfFormat = WF_16BITSTEREO;
					break;
				}
			break;
		}
	wSamplesPerSec = (WORD)(wfWaveFormat.wf.nSamplesPerSec);

	if ((wSamplesPerSec != 11025) &&
		(wSamplesPerSec != 22050) &&
		(wSamplesPerSec != 44100))
		{
#ifdef _DEBUG
		MessageBox(NULL, "Sound at Mac rate!", "Error", MB_OK | MB_TASKMODAL);
		//DebugString("Sound at Mac rate!\n");
#endif // _DEBUG
		if (wSamplesPerSec < 22050)
			wSamplesPerSec = 11025;
		else if (wSamplesPerSec < 44100)
			wSamplesPerSec = 22050;
		else
			wSamplesPerSec = 44100;
		}

	// Ascend out of the format subchunk.
	mmioAscend(hmmio, &mmckinfoSubchunk, 0);

	// Find the data subchunk.
	mmckinfoSubchunk.ckid = mmioFOURCC('d', 'a', 't', 'a');
	mmioDescend(hmmio, &mmckinfoSubchunk, &mmckinfoParent, MMIO_FINDCHUNK);

	// Get the size of the data subchunk.
	dwWaveDataSize = mmckinfoSubchunk.cksize;

	dwNumOfSamples = dwWaveDataSize;

	// Allocate and lock memory for the waveform data.
	pSamples = new UniHandle(dwWaveDataSize);
	hpWaveData = (char *)pSamples->Lock();

	// Read the waveform data subchunk.
	mmioRead(hmmio, hpWaveData, dwWaveDataSize);

	// We're done with the file, close it.
	mmioClose(hmmio, 0);

	delete pmmioInfo;
	pSamples->Unlock();

	if (wfWaveFormat.wBitsPerSample == 8) {
		if (wfWaveFormat.wf.nChannels == 1) {
			wfFormat = WF_8BITMONO;
			dwNumOfSamples = dwWaveDataSize;
			}
		else if (wfWaveFormat.wf.nChannels == 2) {
			wfFormat = WF_8BITSTEREO;
			dwNumOfSamples = dwWaveDataSize / 2;
			}
		}
	else if (wfWaveFormat.wBitsPerSample == 16) {
		if (wfWaveFormat.wf.nChannels == 1) {
			wfFormat = WF_16BITMONO;
			dwNumOfSamples = dwWaveDataSize / 2;
			}
		else if (wfWaveFormat.wf.nChannels == 2) {
			wfFormat = WF_16BITSTEREO;
			dwNumOfSamples = dwWaveDataSize / 4;
			}
		}

	// Convert 16bit to 8bit if necessary.
	if ((btSoundBitSupport == 8) && ((wfFormat == WF_16BITMONO) || (wfFormat == WF_16BITSTEREO)))
	{
		UniHandle *newHandle;
		BYTE *bytePointer;
		WORD *wordPointer;
		DWORD i;

		DebugString("Converting 16bit sound to 8bit.\n");
		switch (wfFormat)
			{
			case WF_16BITMONO:
				newHandle = new UniHandle(dwNumOfSamples);
				wfFormat = WF_8BITMONO;
				break;
			case WF_16BITSTEREO:
				newHandle = new UniHandle(dwNumOfSamples * 2);
				wfFormat = WF_8BITSTEREO;
				break;
			}
		bytePointer = (BYTE *)newHandle->Lock();
		wordPointer = (WORD *)pSamples->Lock();
		for (i = 0; i < dwNumOfSamples; ++i)
			{
			*bytePointer = (*wordPointer >> 8) + 128;
			++bytePointer;
			++wordPointer;
			}
		delete pSamples;
		pSamples = newHandle;
		}

	if (data)
		{
		delete[] data;
		data = NULL;
		}
	PrepareForStaticPlay();
//	delete pSamples;
//	pSamples = NULL;

}

void Sound::PrepareForStaticPlay()
{
	// CONVERT MEMORY IMAGE TO DIRECTSOUNDBUFFER
	FILEINFO fileinfo;
	FILEINFO * pFileInfo = &fileinfo;
	ZeroMemory(pFileInfo, sizeof(FILEINFO));
	pFileInfo->mnSoundID = resourceID;
	pFileInfo->cbSize = pSamples->Size();
	pFileInfo->pwfx = &(wfFormatEx);
	pFileInfo->pbData = (BYTE*)pSamples->Lock();
	

	// new direct sound buffer
    DSBUFFERDESC    dsbd;
    DSBCAPS         dsbc;
    HRESULT         hr;
    BYTE            *pbData         = NULL;
    BYTE            *pbData2        = NULL;
    DWORD           dwLength;
    DWORD           dwLength2;

    // Set up the direct sound buffer. 
    memset(&dsbd, 0, sizeof(DSBUFFERDESC));
    dsbd.dwSize                 = sizeof(DSBUFFERDESC);
    dsbd.dwFlags                = 0;
    dsbd.dwFlags                |= DSBCAPS_STATIC;
    // Use new GetCurrentPosition() accuracy (DirectX 2 feature)
    dsbd.dwFlags                |= DSBCAPS_CTRLPOSITIONNOTIFY | DSBCAPS_CTRLDEFAULT | DSBCAPS_GETCURRENTPOSITION2;
    if (pFileInfo->fSticky)
        dsbd.dwFlags |= DSBCAPS_STICKYFOCUS;
    dsbd.dwBufferBytes               = pFileInfo->cbSize;
    dsbd.lpwfxFormat            = pFileInfo->pwfx;

	LPDIRECTSOUND gpds = pApp->pSoundManager->GetDirectSound();

    if ((hr = gpds->CreateSoundBuffer(
			  &dsbd,
			  &(pFileInfo->pDSB),
			  NULL )) != 0)
	{
	goto ERROR_IN_ROUTINE;
	}

	mpDSB = pFileInfo->pDSB;
	mpDSB->AddRef();

    // Ok, lock the sucker down, and copy the memory to it.
    if ((hr = pFileInfo->pDSB->Lock(
			0,
			pFileInfo->cbSize,
			(void **)&pbData,
			&dwLength,
			(void **)&pbData2,
			&dwLength2,
					    0L)) != 0)
    {
    goto ERROR_IN_ROUTINE;
    }

    //Assert(pbData != NULL);
    memcpy(pbData, pFileInfo->pbData, pFileInfo->cbSize);

	pSamples->Unlock();

    // Ok, now unlock the buffer, we don't need it anymore.
    if ((hr = pFileInfo->pDSB->Unlock(
					      pbData, pFileInfo->cbSize,
					      NULL, 0)) != 0)
    {
    goto ERROR_IN_ROUTINE;
    }

    pbData = NULL;

    if ((hr = pFileInfo->pDSB->SetVolume(
	    MAXVOL_VAL)) != 0)
    {
    goto ERROR_IN_ROUTINE;
    }

    if ((hr = pFileInfo->pDSB->SetPan(
	    MIDPAN_VAL)) != 0)
    {
    goto ERROR_IN_ROUTINE;
    }

    dsbc.dwSize = sizeof(dsbc);
    if (hr = IDirectSoundBuffer_GetCaps(pFileInfo->pDSB, &dsbc))
    {
    goto ERROR_IN_ROUTINE;
    }

    if (dsbc.dwFlags & DSBCAPS_LOCHARDWARE) 
	{
		pFileInfo->fHardware = TRUE;
    } 
	else 
	{
		pFileInfo->fHardware = FALSE;
    }
    return;

ERROR_IN_ROUTINE:
    if (pbData != NULL)
    {
    hr = pFileInfo->pDSB->Unlock(pbData,
						pFileInfo->cbSize, NULL, 0);
    pbData = NULL;
    }

    if (pFileInfo->pDSB != NULL)
    {
		pFileInfo->pDSB->Release();
		pFileInfo->pDSB = NULL;
		mpDSB = NULL;
    }
}

void Sound::PrepareForStreaming()
{
	WAVEINFOCA	 wiWave;
    DSBUFFERDESC dsbd;
    HRESULT      dsRetVal;

    int nChkErr;
	int nRem;

    /* This portion of the WAVE I/O is patterned after what's in DSTRWAVE, which
     * was in turn adopted from WAVE.C which is part of the DSSHOW sample.
     */

	char * filename = (char *)data + 9;

    if(( nChkErr = WaveOpenFile( filename, &wiWave.hmmio, &wiWave.pwfx, &wiWave.mmckInRIFF )) != 0 )
        {
       // ErrorMessageBox( IDS_ERROR_WAVEFILEOPEN, MB_ICONEXCLAMATION );
        return; //( ERR_WAVE_OPEN_FAILED );
        }

    if( wiWave.pwfx->wFormatTag != WAVE_FORMAT_PCM )
        {
       // ErrorMessageBox( IDS_ERROR_WAVENOTPCM, MB_ICONEXCLAMATION );
        WaveCloseReadFile( &wiWave.hmmio, &wiWave.pwfx );
        return; // ( ERR_WAVE_INVALID_FORMAT );
		}
    // Seek to the data chunk. mmck.ckSize will be the size of all the data in the file.
    if(( nChkErr = WaveStartDataRead( &wiWave.hmmio, &wiWave.mmck, &wiWave.mmckInRIFF )) != 0 )
        {
       // ErrorMessageBox( IDS_ERROR_WAVESEEKFAILED, MB_ICONEXCLAMATION );
        WaveCloseReadFile( &wiWave.hmmio, &wiWave.pwfx );
        return; //( ERR_WAVE_CORRUPTED_FILE );
        }
	
    // Calculate a buffer length 3 sec. long. This should be an integral number of the
	// number of bytes in one notification period. 
	wiWave.dwNotifySize = wiWave.pwfx->nSamplesPerSec * 3 * (DWORD)wiWave.pwfx->nBlockAlign;
	wiWave.dwNotifySize = wiWave.dwNotifySize/NUM_PLAY_NOTIFICATIONS;
	// the notify size should be an intergral multiple of the nBlockAlignvalue.
	if ((nRem = wiWave.dwNotifySize%(DWORD)wiWave.pwfx->nBlockAlign) != 0)
	{
		wiWave.dwNotifySize += (wiWave.pwfx->nBlockAlign - nRem);
	}
	wiWave.dwBufferSize = wiWave.dwNotifySize * NUM_PLAY_NOTIFICATIONS;	
    //DPF(0, "BufferSize = %lu ", wiWave.dwBufferSize);

#ifdef DEBUG
    //wsprintf( szDebug, "BufferSize = %lu", wiWave.dwBufferSize );
    //ListBox_AddString( hWndList, szDebug );
#endif

     //Create the secondary DirectSoundBuffer object to receive our sound data.
    memset( &dsbd, 0, sizeof( DSBUFFERDESC ));
    dsbd.dwSize = sizeof( DSBUFFERDESC );
    // Use new GetCurrentPosition() accuracy (DirectX 2 feature)
    dsbd.dwFlags = DSBCAPS_CTRLPOSITIONNOTIFY| DSBCAPS_CTRLDEFAULT | DSBCAPS_GETCURRENTPOSITION2 | DSBCAPS_STICKYFOCUS;
    dsbd.dwBufferBytes = wiWave.dwBufferSize;

    //Set Format properties according to the WAVE file we just opened
    dsbd.lpwfxFormat = wiWave.pwfx;
    
	dsRetVal = pApp->pSoundManager->GetDirectSound()->CreateSoundBuffer( &dsbd,&lpDSBStreamBuffer,NULL);
    if( dsRetVal != DS_OK )
    {
    //ErrorMessageBox( IDS_ERROR_DSBCREATE, MB_ICONEXCLAMATION );
	    return;//( ERR_CREATEDSB_FAILED );
    }

    wiWave.lpDSBStreamBuffer = lpDSBStreamBuffer;
    wiWave.bFoundEnd = FALSE;
	wiWave.dwNextWriteOffset = 0;
    wiWave.bLoopFile = TRUE;

	// now get the pointer to the notification interface.

	dsRetVal = lpDSBStreamBuffer->QueryInterface(IID_IDirectSoundNotify, ((LPLPVOID)&lpDirectSoundNotify));
	if (dsRetVal != DS_OK)
	{	
		//ErrorMessageBox(IDS_ERROR_QINOTIFY_FAILED, MB_ICONEXCLAMATION);
	    return;// (dsRetVal);
    }

	// Fill data in the buffer.
	{
		LPBYTE  lpWrite1, lpWrite2;
		DWORD	dwLen1, dwLen2;
		UINT	uActualBytesWritten;
		int		nChkErr;
		HRESULT	dsRetVal;
		DWORD dwBytes = wiWave.dwBufferSize; 

		// This is the initial read. So we fill the entire buffer.
		// This will not wrap around so the 2nd pointers will be NULL.
		dsRetVal = lpDSBStreamBuffer->Lock( 0, dwBytes,
													(LPLPVOID)&lpWrite1, &dwLen1,
													(LPLPVOID)&lpWrite2, &dwLen2, 0 );
	//	ASSERT(lpWrite2 == NULL);	
	//	ASSERT(dwLen2 == 0);

		if (dwLen1 != dwBytes)
		{
	//		DPF(0, "FillDataBuffer: dwBytes != dwLen1");
		}


		if( dsRetVal != DS_OK )
		{
	//		ASSERT(FALSE);
			return; 
		}

	//	ASSERT(dwLen1);
	//	ASSERT( NULL != lpWrite1 );
	//	ASSERT(wiWave.dwNextWriteOffset < wiWave.dwBufferSize);

		nChkErr = WaveReadFile( wiWave.hmmio, (UINT)dwLen1, lpWrite1,
								&wiWave.mmck, &uActualBytesWritten );
		// if the number of bytes written is less than the 
		// amount we requested, we have a short file.
		if (uActualBytesWritten < dwLen1)
		{
	//		DPF(0, "FillDataBuffer: Actual written is less than dwlen1");
			if (!wiWave.bLoopFile)
			{
				// we set the bFoundEnd flag if the length is less than
				// one notify period long which is when the first notification comes in.
				// The next notification will then call send a message to process a stop. 
				if (uActualBytesWritten < wiWave.dwNotifySize)
					wiWave.bFoundEnd = TRUE;
				// Fill in silence for the rest of the buffer.
	//				DPF(0, "Filling in silence");
					FillMemory(lpWrite1+uActualBytesWritten, dwLen1-uActualBytesWritten, 
							(BYTE)(wiWave.pwfx->wBitsPerSample == 8 ? 128 : 0));
			}
			else
			{
				// we are looping.
				UINT uWritten = uActualBytesWritten;	// from previous call above.
				while (uWritten < dwLen1)
				{	// this will keep reading in until the buffer is full. For very short files.
					nChkErr = WaveStartDataRead( &wiWave.hmmio, &wiWave.mmck, &wiWave.mmckInRIFF );
	//				ASSERT(nChkErr == 0);	// we've already this before so shouldn't fail.
					nChkErr = WaveReadFile(wiWave.hmmio, (UINT)dwLen1-uWritten, 
									lpWrite1 + uWritten, &wiWave.mmck, &uActualBytesWritten);
					uWritten += uActualBytesWritten;
				} // while
	//			ASSERT(wiWave.bFoundEnd == FALSE);
			} // else
		}

			// now unlock the buffer.
		dsRetVal = lpDSBStreamBuffer->Unlock( (LPVOID)lpWrite1, dwLen1,		NULL, 0 );
                                                                             
		wiWave.dwNextWriteOffset += dwLen1;	
	// this is a circular buffer. Do mod buffersize.
		if (wiWave.dwNextWriteOffset >= wiWave.dwBufferSize)
			wiWave.dwNextWriteOffset -= wiWave.dwBufferSize; 

	//	DPF(3, "Setting dwProgress = 0");
		wiWave.dwProgress = 0;
		wiWave.dwLastPos = 0;
	}

//////////////////////////    FillDataBuffer();	

	// we're set to play now.
    wiWave.bDonePlaying = FALSE;

//#ifdef DEBUG
 //   wsprintf( szDebug, "wiWave.dwBufferSize = %lu", wiWave.dwBufferSize );
 //   DPF( 3, "StreamBufferSetup Debug" );
 //  DPF( 3, szDebug );
//#endif

//SendMessage( hWndVol, TBM_SETPOS, TRUE, VOL_MAX );
//    SendMessage( hWndPan, TBM_SETPOS, TRUE, PAN_CENTER );
//    SendMessage( hWndFreq, TBM_SETPOS, TRUE,(LPARAM)wiWave.pwfx->nSamplesPerSec / FREQ_MULTIPLIER );
//    PostMessage( hWndMain, WM_DSSTREAM_PROGRESS, 0L, 0L );
//    UpdateFromControls();
}



Sound::~Sound() {
	}

LPDIRECTSOUNDBUFFER Sound::GetDSB() 
{ 
	LPDIRECTSOUNDBUFFER pDSB;
	pApp->pSoundManager->GetDirectSound()->DuplicateSoundBuffer(mpDSB, &pDSB);
	return pDSB;
}