#include "os.h"

#include "sounder.h"
#include "dsstream.h"
#include "sound.h"
#include "unihand.h"
#include "inifile.h"
#include "scigame.h"
#include "resman.h"
#include "util.h"
#include "game.h"

#include <string.h>

Sounder::Sounder()
	{
	hWaveOut = NULL;
	TurnSoundOn();
	}

Sounder::~Sounder()
	{
	TurnSoundOff();
	}

void Sounder::PlayItNow()
	{
	WAVEFORMATEX wfWaveFormatEx;
	int iError;

	if (!soundOn)
		return;
	// Setup wfWaveFormat.
	wfWaveFormatEx.wFormatTag = WAVE_FORMAT_PCM;
	wfWaveFormatEx.nSamplesPerSec = playingSound->wSamplesPerSec;
	wfWaveFormatEx.cbSize = 0;
	switch (playingSound->wfFormat)
		{
		case WF_8BITMONO:
			wfWaveFormatEx.wBitsPerSample = 8;
			wfWaveFormatEx.nChannels = 1;
			wfWaveFormatEx.nBlockAlign = 1;
			break;
		case WF_8BITSTEREO:
			wfWaveFormatEx.wBitsPerSample = 8;
			wfWaveFormatEx.nChannels = 2;
			wfWaveFormatEx.nBlockAlign = 2;
			break;
		case WF_16BITMONO:
			wfWaveFormatEx.wBitsPerSample = 16;
			wfWaveFormatEx.nChannels = 1;
			wfWaveFormatEx.nBlockAlign = 2;
			break;
		case WF_16BITSTEREO:
			wfWaveFormatEx.wBitsPerSample = 16;
			wfWaveFormatEx.nChannels = 2;
			wfWaveFormatEx.nBlockAlign = 4;
			break;
		}
	wfWaveFormatEx.nAvgBytesPerSec =
			wfWaveFormatEx.nBlockAlign * wfWaveFormatEx.nSamplesPerSec;
                                                            
	if (playingSound->pSamples)
		{
		lpWaveData = (LPSTR)playingSound->pSamples->Lock();
		}
	else
		{
		DebugString("No Samples.\n");
		if (deleteSound)
			playingSound->Free();
		playingSound = NULL;
		return;
		}

	iError = waveOutOpen((LPHWAVEOUT)&hWaveOut, WAVE_MAPPER,
			&wfWaveFormatEx, 0L, 0L, CALLBACK_NULL);

	if (iError)
		{
		DebugString("waveOutOpen error.\n");
		EndSound();
		return;    
		}

	whWaveHdr.lpData = (char *)lpWaveData;
	whWaveHdr.dwBufferLength = wfWaveFormatEx.nBlockAlign * playingSound->dwNumOfSamples;
	whWaveHdr.dwBytesRecorded = 0;
	whWaveHdr.dwUser = 0;
	whWaveHdr.lpNext = 0;
	whWaveHdr.reserved = 0;

	if (playingSound->loop)
		{
		whWaveHdr.dwFlags = WHDR_BEGINLOOP | WHDR_ENDLOOP;
		whWaveHdr.dwLoops = playingSound->loop;
		}
	else
		{
		whWaveHdr.dwFlags = 0L;
		whWaveHdr.dwLoops = 0L;
		}

	iError = waveOutPrepareHeader(hWaveOut, &whWaveHdr, sizeof(WAVEHDR));
	if (iError || !hWaveOut)
		{
		DebugString("waveOutPrepareHeader error.\n");
		headerPrepared = FALSE;
		EndSound();
		return;    
		}
	headerPrepared = TRUE;

	iError = waveOutWrite(hWaveOut, &whWaveHdr, sizeof(WAVEHDR));
	if (iError || !hWaveOut)
		{
		DebugString("waveOutWrite error.\n");
		EndSound();
		return;    
		}

	// Update endtime.
	if (playingSound->loop == -1)
		{
		// forever
		playingSound->itsEndTime = -1;
		}
	else
		{
		playingSound->itsEndTime = game->GetTrueTime();
		playingSound->itsEndTime += (playingSound->loop+1) * (int)((((double)playingSound->dwNumOfSamples) / playingSound->wSamplesPerSec) * 60 + 15);
		}
	}

void Sounder::EndSound()
	{
	if (playingSound == NULL)
		return;

	if (hWaveOut)
		{
		waveOutReset(hWaveOut);
		if (headerPrepared)
			waveOutUnprepareHeader(hWaveOut, &whWaveHdr, sizeof(WAVEHDR));
		}
	playingSound->pSamples->Unlock();
	if (hWaveOut)
		{
		waveOutClose(hWaveOut);
		hWaveOut = NULL;
		}

	if (deleteSound)
		playingSound->Free();
	playingSound = NULL;
	}

void Sounder::TurnSoundOn()
	{
	soundOn = TRUE;
	}

void Sounder::TurnSoundOff()
	{
	soundOn = FALSE;
	EndSound();
	}
