#define INITGUID

#include "os.h"
#include "procs.h"
#include "dsstream.h"
#include "Sound.h"
#include "SoundManager.h"
#include "SoundManagerDirect.h"
#include <math.h>

// EXTERNALS
extern DWORD __stdcall HandleNotifyForStaticSound(LPVOID lpvoid);
extern DWORD __stdcall HandleNotifyForStreamingSound(LPVOID lpvoid);
extern char *szAppName;
extern LPDIRECTSOUNDNOTIFY lpDirectSoundNotify;

// LOCALS
int SoundManagerDirectSound::SMDS_NUMBEROFVOICES = 8;
int StartDSound(FILEINFO *pFileInfo);
CRITICAL_SECTION oCriticalSection1;
#define ENTER_CS1 (EnterCriticalSection(&oCriticalSection1))
#define LEAVE_CS1 (LeaveCriticalSection(&oCriticalSection1))
int localSoundsForThisCycle = 0;
int localThisCycle = 0;
DWORD __stdcall HandleNotifyForStaticSound(LPVOID lpvoid)
{
	ENTER_CS1;
	
	FILEINFO * pFileInfo = (FILEINFO *)lpvoid;
	DWORD hr = DS_OK;
	DWORD hRet = 0;
	HANDLE hNotifyEvent[2];
	WAVEINFOCA &wiWave = pFileInfo->waveinfoca;
	hNotifyEvent[0] = pFileInfo->mhNotifyEvent[0];
	DWORD dwMilliseconds = 200 + (int)(1000 * ((double)pFileInfo->cbSize / (double)pFileInfo->pwfx->nAvgBytesPerSec));
	HANDLE hThread = pFileInfo->mhThread;

	LEAVE_CS1;

	if ((hRet = WaitForMultipleObjects(1, hNotifyEvent, FALSE, dwMilliseconds)) != WAIT_TIMEOUT)
	{
		// signaled
		if (hRet == WAIT_OBJECT_0) 
		{
			// handle notification
			ENTER_CS1;

			FILEINFO * pNext = pApp->pSoundManager->pGetFirst();

			// remove sound info
			while(pNext->pNext)
			{
				if (pNext->pNext == pFileInfo)
				{
					pNext->pNext = pFileInfo->pNext;
					CloseHandle(pFileInfo->mhNotifyEvent[0]);
					delete pFileInfo;
					break;  // deleted sound information so quit tracking
				}
				pNext = pNext->pNext;
			}
			LEAVE_CS1;
		}
	}
	else
	{
		// leaving event does not necessarily indicate problem.
		// just means that sound has taken longer than a half a second to
		// end which means it may have taken that long to start.  This just
		// indicates to clean up later since an event could still be signaled
		// for the sound.
		vDebugOut("possible problem: timeout sound\n");
		ENTER_CS1;
		pFileInfo->mbTimeout = TRUE;
		LEAVE_CS1;
	}
	CloseHandle(hThread);
	return TRUE;
}

DWORD __stdcall HandleNotifyForStreamingSound(LPVOID lpvoid)
{
	FILEINFO * pFileInfo = (FILEINFO *)lpvoid;
	DWORD hr = DS_OK;
	WAVEINFOCA	 &wiWave = pFileInfo->waveinfoca;
	int nCount = 2;
	DWORD hRet = 0;
	HANDLE hThread = pFileInfo->mhThread;

	// STATIC BUFFERS JUST STOP 
	if (pFileInfo->mhNotifyEvent[1] == 0)
		nCount = 1;
		
	while((hRet = WaitForMultipleObjects(nCount, pFileInfo->mhNotifyEvent, FALSE, INFINITE)) != WAIT_FAILED)
	{
		// handle notification
		if (hRet == WAIT_OBJECT_0) 
		{
			LPBYTE lpWrite1;
			DWORD dwWrite1;
			UINT cbActual = 0;

			// If the entire file has been read into the buffer, bFoundEnd will be TRUE.
			if (!wiWave.bFoundEnd)
			{	// read in more of the file.
				// Lock the buffer.
				hr = pFileInfo->pDSB->Lock(		wiWave.dwNextWriteOffset, 
													wiWave.dwNotifySize, 
													(void**)&lpWrite1, &dwWrite1,
													NULL, NULL, 0);
				if (hr != DS_OK)
				{
					// couldn't lock buffer with hr
					CloseHandle(hThread);
					return hr;
				}

				if (dwWrite1 < wiWave.dwNotifySize)
				{
					// Lock returned number of bytes and requested size differ
				}
				
				// now read the file. 
				hr = WaveReadFile(wiWave.hmmio, dwWrite1, lpWrite1, &wiWave.mmck, &cbActual);

				//	if cbActual is less than the amount asked for, we may have
				//	reached end of file. If we are looping, we will simply read the 
				//	file in again until we have written dwWrite1 bytes into the buffer.
				//	If not looping we fill with silence.
				if ((DWORD)cbActual < dwWrite1)
				{
					// ReadFile read less than what we asked for. less by dwWrite1-(DWORD)cbActual
					if (!wiWave.bLoopFile)
					{
						wiWave.bFoundEnd = TRUE;
						// fill up the rest of the buffer with silence.
						FillMemory( lpWrite1+cbActual, (UINT)dwWrite1-cbActual, 
									(BYTE)(wiWave.pwfx->wBitsPerSample == 8 ? 128 : 0));
					}
					else
					{	// we're looping. Read in the file again.
						int nChkErr = 0;
						UINT cbWritten = cbActual;	// from previous call above.
						//Looping so read in file again
						while (cbWritten < (UINT)dwWrite1)	// write only one notification worth.
						{	// this will keep reading in until the buffer is full. For very short files.
							nChkErr = WaveStartDataRead( &wiWave.hmmio, &wiWave.mmck, &wiWave.mmckInRIFF );
							assert(nChkErr == 0);	// we've already this before so shouldn't fail.
							nChkErr = WaveReadFile(wiWave.hmmio, (UINT)dwWrite1-cbWritten, 
													lpWrite1 + cbWritten, &wiWave.mmck, &cbActual);
							assert(nChkErr == 0);	// we've already this before so shouldn't fail.
							cbWritten += cbActual;
						} 
						assert(wiWave.bFoundEnd == FALSE);
					}
				}

				//unlock now
				hr = pFileInfo->pDSB->Unlock((LPVOID)lpWrite1, dwWrite1,  NULL, 0 );
				// setup the nextwrite offset.
				wiWave.dwNextWriteOffset += dwWrite1;
				if (wiWave.dwNextWriteOffset >= wiWave.dwBufferSize)
					wiWave.dwNextWriteOffset -= wiWave.dwBufferSize;
			}// if !bFoundEnd.
			else
			{	
				// We have read in the whole file. We will keep filling in silence
				// in the buffer to allow the remaining sound bytes filled earlier 
				// to be played. 
				
				//ASSERT(wiWave.bLoopFile == FALSE);

				// Allow the rest of the bytes to be played and fill here
				// with silence. The next notification will quit the while loop.
				hr = pFileInfo->pDSB->Lock(	wiWave.dwNextWriteOffset, 
												wiWave.dwNotifySize, 
												(void**)&lpWrite1, &dwWrite1,
												NULL, NULL, 0);
				//Filling in silence
				FillMemory(lpWrite1, dwWrite1,(BYTE)(wiWave.pwfx->wBitsPerSample == 8 ? 128 : 0));

				hr = pFileInfo->pDSB->Unlock((LPVOID)lpWrite1, dwWrite1,  NULL, 0 );

				// We don't want to cut off the sound before it's done playing.
				// If doneplaying is set, the next notification event will post a stop message.
				if (wiWave.mmckInRIFF.cksize > wiWave.dwNotifySize &&
						wiWave.dwProgress >= wiWave.mmckInRIFF.cksize - wiWave.dwNotifySize)
				{
					wiWave.bDonePlaying = TRUE;
				}
				else	// for short files.
					if (wiWave.dwProgress >= wiWave.mmckInRIFF.cksize)
						wiWave.bDonePlaying = TRUE;
			}
		}
		else if (hRet == WAIT_OBJECT_0 + 1) 
		{
			// A stop notification has been received.
			
			wiWave.bDonePlaying = TRUE;

			// call stop.
			if (pFileInfo->pDSB)
			{
				hr = pFileInfo->pDSB->Stop();
				assert(hr == DS_OK);
			}

			CloseHandle(pFileInfo->mhNotifyEvent[0]);
			CloseHandle(pFileInfo->mhNotifyEvent[1]);
			pFileInfo->mhNotifyEvent[0] = pFileInfo->mhNotifyEvent[1] = (HANDLE)NULL;

			// THIS WAS ALLOCATED IN WAVEOPENFILE... JUST DELETE HERE FOR NOW
			if (pFileInfo->waveinfoca.pwfx)
			{
				GlobalFreePtr(pFileInfo->waveinfoca.pwfx);
				pFileInfo->waveinfoca.pwfx = NULL;
			}

			ENTER_CS1;
			FILEINFO * pNext = pApp->pSoundManager->pGetFirst();
			while(pNext->pNext)
			{
				if (pNext->pNext == pFileInfo)
				{
					pNext->pNext = pFileInfo->pNext;
					delete pFileInfo;
					break;
				}
				pNext = pNext->pNext;
			}
			LEAVE_CS1;
		}

	}
	CloseHandle(hThread);
	return TRUE;
}

void SoundManagerDirectSound::GameCycle()
{
	FILEINFO * pNext = FileInfoFirst.pNext;

	if (mbLocationDirty)
	{
		ENTER_CS1;
		// SetPanVolume FOR ALL POSITIONAL SOUNDS
		while(pNext)
		{
			if (pNext->mbPointSourced)
			{
				SetPanVolume(pNext);
			}
			pNext = pNext->pNext;
		}
		LEAVE_CS1;
	}
}

void SoundManagerDirectSound::SetSoundEmitterPos(SoundEmitter * poSoundEmitter, int nLeft, int nTop)
{
	//
	// FOR ALL SOUNDS FROM THIS EMITTER
	FILEINFO * pNext = FileInfoFirst.pNext;

	ENTER_CS1;
	while(pNext)
	{
		if (pNext->mpoObject == poSoundEmitter)
		{
			//
			//		IF THE LAST RECORDED POSITION DIFFERS SIGNIFICANTLY ENOUGH
//			int ndX = abs(pNext->mnX - nLeft);
//			int ndY = abs(pNext->mnY - nTop);

			//		IE. BY MORE THAN 2 FULL NODES  
//			if ((ndX > 34)||(ndY > 20)) 
			{
				// THEN RECORD THE NEW POSITION AND
				pNext->mnX = nLeft;
				pNext->mnY = nTop;

				//	SetPanVolume FOR THIS SOUND
				SetPanVolume(pNext);
			}
		}
		pNext = pNext->pNext;
	}
	LEAVE_CS1;
}

void SoundManagerDirectSound::SetPanVolume(FILEINFO * pFileInfo)
{
	assert(pFileInfo);

	if (pFileInfo->mbPointSourced)
	{
		if (pFileInfo->mnX >= mnViewWindowPositionLeft && pFileInfo->mnX < (mnViewWindowPositionLeft + mnViewWindowWidth) &&
			pFileInfo->mnY >= mnViewWindowPositionTop && pFileInfo->mnY < (mnViewWindowPositionTop + mnViewWindowHeight))
		{
			// SOUND IS WITHIN VIEW WINDOW
			pFileInfo->pDSB->SetVolume(DSBVOLUME_MAX);

			LONG lpan = ((pFileInfo->mnX - mnViewWindowPositionLeft) * ((DSBPAN_RIGHT - DSBPAN_LEFT) / 12)) / mnViewWindowWidth + DSBPAN_LEFT / 12;

			pFileInfo->pDSB->SetPan(lpan);
		}
		else
		{
			// SOUND IS OUTSIDE OF VIEW WINDOW
			int nNearX, nNearY;
			// SET THE VOLUME
			if (pFileInfo->mnX < mnViewWindowPositionLeft)
			{
				nNearX = mnViewWindowPositionLeft;
			}
			else if (pFileInfo->mnX > (mnViewWindowPositionLeft + mnViewWindowWidth))
			{
				nNearX = (mnViewWindowPositionLeft + mnViewWindowWidth);
			}
			else
			{
				nNearX = pFileInfo->mnX;
			}

			if (pFileInfo->mnY < mnViewWindowPositionTop)
			{
				nNearY = mnViewWindowPositionTop;
			}
			else if (pFileInfo->mnY > (mnViewWindowPositionTop + mnViewWindowHeight))
			{
				nNearY = (mnViewWindowPositionTop + mnViewWindowHeight);
			}
			else
			{
				nNearY = pFileInfo->mnY;
			}

			double dFallOff = 1.0 - ((mdMaxFallOff * mdMaxFallOff) - (((pFileInfo->mnX - nNearX)*(pFileInfo->mnX - nNearX) + 4.0*(pFileInfo->mnY - nNearY)*(pFileInfo->mnY - nNearY)))) / (mdMaxFallOff * mdMaxFallOff);
//			vDebugOut("%lf\n", dFallOff);
			dFallOff = max(0, dFallOff);
			pFileInfo->pDSB->SetVolume((long)(DSBVOLUME_MIN * dFallOff/3.8));

			// SET THE PANNING
			if (pFileInfo->mnX < mnViewWindowPositionLeft)
			{
				pFileInfo->pDSB->SetPan(DSBPAN_LEFT / 12);
			}
			else if (pFileInfo->mnX > (mnViewWindowPositionLeft + mnViewWindowWidth))
			{
				pFileInfo->pDSB->SetPan(DSBPAN_RIGHT / 12);
			}
			else
			{
				LONG lpan = ((pFileInfo->mnX - mnViewWindowPositionLeft) * ((DSBPAN_RIGHT - DSBPAN_LEFT) / 12)) / mnViewWindowWidth + DSBPAN_LEFT / 12;
				pFileInfo->pDSB->SetPan(lpan);
			}
		}
	}
}

void SoundManagerDirectSound::SetViewWindowPos(int nLeft, int nTop)
{
	mnViewWindowPositionLeft = nLeft * 17;
	mnViewWindowPositionTop = nTop * 10;

	// UPDATE PANNING/VOLUME FOR TRACKED SOUNDS
	mbLocationDirty = TRUE;
}

void SoundManagerDirectSound::SetWorldCoordinates(int nWidth, int nHeight)
{
	mdMaxFallOff = sqrt((double)(nWidth * 17 * nWidth * 17 + nHeight * 10 * nHeight * 10));
}

void SoundManagerDirectSound::PreInit()
{
	DSBUFFERDESC    dsbd;
    HRESULT         hr;
    DWORD           dw;
	HWND			hWnd = pApp->pWindow->hWnd;

#define PIXELWIDTHOFMAPWINDOW 636
#define PIXELHEIGHTOFMAPWINDOW 360

	mnViewWindowWidth = PIXELWIDTHOFMAPWINDOW;
	mnViewWindowHeight = PIXELHEIGHTOFMAPWINDOW;

    // Set up the file info header
    FileInfoFirst.pNext = NULL;
    FileInfoFirst.pwfx = NULL;

	InitializeCriticalSection(&oCriticalSection1);
		
    // DIRECT SOUND INITIALIZATION

    // GET THE LARGEST WAVEFORMATEX STRUCTURE.
    if (MMSYSERR_NOERROR != acmMetrics(NULL, ACM_METRIC_MAX_SIZE_FORMAT, &dw))
    {
		MessageBox(NULL, "ACM Metrics failed, aborting", "DirectSound Initialization",
			   MB_OK|MB_ICONSTOP);
		return;
    }

    // SETUP THE FORMAT, FREQUENCY, VOLUME, ETC.
    if ((FileInfoFirst.pwfx = (WAVEFORMATEX *)GlobalAllocPtr(GPTR, dw)) == NULL)
    {
		MessageBox(hWnd, "Out of Memory", "DirectSound Initialization",
			   MB_OK|MB_ICONSTOP);
		return;
    }

    FileInfoFirst.pwfx->wFormatTag = WAVE_FORMAT_PCM;
    FileInfoFirst.pwfx->nChannels = 2;
    FileInfoFirst.pwfx->nSamplesPerSec = 44100;
    FileInfoFirst.pwfx->nAvgBytesPerSec = 44100*2*2;
    FileInfoFirst.pwfx->nBlockAlign = 4;
    FileInfoFirst.pwfx->wBitsPerSample = 16;
    FileInfoFirst.pwfx->cbSize = 0;

	if (!SUCCEEDED(CoInitialize(NULL))) {
		MessageBox(NULL, "Failed to initialize COM library", "DirectSound Initialization", MB_OK | MB_ICONSTOP);
		return;
	}
    
	// Create the direct sound object.
	hr = CoCreateInstance(CLSID_DirectSound, NULL, CLSCTX_INPROC_SERVER,
			  IID_IDirectSound, (void**)&gpds);

	if (SUCCEEDED(hr) && (NULL != gpds))
	{
		hr = gpds->Initialize(NULL);
		//hr = IDirectSound_Initialize(gpds, fUseGuid ? &guID : NULL);
		if (SUCCEEDED(hr)) 
		{
			// Note we need to set the level to be priority to set the
			// format of the primary buffer
			hr = gpds->SetCooperativeLevel(pApp->pWindow->hWnd, DSSCL_EXCLUSIVE );
			if (SUCCEEDED(hr)) 
			{

				// Set up the primary direct sound buffer.
				ZeroMemory(&dsbd, sizeof(dsbd));
				dsbd.dwSize = sizeof(dsbd);
				dsbd.dwFlags = DSBCAPS_PRIMARYBUFFER;

				hr = gpds->CreateSoundBuffer(&dsbd, &(FileInfoFirst.pDSB), NULL);
				if (SUCCEEDED(hr)) 
				{
					hr = FileInfoFirst.pDSB->SetFormat(FileInfoFirst.pwfx);
					hr = FileInfoFirst.pDSB->SetFrequency(44100);
					
					hr = FileInfoFirst.pDSB->Play(0, 0, DSBPLAY_LOOPING); 
					if (!SUCCEEDED(hr)) 
					{
						MessageBox(NULL, "Cannot play primary buffer", szAppName, MB_OK|MB_ICONSTOP);
						FileInfoFirst.pDSB->Release(); 
						FileInfoFirst.pDSB = NULL;
					}		
				} 
				else 
				{
					MessageBox(NULL, "Cannot create primary buffer", szAppName, MB_OK|MB_ICONSTOP);
				}
			} 
			else 
			{
				MessageBox(NULL, "DirectSound SetCooperativeLevel failed",szAppName, MB_OK|MB_ICONSTOP);
			}
		} 
		else 
		{
			MessageBox(NULL, "Failed to Initialize DirectSound object", szAppName, MB_OK | MB_ICONSTOP);
		}

		if (!SUCCEEDED(hr)) 
		{
			gpds->Release();  //IDirectSound_Release();
			gpds = NULL;
		}

	} 
	else 
	{
		MessageBox(hWnd, "Failed to create DirectSound COM object",
			   szAppName, MB_OK | MB_ICONSTOP);
	}

	if (SUCCEEDED(hr)) 
	{
		eStatus = STATUS_READY;
		return;
	} 
	else 
	{
		CoUninitialize();
		return;
	}
}


void SoundManagerDirectSound::SoundEmitterDestroyed(SoundEmitter * pSoundEmitter)
{
	// SCAN AND UN-REFERENCE
	ENTER_CS1;

	FILEINFO * pNext = pApp->pSoundManager->pGetFirst();

	pNext = pNext->pNext;

	// remove sound info
	while(pNext)
	{
		// THIS SOUND EMITTED BY THIS OBJECT
		if ((pNext->mbTrackedObject) && (pNext->mpoObject == pSoundEmitter))
		{
			if (pNext->pDSB)
			{
				FILEINFO * pNextNext = pNext->pNext;
				pNext->pDSB->Stop();
				LEAVE_CS1;

				// LET IT PASS
				Sleep(0);

				ENTER_CS1;
				pNext = pNextNext;
				continue;		// NEXT SOUND
			}
		}
		pNext = pNext->pNext;
	}
	LEAVE_CS1;
}

void SoundManagerDirectSound::Play(int nSoundResourceID, int volume, int SoundPriority, sciScript *whoToCue)
{
	int nReturn = 0;
	Play(&nReturn, nSoundResourceID, volume, SoundPriority, whoToCue, FALSE);
	return;
}

void SoundManagerDirectSound::Play(int * nReturnSoundNumber, int nSoundResourceID, int volume, int SoundPriority, sciScript *whoToCue, BOOL bLooped)
{
	Emit(NULL, FALSE, 0, 0, nReturnSoundNumber, nSoundResourceID, volume, SoundPriority, whoToCue, FALSE);
	return;
}


void SoundManagerDirectSound::PrepareForStaticPlay(Sound * pSound, FILEINFO * pFileInfo)
{
	int hr = DS_OK;
	DSBPOSITIONNOTIFY dsbPosNotify[1];	
	DWORD dwThreadId;

	// CREATE EVENT FOR STOP
	pFileInfo->mhNotifyEvent[0] = CreateEvent(NULL, FALSE, FALSE, NULL);
	pFileInfo->mhNotifyEvent[1] = 0;
	// set the stop notification.
	dsbPosNotify[0].dwOffset = DSBPN_OFFSETSTOP;
	dsbPosNotify[0].hEventNotify = pFileInfo->mhNotifyEvent[0];

	// Now create the thread to wait on the events created.
	pFileInfo->mhThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)HandleNotifyForStaticSound, pFileInfo, CREATE_SUSPENDED, &dwThreadId);
	SetThreadPriority(pFileInfo->mhThread, THREAD_PRIORITY_TIME_CRITICAL);
	ResumeThread(pFileInfo->mhThread);

	if (!pFileInfo->mhThread)
	{
		goto HandleErr;
	}

	LPDIRECTSOUNDNOTIFY lpDirectSoundNotify;
	lpDirectSoundNotify = NULL;
	pFileInfo->pDSB->QueryInterface(IID_IDirectSoundNotify, ((LPLPVOID)&lpDirectSoundNotify));
	// setup notification
	hr = lpDirectSoundNotify->SetNotificationPositions(1, dsbPosNotify);

	HandleErr:
		if (hr != DS_OK)
		{	
			// TBD - HANDLE THESE
			// DSERR_INVALIDPARAM
			// DSERR_OUTOFMEMORY
			CloseHandle(pFileInfo->mhNotifyEvent[0]);
			pFileInfo->mhNotifyEvent[0] = (HANDLE)NULL;
		}
}

void SoundManagerDirectSound::PrepareForStreaming(Sound * pSound, FILEINFO * pFileInfo)
{
	WAVEINFOCA	 &wiWave = pFileInfo->waveinfoca;
    DSBUFFERDESC dsbd;
    HRESULT      dsRetVal;

    int nChkErr;
	int nRem;

    /* This portion of the WAVE I/O is patterned after what's in DSTRWAVE, which
     * was in turn adopted from WAVE.C which is part of the DSSHOW sample.
     */

	char * filename = (char *)pSound->GetData() + 9;

    if(( nChkErr = WaveOpenFile( filename, &wiWave.hmmio, &wiWave.pwfx, &wiWave.mmckInRIFF )) != 0 )
        {
       // ErrorMessageBox( IDS_ERROR_WAVEFILEOPEN, MB_ICONEXCLAMATION );
        return; //( ERR_WAVE_OPEN_FAILED );
        }

    if( wiWave.pwfx->wFormatTag != WAVE_FORMAT_PCM )
        {
       // ErrorMessageBox( IDS_ERROR_WAVENOTPCM, MB_ICONEXCLAMATION );
        WaveCloseReadFile( &wiWave.hmmio, &wiWave.pwfx );
        return; // ( ERR_WAVE_INVALID_FORMAT );
		}
    // Seek to the data chunk. mmck.ckSize will be the size of all the data in the file.
    if(( nChkErr = WaveStartDataRead( &wiWave.hmmio, &wiWave.mmck, &wiWave.mmckInRIFF )) != 0 )
        {
       // ErrorMessageBox( IDS_ERROR_WAVESEEKFAILED, MB_ICONEXCLAMATION );
        WaveCloseReadFile( &wiWave.hmmio, &wiWave.pwfx );
        return; //( ERR_WAVE_CORRUPTED_FILE );
        }
	
    // Calculate a buffer length 3 sec. long. This should be an integral number of the
	// number of bytes in one notification period. 
	wiWave.dwNotifySize = wiWave.pwfx->nSamplesPerSec * 3 * (DWORD)wiWave.pwfx->nBlockAlign;
	wiWave.dwNotifySize = wiWave.dwNotifySize/NUM_PLAY_NOTIFICATIONS;
	// the notify size should be an intergral multiple of the nBlockAlignvalue.
	if ((nRem = wiWave.dwNotifySize%(DWORD)wiWave.pwfx->nBlockAlign) != 0)
	{
		wiWave.dwNotifySize += (wiWave.pwfx->nBlockAlign - nRem);
	}
	wiWave.dwBufferSize = wiWave.dwNotifySize * NUM_PLAY_NOTIFICATIONS;	
    //DPF(0, "BufferSize = %lu ", wiWave.dwBufferSize);

#ifdef DEBUG
    //wsprintf( szDebug, "BufferSize = %lu", wiWave.dwBufferSize );
    //ListBox_AddString( hWndList, szDebug );
#endif

     //Create the secondary DirectSoundBuffer object to receive our sound data.
    memset( &dsbd, 0, sizeof( DSBUFFERDESC ));
    dsbd.dwSize = sizeof( DSBUFFERDESC );
    // Use new GetCurrentPosition() accuracy (DirectX 2 feature)
    dsbd.dwFlags = DSBCAPS_CTRLPOSITIONNOTIFY| DSBCAPS_CTRLDEFAULT | DSBCAPS_GETCURRENTPOSITION2;
    dsbd.dwBufferBytes = wiWave.dwBufferSize;

    //Set Format properties according to the WAVE file we just opened
    dsbd.lpwfxFormat = wiWave.pwfx;
    
	//dsRetVal = pApp->pSoundManager->GetDirectSound()->CreateSoundBuffer( &dsbd,&lpDSBStreamBuffer,NULL );
	dsRetVal = pApp->pSoundManager->GetDirectSound()->CreateSoundBuffer( &dsbd,&(pFileInfo->pDSB),NULL );

    if( dsRetVal != DS_OK )
    {
    //ErrorMessageBox( IDS_ERROR_DSBCREATE, MB_ICONEXCLAMATION );
	    return;//( ERR_CREATEDSB_FAILED );
    }

    wiWave.lpDSBStreamBuffer = pFileInfo->pDSB; // lpDSBStreamBuffer;
    wiWave.bFoundEnd = FALSE;
	wiWave.dwNextWriteOffset = 0;
    wiWave.bLoopFile = TRUE;

	// now get the pointer to the notification interface.

	//dsRetVal = lpDSBStreamBuffer->QueryInterface(IID_IDirectSoundNotify, ((LPLPVOID)&lpDirectSoundNotify));
	dsRetVal = pFileInfo->pDSB->QueryInterface(IID_IDirectSoundNotify, ((LPLPVOID)&lpDirectSoundNotify));
	
	if (dsRetVal != DS_OK)
	{	
		MessageBox(pApp->pWindow->hWnd, "Direct Sound 5 required for background music streaming.", NULL, MB_ICONEXCLAMATION);
	    return;// (dsRetVal);
    }

	// Fill data in the buffer.
	{
		LPBYTE  lpWrite1, lpWrite2;
		DWORD	dwLen1, dwLen2;
		UINT	uActualBytesWritten;
		int		nChkErr;
		HRESULT	dsRetVal;
		DWORD dwBytes = wiWave.dwBufferSize; 

		// This is the initial read. So we fill the entire buffer.
		// This will not wrap around so the 2nd pointers will be NULL.
		dsRetVal = pFileInfo->pDSB->Lock( 0, dwBytes,
													(LPLPVOID)&lpWrite1, &dwLen1,
													(LPLPVOID)&lpWrite2, &dwLen2, 0 );
	//	ASSERT(lpWrite2 == NULL);	
	//	ASSERT(dwLen2 == 0);

		if (dwLen1 != dwBytes)
		{
	//		DPF(0, "FillDataBuffer: dwBytes != dwLen1");
		}


		if( dsRetVal != DS_OK )
		{
	//		ASSERT(FALSE);
			return; 
		}

	//	ASSERT(dwLen1);
	//	ASSERT( NULL != lpWrite1 );
	//	ASSERT(wiWave.dwNextWriteOffset < wiWave.dwBufferSize);

		nChkErr = WaveReadFile( wiWave.hmmio, (UINT)dwLen1, lpWrite1,
								&wiWave.mmck, &uActualBytesWritten );
		// if the number of bytes written is less than the 
		// amount we requested, we have a short file.
		if (uActualBytesWritten < dwLen1)
		{
	//		DPF(0, "FillDataBuffer: Actual written is less than dwlen1");
			if (!wiWave.bLoopFile)
			{
				// we set the bFoundEnd flag if the length is less than
				// one notify period long which is when the first notification comes in.
				// The next notification will then call send a message to process a stop. 
				if (uActualBytesWritten < wiWave.dwNotifySize)
					wiWave.bFoundEnd = TRUE;
				// Fill in silence for the rest of the buffer.
	//				DPF(0, "Filling in silence");
					FillMemory(lpWrite1+uActualBytesWritten, dwLen1-uActualBytesWritten, 
							(BYTE)(wiWave.pwfx->wBitsPerSample == 8 ? 128 : 0));
			}
			else
			{
				// we are looping.
				UINT uWritten = uActualBytesWritten;	// from previous call above.
				while (uWritten < dwLen1)
				{	// this will keep reading in until the buffer is full. For very short files.
					nChkErr = WaveStartDataRead( &wiWave.hmmio, &wiWave.mmck, &wiWave.mmckInRIFF );
	//				ASSERT(nChkErr == 0);	// we've already this before so shouldn't fail.
					nChkErr = WaveReadFile(wiWave.hmmio, (UINT)dwLen1-uWritten, 
									lpWrite1 + uWritten, &wiWave.mmck, &uActualBytesWritten);
					uWritten += uActualBytesWritten;
				} // while
	//			ASSERT(wiWave.bFoundEnd == FALSE);
			} // else
		}

			// now unlock the buffer.
		dsRetVal = pFileInfo->pDSB->Unlock( (LPVOID)lpWrite1, dwLen1,		NULL, 0 );
                                                                             
		wiWave.dwNextWriteOffset += dwLen1;	
	// this is a circular buffer. Do mod buffersize.
		if (wiWave.dwNextWriteOffset >= wiWave.dwBufferSize)
			wiWave.dwNextWriteOffset -= wiWave.dwBufferSize; 

	//	DPF(3, "Setting dwProgress = 0");
		wiWave.dwProgress = 0;
		wiWave.dwLastPos = 0;
	}

//////////////////////////    FillDataBuffer();	

	// we're set to play now.
    wiWave.bDonePlaying = FALSE;

//#ifdef DEBUG
 //   wsprintf( szDebug, "wiWave.dwBufferSize = %lu", wiWave.dwBufferSize );
 //   DPF( 3, "StreamBufferSetup Debug" );
 //  DPF( 3, szDebug );
//#endif

//SendMessage( hWndVol, TBM_SETPOS, TRUE, VOL_MAX );
//    SendMessage( hWndPan, TBM_SETPOS, TRUE, PAN_CENTER );
//    SendMessage( hWndFreq, TBM_SETPOS, TRUE,(LPARAM)wiWave.pwfx->nSamplesPerSec / FREQ_MULTIPLIER );
//    PostMessage( hWndMain, WM_DSSTREAM_PROGRESS, 0L, 0L );
//    UpdateFromControls();


	/////////////////////////////
	{
		int hr = DS_OK;
		DSBPOSITIONNOTIFY dsbPosNotify[NUM_PLAY_NOTIFICATIONS + 1 ];	
		DWORD dwSize = wiWave.dwNotifySize;	
		DWORD dwThreadId;
		int i;

		// Create the 2 events. One for Play one for stop.
		pFileInfo->mhNotifyEvent[0] = CreateEvent(NULL, FALSE, FALSE, NULL);
		//ASSERT(hNotifyEvent[0]);
		pFileInfo->mhNotifyEvent[1] = CreateEvent(NULL, FALSE, FALSE, NULL);
		//ASSERT(hNotifyEvent[1]);

			// setup the first one.	
		dsbPosNotify[0].dwOffset = dwSize;
		dsbPosNotify[0].hEventNotify = pFileInfo->mhNotifyEvent[0];
		//DPF(0, "Set notifies for position %lu", dsbPosNotify[0].dwOffset);

		for (i = 1; i < NUM_PLAY_NOTIFICATIONS; i++)
		{
			dsbPosNotify[i].dwOffset = dsbPosNotify[i-1].dwOffset + dwSize;
			dsbPosNotify[i].hEventNotify = pFileInfo->mhNotifyEvent[0];				
			//DPF(0, "Set notifies for positions %lu", dsbPosNotify[i].dwOffset);
		}

		dsbPosNotify[i-1].dwOffset -= 1;
		// set the stop notification.
		dsbPosNotify[i].dwOffset = DSBPN_OFFSETSTOP;
		dsbPosNotify[i].hEventNotify = pFileInfo->mhNotifyEvent[1];

		
		// Now create the thread to wait on the events created.
		pFileInfo->mhThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)HandleNotifyForStreamingSound, pFileInfo, CREATE_SUSPENDED, &dwThreadId);
		SetThreadPriority(pFileInfo->mhThread, THREAD_PRIORITY_TIME_CRITICAL);
		ResumeThread(pFileInfo->mhThread);

		if (!pFileInfo->mhThread)
		{
			goto HandleErr;
		}

		// setup notification
		hr = lpDirectSoundNotify->SetNotificationPositions(	NUM_PLAY_NOTIFICATIONS +1,
																	dsbPosNotify);

	HandleErr:
		if (hr != DS_OK)
		{	if (hr == DSERR_INVALIDPARAM);
				//DPF(0, "SetupNotificationPos failed. Invalid Param");
			if (hr == DSERR_OUTOFMEMORY);
				//DPF(0, "SetupNotificationPos failed. OutOfMemory");
			CloseHandle(pFileInfo->mhNotifyEvent[0]);
			CloseHandle(pFileInfo->mhNotifyEvent[1]);
			pFileInfo->mhNotifyEvent[0] = pFileInfo->mhNotifyEvent[1] = (HANDLE)NULL;
		}
	}
}

BOOL SoundManagerDirectSound::bEstablishPriority(FILEINFO * pFileInfo, int nPriority)
{
	int nCount = 0;
	FILEINFO * pFileInfoNextAvailable = NULL;
	int nMinPriority = 200000;
	//int nMinTime = MAX_INT;
	pFileInfo->mnPriority = nPriority;

	ENTER_CS1;
	FILEINFO * pFileInfoNext = FileInfoFirst.pNext;
	if (pFileInfoNext == NULL)
	{
		FileInfoFirst.pNext = pFileInfo;  // PRIORITY ESTABLISHED
		LEAVE_CS1;
		return TRUE;
	}
	
	while (nCount < SMDS_NUMBEROFVOICES && pFileInfoNext != NULL) 
	{
		nCount++;
		// if entry available use it
		if (pFileInfoNext->pNext == NULL)
		{
			pFileInfoNext->pNext = pFileInfo;  // PRIORITY ESTABLISHED
			LEAVE_CS1;
			return TRUE;
		}
		pFileInfoNext = pFileInfoNext->pNext;
		// else if at least one entry has same priority or less
		if (pFileInfoNext->mnPriority <= nPriority)
		{
			// cancel oldest playing sound of lowest priority
			if (pFileInfoNext->mnPriority < nMinPriority)
			{
				nMinPriority = pFileInfoNext->mnPriority;
				pFileInfoNextAvailable = pFileInfoNext;
			}
		}
	} 
	

	if (pFileInfoNextAvailable)
	{
		// INSERT NEW SOUND AFTER FIRST
		FILEINFO * pFileInfoInsert = FileInfoFirst.pNext;
		FileInfoFirst.pNext = pFileInfo;
		pFileInfo->pNext = pFileInfoInsert;
		LEAVE_CS1;

		pFileInfoNextAvailable->pDSB->Stop();
		return TRUE;
	}
	else 
	{
		LEAVE_CS1;
		// ignore play sound
		return FALSE;
	}
}

BOOL SoundManagerDirectSound::bSoundPlaying(int nSoundResourceID, SoundEmitter * poSoundEmitter)
{
	BOOL bReturn = FALSE;
	ENTER_CS1;

	FILEINFO * pNext = pApp->pSoundManager->pGetFirst();

	// remove sound info
	if (nSoundResourceID)
	{
		while(pNext->pNext)
		{
			pNext = pNext->pNext;
			if (pNext->mbTimeout) continue;
			if (pNext->mnSoundID == nSoundResourceID)
			{
				if (poSoundEmitter)
				{
					if (pNext->mbTrackedObject && pNext->mpoObject==poSoundEmitter)
					{
						bReturn = TRUE;
						break;
					}
				}
				else
				{
					bReturn = TRUE;
					break;
				}
			}
		}
	}
	else if (poSoundEmitter)
	{
		while(pNext->pNext)
		{
			pNext = pNext->pNext;
			if (pNext->mbTimeout) continue;
			if (pNext->mbTrackedObject && pNext->mpoObject==poSoundEmitter)
			{
				bReturn = TRUE;
				break;
			}
		}
	}
	LEAVE_CS1;
	return bReturn;
}

void SoundManagerDirectSound::EmitFromObject(SoundEmitter * poSoundEmitter, int nSoundResourceID, int SoundPriority, BOOL bLooped, int volume, BOOL bCancelIfPlaying)
{
	int nReturnSoundNumber;
	if (bCancelIfPlaying)
	{
		if (bSoundPlaying(0, poSoundEmitter))
		{
			return;
		}
	}
	SoundEmitterDestroyed(poSoundEmitter);
	Emit(poSoundEmitter, FALSE, 0, 0, &nReturnSoundNumber, nSoundResourceID, volume, SoundPriority, NULL, bLooped);
}

void SoundManagerDirectSound::EmitFromPosition(int nPixX, int nPixY, int nSoundResourceID, int SoundPriority, BOOL bLooped, int volume)
{
	int nReturnSoundNumber;
	Emit(NULL, TRUE, nPixX, nPixY, &nReturnSoundNumber, nSoundResourceID, volume, SoundPriority, NULL, bLooped);
}

void SoundManagerDirectSound::EmitAmbient(int nSoundResourceID, int SoundPriority, BOOL bLooped, int volume)
{
	int nReturnSoundNumber;
	Emit(NULL, FALSE, 0, 0, &nReturnSoundNumber, nSoundResourceID, volume, SoundPriority, NULL, bLooped);
}

void SoundManagerDirectSound::Emit(SoundEmitter * poSoundEmitter, BOOL bUsePosition, int nPixX, int nPixY, int * nReturnSoundNumber, int nSoundResourceID, int volume, int SoundPriority, sciScript *whoToCue, BOOL bLooped)
{
	int nCurCycle = game->nGameCycle();
	if ((nCurCycle >= localThisCycle) && (nCurCycle < (localThisCycle + 3) ))
	{
		localSoundsForThisCycle++;
	}
	else
	{
		localThisCycle = nCurCycle;
		localSoundsForThisCycle = 1;
	}
	if (localSoundsForThisCycle > 3) return;

	//  GET LOCKED POINTER TO SOUND BUFFER

	// requests pointer to sample of resource nSoundResourceID
	if (resourceManager == NULL)
	   	resourceManager = pApp->pResourceManager;

	Sound * sound = (Sound *)resourceManager->LoadResource(RES_SOUND, nSoundResourceID);
	
	if (sound)
	{
		if (sound->bStreaming())
		{
			// STREAM THIS SOUND

			// CREATE A FILEINFO
			FILEINFO * pFileInfo = new FILEINFO;
			ZeroMemory(pFileInfo, sizeof(FILEINFO));
			pFileInfo->mnSoundID = nSoundResourceID;
			if (poSoundEmitter)
			{
				pFileInfo->mbPointSourced = TRUE;
				pFileInfo->mbTrackedObject = TRUE; 
				pFileInfo->mpoObject = poSoundEmitter;
				UINT uX, uY;
				poSoundEmitter->GetPlayPosition(&uX, &uY);
				pFileInfo->mnX = (int) uX;
				pFileInfo->mnY = (int) uY;
			}
			else if (bUsePosition)
			{
				pFileInfo->mbPointSourced = TRUE;
				pFileInfo->mbTrackedObject = FALSE; 
				pFileInfo->mpoObject = NULL;
								
				pFileInfo->mnX = nPixX;
				pFileInfo->mnY = nPixY;
			}

			// CREATE A STREAMING BUFFER AND ATTACH TO FILEINFO
			// PREP THE BUFFER WITH INITIAL DATA
			if (bEstablishPriority(pFileInfo, SoundPriority))
			{
				PrepareForStreaming(sound, pFileInfo);
				pFileInfo->pDSB->SetCurrentPosition( 0 );
				SetPanVolume(pFileInfo);

				// START PLAYING SOUND 
				pFileInfo->pDSB->Play( 0, 0, DSBPLAY_LOOPING );                                                      
			}
			else
			{
				delete pFileInfo;
			}
		}
		else
		{
			// TRACK SOUND
			FILEINFO * pFileInfo = new FILEINFO;
			ZeroMemory(pFileInfo, sizeof(FILEINFO));
			pFileInfo->mnSoundID = nSoundResourceID;

			//?? IS THIS STUFF USED ANYWHERE ??
			pFileInfo->cbSize = sound->pSamples->Size();
			pFileInfo->pwfx = &(sound->wfFormatEx);
			//pFileInfo->pbData = (BYTE*)sound->pSamples->Lock();
			pFileInfo->fLooped = bLooped;

			if (poSoundEmitter)
			{
				pFileInfo->mbPointSourced = TRUE;
				pFileInfo->mbTrackedObject = TRUE; 
				pFileInfo->mpoObject = poSoundEmitter;
				UINT uX, uY;
				poSoundEmitter->GetPlayPosition(&uX, &uY);
				pFileInfo->mnX = (int) uX;
				pFileInfo->mnY = (int) uY;
			}
			else if (bUsePosition)
			{
				pFileInfo->mbPointSourced = TRUE;
				pFileInfo->mbTrackedObject = FALSE; 
				pFileInfo->mpoObject = NULL;
								
				pFileInfo->mnX = nPixX;
				pFileInfo->mnY = nPixY;
			}

			pFileInfo->pDSB = sound->GetDSB();

			if (bEstablishPriority(pFileInfo, SoundPriority))//  && (!NewDirectSoundBuffer(pFileInfo)))
			{
				PrepareForStaticPlay(sound, pFileInfo);
				StartDSound(pFileInfo);
				SetPanVolume(pFileInfo);
			}
			else
			{
				delete pFileInfo;
			}
		}
	}
}

void SoundManagerDirectSound::Stop(int nSoundNumber)
{
	return;
}


void SoundManagerDirectSound::Init(BOOL * bOkay)
{
}

void SoundManagerDirectSound::ShutDown()
{
	if (eStatus == STATUS_READY)
	{
		// STOP ALL SOUNDS

		FILEINFO * pNext = FileInfoFirst.pNext;

		ENTER_CS1;///////////////////////
		while(pNext)
		{
			FILEINFO * pLast = pNext;

			LEAVE_CS1;///////////////////////

			pNext->pDSB->Stop();
			
			// LET IT PASS
			Sleep(0);
		
			while (pNext == pLast) 
			{
				ENTER_CS1;///////////////////////
				pNext = FileInfoFirst.pNext;
				
				// WAS THIS NOT CLEANED UP IN LIEU OF PENDING PLAYTIME?
				if (pNext && pNext->mbTimeout)
				{
					CloseHandle(pNext->mhNotifyEvent[0]);
					FileInfoFirst.pNext = pNext->pNext;
					delete pNext;
					pNext = FileInfoFirst.pNext;
					LEAVE_CS1;///////////////////////
					break;
				}
				LEAVE_CS1;///////////////////////
				Sleep(0);
			}
			ENTER_CS1;
		}
		LEAVE_CS1;
		if (gpds)
		{
			gpds->Release();
			gpds = NULL;
		}
		CoUninitialize();
		eStatus = STATUS_SHUTDOWN;
		DeleteCriticalSection(&oCriticalSection1);
		if (FileInfoFirst.pwfx)
		{
			GlobalFreePtr(FileInfoFirst.pwfx);
			FileInfoFirst.pwfx = NULL;
		}
	}
}

// =======================================================================
/*  This routine will initialize a new direct sound buffer,
    set the data in the buffer, 
    set the rate, format, etc...

    Input:
	pFileInfo   -   Pointer to file info with all
	nessecary info filled, 
	like pbData, cbData, etc...

    Output:
	0 if successful, else the error code.

*/
// =======================================================================
int SoundManagerDirectSound::NewDirectSoundBuffer(FILEINFO *pFileInfo)
{

    DSBUFFERDESC    dsbd;
    DSBCAPS         dsbc;
    HRESULT         hr;
    BYTE            *pbData         = NULL;
    BYTE            *pbData2        = NULL;
    DWORD           dwLength;
    DWORD           dwLength2;

    // Set up the direct sound buffer. 
    memset(&dsbd, 0, sizeof(DSBUFFERDESC));
    dsbd.dwSize                 = sizeof(DSBUFFERDESC);
    dsbd.dwFlags                = 0;
    dsbd.dwFlags                |= DSBCAPS_STATIC;
    // Use new GetCurrentPosition() accuracy (DirectX 2 feature)
    dsbd.dwFlags                |= DSBCAPS_CTRLPOSITIONNOTIFY | DSBCAPS_CTRLDEFAULT | DSBCAPS_GETCURRENTPOSITION2;
    if (pFileInfo->fSticky)
        dsbd.dwFlags |= DSBCAPS_STICKYFOCUS;
    dsbd.dwBufferBytes               = pFileInfo->cbSize;
    dsbd.lpwfxFormat            = pFileInfo->pwfx;
    if ((hr = gpds->CreateSoundBuffer(
			  &dsbd,
			  &(pFileInfo->pDSB),
			  NULL )) != 0)
	{
	goto ERROR_IN_ROUTINE;
	}

    // Ok, lock the sucker down, and copy the memory to it.
    if ((hr = pFileInfo->pDSB->Lock(
			0,
			pFileInfo->cbSize,
			(void **)&pbData,
			&dwLength,
			(void **)&pbData2,
			&dwLength2,
					    0L)) != 0)
    {
    goto ERROR_IN_ROUTINE;
    }

    //Assert(pbData != NULL);
    memcpy(pbData, pFileInfo->pbData, pFileInfo->cbSize);

    // Ok, now unlock the buffer, we don't need it anymore.
    if ((hr = pFileInfo->pDSB->Unlock(
					      pbData, pFileInfo->cbSize,
					      NULL, 0)) != 0)
    {
    goto ERROR_IN_ROUTINE;
    }

    pbData = NULL;

    if ((hr = pFileInfo->pDSB->SetVolume(
	    MAXVOL_VAL)) != 0)
    {
    goto ERROR_IN_ROUTINE;
    }

    if ((hr = pFileInfo->pDSB->SetPan(
	    MIDPAN_VAL)) != 0)
    {
    goto ERROR_IN_ROUTINE;
    }

    dsbc.dwSize = sizeof(dsbc);
    if (hr = IDirectSoundBuffer_GetCaps(pFileInfo->pDSB, &dsbc))
    {
    goto ERROR_IN_ROUTINE;
    }

    if (dsbc.dwFlags & DSBCAPS_LOCHARDWARE) {
    pFileInfo->fHardware = TRUE;
    } else {
    pFileInfo->fHardware = FALSE;
    }

    goto DONE_ROUTINE;

ERROR_IN_ROUTINE:
    if (pbData != NULL)
    {
    hr = pFileInfo->pDSB->Unlock(pbData,
						pFileInfo->cbSize, NULL, 0);
    pbData = NULL;
    }

    if (pFileInfo->pDSB != NULL)
    {
    pFileInfo->pDSB->Release();
    pFileInfo->pDSB = NULL;
    }
    
DONE_ROUTINE:

    return(hr); 

}

int StartDSound(FILEINFO *pFileInfo)
{
    HRESULT     hr = 0;
    DWORD       dwLooped;
    DWORD       dwStatus = 0;
	HWND		hWnd = pApp->pWindow->hWnd;

	// ALREADY PLAYING? 
    // Start sound here....
    dwLooped = 0;
    if (pFileInfo->fLooped)
	{
		dwLooped = DSBPLAY_LOOPING;
    }
		    

    if ((hr = pFileInfo->pDSB->GetStatus(&dwStatus)) == 0)
    {
		if ((dwStatus&DSBSTATUS_PLAYING) == DSBSTATUS_PLAYING)
		{
			// Don't bother playing, just restart
			if ((hr = pFileInfo->pDSB->SetCurrentPosition(0)) != 0)
			{
				MessageBox(hWnd, "Cannot set current position",
					   "Direct Sound Error", MB_OK);
			}
	    }
		else 
			goto PLAY_THE_THING;           
    }
    else
    {
PLAY_THE_THING:
		if ((hr = pFileInfo->pDSB->Play(0, 0, dwLooped)) != 0)
		{
			MessageBox(hWnd, "Cannot start playing",
			   "Direct Sound Error", MB_OK);
		}
		else
			pFileInfo->fPlaying = TRUE;
    }
    return(hr);
}

FILEINFO::FILEINFO()
{
	mbTimeout = FALSE;
	pNext = NULL;
	pwfx = NULL; 
}

FILEINFO::~FILEINFO() 
{
	if (pDSB) 
	{
		pDSB->Release();
		pDSB = NULL;
	}
}
