#ifndef _SoundManagerDirect_h_
#define _SoundManagerDirect_h_

#ifndef _SoundManager
#include "SoundManager.h"
#endif

#ifndef __DSSTREAM_INCLUDED__
#include "dsstream.h"
#endif

#ifndef _SOUNDEMITTER_H_
#include "SoundEmitter.h"
#endif

#include <mmsystem.h>
#include <mmreg.h>
#include <msacm.h>
#include <dsound.h>
#include <windowsx.h>


typedef enum _TAG_SMDS_
{
	SMDS_UNINITIALIZED,
	SMDS_READY,
	SMDS_SHUTDOWN,
} SMDS_STATUS;


class FILEINFO
    {
public:
	FILEINFO();
	~FILEINFO();

    BYTE                *pbData;        // Pointer to actual data of file.
    UINT                cbSize;         // Size of data.
    WAVEFORMATEX	*pwfx;          // Pointer to waveformatex structure.

    DWORD               dwFreq;         // Frequency.
    DWORD               dwPan;          // Panning info.
    DWORD               dwVol;          // Total volume.
    BOOL                fLooped;        // Looped?

    BOOL                fPlaying;       // Is this one playing?
    BOOL                fLost;          // Is this one lost?
    BOOL                fHardware;      // Is this a hardware buffer?
    BOOL                fSticky;        // Is this a sticky buffer?

    int			cox;            
    int			coy;            

    LPDIRECTSOUNDBUFFER pDSB;               // Pointer to direct sound buffer.

    int                 nFileName;          // Index to filename, not including dir.
    char                szFileName[MAX_PATH];
    FILEINFO			* pNext;            // Pointer to next file.

	WAVEINFOCA			waveinfoca;
	HANDLE				mhNotifyEvent[2];
	int					mnPriority;
	
	BOOL				mbPointSourced;
	BOOL				mbTrackedObject;
	SoundEmitter	*	mpoObject;
	int					mnX;
	int					mnY;  // POSITION OF SOUND IN PIXELS
	HANDLE				mhThread;
	BOOL				mbTimeout;
	int					mnSoundID;
    };

class Sound;

class SoundManagerDirectSound: public SoundManager
{
	static int SMDS_NUMBEROFVOICES;
public:
	SoundManagerDirectSound()
	{
		SetViewWindowPos(0, 0);
		gpds = NULL;
		status = SMDS_UNINITIALIZED;
	}
	~SoundManagerDirectSound()
	{
		ShutDown();
	}
	void PreInit();
	void Init(BOOL * bOkay);
	
	void EmitFromObject(SoundEmitter * poSoundEmitter, int nSoundResourceID, int SoundPriority = SOUNDPRIORITY_UNITSOUNDS, BOOL bLooped = FALSE, int volume = 128, BOOL bCancelIfPlaying = FALSE);
	void EmitFromPosition(int nPixX, int nPixY, int nSoundResourceID, int SoundPriority = SOUNDPRIORITY_UNITSOUNDS, BOOL bLooped = FALSE, int volume = 128);
	void EmitAmbient(int nSoundResourceID, int SoundPriority = SOUNDPRIORITY_AMBIENTSOUNDS, BOOL bLooped = TRUE, int volume = 128);

	void Play(int nSoundResourceID, int volume, int SoundPriority, sciScript *whoToCue = NULL);
	void Stop(int nSoundNumber);
	void ShutDown();

	void SoundEmitterDestroyed(SoundEmitter * pSoundEmitter);
	void SetViewWindowPos(int nLeft, int nTop);
	void SetSoundEmitterPos(SoundEmitter * poSoundEmitter, int nLeft, int nTop);
	void SetWorldCoordinates(int nWidth, int nHeight);

	void GameCycle();
	BOOL bSoundPlaying(int nSoundResourceID, SoundEmitter * poSoundEmitter = NULL);

	LPDIRECTSOUND GetDirectSound() { return gpds; }
	FILEINFO * pGetFirst() { return &FileInfoFirst; }
protected:
	void Emit(SoundEmitter * poSoundEmitter, BOOL bUsePosition, int nPixX, int nPixY, int * nReturnSoundNumber, int nSoundResourceID, int volume, int SoundPriority, sciScript *whoToCue, BOOL bLooped);
	void Play(int * nReturnSoundNumber, int nSoundResourceID, int volume, int SoundPriority, sciScript *whoToCue = NULL, BOOL bLooped = FALSE);

private:
	void PrepareForStreaming(Sound * pSound, FILEINFO * pFileInfo);
	int NewDirectSoundBuffer(FILEINFO *pFileInfo);
	BOOL bEstablishPriority(FILEINFO * pFileInfo, int nPriority);
	void PrepareForStaticPlay(Sound * pSound, FILEINFO * pFileInfo);
	void SetPanVolume(FILEINFO * pFileInfo);

	/*  This is the main head of the linked list, but its only used for the
		pNext which will point to the first FILEINFO structure, or NULL if there
		are no files loaded  */
	FILEINFO        FileInfoFirst;              // Start of linked list.
	LPDIRECTSOUND   gpds;
	SMDS_STATUS		status;

	BOOL	mbLocationDirty;
	int		mnViewWindowPositionLeft;
	int 	mnViewWindowPositionTop;
	int		mnViewWindowWidth;
	int		mnViewWindowHeight;
	double	mdMaxFallOff;
};

#define MINFREQ_TB              0
#define MAXFREQ_TB              100000
#define MINPAN_TB               0
#define MIDPAN_TB               400
#define MAXPAN_TB               800
#define SHIFTPAN_TB             (-400)
#define MULTPAN_TB              (10L)
#define MINPAN_VAL              (-400)
#define MIDPAN_VAL              0
#define MAXPAN_VAL              400
#define MINVOL_TB               0
#define MAXVOL_TB               400
#define SHIFTVOL_TB             (-400)
#define MULTVOL_TB              (10L)
#define MINVOL_VAL              -400
#define MAXVOL_VAL              0

#endif