#ifndef _texture
#define _texture

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define TRANSPARENT_FLAG	0x01
#define OPAQUE_FLAG			0x02
#define ALPHA_FLAG			0x04
#define RLE_FLAG			0x08

typedef struct
	{
	short	numPixels;
	} TRANSPARENT_BLOCK;

typedef struct
	{
	short	numPixels;
	BYTE	*pixels;
	} OPAQUE_BLOCK;

typedef struct
	{
	BYTE	alpha;
	BYTE	pixel;
	} ALPHA_PIXEL;

typedef struct
	{
	short	numPixels;
	ALPHA_PIXEL	*pixels;
	} ALPHA_BLOCK;

typedef struct
	{
	short	numPixels;
	BYTE	pixel;
	} RLE_BLOCK;

typedef struct
	{
	BYTE	flags;
	union
		{
		TRANSPARENT_BLOCK	tb;
		OPAQUE_BLOCK		ob;
		ALPHA_BLOCK			ab;
		RLE_BLOCK			rb;
		} SOME_BLOCK;
	} BLOCK;

typedef struct
	{
	short length;
	BLOCK *blocks;	// variable length blocks
	} ROW;

typedef struct
	{
	short	width;
	short	height;
	short	xOrigin;
	short	yOrigin;
	short	x;
	short	y;
//	ROW		*rows;
	BYTE	**rows;  // used only in in-memory representation, not in disk file
	void	*pData;  // used only in-memory pointer to disk representation of this struct
	} CEL;

#pragma pack(1)
typedef struct
{
	short	width;
	short	height;
	short	xOrigin;
	short	yOrigin;
	short	x;
	short	y;
} DISK_CEL;
#pragma pack()

typedef struct
	{
	short	numCels;
	CEL		*cels;
	} LOOP;

typedef struct
	{
	BYTE	red;
	BYTE	green;
	BYTE	blue;
	} RGB;

typedef struct
	{
	int		sig;	// 'tex ', i.e. low byte = 't'
	int		version;	// High Byte = major version, etc.
	RGB		colors[256];
	short	numLoops;
	LOOP	*loops;
	} TEXTURE;

#define FASTCALL __fastcall

//void FASTCALL TextureRenderRow(BYTE *destRow, BYTE *compressedRow, int start, int length, BYTE alpha[256][256], RGBQUAD *colors);
void FASTCALL TextureRenderRow(BYTE * pjDestRow, BYTE * pjCompressedRow, 
							   int nStart, int nLength);
void FASTCALL TextureRenderRow16(WORD * pwDestRow, BYTE * pjCompressedRow, int nStart, int nLength, WORD * pwPalette);
void FASTCALL TextureRenderRowFlip16(WORD * pwDestRow, BYTE * pjCompressedRow, int nStart, int nLength, WORD * pwPalette);
void FASTCALL TextureRenderRowAlpha16(WORD * pwDestRow, BYTE * pjCompressedRow, int nStart, int nLength, WORD * pwPalette, BYTE jAlpha, BOOL bFlip);
BYTE *TextureCompressRow(BYTE *sourceRow, int length, int *lengthOfRow, BYTE *alphaRow=NULL);
void Convert2BMPsToTEX(char *inColorFilename, char *inAlphaFilename, char *outFilename, int height);
void BitmapCopyRLEPixels(WORD wWidthBytes, BYTE *pb, BYTE *prle);
void BitmapCopyPixels(WORD wWidthBytes, BYTE *pb, BYTE *prle, int wWidthInBytes, int wHeight);
BYTE PaletteCalculateAlpha(RGBQUAD colors[256], BYTE f, BYTE b, BYTE alpha);
BYTE PaletteFindNearest(RGBQUAD colors[256], BYTE red, BYTE green, BYTE blue);
void WriteTEX(FILE *outFile, BYTE *inColorPixels, RGBQUAD *inColorPalette, BYTE *inAlphaPixels, int inWidth, int inHeight, int height, int inWidthInBytes);

#endif