// UNIT.CPP -- generic unit code
//
// DJM

#include "scimain.h"
#include "mapnode.h"
#include "mapplane.h"
#include "unit.h"
#include "building.h"
#include "command.h"
#include "objectdb.h"
#include "path.h"
#include "template.h"
#include "mapedit.h"
#include "guard.h"
#include "movement.h"
#include "procs.h"
#include "math.h"
#include "intplane.h"
#include "vampire.h"
#include "item.h"
#include "mob.h"
#include "human.h"
#include "HumanPeasant.h"
#include "HumanVampireHunter.h"
#include "HumanSoldier.h"
#include "HumanGypsy.h"
#include "HumanMerchant.h"
#include "cycler.h"
#include "spell.h"
//#include "walker.h"

BYTE Unit::mjFullRed = 0;
BYTE Unit::mjFullGreen = 0;
BYTE Unit::mjFullBlue = 0;
BYTE Unit::mjEmptyRed = 0;
BYTE Unit::mjEmptyGreen = 0;
BYTE Unit::mjEmptyBlue = 0;

int Unit::mclassstaticLastPlayedClassSound = 0;
int Unit::mclassstaticClassSounds = 0;

short Unit::mnRelHealthPercent = 0;

#define MAX_STOPPED_CYCLES	10
#define MAX_BLOCKED_CYCLES	6

// how wide is the fuzzy border around the visibility oval
#ifdef _DEBUG
#define VIS_FUZZY_BORDER 0
#else
#define VIS_FUZZY_BORDER 5
#endif

extern Boolean gbMapEditing;
extern InterfacePlane *gpoInterface;

BYTE gajCommandTag[] = "CommandTag";


Unit::Unit(Player *poOwner, MapPlane *poPlane, MapNode *poNode, int nTemplateID)
:DrawObj(poPlane, nTemplateID), mpoGuardCoordinator(NULL), mpoGuardedUnit(NULL), 
	mpoMovement(NULL), mpoSequencer(NULL), mpoThrallOwner(NULL), mnUnderAttackCycles(0),
	mpoAttacker(0), mnThrallCyclesRemaining(0), mpoMobCoordinator(NULL),
	mnCurrentSpell(0)
{
	mnLastRetreatDirection = -1;

	mnCyclesBeforeCast = 0;
	mnFog = 0;
	mbInvisible = FALSE;
	mwSpyFlags = 0;
	mnAdrenalineCycleTicker = 0;
	for (int i=0; i < ALLPLAYERINDICES; i++)
	{
		mnVisibleTotal[i].attack = 0;
		mnVisibleTotal[i].defense = 0;
		mnVisibleTotal[i].level = 0;
		mnVisibleTotal[i].count = 0;
		mnVisibleTotal[i].relativehealth = 0;
	}
	mnAlertCyclesRemaining = 0;

	mpnFlags = NULL;
	//assert(poOwner);
	assert(poPlane);
	assert(poNode);
	assert(!poNode->mpoUnit);

	
	meClass = OC_UNIT;
	mdwFlags = 0;

	mbCloaked = mpoTemplate->lGetValue(TAG_clok, FALSE);

	meType = (UNITTYPE)mpoTemplate->lGetValue(TAG_unit, UT_VAMPIRE); // TBD -- fill in

	mnLevel = (short)mpoTemplate->lGetValue(TAG_levl, 1);

	// if he has an owner, he's owned by a player. Otherwise, he's a NPC (peasant)
	mpoOwner = poOwner;
	if (mpoOwner) 
	{
		mpoOwner->vAddUnit(this);
	}

	mpoHouse = NULL;
	
//	mpoWalker = NULL;
//	if (mpoCycler && mpoCycler->eCycleType() == CT_MOVER)
//	{
//		mpoWalker = goWalkerDB.poAddObject(this,nTemplateID);
//	}

	mnConvertCost = 50;				// IN POWER UNITS
	

	meDefaultAttack = DA_HANDTOHAND;
	meBehavior = BV_GUARD;

	mTenacity = SETTING_HIGH; //(SETTING_LMH)mpoTemplate->lGetValue(TAG_tena, SETTING_LOW);
	mAggression = SETTING_HIGH; //(SETTING_LMH)mpoTemplate->lGetValue(TAG_aggr, SETTING_MEDIUM);
	mConversion = SETTING_LOW;

	mpoLastNode = NULL;		// the one we left
	mpoCurNode = poNode;	// the one we're on

	poNode->mpoUnit = this;

	mlPixelX = mpoCurNode->nPixelX();	// pixel-accurate position
	mlPixelY = mpoCurNode->nPixelY();

	mbAlive = TRUE;

	mnPartialMove = 0; // how many cycles in this move have we done?
	mbBetweenPoints = FALSE; // are we moving between points?
	mbMoving = FALSE;		// are we moving?
	mnStoppedCycles = 0;
	mnTempBlockedCycles = 0;

	// attack-related
	mnAttackee = 0;
	mnLastTargetX = 0;
	mnLastTargetY = 0;
	mnLastAttackCycle = 0;
	mnLastThinkCycle = 0;
	mnLastPowerConvCycle = 0;
	mnStartWaitCycle = 0;
	mnDeathCycle = 0;
	
	//setup shadow
	mpoShadow = NULL;
	int nShadowTemplate = mpoTemplate->lGetValue(TAG_shad, 0);
	if (nShadowTemplate)
	{
		mpoShadow = new DrawObj(mpoPlane, nShadowTemplate);
	}


	mnAttackRange = (short)mpoTemplate->lGetValue(TAG_arng, 1);
	mnAttackSpeed = (short)mpoTemplate->lGetValue(TAG_aspd, mnAttackSpeed);
	mnThinkSpeed = (short)mpoTemplate->lGetValue(TAG_tspd, 4);
	mfFeedRate = (float)mpoTemplate->lGetValue(TAG_frat, 5) / (float)20.0;
	mnConvertCycleCount = 0;

	// power conversion
	mnPowerConvSpeed = (short)mpoTemplate->lGetValue(TAG_p2hs, 50);
	

	// power regeneration
	mnLastPowerRegenCycle = 0;
	mnPowerRegenSpeed = (short)mpoTemplate->lGetValue(TAG_pwrs, 50);
	mjPlayerLink = 0;

	meDirection = D_SOUTH;
	vFidget();

}

Unit::~Unit()
{
//	mpoWalker = NULL;

	if (mpoShadow)
	{
		delete mpoShadow;
		mpoShadow = NULL;
	}

	while (mpnFlags)
	{
		FLAGENTRY * pn = mpnFlags->next;
		delete mpnFlags;
		mpnFlags = pn;
	}
	ClearThought();

	if (mpoSequencer)
	{
		delete mpoSequencer;
		mpoSequencer = NULL;
	}

	if (bIsMobMember())
	{	
		BOOL bOkay = FALSE;
		mpoMobCoordinator->RemoveUnit(&bOkay, this);
		if (!mpoMobCoordinator->GetSize())
		{
			delete mpoMobCoordinator;
			mpoMobCoordinator = NULL;
		}
	}
	if (mpoGuardCoordinator)
	{
		delete mpoGuardCoordinator;
		mpoGuardCoordinator = NULL;
	}
	if (mpoOwner) {
		mpoOwner->vDeleteUnit(this);
		mpoOwner = NULL;
	}
	if (mpoCurNode) 
	{
		// ONLY CLEAR THE UNIT FROM THE NODE IF
		// IT IS THIS UNIT.  DEAD UNITS ARE IN THE
		// DRAWN LIST BUT ARE NOT THE PRIMARY UNIT
		// ON THE NODE.
		if (mpoCurNode->mpoUnit == this)
		{
			mpoCurNode->mpoUnit = NULL;
		}
		mpoCurNode = NULL;
	}
	mnAttackee = NULL;
}


void Unit::SetThrallOwner(Unit * poOwnerUnit, int nCycles)
{
	if (!bIsVampire())
	{
		if (vCanBeEnthralled())
		{
			if (mpoThrallOwner)
			{
				mpoThrallOwner = NULL;
			}
			mpoThrallOwner = poOwnerUnit;
			mnThrallCyclesRemaining = nCycles;
		}
	}
	else
	{
		mpoThrallOwner = NULL;
		mnThrallCyclesRemaining = 0;
	}
}

long Unit::lGetChecksum(void)
{
	// checksum all member variables that are important to net synch

	long lCheck = 0;

	// do superclass
	lCheck += DrawObj::lGetChecksum();

	// do self
	lCheck += 
		(long)mbAlive + 
		(long)mbBetweenPoints + 
		(long)mbCloaked + 
		(long)mbMoving +
		(long)meBehavior + 
		(long)meType + 
		mlPixelX + 
		mlPixelY + 
		mnAttackRange + 
		mnAttackSpeed + 
		mnThinkSpeed +
		mnLastThinkCycle +
		mnDeathCycle + 
		(long)mfHealth +
		mnLastAttackCycle + 
		mnLastTargetX + 
		mnLastTargetY + 
		(long)mfMaxHealth + 
		(long)mfMaxPower +
		mnPartialMove + 
		(long)mfPower + 
		mnSpeed + 
		mnStoppedCycles +
		mnTempBlockedCycles +
		mAggression +
		mTenacity +
		(long)moCurrentCommand.meType;
	
	return lCheck;
}

void Unit::vDumpSyncText(FILE *fileDump)
{
	// do superclass
	DrawObj::vDumpSyncText(fileDump);

	// do self
	fprintf(fileDump, " *Unit\n");
	fprintf(fileDump, "  Alive?: %d\n", (int)mbAlive);
	fprintf(fileDump, "  Between points?: %d\n", (int)mbBetweenPoints);
	fprintf(fileDump, "  Cloaked?: %d\n", (int)mbCloaked);
	fprintf(fileDump, "  Moving?: %d\n", (int)mbMoving);
	fprintf(fileDump, "  Behavior: %d\n", (int)meBehavior);
	fprintf(fileDump, "  Type: %d\n", (int)meType);
	fprintf(fileDump, "  PixelX: %ld\n", mlPixelX);
	fprintf(fileDump, "  PixelY: %ld\n", mlPixelY);
	fprintf(fileDump, "  Attack Range: %d\n", mnAttackRange);
	fprintf(fileDump, "  Attack Speed: %d\n", mnAttackSpeed);
	fprintf(fileDump, "  Died on cycle: %d\n", mnDeathCycle);
	fprintf(fileDump, "  Health: %f\n", mfHealth);
	fprintf(fileDump, "  Last attacked on cycle: %d\n", mnLastAttackCycle);
	fprintf(fileDump, "  Last target X: %d\n", mnLastTargetX);
	fprintf(fileDump, "  Last target Y: %d\n", mnLastTargetY);
	fprintf(fileDump, "  Max Health: %f\n", mfMaxHealth);
	fprintf(fileDump, "  Max Power: %f\n", mfMaxPower);
	fprintf(fileDump, "  Partial Move: %d\n", mnPartialMove);
	fprintf(fileDump, "  Power: %f\n", mfPower);
	fprintf(fileDump, "  Speed: %d\n", mnSpeed);
	fprintf(fileDump, "  Stopped for cycles: %d\n", mnStoppedCycles);
	fprintf(fileDump, "  Temporarily blocked for cycles: %d\n", mnTempBlockedCycles);
	fprintf(fileDump, "  Aggression: %d\n", mnSpeed);
	fprintf(fileDump, "  Tenacity: %d\n", mnSpeed);
	fprintf(fileDump, "  Current command type: %d\n", (int)moCurrentCommand.meType);
}

long Unit::lDistToSpace(int nX, int nY)
{
	int nDX = abs(nX - mpoCurNode->mnXPos);
	int nDY = abs(nY - mpoCurNode->mnYPos);
	return max(nDX, nDY);
}

void Unit::vSetPos(WORD wX, WORD wY)
{
	GameMap *poMap = mpoPlane->mpoMap;
	MapNode *poNode = poMap->poGetNode(wX, wY);

	if (poNode) 
	{
		mpoCurNode = poNode;
		mbBetweenPoints = FALSE;
		vSetPixelPos(poNode->nPixelX(), poNode->nPixelY());
	}
}

void Unit::vSetPixelPos(int nX, int nY)
{
	mlPixelX = nX;
	mlPixelY = nY;

	vMarkUpdate();
}

BOOL Unit::bStuck()
{
	return (mnStoppedCycles || mnTempBlockedCycles);
}

void Unit::SetFrozen()
{
	mjAlpha = 191;
	mbAlive = FALSE;
}

void Unit::ClearFrozen()
{
	mjAlpha = 255;
	mbAlive = TRUE;
}

void Unit::vCommandNone()
{
	if (meAction == ACT_WALK || meAction == ACT_RUN || meAction == ACT_FEED)
	{
		vFidget();
	//	vSetAction(ACT_FIDGET_CLOAKED);
	}

	moCurrentCommand.meType = CM_NONE;
	mdwFlags = mdwFlags & (~(DWORD)UNIT_FLAG_WAITING );
	moCurrentRoute.RemoveAll();
	
	if (mnAttackee) 
	{
		Unit * poAttackee = (Unit*)goObjectDB.poGetObject(mnAttackee);
		if (poAttackee)
		{
			if (poAttackee->bAlive())
			{
				poAttackee->vFidget();
			}
			poAttackee->vRemoveAttacker(this);
			mnAttackee = NULL;
		}
	}
}

void Unit::vFidget()
{
	if (bIsVampire())
	{
		if(bIsCloaked())
		{
			vSetAction(ACT_FIDGET_CLOAKED);
		}
		else {
			vSetAction(ACT_FIDGET_UNCLOAKED);
		}
	} else {
		//we're a peassant
		vSetAction(ACT_FIDGET_CLOAKED);
	}
	
}

BOOL Unit::bSpaceVisible(int nX, int nY)
{
	return TRUE;
}

//make all other non-local units inside same building visible to player
void Unit::vCheckVisibility (void) 
{
	DrawObj *poCover = mpoCurNode->poGetCover();

	//
	// NOTE: This function assumes that the decision as to whether
	// a building is open or closed has already been made.
	//

	if (poCover && poCover->bIsBuilding()) 
	{
		if (((Building*)poCover)->bIsOpen()) 
		{

			// if we're in an open building, we're visible
			vSetVisibility(TRUE);
		}
		else 
		{

			// if we're in a closed building, we're not
			vSetVisibility(FALSE);
		}
	
	}
	else 
	{
		// else we're not in a building so for now just set to TRUE
		vSetVisibility(TRUE);
	}
}

BOOL Unit::GetFlag(int nFlag)
{
	if (nFlag != -1)
	{
		FLAGENTRY * pn = mpnFlags;
		while (pn)
		{
			if (pn->value == nFlag)
				return TRUE;
			pn = pn->next;
		}
	}
	return FALSE;
}

void Unit::SetFlag(int nFlag)
{
	if (nFlag != -1)
	{
		FLAGENTRY * pn = new FLAGENTRY;
		
		pn->value = nFlag;
		pn->next = mpnFlags;
		mpnFlags = pn;
	}	
}

void Unit::ClearFlag(int nFlag)
{
	if (nFlag != -1)
	{
		FLAGENTRY * pn = mpnFlags;
		FLAGENTRY * pnLast = pn;
		FLAGENTRY * pnNext = pn->next;

		while (pn)
		{
			if (pn->value == nFlag)
			{
				if (pnLast != pn)
				{
					pnLast->next = pn->next;
				}
				else
				{
					mpnFlags = pn->next;
				}
				delete pn;
				return;
			}
			pnLast = pn;
			pn = pn->next;
		}
	}
}

void Unit::vAttack()
{
	int nCurCycle = game->nGameCycle();
	if (nCurCycle - mnLastAttackCycle < mnAttackSpeed) 
	{
		return;
	}
	mnLastAttackCycle = nCurCycle;

	if (!mnAttackee) 
	{
		return;
	}
	Unit * poAttackee = (Unit *)goObjectDB.poGetObject(mnAttackee);
	if (!poAttackee) return;

	poAttackee->SetAttacker(this, 20);	
	int nAttackBonus = 0;
	int nDefendBonus = 0;

	if (poAttackee->bNotEngagedWithAttacker(this))
	{
		nAttackBonus = 1;
		nDefendBonus = -1;
	}
	int nMinimum = (int)(((mnLevel + nAttackBonus + GetUpgradeAttack()) - (poAttackee->nLevel() + nDefendBonus + poAttackee->GetUpgradeDefend())) * vGetClan()->GetXMinDamage());
	int nMaximum = max(10, (int)(((mnLevel + nAttackBonus + GetUpgradeAttack()) * vGetClan()->GetXMaxDamageAttack()) - ((poAttackee->nLevel() + nDefendBonus + poAttackee->GetUpgradeDefend()) * poAttackee->vGetClan()->GetXMaxDamageDefend())));
	int nDamage = nMinimum + GlobalRand(nMaximum - nMinimum);
	vDebugOut("min%d max%d damage%d", nMinimum, nMaximum, nDamage);

	if (nDamage > 0)
	{
		if (mpoOwner == game->poLocalPlayer())		// HIT
		{
			vPlayRandomSound(TAG_Satk);
		}
		poAttackee->vDamage((float)nDamage, this);	
	}
	else
	{
		if (mpoOwner == game->poLocalPlayer())		// MISS
		{
			vPlayRandomSound(TAG_Smis);
		}
	}
}

BOOL Unit::bNotEngagedWithAttacker(Unit * poUnitAttacker)
{
	// UNABLE TO RESPOND
	if (mnThrallCyclesRemaining)
	{
		return TRUE;
	}
	// ENGAGED
	if (moCurrentCommand.meType == CM_ATTACK)
	{
		//  WITH ANOTHER UNIT
		if (moCurrentCommand.mnUnit != poUnitAttacker->nGetID())
			return TRUE;
	}
	return FALSE;
}

void Unit::SetAttacker(Unit * poAttacker, int nCycles)
{
	assert(poAttacker);
	assert(nCycles >= 0);
	mnUnderAttackCycles = nCycles;
	mpoAttacker = poAttacker;
}

FLOAT Unit::vDrainPower(FLOAT fDamage)
{
	FLOAT fReturnDifference = mfPower;

	mfPower = max(0, mfPower - fDamage);
	fReturnDifference = fReturnDifference - mfPower;
	return fReturnDifference;
}

void Unit::IncreasePower(FLOAT fPower)
{
	FLOAT fPowerDesired = mfMaxPower - mfPower;

	// ADJUST TO CLAN/UNIT RANGES 
	fPower = min(fPowerDesired, fPower);

	mfPower = mfPower + fPower;
}

FLOAT Unit::vDamage(FLOAT fDamage, Unit * poUnitFrom)
{
	FLOAT fReturnDifference = mfHealth;

	mfHealth = max(0, mfHealth - fDamage);
	fReturnDifference = fReturnDifference - mfHealth;
	
	if (mfHealth == 0) 
	{
		mnDeathCycle = game->nGameCycle();
		mbAlive = FALSE;
		mpoCurNode->mpoUnit = NULL;
		vSetAction(ACT_DIE);
		
		vPlayRandomSound(TAG_Sdth);
		if (bIsMobMember())
		{
			BOOL bOk = FALSE;
			mpoMobCoordinator->RemoveUnit(&bOk, this);

			// GO AHEAD AND DELETE IF EMPTY
			if (!mpoMobCoordinator->GetSize())
			{
				delete mpoMobCoordinator;
			}

			// DEFINITELY REMOVE REFERENCE SINCE UNIT IS NOT DELETED
			mpoMobCoordinator = NULL;
		}
		//vSetBitmap(mpoTemplate->lGetValue(TAG_Bded, nBitmapID()));
		vSetLayer(DP_ITEMS);

		//Missile * poMissile = new Missile(MISSILE_TIMERFADEDRAWOBJ, gpoMap->mpoPlane, this, mpoCurNode, mpoCurNode, 6019, 0, FALSE, 1.18f, NULL, 10.0, 10.0, 2, 0);
	
		mpoCurNode->vAddObjToFront(this);
		vCommandNone();
		mbSelectable = FALSE;
		if (mbSelected) 
		{
			gpoMap->vRemoveSelection(this);
			gpoMap->vCommitSelection();
			vMarkUpdate();
		}

		// BLAST ANXIETY
		if (poUnitFrom && poUnitFrom->bIsVampire())
		{
			Vampire * poVampire = (Vampire *)poUnitFrom;
			poVampire->FireAnxiety(((Clan *)(poVampire->vGetClan()))->GetAnxietyBlastForKilling());
		}

	} else {
		if (meAction != ACT_FEED) {
			vSetAction(ACT_REACT);
		}
	}

	return fReturnDifference;
}
/*
#define MAXSCANDISTANCE 20

void Unit::vMoveToNearNode(MapNode * poNode)
{
	// SHOULD BE A VALID POINTER
	assert(poNode);

	// SHOULD HAVE A NODE
	assert(mpoCurNode);

	// SHOULD NOT HAVE A UNIT AT NODE
	assert(poNode->mpoUnit);	

	// NODE SHOULD BE ONE UNIT AWAY
	assert(abs(poNode->mnX - mpoCurNode->mnX)<=1);	
	assert(abs(poNode->mnY - mpoCurNode->mnY)<=1);

	// OKAY SCAN THE OLD LINES AND REMOVE UNITS
}
void Unit::vSetNode(MapNode * poNode)
{
	// SHOULD NOT HAVE A UNIT AT NODE
	assert(poNode->mpoUnit);	
	// SHOULD NOT CURRENTLY HAVE A NODE
	assert(mpoCurNode);

	// SCAN THE WHOLE MAXSCANDISTANCE AREA
	// CLIPPED TO MAP EDGES
}
*/

void Unit::vScan()
{
	// CREATE LIST OF OBJECTS WITHIN VISIBLE RANGE
	int nCenterX = nGetX();
	int nCenterY = nGetY();
	int nVisibleRange = vGetVisibleRange();

	int nLeft = max(0, nCenterX - nVisibleRange);
	int nRight = min(gpoMap->nGetWidth()-1, nCenterX + nVisibleRange);
	int nTop = max(0, nCenterY - nVisibleRange);
	int nBottom = min(gpoMap->nGetHeight()-1, nCenterY + nVisibleRange);

	moVisibleUnits.RemoveAll();

	for (int i=0; i < ALLPLAYERINDICES; i++)
	{
		mnVisibleTotal[i].attack = 0;
		mnVisibleTotal[i].defense = 0;
		mnVisibleTotal[i].level = 0;
		mnVisibleTotal[i].count = 0;
		mnVisibleTotal[i].relativehealth = 0;
	}

	int nPlayer;
	if (mpoOwner)
	{
		nPlayer = mpoOwner->mnPlayer;
	}
	else
	{
		nPlayer = NONPLAYERINDEX;
	}

	// THIS
	mnVisibleTotal[nPlayer].count++;

	// ACCUMULATE TOTALS
	// TBD - JUST USE VISIBLE LEVELS 
	mnVisibleTotal[nPlayer].level += mnLevel; 
	mnVisibleTotal[nPlayer].attack += GetUpgradeAttack();
	mnVisibleTotal[nPlayer].defense += GetUpgradeDefend();
	mnVisibleTotal[nPlayer].relativehealth += vGetRelativeHealth(mpoOwner);

	// 
	//  TBD - BETTER SCAN ALGORITHM.  ALSO, INCLUDE VISIBILITY,
	//			LINE OF SIGHT,
	//
#if 0
	int nRange = 1;
	while (nRange < nVisibleRange)
	{
		MapNode * poNode;

		int y = nTop;
		int l = max(nLeft, nCenterX - nRange);
		int r = min(nRight, nCenterX + nRange);
		while(y <= nBottom)
		{	
			poNode = gpoMap->poGetNode(l, y);
			CheckNodeForVisibleUnits(poNode, nRange);
			poNode = gpoMap->poGetNode(r, y);
			CheckNodeForVisibleUnits(poNode, nRange);
			y++;
		}
		int x = nLeft + 1;
		int b = min(nBottom, nCenterY + nRange);
		int t = max(nTop, nCenterY - nRange);
		while (x < nRight)
		{
			poNode = gpoMap->poGetNode(x, b);
			CheckNodeForVisibleUnits(poNode, nRange);
			poNode = gpoMap->poGetNode(x, t);
			CheckNodeForVisibleUnits(poNode, nRange);
			x++;
		}
		nRange++;
	}
#endif
	// we fudge a little on the squaring so as to get a more pleasing circle
	int nSquareRange = nVisibleRange * nVisibleRange + nVisibleRange;
	MapNode *poNode;
	int j;
	int nRange;
	int nVisibility;
	int nLocalMask;
	if (mpoOwner && mpoOwner == game->poLocalPlayer()) 
	{
		nLocalMask = 0xFFFFFFFF;
	} 
	else if (mwSpyFlags & ((WORD)1 << game->poLocalPlayer()->mnPlayer))
	{
		nLocalMask = 0xFFFFFFFF;
	} 
	else
	{
		nLocalMask = 0x0;
	}

	//
	// we do a 1/8th triangle of the full range area
	// then use rotations and reflections to check the other nodes
	for (i = -nVisibleRange; i <= 0; ++i) {
		for (j = i; j <= 0; ++j) {
			int nSquare = i * i + j * j;
			if (nSquare <= nSquareRange) {
				double fRange = sqrt((double)nSquare);
				nRange = (int)fRange;

#if VIS_FUZZY_BORDER
				nVisibility = (int)(((nVisibleRange + 0.5) - fRange) * (255.0 / VIS_FUZZY_BORDER) + 0.5);
				nVisibility = min(255, max(0, nVisibility));
#else
				nVisibility = 255;
#endif
				// this sets non-local players' visibility to zero without
				// using an extra if -- test is above, outside the loop
				nVisibility &= nLocalMask;

				// do all 8 reflected combinations
				poNode = gpoMap->poGetNodeRC(nCenterX + i, nCenterY + j);
				if (poNode) {
					CheckNodeForVisibleUnits(poNode, nRange);
					poNode->vSetVisibility((BYTE)nVisibility);
				}
				if (i != j) {
					poNode = gpoMap->poGetNodeRC(nCenterX + j, nCenterY + i);
					if (poNode) {
						CheckNodeForVisibleUnits(poNode, nRange);
						poNode->vSetVisibility((BYTE)nVisibility);
					}
				}

				if (j) {
					// don't bother doing -j if j == 0
					poNode = gpoMap->poGetNodeRC(nCenterX + i, nCenterY - j);
					if (poNode) {
						CheckNodeForVisibleUnits(poNode, nRange);
						poNode->vSetVisibility((BYTE)nVisibility);
					}
					if (i != -j) {
						poNode = gpoMap->poGetNodeRC(nCenterX - j, nCenterY + i);
						if (poNode) {
							CheckNodeForVisibleUnits(poNode, nRange);
							poNode->vSetVisibility((BYTE)nVisibility);
						}
					}
				}
				if (i) {
					// don't bother doing -i if i == 0
					poNode = gpoMap->poGetNodeRC(nCenterX - i, nCenterY + j);
					if (poNode) {
						CheckNodeForVisibleUnits(poNode, nRange);
						poNode->vSetVisibility((BYTE)nVisibility);
					}
					if (i != -j) {
						poNode = gpoMap->poGetNodeRC(nCenterX + j, nCenterY - i);
						if (poNode) {
							CheckNodeForVisibleUnits(poNode, nRange);
							poNode->vSetVisibility((BYTE)nVisibility);
						}
					}
					if (j) {
						// don't bother doing -j if j == 0
						poNode = gpoMap->poGetNodeRC(nCenterX - i, nCenterY - j);
						if (poNode) {
							CheckNodeForVisibleUnits(poNode, nRange);
							poNode->vSetVisibility((BYTE)nVisibility);
						}
						if (i != j) {
							poNode = gpoMap->poGetNodeRC(nCenterX - j, nCenterY - i);
							if (poNode) {
								CheckNodeForVisibleUnits(poNode, nRange);
								poNode->vSetVisibility((BYTE)nVisibility);
							}
						}
					}
				}
			}
		}
	}
}

void Unit::MarkUnit(Unit * poUnit)
{
	assert(poUnit);
	assert(poUnit->bIsVampire());

	Vampire * pVampire = (Vampire *)poUnit;

	pVampire->vSetMarked(mpoOwner->mnPlayer, poUnit->vGetRelativeHealth(mpoOwner), poUnit->mnLevel);
	pVampire->moStatus.vUpdateHealth(poUnit->vGetRelativeHealth(mpoOwner));
	pVampire->moStatus.vShow();
}

void Unit::CheckNodeForVisibleUnits(MapNode * poNode, int nDistance)
{
	assert(poNode);
	if (poNode->mpoUnit)
	{
		CheckNodeThatHasUnits(poNode, nDistance);
	}
}

void Unit::CheckNodeThatHasUnits(MapNode * poNode, int nDistance)
{
	if (poNode->mpoUnit->IsInvisible()) return;

	BOOL bMarkAsVampire = FALSE;
	BOOL bIgnoreFriendly = FALSE;
	BOOL bIsFriendly = FALSE;

	if (poNode->mpoUnit->bIsVampire())			// VAMPIRE
	{
		if (mpoOwner)								
		{
			if (poNode->mpoUnit->mpoOwner != mpoOwner)	// NOT ON THE SAME TEAM
			{
				if (!poNode->mpoUnit->mbCloaked)		// NOT CLOAKED
				{
					MarkUnit(poNode->mpoUnit);
					bMarkAsVampire = TRUE;
				}
				else									// CLOAKED
				{
					// UNIT IS DETECTABLE
					int nVisibleRangeIndex = (mnLevel + GetUpgradeDetection()) - (poNode->mpoUnit->mnLevel + poNode->mpoUnit->GetUpgradeStealth());
					if (nVisibleRangeIndex>=0)
					{
						if (nVisibleRangeIndex > 2)
						{
							nVisibleRangeIndex = 2;
						}
						if (nDistance <= mnVampireRange[nVisibleRangeIndex])  
						{
							// DETECTED
							MarkUnit(poNode->mpoUnit);
							bMarkAsVampire = TRUE;
						}
					}
					if (!bMarkAsVampire)
					{
						if (((Vampire *)(poNode->mpoUnit))->bMarked(mpoOwner->mnPlayer))
						{
							// ALREADY DETECTED
							// TBD - FOR NOW JUST MARK AS VAMPIRE -- SEE TAG 1 BELOW
							bMarkAsVampire = TRUE;
						}
					}
				}
			}
			else
			{
				bIsFriendly = TRUE;
				bMarkAsVampire = TRUE;
				if (nDistance > vGetClan()->GetDistanceIncludeFriendlies())
				{
					bIgnoreFriendly = TRUE;
				}
			}
		}
		else
		{
			if (!poNode->mpoUnit->mbCloaked)		// NOT CLOAKED
			{
				bMarkAsVampire = TRUE;
			}
		}
	}

	int nEnemyIndex;

	if (bMarkAsVampire)
	{
		if (!bIsFriendly)
		{
			// ENEMY VAMPIRE!
			nEnemyIndex = poNode->mpoUnit->mpoOwner->mnPlayer;

			if (bIsVampire())
				AlertClan(nEnemyIndex, TRUE);

			int nApparentHealth = 0;
			int nApparentLevel = 0;
			if (mpoOwner)
			{
				((Vampire*)(poNode->mpoUnit))->vGetMarked(mpoOwner->mnPlayer, &nApparentHealth, &nApparentLevel);
			}
			else
			{
				nApparentHealth = poNode->mpoUnit->vGetRelativeHealth(mpoOwner);
				nApparentLevel = poNode->mpoUnit->mnLevel;
			}
			mnVisibleTotal[nEnemyIndex].level += nApparentLevel;
			mnVisibleTotal[nEnemyIndex].attack += poNode->mpoUnit->GetUpgradeAttack();
			mnVisibleTotal[nEnemyIndex].defense += poNode->mpoUnit->GetUpgradeDefend();
			mnVisibleTotal[nEnemyIndex].relativehealth += nApparentHealth;	
		}
		else
		{
			// FRIENDLY
			nEnemyIndex = poNode->mpoUnit->mpoOwner->mnPlayer;
			if (!bIgnoreFriendly)
			{
				mnVisibleTotal[nEnemyIndex].level += poNode->mpoUnit->mnLevel;
				mnVisibleTotal[nEnemyIndex].attack += poNode->mpoUnit->GetUpgradeAttack();
				mnVisibleTotal[nEnemyIndex].defense += poNode->mpoUnit->GetUpgradeDefend();
				mnVisibleTotal[nEnemyIndex].relativehealth += poNode->mpoUnit->vGetRelativeHealth(mpoOwner);
			}
		}
	}
	else
	{
		// LOOKS LIKE A PEASANT
		nEnemyIndex = NONPLAYERINDEX;
		mnVisibleTotal[nEnemyIndex].relativehealth += poNode->mpoUnit->vGetRelativeHealth(mpoOwner);
	}
	mnVisibleTotal[nEnemyIndex].count++;

	// ACCUMULATE TOTALS
	// TBD - JUST USE VISIBLE LEVELS - SEE TAG 1 ABOVE
	// TBD - HAVE TO TRACK LAST SEEN UPGRADES TOO.

	//moVisibleUnits.AddToEnd(poNode->mpoUnit);
	moVisibleUnits.AddToEnd(poNode->mpoUnit->nGetID());
}

void Unit::CallForHelp()
{
}

BOOL Unit::OutGunnedByVisibleUnits(Unit * poUnit)
{
	int nEnemyPlayer;
	int nFriendlyPlayer;
	if (mpoOwner)
	{
		nFriendlyPlayer = mpoOwner->mnPlayer;
	}
	else
	{
		nFriendlyPlayer = NONPLAYERINDEX;
	}

	int nStrengthFriendly = (mnVisibleTotal[nFriendlyPlayer].level + (( mnVisibleTotal[nFriendlyPlayer].attack + mnVisibleTotal[nFriendlyPlayer].defense) / 2)) * mnVisibleTotal[nFriendlyPlayer].relativehealth / mnVisibleTotal[nFriendlyPlayer].count;

	if (poUnit->mpoOwner)
	{
		nEnemyPlayer = poUnit->mpoOwner->mnPlayer;
	}
	else
	{
		nEnemyPlayer = NONPLAYERINDEX;
	}

	int nStrengthEnemy = (mnVisibleTotal[nEnemyPlayer].level + mnVisibleTotal[nEnemyPlayer].count / 2) * (mnVisibleTotal[nEnemyPlayer].relativehealth + poUnit->vGetRelativeHealth(mpoOwner)) / (mnVisibleTotal[nEnemyPlayer].count + 1);
	
	if (nStrengthFriendly < (3 * nStrengthEnemy / 2))
	{
		return TRUE;
	}
	return FALSE;
}

BOOL Unit::bShouldRetreat()
{
	// VAMPIRE
	if (bIsVampire())
	{
		if (mnUnderAttackCycles || moCurrentCommand.meType == CM_ATTACK)
		{
			// HEALTH BELOW TENACITY
			if (!bStayAndFight())
			{
				return TRUE;
			}
		}
		else
		{
			// HEALTH BELOW TENACITY
			if (!bStayAndFight())
			{
				// IF ENEMY WITHIN 4 UNITS
				Unit * poUnit;
				int i = moVisibleUnits.GetSize();
				while(i)
				{
					i--;
					poUnit = (Unit*)goObjectDB.poGetObject(moVisibleUnits.GetItem(i));

					if ((poUnit == NULL)||(poUnit == this)) continue;
					
					if (poUnit->bIsVampire() || poUnit->bIsMobMember() || !poUnit->bNotEngagedWithAttacker(this))
					{
						if (4 >= poUnit->lDistToSpace(nGetX(), nGetY()))
							return TRUE;
					}
				}
			}
		}
	}
	return FALSE;
}

void Unit::Retreat()
{
	/*
	Command oCommand;
	oCommand.mnX = 90 * mpoOwner->mnPlayer + 5;
	oCommand.mnY = 90 * mpoOwner->mnPlayer + 5;
	oCommand.meType = CM_MOVE;
	vCommand(oCommand, TRUE);
	*/
	CommandSequenceClear();

	Command oCommand;
	oCommand.meType = CM_RETREAT;
	vCommand(oCommand, TRUE);
}

BOOL Unit::bStayAndFight()
{
	switch (mTenacity)
	{
	case SETTING_LOW:
		// RETREAT IF 2/3 HEALTH REMAINS
		if (mfHealth < (mfMaxHealth * 2/3))
		{
			return FALSE;
		}
		break;
	case SETTING_MEDIUM:
		// RETREAT IF 1/3 HEALTH REMAINS
		if (mfHealth < (mfMaxHealth / 3))
		{
			return FALSE;
		}
		break;
	}
	return TRUE;
}

BOOL Unit::bAttackEnemyUnit(Unit * poUnit)
{
	// ATTACK
	switch (mAggression)
	{
	case SETTING_MEDIUM:
		if (!OutGunnedByVisibleUnits(poUnit))
		{
			return TRUE;
		}
		break;
	case SETTING_HIGH:
		return TRUE;
	}
	return FALSE;
}

BOOL Unit::bIsInside()
{
	DrawObj * poCover = mpoCurNode->poGetCover();
	if (poCover && poCover->bIsBuilding()) 
	{
		return TRUE;
	}
	return FALSE;
}

#define ACTION_INSIDE_1 0
#define ACTION_INSIDE_2 1
#define ACTION_INSIDE_3 2
#define ACTION_INSIDE_COUNT 3

#define ACTION_OUTSIDE_1 0
#define ACTION_OUTSIDE_2 1
#define ACTION_OUTSIDE_3 2
#define ACTION_OUTSIDE_COUNT 3

void Unit::SelectNewRandomBehavior()
{
	if (bIsInside())
	{
		int nAction = (GlobalRand(ACTION_INSIDE_COUNT));
		{
			Sequence_Clear();
			switch(nAction)
			{
			case ACTION_INSIDE_1:
				{
					Building * pBuilding = (Building *)mpoCurNode->poGetCover();
					Sequence_WalkToBuildingOutside(pBuilding);	// walk outside
					Sequence_WaitLong();						// wait long
				}
				break;
				
			case ACTION_INSIDE_2:
				{
					Building * pBuildingA = (Building *)mpoCurNode->poGetCover();
					Building * pBuildingB = NULL;
					AcquireNewBuilding(&pBuildingB, pBuildingA);
					assert(pBuildingB);
					Sequence_WalkToBuildingInside(pBuildingB);
					Sequence_WaitShort();					
					Sequence_WalkToBuildingOutside(pBuildingB);
					Sequence_WaitLong();					
					Sequence_WalkToBuildingInside(pBuildingA);
					Sequence_WaitLong();					
				}
				break;
			case ACTION_INSIDE_3:
				{
					Building * pBuildingA = (Building *)mpoCurNode->poGetCover();
					Building * pBuildingB = NULL;
					AcquireNewBuilding(&pBuildingB, pBuildingA);
					assert(pBuildingB);
					// walk to other building
					Sequence_WalkToBuildingOutside(pBuildingB);	// walk to inside building B
					// stand outside medium
					Sequence_WaitLong();						// wait long
					// walk inside short
					Sequence_WalkToBuildingInside(pBuildingB);	// walk to inside building B
					Sequence_WaitMedium();						// wait short
					Sequence_WalkToBuildingOutside(pBuildingB);	// walk outside
					Sequence_WaitLong();						// wait long
				}
				break;
			}
		}
	}
	else
	{
		int nAction = (GlobalRand(ACTION_OUTSIDE_COUNT));
		{
			Sequence_Clear();
			switch(nAction)
			{
			case ACTION_OUTSIDE_1:
				{
					// wander in general direction for [n]while
					// pause to rest periodically
					Sequence_WaitLong();						// wait long
					Sequence_WaitLong();						// wait long
					Sequence_WanderMethodically(2);
					Sequence_WaitMedium();						// wait long
				}
				break;
			case ACTION_OUTSIDE_2:
				{
					// wander methodically
					Sequence_WaitLong();						// wait long
					Sequence_WaitLong();						// wait long
					Sequence_WanderMethodically(5);
					Sequence_WaitMedium();
				}
				break;
			case ACTION_OUTSIDE_3:
				{
					Sequence_WaitMedium();
					Building * pBuildingA = (Building *)mpoCurNode->poGetCover();
					Building * pBuildingB = NULL;
					// walk to inside building
					AcquireNewBuilding(&pBuildingB, pBuildingA);
					//assert(pBuildingB);
					if (pBuildingB != NULL && pBuildingB->bHasInterior())
					{
						Sequence_WalkToBuildingInside(pBuildingB);
						Sequence_WaitLong();						// wait long
					}
					else
					{
						Sequence_WanderMethodically(5);
					}
				}
				break;
			}
		}
	}
}

void Unit::Sequence_Clear()
{
	//vDebugOut("Unit::Sequence_Clear\n");
	CommandSequenceClear();
	assert(!mpoSequencer);
	mpoSequencer = new SequentialCommands(*this);
}

void Unit::Sequence_WaitShort()
{
	//vDebugOut("Unit::Sequence_WaitShort\n");
	assert(mpoSequencer);

	// queue command to wait
	Command oCommand;
	oCommand.meType = CM_WAIT_500MS;
	mpoSequencer->Add(&oCommand);
}

void Unit::Sequence_WaitMedium()
{
	//vDebugOut("Unit::Sequence_WaitMedium\n");
	assert(mpoSequencer);

	// queue command to wait
	Command oCommand;
	oCommand.meType = CM_WAIT_500MS;
	for (int i=0;i<5;i++)
		mpoSequencer->Add(&oCommand);
}

void Unit::Sequence_WaitLong()
{
	//vDebugOut("Unit::Sequence_WaitLong\n");
	assert(mpoSequencer);

	// queue command to wait
	Command oCommand;
	oCommand.meType = CM_WAIT_500MS;
	for (int i=0;i<15;i++)
		mpoSequencer->Add(&oCommand);
}


void Unit::Sequence_WalkToBuildingOutside(Building * pBuilding)
{
	//vDebugOut("Unit::Sequence_WalkToBuildingOutside\n");
	assert(mpoSequencer);
	assert(pBuilding);

	// determine position outside of this building
	MapNode * poNode = pBuilding->GetPositionOutside();

	// queue command to move here
	Command oCommand;
	oCommand.mnX = poNode->mnXPos;
	oCommand.mnY = poNode->mnYPos;
	oCommand.meType = CM_MOVE;

	mpoSequencer->Add(&oCommand);
}

void Unit::Sequence_WanderWithoutRest(int nDistance)
{
	assert(mpoSequencer);

	Command oCommand;
	Command oCommandWait;

	oCommand.meType = CM_MOVE;
	oCommandWait.meType = CM_WAIT_500MS;

	// determine number of steps
	int nCountSteps = nDistance;

	int nDirectionX = GlobalRand(11)-5;
	int nDirectionY = GlobalRand(11)-5;
	int nNowX = nGetX();
	int nNowY = nGetY();

	// wander for number of steps
	while(nCountSteps>0)
	{
		nCountSteps--;

		// DETERMINE LOCATION
		oCommand.mnX = nNowX + nDirectionX;
		oCommand.mnY = nNowY + nDirectionY;

		if (oCommand.mnX < 0)
		{
			oCommand.mnX = 0;
			nDirectionX = -nDirectionX;
		}
		else if (oCommand.mnX > gpoMap->nGetWidth()-1)
		{
			oCommand.mnX = gpoMap->nGetWidth()-1;
			nDirectionX = -nDirectionX;
		}
		if (oCommand.mnY < 0)
		{
			oCommand.mnY = 0;
			nDirectionY = -nDirectionY;
		}
		else if (oCommand.mnY > gpoMap->nGetHeight()-1)
		{
			oCommand.mnY = gpoMap->nGetHeight()-1;
			nDirectionY = -nDirectionY;
		}

		nNowX = oCommand.mnX;
		nNowY = oCommand.mnY;

		// QUEUE THE COMMAND
		mpoSequencer->Add(&oCommand);

		// ADJUST THE DIRECTION
		nDirectionX = nDirectionX + GlobalRand(3) - 1;
		nDirectionX = min(5, nDirectionX);
		nDirectionX = max(-5, nDirectionX);

		nDirectionY = nDirectionY + GlobalRand(3) - 1;
		nDirectionY = min(5, nDirectionY);
		nDirectionY = max(-5, nDirectionY);

		// OPTIONALLY WAIT AT EACH STEP FOR RANDOM TIME
		if (GlobalRand(7) == 1)
		{
			mpoSequencer->Add(&oCommandWait);
		}
	}
}

void Unit::Sequence_WanderMethodically(int nPauseFreq, int nDirectionX, int nDirectionY)
{
	if (nDirectionX == -10000)
	{
		nDirectionX = GlobalRand(11)-5;
	}
	if (nDirectionY == -10000)
	{
		nDirectionY = GlobalRand(11)-5;
	}
	//vDebugOut("Unit::Sequence_WanderMethodically\n");
	assert(mpoSequencer);

	Command oCommand;
	Command oCommandWait;

	oCommand.meType = CM_MOVE;
	oCommandWait.meType = CM_WAIT_500MS;

	// determine number of steps
	int nCountSteps = GlobalRand(10) + 25;

	int nNowX = nGetX();
	int nNowY = nGetY();

	// wander for number of steps
	while(nCountSteps>0)
	{
		nCountSteps--;

		// DETERMINE LOCATION
		oCommand.mnX = nNowX + nDirectionX;
		oCommand.mnY = nNowY + nDirectionY;

		oCommand.mnX = max(0, oCommand.mnX);
		oCommand.mnX = min(gpoMap->nGetWidth()-1, oCommand.mnX);

		oCommand.mnY = max(0, oCommand.mnY);
		oCommand.mnY = min(gpoMap->nGetHeight()-1, oCommand.mnY);

		nNowX = oCommand.mnX;
		nNowY = oCommand.mnY;

		// QUEUE THE COMMAND
		mpoSequencer->Add(&oCommand);

		// ADJUST THE DIRECTION
		nDirectionX = nDirectionX + GlobalRand(5) - 2;
		nDirectionX = min(5, nDirectionX);
		nDirectionX = max(-5, nDirectionX);

		nDirectionY = nDirectionY + GlobalRand(5) - 2;
		nDirectionY = min(5, nDirectionY);
		nDirectionY = max(-5, nDirectionY);

		// OPTIONALLY WAIT AT EACH STEP FOR RANDOM TIME
		if (nPauseFreq)
		{
			if (GlobalRand(nPauseFreq))
			{
				mpoSequencer->Add(&oCommandWait);
			}
		}
	}
}

void Unit::Sequence_WalkToBuildingInside(Building * pBuilding)
{
	//vDebugOut("Unit::Sequence_WalkToBuildingInside\n");
	assert(mpoSequencer);
	assert(pBuilding);

	// determine position inside of this building
	MapNode * poNode = pBuilding->GetPositionInside();
	if (!poNode)
	{
		poNode = pBuilding->GetPositionOutside();
	}

	// queue command to move here
	Command oCommand;
	oCommand.mnX = poNode->mnXPos;
	oCommand.mnY = poNode->mnYPos;
	oCommand.meType = CM_MOVE;

	mpoSequencer->Add(&oCommand);
}

int Unit::vGetVisibleRange()
{
	return vGetClan()->GetBaseSight();
}

void Unit::GetStatusIndicatorOffset(int * px, int * py)
{
	*px = mpoTemplate->lGetValue(TAG_stiX, vGetClanTemplate()->lGetValue(TAG_stiX, 0));
	*py = mpoTemplate->lGetValue(TAG_stiY, vGetClanTemplate()->lGetValue(TAG_stiY, 0));
}

void Unit::SetSpeed(SPEED_ENUMTYPE eSpeed)
{
	int n;
	switch (eSpeed)
	{
	case SPEED_NORMAL:
		n = DEFAULT_SPEED_NORMAL_JUSTINCASE;
		vOverride(&n, TAG_sped);
		break;
	case SPEED_CLOAKED:
		vOverride(&n, TAG_cspd);
		break;
	case SPEED_ADRENALINE:
		n = DEFAULT_SPEED_NORMAL_JUSTINCASE;
		vOverride(&n, TAG_sped);
		SetSpeedAdrenalineBoost();
		break;
	case SPEED_MOBBING:
		n = 5;
		break;
	}
	mnSpeed = n;
	mnSpeed = mnSpeed - GetUpgradeSpeed();
}

Bitmap * Unit::GetHighlightBitmap()
{
	return game->GetHighlightBitmap();
}

int Unit::FindDirectionToOffset(int x, int y)
{
	assert(x!=0 || y!=0);  // NOT SAME NODE
	double dX = x;
	double dY = y;
	if (y == 0)
	{
		if (x > 0)
			return 0;
		else
			return 4;
	}
	else if (x == 0)
	{
		if (y > 0)
			return 6;
		else
			return 2;
	}
	double dAngle = atan(fabs((double)dY / (double)dX));

	// ADJUST dAngle
	if (y < 0)
	{
		if (x > 0)
		{
			
		}
		else
		{
			dAngle = PI - dAngle;
		}
	}
	else
	{
		if (x > 0)
		{
			dAngle = 2*PI - dAngle;
		}
		else
		{
			dAngle = PI + dAngle;
		}
	}
	dAngle = dAngle + 2*PI / 16;
	if (dAngle> 2*PI) dAngle = dAngle - 2*PI;
	int nReturn = (int)(dAngle / (2*PI / 8));
	assert(nReturn >= 0 && nReturn < 8);
	return nReturn;
}


#define ADJUSTTOPERCENT(n, p) {n=(n)*(p);}
void Unit::FindRetreatLocation(int nDistance, int * pnReturnX, int * pnReturnY)
{

	int aOffset[] = {	 1,  0, 
						 1, -1,
						 0, -1,
						-1, -1, 
						-1,  0,
						-1,  1,
						 0,  1,
						 1,  1 };


	double afDirectionScore[8] = { 1, 1, 1, 1, 1, 1, 1, 1 };

	//321	
	//4d0
	//567

	// ADJUST FOR GAME MAP SURROUNDINGS
	if (nGetX() == 0)
	{
		afDirectionScore[3] = 0;
		afDirectionScore[4] = 0;
		afDirectionScore[5] = 0;
	}
	else if (nGetX() == gpoMap->nGetWidth() - 1)
	{
		afDirectionScore[1] = 0;
		afDirectionScore[0] = 0;
		afDirectionScore[7] = 0;
	}

	if (nGetY() == 0)
	{
		afDirectionScore[1] = 0;
		afDirectionScore[2] = 0;
		afDirectionScore[3] = 0;
	}
	else if (nGetY() == gpoMap->nGetHeight() - 1)
	{
		afDirectionScore[7] = 0;
		afDirectionScore[6] = 0;
		afDirectionScore[5] = 0;
	}
	

	int nX = nGetX();
	int nY = nGetY();
	int j;
	for (j = 0; j<8; j++)
	{
		if (afDirectionScore[j] == 1)
		{
			MapNode * poNode = gpoMap->poGetNode(nX + aOffset[j*2], nY + aOffset[j*2 + 1]);
			TERRAINTYPE t = poNode->eTerrainType();
			switch (t)
			{
			case TT_CLEAR:
				break;
			case TT_OBSTACLE:
				afDirectionScore[j] = 0;
				break;
			default:
				afDirectionScore[j] = 0.5;
				break;
			}
		}
	}

	if (mnLastRetreatDirection != -1)
	{
		ADJUSTTOPERCENT(afDirectionScore[(mnLastRetreatDirection+4)%8], 0.25);
	}

	// ADJUST FOR GAME MAP POSITION
	if ((double)nGetX() > (0.9 * ((double)gpoMap->nGetWidth()-1)))
	{
		ADJUSTTOPERCENT(afDirectionScore[2], 0.07);
		ADJUSTTOPERCENT(afDirectionScore[1], 0.03);
		ADJUSTTOPERCENT(afDirectionScore[0], 0.015);
		ADJUSTTOPERCENT(afDirectionScore[7], 0.03);
		ADJUSTTOPERCENT(afDirectionScore[6], 0.07);
	}
	else if ((double)nGetX() < (0.1* ((double)gpoMap->nGetWidth()-1)))
	{
		ADJUSTTOPERCENT(afDirectionScore[6], 0.07);
		ADJUSTTOPERCENT(afDirectionScore[5], 0.03);
		ADJUSTTOPERCENT(afDirectionScore[4], 0.015);
		ADJUSTTOPERCENT(afDirectionScore[3], 0.03);
		ADJUSTTOPERCENT(afDirectionScore[2], 0.07);
	}
	if ((double)nGetY() < (0.1 * ((double)gpoMap->nGetHeight()-1)))
	{
		ADJUSTTOPERCENT(afDirectionScore[0], 0.07);
		ADJUSTTOPERCENT(afDirectionScore[1], 0.03);
		ADJUSTTOPERCENT(afDirectionScore[2], 0.015);
		ADJUSTTOPERCENT(afDirectionScore[3], 0.03);
		ADJUSTTOPERCENT(afDirectionScore[4], 0.07);
	}
	else if ((double)nGetY() > (0.9 * ((double)gpoMap->nGetHeight()-1)))
	{
		ADJUSTTOPERCENT(afDirectionScore[4], 0.07);
		ADJUSTTOPERCENT(afDirectionScore[5], 0.03);
		ADJUSTTOPERCENT(afDirectionScore[6], 0.015);
		ADJUSTTOPERCENT(afDirectionScore[7], 0.03);
		ADJUSTTOPERCENT(afDirectionScore[0], 0.07);
	}

	int i = moVisibleUnits.GetSize();
	while(i)
	{
		i--;
		Unit * poUnit = (Unit*)goObjectDB.poGetObject(moVisibleUnits.GetItem(i));
		if (poUnit == NULL) continue;
		if ((poUnit == this) || !poUnit->bAlive()) continue;
		if (!poUnit->bIsVampire()) continue;

		int nDistance = poUnit->lDistToSpace(nGetX(), nGetY());
		int nDirection = FindDirectionToOffset(poUnit->nGetX() - nGetX(), poUnit->nGetY() - nGetY());
		if (nDistance == 1)
		{
			// THIS DIRECTION SCORE SHOULD BE MULTIPLIED BY 0
			afDirectionScore[nDirection] = 0.0;
			ADJUSTTOPERCENT(afDirectionScore[(nDirection + 4) % 8], 1.2);

			// ADJACENT DIRECTIONS SHOULD BE MULTIPLIED BY 0.5
			int nOtherDirection;

			nOtherDirection = nDirection + 1;
			if (nOtherDirection == 8) nOtherDirection = 0;
			afDirectionScore[nOtherDirection] = afDirectionScore[nOtherDirection] * 0.5 / (1 + poUnit->bIsVampire());

			nOtherDirection = nDirection - 1;
			if (nOtherDirection == -1) nOtherDirection = 7;
			afDirectionScore[nOtherDirection] = afDirectionScore[nOtherDirection] * 0.5 / (1 + poUnit->bIsVampire());
			continue;
		}
		else// if (nDistance <= 4)
		{
			// THIS DIRECTION SCORE SHOULD BE MULTIPLIED BY 0.25
			afDirectionScore[nDirection] = afDirectionScore[nDirection] * 0.25;
			ADJUSTTOPERCENT(afDirectionScore[(nDirection + 4) % 8], 1.5);

			// ADJACENT DIRECTIONS SHOULD BE MULTIPLIED BY 0.75
			int nOtherDirection;

			nOtherDirection = nDirection + 1;
			if (nOtherDirection == 8) nOtherDirection = 0;
			afDirectionScore[nOtherDirection] = afDirectionScore[nOtherDirection] * 0.75;

			nOtherDirection = nDirection - 1;
			if (nOtherDirection == -1) nOtherDirection = 7;
			afDirectionScore[nOtherDirection] = afDirectionScore[nOtherDirection] * 0.75;
			continue;
		}
		/*
		else
		{
			// THIS DIRECTION SCORE SHOULD BE MULTIPLIED BY 0.5 * nDistance / (1+2*poUnitEnemy->bIsVampire())
			afDirectionScore[nDirection] = afDirectionScore[nDirection] * 0.5 * (nDistance / mnVisibleRange) / 2*poUnit->bIsVampire(); 
			continue;
		}*/
	}
	double dMax = -1;
	int nDirection = -1;
	int nCountOfMaxScores;
	nCountOfMaxScores = 0;
	
	for (i=0; i<8; i++)
	{
		if (afDirectionScore[i] == dMax)
		{
			nCountOfMaxScores++;
		} 
		else if (afDirectionScore[i] > dMax)
		{
			nDirection = i;
			nCountOfMaxScores = 1;
			dMax = afDirectionScore[i];
		}
	}
	if (mnLastRetreatDirection != -1)
	{
		// DIFFERENT DIRECTION
		if (nDirection != mnLastRetreatDirection)
		{
			// NO IMPROVEMENT SO USE SAME DIRECTION
			if (afDirectionScore[mnLastRetreatDirection] >= afDirectionScore[nDirection])
			{
				nDirection = mnLastRetreatDirection;
			}
			else
			{
				// IMPROVEMENT SO SELECT FURTHEST COANGULAR ONE (UNLESS THERE's JUST ONE)
				if (nCountOfMaxScores>1)
				{
					// int nPickOne = GlobalRand(nCountOfMaxScores);
					int nIn = mnLastRetreatDirection + 1;
					int nOut = mnLastRetreatDirection - 1;
					i = 8;
					while (i)
					{
						if (nIn == 8) nIn = 0;
						if (nOut == -1) nIn = 7;
						i--;
						if (afDirectionScore[nIn] == dMax)
						{
							nDirection = nIn;
							break;
						}
						i--;
						if (afDirectionScore[nOut] == dMax)
						{
							nDirection = nOut;
							break;
						}
						nOut--;
						nIn++;
					}
				}
			}
		}
	}
	mnLastRetreatDirection = nDirection;
	switch(nDirection)
	{
		case -1:
			*pnReturnX = 0;
			*pnReturnY = 0;
			break;
		default:
			*pnReturnX = aOffset[nDirection*2];
			*pnReturnY = aOffset[nDirection*2 + 1];
			break;
	}
}


void Unit::AcquireNewBuilding(Building ** ppOutBuilding, Building * pInOldBuilding)
{
	// determine current building (if any)
	Building * pBuilding = pInOldBuilding;

	// ACQUIRE ANOTHER BUILDING
	GameMap *poMap = mpoPlane->mpoMap;
	*ppOutBuilding = poMap->GetAnotherRandomBuilding(pBuilding);
}

void Unit::vSetMobbing(BOOL bVal, MobCoordinator * poMobCoordinator)
{
	// FIRST REMOVE THE UNIT FROM AN EXISTING MOB
	if (bIsMobMember())
	{	
		// ASSIGNING UNIT TO SAME MOB
		if (poMobCoordinator == mpoMobCoordinator)
		{
//			assert(0);
			return;
		}
		BOOL bOkay = FALSE;
		mpoMobCoordinator->RemoveUnit(&bOkay, this);
		if (!mpoMobCoordinator->GetSize())
		{
			delete mpoMobCoordinator;
		}
		mpoMobCoordinator = NULL;
	}
	// ADJUST UNIT
	if (bVal) 
	{
		SetSpeed(SPEED_MOBBING);
		mpoMobCoordinator = poMobCoordinator;
	}
	else
	{
		SetSpeed(SPEED_NORMAL);
	}
}

void Unit::vInternalCommand_Attack(Unit * poTargetUnit)
{
	if ((moCurrentCommand.meType == CM_CAST)||(moCurrentCommand.meType == CM_ATTACK)) return;

	assert(poTargetUnit);
	Command oCommand;
	oCommand.meType = CM_CAST;
	oCommand.mnUnit = poTargetUnit->nGetID();
	oCommand.mbTargetUnit = TRUE;
	oCommand.mbInternal = TRUE;
	Spell * poSpell = game->GetSpell(mnCurrentSpell);

	if (meDefaultAttack == DA_SPELL && poSpell->bIsAttackUnitSpell() && bAbleToPerformCommand(oCommand))
	{
		vCommand(oCommand, TRUE);
		return;
	}
	oCommand.meType = CM_ATTACK;
	vCommand(oCommand, TRUE);
}

void Unit::vThink()
{
	if (!mbAlive) return;
	int nCurCycle = game->nGameCycle();
/*

	// THINK AT THINK SPEED
	if (nCurCycle - mnLastThinkCycle < mnThinkSpeed) {
		return;
	}
*/
	int nVisibleCount = 0;
	int nVisible = 0;

	if (bIsMobLeader())
	{
		// TBD - UPPER & LOWER SIZE LIMITS OF A MOB
		if (mpoMobCoordinator->GetSize() < 40)  // mpoMobCoordinator->GetSize() > 3 &&
		{
			int nRange = mnVisibleRange;

			mnVisibleRange = 15;

			vScan();

			nVisibleCount = moVisibleUnits.GetSize();
			nVisible = nVisibleCount;
			while (nVisible > 0)
			{	
				nVisible--;
				
				Unit * poUnit = (Unit*)goObjectDB.poGetObject(moVisibleUnits.GetItem(nVisible));
				if (poUnit == NULL) continue;

				if (!poUnit->bIsVampire() && poUnit->vGetClan() == vGetClan())
				{
					if (!poUnit->bIsMemberOfThisMob(mpoMobCoordinator))
					//if (!mpoMobCoordinator->bHasMember(poUnit))
					{
						// if (mpoMobCoordinator->GetSize() > 3)
						{
							if ((poUnit->moCurrentCommand.meType == CM_NONE)||(poUnit->moCurrentCommand.meType == CM_MOVE))
							{
								mpoMobCoordinator->AddUnitAsFollower(poUnit, mpoMobCoordinator->GetUnitToFollow());
				
								break;
							}
						}
					}
				}
			}
			mnVisibleRange = nRange;
			vScan();
		}
	}

	mnLastThinkCycle = nCurCycle;

	nVisibleCount = moVisibleUnits.GetSize();
	nVisible = 0;

	if (bIsVampire())	// I AM A VAMPIRE
	{
		// ATTACKING?
		//if (moCurrentCommand.meType == CM_ATTACK)
		if (bShouldRetreat())
		{
			CallForHelp();
			Retreat();
		}
		// UNDER ATTACK AND NOT ATTACKING?
		else if (mnUnderAttackCycles)
		{
			BOOL bDefend = TRUE;
			switch (mAggression)
			{
			case SETTING_LOW:
			case SETTING_MEDIUM:
				// CHECK TO SEE IF UNIT ALREADY HAS ORDERS
				if (moCurrentCommand.meType != CM_NONE)
				{
					bDefend = FALSE;
				}
				break;
			}
			if (bDefend)
			{
				// TENACITY OKAY TO STAY AND FIGHT?
				if (bStayAndFight())
				{
					/*
					// ATTACK THE ATTACKER
					Command oCommand;
					oCommand.meType = CM_ATTACK;
					oCommand.mbTargetUnit = TRUE;
					oCommand.mnUnit = mpoAttacker->mnObjID;
					vCommand(oCommand, TRUE);
					*/
					if (OutGunnedByVisibleUnits(mpoAttacker))
					{
						CallForHelp();
					}

					if (mpoAttacker)
					{
						vInternalCommand_Attack(mpoAttacker);
					}
				}
				else
				{
					CallForHelp();
					Retreat();
				}
			}
		}
		else
		{
			// SCAN VISIBLE UNITS
			while (nVisible < nVisibleCount)
			{
				Vampire * poUnit = (Vampire *)goObjectDB.poGetObject(moVisibleUnits.GetItem(nVisible));
				nVisible++;
				if (poUnit == NULL) continue;

				if (poUnit->bAlive())
				{
					// JUST A PEASANT
					if (!poUnit->bIsVampire())
					{
						if (moCurrentCommand.meType == CM_NONE)
						{
							if (poUnit->mpoThrallOwner == NULL)
							{
								// CONVERT
								if (mConversion == SETTING_HIGH)
								{
									Command oCommand;
									oCommand.meType = CM_CONVERT;
									oCommand.mbTargetUnit = TRUE;
									oCommand.mnUnit = poUnit->mnObjID;
									vCommand(oCommand, TRUE);
								}
								break;
								
							}
						}
					}
					// IS IT A KNOWN ENEMY?
					else if (poUnit->mpoOwner && poUnit->mpoOwner != mpoOwner && poUnit->bMarked(mpoOwner->mnPlayer))
					{
						if (OutGunnedByVisibleUnits(poUnit))
						{
							CallForHelp();
						}

						// DOING NOTHING?
						if (moCurrentCommand.meType == CM_NONE)
						{
							// IF OKAY TO ATTACK
							BOOL bAttack = bAttackEnemyUnit(poUnit);
							if (bAttack && bStayAndFight())
							{
								if (poUnit)
								{
									vInternalCommand_Attack(poUnit);
								}
								break;
							}
							else if (!bAttack)
							{
								CallForHelp();
								break;
							}
						}
					}
				}
			}
		}
	}
	else			// I AM A PEASANT
	{
		// IF MOBBING 
		if (bIsMobLeader())
		{
			//if ((moCurrentCommand.meType != CM_ATTACK)&&(moCurrentCommand.meType != CM_MOVE_INVITE))
			if (moCurrentCommand.meType != CM_ATTACK)
			{
				while (nVisible<nVisibleCount)
				{
					Unit * poUnit = (Unit*)goObjectDB.poGetObject(moVisibleUnits.GetItem(nVisible));

					nVisible++;
					if (poUnit == NULL) continue;
					// VISIBLE VAMPIRES?
					if (poUnit->bIsVampire())
					{
						if (mpoMobCoordinator->GetSize() > 3)
						{
							CommandSequenceClear();
							mpoMobCoordinator->AttackUnit(poUnit);
							break;
						}
					}
				}
					/*
					else if (!poUnit->poOwner())
					{
						if (!poUnit->bIsMemberOfThisMob(mpoMobCoordinator))
						{
							if (mpoMobCoordinator->GetSize() <= 3)
							{
								CommandSequenceClear();
								Command oCommand;
								oCommand.meType = CM_MOVE_INVITE;
								oCommand.mbTargetUnit = TRUE;
								oCommand.mnUnit = poUnit->nGetID();
								vCommand(oCommand, TRUE);
								break;
							}
						}
					}
					*/
				
				if (!mpoSequencer)
				{
					Sequence_Clear();
					Sequence_WanderWithoutRest(25);
				}
			}
		}
		else if (moCurrentCommand.meType == CM_MOVE_FOLLOW)
		{
			while (nVisible<nVisibleCount)
			{
				Unit * poUnit = (Unit*)goObjectDB.poGetObject(moVisibleUnits.GetItem(nVisible));
				nVisible++;
				if (poUnit == NULL) continue;
				// VISIBLE VAMPIRES?
				if (poUnit->bIsVampire())
				{
					if (bIsMobMember())
					{
						mpoMobCoordinator->AlertLeaderOfVampireLocation(poUnit);
					}
					break;
				}
			}
		}
		// NOT IN A MOB AND NOT HYPNOTIZED
		else if (!mnThrallCyclesRemaining && !bIsMobMember())
		{

			while (nVisible<nVisibleCount)
			{
				Unit * poUnit = (Unit*)goObjectDB.poGetObject(moVisibleUnits.GetItem(nVisible));
				nVisible++;
				if (poUnit == NULL) continue;
				// VISIBLE VAMPIRES?
				if (poUnit->bIsVampire() && !poUnit->bCloaked())
				{
					Clan * poClan = (Clan*)((Vampire*)poUnit)->vGetClan();
					vIncreaseAnxiety(poClan->GetAnxietyCausedByUncloaking());

					// RETREAT FROM VAMPIRES
					Retreat();

					break;
				}
			}
		}
	}
}

void Unit::SwapPositions(Unit & unit1, Unit & unit2)
{
	Command oCommand1;
	oCommand1.meType = CM_GUARD_MOVEINTERNAL;
	oCommand1.mnX = unit2.nGetX();
	oCommand1.mnY = unit2.nGetY();
	Command oCommand2;
	oCommand2.meType = CM_GUARD_MOVEINTERNAL;
	oCommand2.mnX = unit1.nGetX();
	oCommand2.mnY = unit1.nGetY();
	unit1.vCommand(oCommand1, TRUE);
	unit2.vCommand(oCommand2, TRUE);
}

void Unit::vPowerToHealth()
{
}


void Unit::GuardStop()
{
	if (mpoGuardedUnit)	// guarding a unit
	{
		mpoGuardedUnit->vRemoveGuard(this);		// stop guarding that unit
		mpoGuardedUnit = NULL;
	}
}


void Unit::vMoveToNextNode(MapNode ** pReturnMapNode)
{
	// nab the next node in the move list
	*pReturnMapNode = poGetNextMoveNode();

	// if we don't have a node, we're done.
	if (!(*pReturnMapNode)) {
		return;
	}

	// so now we have a node to move to. Let's do it.
	vMoveToNode(*pReturnMapNode);
}

void Unit::vMoveToNextNode(void)
{
	// nab the next node in the move list
	MapNode *poTempNode = poGetNextMoveNode();

	// if we don't have a node, we're done.
	if (!poTempNode) {
		return;
	}

	// so now we have a node to move to. Let's do it.
	vMoveToNode(poTempNode);
	pApp->pSoundManager->SetSoundEmitterPos(this, nPixelX(), nPixelY());
}

void Unit::vMoveToNextNodeWithOffset(void)
{
	// nab the next node in the move list
	MapNode *poTempNode = poGetNextMoveNodeWithOffset();

	// if we don't have a node, we're done.
	if (!poTempNode) {
		return;
	}

	// so now we have a node to move to. Let's do it.
	vMoveToNode(poTempNode);
}

void Unit::SetNode(MapNode * poNode)
{
	mpoCurNode = poNode;
}

void Unit::vMoveToNode(MapNode *poNode)
{
	mnStoppedCycles = 0;
	mnTempBlockedCycles = 0;

#if 0
	{
		// check to make sure we've only moved 1 node
		int nDX = poNode->mnXPos - mpoCurNode->mnXPos;
		int nDY = poNode->mnYPos - mpoCurNode->mnYPos;
		assert(abs(nDX) <= 1);
		assert(abs(nDY) <= 1);
		assert(poNode != mpoCurNode);
	}
#endif
	mpoLastNode = mpoCurNode;
	assert(!poNode->mpoUnit);
	assert(mpoCurNode->mpoUnit == this);
	mpoCurNode->mpoUnit = NULL;
	mpoCurNode = poNode;
	mpoCurNode->mpoUnit = this;
	mbBetweenPoints = TRUE;
	mnPartialMove = 0;
	vMoveOneCycle();
}

void Unit::vMoveOneCycle()
{
	assert(mnSpeed > 0);
	assert(mpoCurNode->mpoUnit == this);

	int nBaseX = mpoLastNode->nPixelX();
	int nBaseY = mpoLastNode->nPixelY();
	int nDeltaX = mpoCurNode->nPixelX() - nBaseX;
	int nDeltaY = mpoCurNode->nPixelY() - nBaseY;

	++mnPartialMove;
	if (mnPartialMove > mnSpeed) {
		mnPartialMove = mnSpeed;
	}

	int nNewX = nBaseX + ((mnPartialMove * nDeltaX) / mnSpeed);
	int nNewY = nBaseY + ((mnPartialMove * nDeltaY) / mnSpeed);

//	DIRECTION eDir = eDirection(nBaseX,nBaseY,nNewX,nNewY);
	DIRECTION eDir = eDirection(mpoLastNode->mnXPos, mpoLastNode->mnYPos, mpoCurNode->mnXPos, mpoCurNode->mnYPos);
	if (eDir != D_NOCHANGE && meDirection != eDir)
	{
		meDirection = eDir;
		vSetActionLoop();

		if (mpoCycler) {
			mpoCycler->vReset();
		}
	}
//	meDirection =  eDirection(nBaseX,nBaseY,nNewX,nNewY);
	
	//need some way of determining if we are running/mobbing. This is it for now
	//TBD - what about vamp hunters? how do they run?
	if ( (bIsVampire() && !bIsCloaked()) || bIsMobMember() )
		vSetAction(ACT_RUN);
	else
		vSetAction(ACT_WALK);

	vSetPixelPos(nNewX, nNewY);

	if (mnPartialMove == mnSpeed) {
		mbBetweenPoints = FALSE;
	}

	// we must be moving!
	mbMoving = TRUE;
}

BOOL Unit::CheckRoute(MapNode * poGoal, MapNode ** ppoDest)
{
	GameMap *poMap = mpoPlane->mpoMap;
	MapNode *poDest = NULL;
	return bFindPath(poMap, mpoCurNode, poGoal, &moCurrentRoute, ppoDest);
}

void Unit::vCalcRoute(int nOffsetX, int nOffsetY)
{
	GameMap *poMap = mpoPlane->mpoMap;
	MapNode *poGoal = NULL;
	MapNode *poDest = NULL;

	moCurrentRoute.RemoveAll();

	int nTargetX, nTargetY;
	moCurrentCommand.vGetTargetPos(&nTargetX, &nTargetY);
	
	mnLastTargetX = nOffsetX;
	mnLastTargetY = nOffsetY;

	poGoal = gpoMap->poGetNode(nTargetX + mnLastTargetX, nTargetY + mnLastTargetY);

	if (poGoal) {
		bFindPath(poMap, mpoCurNode, poGoal, 
			  &moCurrentRoute, &poDest);
	}
}

void Unit::vCalcRoute()
{
	GameMap *poMap = mpoPlane->mpoMap;
	MapNode *poGoal = NULL;
	MapNode *poDest = NULL;

	moCurrentRoute.RemoveAll();
	//
	// get the goal either from the unit we're targeting or the
	// locationi
	//
	{
		int nTargetX, nTargetY;
		moCurrentCommand.vGetTargetPos(&nTargetX, &nTargetY);
		mnLastTargetX = nTargetX;
		mnLastTargetY = nTargetY;
	}

	poGoal = moCurrentCommand.poGetTargetNode();
	if (poGoal) {
		bFindPath(poMap, mpoCurNode, poGoal, 
			  &moCurrentRoute, &poDest);
	}

#if 0
	// test the path for sanity
	{
		MapNode *poTestNode = moCurrentRoute.GetLast();
		MapNode *poCurNode = mpoCurNode;
		while (poTestNode) {
			// check to make sure we've only moved 1 node
			int nDX = poCurNode->mnXPos - poTestNode->mnXPos;
			int nDY = poCurNode->mnYPos - poTestNode->mnYPos;
			assert(abs(nDX) <= 1);
			assert(abs(nDY) <= 1);
			assert(poTestNode != poCurNode);
			
			poCurNode = poTestNode;
			poTestNode = moCurrentRoute.GetPrevious();
		}
	}
#endif			
	
}

MapNode *Unit::poPopOneNode()
{
	MapNode *poNode;
	// make sure to pop off the current node, if it's in the list
	poNode = moCurrentRoute.PopLast();

#ifdef _DEBUG
	if (poNode) {
		assert(poNode != mpoCurNode);

		// check to make sure we've only moved 1 node
		int nDX = poNode->mnXPos - mpoCurNode->mnXPos;
		int nDY = poNode->mnYPos - mpoCurNode->mnYPos;
		assert(abs(nDX) <= 1);
		assert(abs(nDY) <= 1);
		assert(poNode != mpoCurNode);
	}
#endif
	return poNode;
}

MapNode *Unit::poGetNextMoveNodeWithOffset()
{
	MapNode *poTempNode = poPopOneNode();

	// if there's no current node to move to, try to make a route
	if (!poTempNode) {
	//	vCalcRoute(mnLastTargetX, mnLastTargetY);
	//	poTempNode = poPopOneNode();
	//	// if the route fails, return, and possibly give up
	//	if (!poTempNode) {
	//		if (++mnStoppedCycles > MAX_STOPPED_CYCLES) {
	//			// we're giving up
	//			vCommandNone();
	//		}

			// no moving in any case
			return NULL;
		}
	
	BOOL bTemp;

	// if we're blocked, how badly?
	if (poTempNode->bBlocked(&bTemp)) {

		// is it just temporary? Then wait a bit
		if (bTemp) 
		{
			if (++mnTempBlockedCycles > MAX_BLOCKED_CYCLES) 
			{
				// it's been too long; recalc the route
				vCalcRoute(mnLastTargetX, mnLastTargetY);
				poTempNode = poPopOneNode();
			} 
			else 
			{
				// no, just wait some more

				// push back the node
				moCurrentRoute.AddToEnd(poTempNode);
				return NULL;
			}
		} 
		else 
		{

			// we're permanently blocked -- find a new route
			vCalcRoute(mnLastTargetX, mnLastTargetY);
			poTempNode = poPopOneNode();
		}

		// if after all this we don't have a node, then give up and bail
		if (!poTempNode) 
		{
			vCommandNone();
			return NULL;
		}
	}
	return poTempNode;
}


MapNode *Unit::poGetNextMoveNode()
{
	MapNode *poTempNode = poPopOneNode();

	// if there's no current node to move to, try to make a route
	if (!poTempNode) {
		vCalcRoute();
		poTempNode = poPopOneNode();

		// if the route fails, return, and possibly give up
		if (!poTempNode) {
			if (++mnStoppedCycles > MAX_STOPPED_CYCLES) {
				// we're giving up
				vCommandNone();
			}

			// no moving in any case
			return NULL;
		}
	}

	BOOL bTemp;

	// if we're blocked, how badly?
	if (poTempNode->bBlocked(&bTemp)) {

		// is it just temporary? Then wait a bit
		if (bTemp) {
			if (++mnTempBlockedCycles > MAX_BLOCKED_CYCLES) {
				// it's been too long; recalc the route
				vCalcRoute();
				poTempNode = poPopOneNode();
			} else {
				// no, just wait some more

				// push back the node
				moCurrentRoute.AddToEnd(poTempNode);
				return NULL;
			}
		} else {

			// we're permanently blocked -- find a new route
			vCalcRoute();
			poTempNode = poPopOneNode();
		}

		// if after all this we don't have a node, then give up and bail
		if (!poTempNode) {
			vCommandNone();
			return NULL;
		}
	}
	return poTempNode;
}


// vSetGuardedUnit  sets which unit to guard
void Unit::vSetGuardedUnit(Unit *poUnit)
{
	if (mpoGuardedUnit == poUnit)				// already guarding this unit
	{
		return;	
	}
	if (mpoGuardedUnit)	// guarding a unit
	{
		mpoGuardedUnit->vRemoveGuard(this);		// stop guarding that unit
	}
	mpoGuardedUnit = poUnit;					// guard this unit instead
	if (mpoGuardedUnit)							// is it something?
	{
		mpoGuardedUnit->vAddGuard(this);		// guard it 
	}
}


void Unit::Message(Unit * poFrom, DWORD dwMessage)
{
	if (mpoGuardCoordinator)
	{
		mpoGuardCoordinator->Message(poFrom, (enumGuardMessage)dwMessage);
	}
}

// poUnit is Guarding me
// this should be private... multiple guardcoordinators cannot reference same guard
void Unit::vAddGuard(Unit * poUnit)
{
	if (!mpoGuardCoordinator)
	{
		mpoGuardCoordinator = new GuardCoordinator(*this);
	}
	assert(mpoGuardCoordinator);
	mpoGuardCoordinator->vAddGuard(poUnit);
}

void Unit::vRemoveGuard(Unit * poUnit)
{
	assert(mpoGuardCoordinator);  // called this one too many times
	if (mpoGuardCoordinator->vRemoveGuard(poUnit) == 0)
	{
		delete mpoGuardCoordinator;
		mpoGuardCoordinator = NULL;
	}
}

void Unit::vSetAttackee(Unit *poUnit)
{
	int nNewID;
	if (poUnit == NULL) 
	{
		nNewID = 0;
	}
	else
	{
		nNewID = poUnit->nGetID();
	}
	if (mnAttackee == nNewID) 
	{
		return;
	}
	if (mnAttackee) {
		Unit * poAttackee = (Unit*)goObjectDB.poGetObject(mnAttackee);
		if (poAttackee)
			poAttackee->vRemoveAttacker(this);
	}
	mnAttackee = nNewID;
	if (mnAttackee) {
		poUnit->vAddAttacker(this);
	}
}

void Unit::vAddAttacker(Unit *poUnit)
{
	moAttackers.AddToEnd(poUnit);
}

void Unit::vRemoveAttacker(Unit *poUnit)
{
	moAttackers.Remove(poUnit);
}

void Unit::vOverride(float * pfVar, long lTag)
{
	float nOverride;

	// TRY THE UNIT TEMPLATE FIRST
	nOverride = (float)mpoTemplate->fGetValue(lTag, -12345);
	if (nOverride != -12345)
	{
		*pfVar = nOverride;
		return;
	}

	// TRY THE CLAN TEMPLATE
	Template * poClanTemplate = vGetClanTemplate();
	nOverride = (float)poClanTemplate->fGetValue(lTag, -12345);
	if (nOverride != -12345)
	{
		*pfVar = nOverride;
	}
}

void Unit::vOverride(int * pnVar, long lTag)
{
	int nOverride;
	// TRY THE UNIT TEMPLATE FIRST
	nOverride = mpoTemplate->lGetValue(lTag, -12345);
	if (nOverride != -12345)
	{
		*pnVar = nOverride;
		return;
	}

	// TRY THE CLAN TEMPLATE
	Template * poClanTemplate = vGetClanTemplate();
	nOverride = poClanTemplate->lGetValue(lTag, -12345);
	if (nOverride != -12345)
	{
		*pnVar = nOverride;
	}
}

void Unit::AddPatrolWaypoint( int nX, int nY )
{
	if (!mpoMovement)
	{
		mpoMovement = new SequentialMovement(*this);
	}

	GameMap *poMap = mpoPlane->mpoMap;
	MapNode *poNode = poMap->poGetNode(nX, nY);

	if (poNode)
	{
		mpoMovement->Add( poNode );
	}
}

void Unit::Patrol()
{
	// continue moving 
	assert(mpoMovement);  // shouldn't be calling this
							// unnecessarily
	if (mpoMovement)
	{
		Command oCommand;
		MapNode * poNode;
		mpoMovement->Get(&poNode);
		oCommand.mnX = poNode->mnXPos;
		oCommand.mnY = poNode->mnYPos;
		oCommand.meType = CM_MOVE;
		vCommand(oCommand, TRUE);
	}
}

void Unit::ClearThought()
{
	if (mpoMovement)
	{
		delete mpoMovement;
		mpoMovement = NULL;
	}
}

void Unit::vCommand(Command &oCommand, BOOL bInternal)
{
	// dead units don't accept commands
	if (!mbAlive) {
		return;
	}

	if (!bInternal)
	{
		if (mpoPlane->mpoMap->GetMode() == UIMODE_ENTERCOMMANDS)
		{
			if (!mpoSequencer)
			{
				assert(!mpoSequencer);
				mpoSequencer = new SequentialCommands(*this);
			}
			mpoSequencer->Add(&oCommand);
			return;
		}
	}

	// unowned units can't accept outside commands
	if (!bInternal && !mpoOwner) 
	{
		if (oCommand.meType == CM_ATTR)
		{
			switch(oCommand.mnAttributeID)
			{
			case ATTR_ANXIETY:
				vSetAnxiety(100 * oCommand.mnNewValue);
				break;
			default:
				break;
			}
		}
		return;
	}

	// we don't accept commands from other players, either
	if (!bInternal && mpoOwner->mnPlayer != oCommand.mnPlayer) {
		return;
	}


	switch(oCommand.meType) {
	case CM_ATTR:
		assert(0);
		break;

	case CM_GUARD_MOVEINTOPOSITION:
		moCurrentCommand = oCommand;
		vCalcRoute();
		break;
	case CM_GUARD_MOVEINTERNAL:				// GENERAL MOVE COMMAND
	case CM_GUARD_MOVETOVERIFY:
		// the old route is certainly dead
		moCurrentCommand = oCommand;
		vCalcRoute();
		break;

	case CM_PATROL_WAYPOINT:
		// create waypoint
		// continue waypoint thought
		AddPatrolWaypoint(oCommand.mnX, oCommand.mnY);
		Patrol();
		break;

	case CM_MOVE_FOLLOW:
	case CM_MOVE_INVITE:
	case CM_MOVE:							// GENERAL MOVE COMMAND
		// the old route is certainly dead
		if (!bInternal)
		{
			ClearThought();
		}
		GuardStop();
		moCurrentCommand = oCommand;
		vCalcRoute();
		if (oCommand.meType == CM_MOVE_FOLLOW)
		{
			moCurrentCommand.mnNewValue = 0;
		}
		break;

	case CM_FORMATION:						// FORMATION RELATIVE TO UNIT
		// determine formation unit
		// determine formation offset
		// store formation offset
		// store formation unit
		
		break;


	case CM_GUARD:							// GUARD A UNIT
		{
			moCurrentCommand = oCommand;
			Unit *poObj = (Unit *)moCurrentCommand.poGetTarget();
			assert(poObj->eClass() == OC_UNIT);
			vSetGuardedUnit((Unit*)poObj);
			//vCalcRoute(nGetX() - poObj->nGetX(), nGetY() - poObj->nGetY());
		}
		break;
	case CM_RETREAT:
		{
			moCurrentCommand = oCommand;
			//mnLastRetreatDirection = -1;
		}
		break;

	case CM_ATTACK:							// DESIRE TO ATTACK
		{
			GuardStop();

			moCurrentCommand = oCommand;
			moCurrentCommand.mnNewValue = bInternal;

			Unit * poUnit = (Unit *)moCurrentCommand.poGetTarget();
			assert(poUnit->eClass() == OC_UNIT);

			vSetAttackee(poUnit);
			vCalcRoute();
		}
		break;

	case CM_DROP:
	case CM_GIVE:
	case CM_TAKE:
		{
			moCurrentCommand = oCommand;
			vCalcRoute();
		}
		break;
	case CM_CONVERT:
		assert(0);
		break;
	case CM_FEED:
		assert(0);
		break;
	case CM_CLOAK:
		assert(0);
		break;
	case CM_WAIT_500MS:
		moCurrentCommand = oCommand;
		break;
	default:
		// TBD -- handle other commands
		assert(0); // command not handled
		break;
	}
}

void Unit::vSaveMap(FILE *file, WORD wVersion) 
{
	int nResID = mpoTemplate->resourceID;	
	BYTE jPlayer = mjPlayerLink;
	int noLevel = 0;

	switch (wVersion) {
	case 1:
		// resourceID of the template

		if (mpoOwner)
		{
			// Level of the Unit
			fwrite(&noLevel, sizeof(int), 1, file);			
			fwrite(&mnLevel, sizeof(int), 1, file);			
		}
		else
		{
			fwrite(&nResID, sizeof(int), 1, file);			
			fwrite(&noLevel, sizeof(int), 1, file);			
		}

		// player link for the unit
		fwrite(&jPlayer, sizeof(BYTE), 1, file);		
		
		if (mpoSequencer)
		{
			// command tag
			fwrite(gajCommandTag, sizeof(gajCommandTag), 1, file);

			// command sequence
			mpoSequencer->vSaveMap(file, wVersion);
		}
	
		break;

	}
}

void Unit::vLoadMap(FILE *file, WORD wVersion) 
{
	// TBD - rearrange this

	if (mpoSequencer)
	{
		delete mpoSequencer;
		mpoSequencer = NULL;
	}
	assert(!mpoSequencer);
	mpoSequencer = new SequentialCommands(*this);
	mpoSequencer->vLoadMap(file, wVersion);

}

// For adding to the map.
// The unit object being dragged around makes a copy of itself and inserts it in into the
// node nearest the event
void Unit::vCopySelf(sciEvent &oEvent) {
	MapNode *poNode = mpoPlane->mpoMap->poNodeFromEvent(oEvent);
	if (poNode) {

		// If self is in the clipboard and a unit is in spot where we want a copy to go,
		// then we can't allow it as only 1 unit per node is allowed. This condition can and will
		// occur when you add a unit with the right mouse button, don't move it, and click the 
		// button again.
		//		OR
		// we right click on a unit that already exits (which is not equal to our self)
		if ( (gpoMapEdit->mpoMapNode->mpoUnit) && (poNode->mpoUnit) ||
			 (this != poNode->mpoUnit) ) {
			vDebugOut("can't add 2 units in same space\n");
		} else {
			// clear self out of node the copy is suppose to go into and put back onto clipboard
			mpoCurNode->mpoUnit = NULL;					//clear self out of node
			gpoMapEdit->mpoMapNode->mpoUnit = this;		//put onto clipboard
			mpoCurNode = gpoMapEdit->mpoMapNode;		//update back reference

			// now create copy and place into node
			// Unit *poCopy = new Unit(NULL, mpoPlane, poNode, nTemplateID());	
			Unit * poCopy = NULL;
			Unit::Create(&poCopy, NULL, mpoPlane, poNode, nTemplateID());
			poCopy->mjPlayerLink = mjPlayerLink;
		}
	}
}

void Unit::vDrag(sciEvent &oEvent) {
	MapNode *poNode = mpoPlane->mpoMap->poNodeFromEvent(oEvent);
	if (poNode) {
		//vDebugOut ("x: %d y:%d\n", poNode->mnXPos, poNode->mnYPos);
		if (mpoCurNode != poNode) {
			// the mouse is in a different node than the current object
			if (poNode->mpoUnit == NULL) {
				//we can't drag one unit on top of another
				mpoCurNode->mpoUnit = NULL;
				poNode->mpoUnit = this;
				vSetPos(poNode->mnXPos, poNode->mnYPos); //this call results in a MarkUpdate
				oEvent.claimed = TRUE;	
			}
		}
	}
}

// For adding to the map.
// The unit object being dragged around inserts ITSELF in the node nearest its event
void Unit::vInsertSelf(sciEvent &oEvent) {
	MapNode *poNode = mpoPlane->mpoMap->poNodeFromEvent(oEvent);
	if (poNode) {

		// if we're on the clipboard
		if (gpoMapEdit->mpoMapNode->mpoUnit) {
			if (poNode->mpoUnit == NULL) {
				//most likely vDrag will have already inserted us into the node but if it didn't
				//we need to catch it here.
				mpoCurNode->mpoUnit = NULL;
				poNode->mpoUnit = this;
				vSetPos(poNode->mnXPos, poNode->mnYPos); //this call results in a MarkUpdate
//				mjPlayerLink = 0; //now done in mapedit.cpp
			} else {
				vDebugOut("can't add 2 units in same space\n"); //TBD - add some real error message here - TW
				delete this;
			}
		}
		
	} else {
		// we clicked somewhere not on the game map so get rid of self
		delete this;
	}

}


// For dragging a building around that has units inside. Moves unit from previous location to 
// a new node relative the the buildings position.(calculated from the MapEditPlane)
void Unit::vMapShift(MapNode *poTo) {
	poTo->vAddCover(mpoCurNode->poGetCover());	//add covering to the node we're going into
	poTo->meType = TT_INTERIOR;					//set the terrain type for the node we're going into

	mpoCurNode->vClearCover();				//clear the covering from the node we left
	mpoCurNode->meType = TT_CLEAR;			//clear the terrain type from the node we left
	
	mpoCurNode->mpoUnit = NULL;				//clear the node the unit WAS in
	poTo->mpoUnit = this;					//insert into new node
	vSetPos(poTo->mnXPos, poTo->mnYPos);	//update back reference & do a MarkUpdate
}


Boolean Unit::HandleEvent(sciEvent & oEvent)
{
	if (gbMapEditing) {
		DrawObj::HandleEvent(oEvent);
		return oEvent.claimed;
	}

	if (!gbMapEditing)
	{
		enumGameMode eGameMode = mpoPlane->mpoMap->GetMode();

		// normal operation
		if (mbSelectable && (bOnMe(oEvent))) 
		{
			if (oEvent.modifiers & RIGHT_MOUSE)			// RIGHT MOUSE
			{
				switch(oEvent.type)
				{
				case MOUSE_DOWN:
					if (eGameMode != UIMODE_SELECT)
					{
						mpoPlane->mpoMap->SetMode(UIMODE_SELECT);
						oEvent.claimed = TRUE;
						break;
					}
					// LIVING PEASANT
					if (!poOwner() && bAlive())  
					{
						// FEED
						Command *poCommand = new Command;
						poCommand->meType = CM_FEED;
						poCommand->mbTargetUnit = TRUE;
						poCommand->mnUnit = mnObjID;
						game->poLocalPlayer()->vPostCommand(poCommand);
						oEvent.claimed = TRUE;
						//gpoMap->vAudioConfirmCommand();
					}
					// ENEMY VAMPIRE
					else if (poOwner() != game->poLocalPlayer() && bAlive()) 
					{
						// ATTACK
						Command *poCommand = new Command;
						poCommand->meType = CM_ATTACK;
						poCommand->mbTargetUnit = TRUE;
						poCommand->mnUnit = mnObjID;
						game->poLocalPlayer()->vPostCommand(poCommand);
						oEvent.claimed = TRUE;
						//gpoMap->vAudioConfirmCommand();
					}
					// LOCAL VAMPIRE
					else if (poOwner() == game->poLocalPlayer() && bAlive()) 
					{
						// GUARD
						Command *poCommand = new Command;
						poCommand->meType = CM_GUARD;
						poCommand->mbTargetUnit = TRUE;
						poCommand->mnUnit = mnObjID;
						game->poLocalPlayer()->vPostCommand(poCommand);
						oEvent.claimed = TRUE;
						//gpoMap->vAudioConfirmCommand();
					}
				}
			}
			else										// LEFT MOUSE
			{
				switch(oEvent.type)
				{
				case MOUSE_DOWN:
					if (eGameMode == UIMODE_PATROL)
					{
						eGameMode = UIMODE_SELECT;
					}

					switch(eGameMode)
					{
					case UIMODE_CONVERT:
						if (!poOwner() && bAlive())  // LIVING PEASANT
						{
							Command *poCommand = new Command;
							poCommand->meType = CM_CONVERT;
							poCommand->mbTargetUnit = TRUE;
							poCommand->mnUnit = mnObjID;
							game->poLocalPlayer()->vPostCommand(poCommand);
							oEvent.claimed = TRUE;
							//gpoMap->vAudioConfirmCommand();
						}
						break;
					case UIMODE_GIVE:
						if (poOwner() == game->poLocalPlayer() && bAlive()) 
						{
							Command *poCommand = new Command;
							poCommand->meType = CM_GIVE;
							poCommand->mbTargetUnit = TRUE;
							poCommand->mnUnit = mnObjID;
							poCommand->mnNewValue = (int)gpoMap->mnSetAttributeID;
							game->poLocalPlayer()->vPostCommand(poCommand, FALSE);
							oEvent.claimed = TRUE;
							//gpoMap->vAudioConfirmCommand();
						}
						break;

					case UIMODE_FEED:
						if (!poOwner() && bAlive())  // LIVING PEASANT
						{
							Command *poCommand = new Command;
							poCommand->meType = CM_FEED;
							poCommand->mbTargetUnit = TRUE;
							poCommand->mnUnit = mnObjID;
							game->poLocalPlayer()->vPostCommand(poCommand, FALSE);
							oEvent.claimed = TRUE;
							//oMap->vAudioConfirmCommand();
						}
						break;
					case UIMODE_CAST:
						if (poOwner() != game->poLocalPlayer() && bAlive()) 
						{
							Command *poCommand = new Command;
							poCommand->meType = CM_CAST;
							poCommand->mbTargetUnit = TRUE;
							poCommand->mnUnit = mnObjID;
							poCommand->mnNewValue = 0;	// SPELL NUMBER
							game->poLocalPlayer()->vPostCommand(poCommand, FALSE);
							oEvent.claimed = TRUE;
							//gpoMap->vAudioConfirmCommand();
						}
						break;

					case UIMODE_ATTACK:
						if (poOwner() != game->poLocalPlayer() && bAlive()) 
						{
							// we want to attack/convert this unit
							Command *poCommand = new Command;
							poCommand->meType = CM_ATTACK;
							poCommand->mbTargetUnit = TRUE;
							poCommand->mnUnit = mnObjID;
							game->poLocalPlayer()->vPostCommand(poCommand);
							oEvent.claimed = TRUE;
							//gpoMap->vAudioConfirmCommand();
						}
						break;
					
					case UIMODE_GUARD:
						if (poOwner() == game->poLocalPlayer() && bAlive()) 
						{
							// we want to guard this unit
							Command *poCommand = new Command;
							poCommand->meType = CM_GUARD;
							poCommand->mbTargetUnit = TRUE;
							poCommand->mnUnit = mnObjID;
							game->poLocalPlayer()->vPostCommand(poCommand);
							oEvent.claimed = TRUE;
							//gpoMap->vAudioConfirmCommand();
						}
						break;
					case UIMODE_UNDEFINED:
						break;
					case UIMODE_SELECT:
						if (oEvent.modifiers & SHIFT_DOWN) 
						{
							if (mbSelected) {
								mpoPlane->mpoMap->vRemoveSelection(this);
							} 
							else 
							{
								mpoPlane->mpoMap->vAddSelection(this);
							}
							mpoPlane->mpoMap->vCommitSelection();														
						} 
						//else if (poOwner() != game->poLocalPlayer() && bAlive()) 
						//{
						//	eGameMode = UIMODE_ATTACK;
						//}
						else 
						{ 
							// just make a single selection
							mpoPlane->mpoMap->vClearAllSelections();
							mpoPlane->mpoMap->vAddSelection(this);
							mpoPlane->mpoMap->vCommitSelection();

							if (poOwner() == game->poLocalPlayer()) 
							{
								vPlayRandomSound(TAG_Syes);
							}
						}
						// invalidate my area
						vMarkUpdate();
						break;
					}
					oEvent.claimed = TRUE;
					if (mpoPlane->mpoMap->GetMode() == UIMODE_CAST)
					{
						if (oEvent.modifiers & CTRL_DOWN)
						{
							break;
						}
					}
					mpoPlane->mpoMap->SetMode(UIMODE_SELECT);
				}
			}
		}
	}
	if (!oEvent.claimed)
	{
		return DrawObj::HandleEvent(oEvent);
	}
	else
	{
		return oEvent.claimed;
	}
}


void Unit::vDraw() {
//	if (mpoWalker)
//		mpoBitmap = mpoWalker->poGetBitmap(this, meDirection);

	if (!mbMoving && (meAction == ACT_WALK || meAction == ACT_RUN))
		vSetAction(ACT_FIDGET_CLOAKED);

	//draw the shadow if unit is visible
	if (mpoShadow && mbVisible)
	{
		if (mbMoving && (meAction == ACT_WALK || meAction == ACT_RUN)) {
			mpoShadow->vIncrementCel();
		}

		mpoShadow->vDraw(nPixelX() - mpoPlane->nMapOrgPixelX(), nPixelY() - mpoPlane->nMapOrgPixelY());
	}

	//call the super
	DrawObj::vDraw();

//	if (mpoOwner && bMarked(game->poLocalPlayer()->mnPlayer) && mbAlive) {
//	if (mbAlive) {
		//right now just draw everyone's status
//		moStatus.vDraw(nPixelX() - mpoPlane->nMapOrgPixelX() - mpoBitmap->nOriginX, 
//					   nPixelY() - mpoPlane->nMapOrgPixelY() - mpoBitmap->nOriginY );
//	}

}

void Unit::vInitStatics() {
	mjFullRed    = (BYTE)game->mpoTemplate->lGetValue(TAG_Fred, 0);
	mjFullGreen  = (BYTE)game->mpoTemplate->lGetValue(TAG_Fgre, 255);
	mjFullBlue   = (BYTE)game->mpoTemplate->lGetValue(TAG_Fblu, 0);
	mjEmptyRed   = (BYTE)game->mpoTemplate->lGetValue(TAG_Ered, 0);
	mjEmptyGreen = (BYTE)game->mpoTemplate->lGetValue(TAG_Egre, 255); 
	mjEmptyBlue  = (BYTE)game->mpoTemplate->lGetValue(TAG_Eblu, 0);
	mnRelHealthPercent = (short)game->mpoTemplate->lGetValue(TAG_rhpe, 20);
}

void Unit::CommandSequenceClear()
{
	if (mpoSequencer)
	{
		delete mpoSequencer;
		mpoSequencer = NULL;
	}
}


void Unit::SetAttribute(Unit * poThis, ATTR_CODE mnSetAttributeID, int setting) 
{ 
	Command *poCommand = new Command;
	poCommand->meType = CM_ATTR;
	poCommand->mnAttributeID = (int)mnSetAttributeID;
	poCommand->mnNewValue = (int)setting;
	if (poThis)
	{
		poCommand->mnUnit = poThis->mnObjID;
		poCommand->mbTargetUnit = TRUE;
	}
	game->poLocalPlayer()->vPostCommand(poCommand);
}



long Unit::lBitmapNormal() 
{
	return mpoTemplate->lGetValue(TAG_Bnor, nBitmapID());
}

long Unit::lBitmapRunning() 
{
	return mpoTemplate->lGetValue(TAG_Brun, nBitmapID());
}

long Unit::lBitmapReacting() 
{
	return mpoTemplate->lGetValue(TAG_Brea, nBitmapID());
}

long Unit::lBitmapAttacking() 
{
	return mpoTemplate->lGetValue(TAG_Batt, nBitmapID());
}

long Unit::lBitmapFeeding() 
{
	return mpoTemplate->lGetValue(TAG_Bfee, nBitmapID());
}

long Unit::lBitmapCastingSpell()
{
	return mpoTemplate->lGetValue(TAG_Bcas, nBitmapID());
}

long Unit::lBitmapDying() 
{
	return mpoTemplate->lGetValue(TAG_Bded, nBitmapID());
}

long Unit::lBitmapPortrait() 
{
	return mpoTemplate->lGetValue(TAG_Bpor, nBitmapID());
}

long Unit::lBitmapSoloPortrait() 
{
	return mpoTemplate->lGetValue(TAG_Bspo, nBitmapID());
}

void Unit::vRegeneratePower() 
{
	mfHealth = min(mfHealth + 0.005f, mfMaxHealth);
}

void Unit::vUpdateAttributes() 
{
	// TBD - INCLUDE IN PEASANT CLAN TEMPLATE
	mnAttackSpeed = vGetClanTemplate()->lGetValue(TAG_aspd, 20) - (mnLevel + (GetUpgradeSpeed() * 2));

	// TBD - INCLUDE IN PEASANT CLAN TEMPLATE
	// SETUP HEALTH
	mfMaxHealth = 25;
	vOverride(&mfMaxHealth, TAG_mhlt);

	// SETUP POWER
	mfMaxPower = 0;
	vOverride(&mfMaxPower, TAG_mpwr);
	
	mnVisibleRange = vGetVisibleRange();
}

void Unit::vSetLevel(int nLevel) 
{
	assert(nLevel<8);
	assert(mnLevel>=0);

	nLevel = max(0, min(nLevel, 8));
	mnLevel = nLevel;

	vUpdateAttributes();
}


BOOL Unit::bIsMobLeader()
{
	if (mpoMobCoordinator)
	{
		if (this == mpoMobCoordinator->GetLeader())
		{
			return TRUE;
		}
	}
	return FALSE;
}

BOOL Unit::bIsMemberOfThisMob(MobCoordinator * poMobCoordinator)
{
	if (mpoMobCoordinator == poMobCoordinator)
	{
		return TRUE;
	}
	return FALSE;
}

// BOOL Unit::InviteToMob(MobCoordinator * poMobCoor

BOOL Unit::bIsMobMember()
{
	return (mpoMobCoordinator != FALSE);
}


HRESULT Unit::Create(Unit ** ppoUnit, Player *poOwner, MapPlane *poPlane, MapNode *poNode, int nTemplateID)
{
	// LOAD THE TEMPLATE
	Template * poTemplate = (Template*)pApp->pResourceManager->LoadResource(RES_TEMPLATE, nTemplateID);

	int nVal = poTemplate->lGetValue(TAG_clan, (int)NPC_PEASANT);

	ENUM_NPC eClan;
	if ((nVal < (int)ENUM_NPC_COUNT)&&(nVal >= 0))
		eClan = (ENUM_NPC)nVal;
	else
		eClan = NPC_OTHER;
	
	// CREATE THE APPROPRIATE UNIT USING NPCCLASS
	switch(eClan)
	{
	case NPC_PEASANT:
		*ppoUnit = new HumanPeasant(poOwner, poPlane, poNode, nTemplateID);
		break;	
	case NPC_VAMPIREHUNTER:
		*ppoUnit = new HumanVampireHunter(poOwner, poPlane, poNode, nTemplateID);
		break;	
	case NPC_SOLDIER:
		*ppoUnit = new HumanSoldier(poOwner, poPlane, poNode, nTemplateID);
		break;	
	case NPC_GYPSY:
		*ppoUnit = new HumanGypsy(poOwner, poPlane, poNode, nTemplateID);
		break;	
	case NPC_MERCHANT:
		*ppoUnit = new HumanMerchant(poOwner, poPlane, poNode, nTemplateID);
		break;	
	case NPC_OTHER:
	default:
		return E_FAIL;
	}
	if (*ppoUnit)
		return S_OK;
	else
		return E_OUTOFMEMORY;
}

ENUM_SEX Unit::eGetSex()
{
	if (mpoTemplate->lGetValue(TAG_male, 1))
	{
		return SEX_MALE;
	}
	else
	{
		return SEX_FEMALE;
	}

}


void Unit::vChangeFacing(int nCurX, int nCurY, int nNewX, int nNewY)
{
	DIRECTION eDir = eDirection(nCurX, nCurY, nNewX, nNewY);
	//vDebugOut("dir:%d\n",(int)eDir);
	if (eDir != D_NOCHANGE)
	{
		meDirection = eDir;
	}
}

DIRECTION Unit::eDirection(int nCurX, int nCurY, int nNewX, int nNewY)
{
	//calculate any differences in h/v movements
	Boolean bHorizontalChange = (nCurX != nNewX);
	Boolean bVerticalChange = (nCurY != nNewY);
	DIRECTION eDirection;

	if (!bHorizontalChange && !bVerticalChange)
		//return meDirection;
		return D_NOCHANGE;

	if (nNewX < nCurX)
	{
		//we moved SW, W or NW
		if (bVerticalChange)
		{
			if (nNewY < nCurY)
				eDirection = D_NORTHWEST;
			else
				eDirection = D_SOUTHWEST;

		} else {

			eDirection = D_WEST;
		}

	}

	else if (nNewX > nCurX) 
	{
		//we moved SE, E or NE
		if (bVerticalChange)
		{
			if (nNewY < nCurY)
				eDirection = D_NORTHEAST;
			else
				eDirection = D_SOUTHEAST;

		} else {

			eDirection = D_EAST;
		}

	}

	else 
	{
		//we moved in either N or S
		if (nNewY < nCurY)
			eDirection = D_NORTH;
		else
			eDirection = D_SOUTH;
	}
	//vDebugOut ("curD:%d newD%d cx:%d cy:%d nx:%d ny:%d\n", meDirection, eDirection, nCurX, nCurY, nNewX, nNewY);
	
	return eDirection;
	
}

//always set meAction through this call!!!! 
void Unit::vSetAction(ACTIONS eAction)
{
#ifdef _DEBUG
	//assert on wrong actions
	switch (meType)
	{
	case UT_VAMPIRE:
		assert(eAction != ACT_SPECIAL); 
		break;
	case UT_VAMPIREHUNTER:
		assert(eAction != ACT_REACT); //TBD - this may change
		assert(eAction != ACT_FIDGET_UNCLOAKED);
		assert(eAction != ACT_SPECIAL); 
		break;
//			uncomment when revenats get put in
//	case UT_REVENANT:
//		assert(eAction != ACT_RUN);
//		assert(eAction != ACT_FEED);
//		assert(eAction != ACT_FIDGET_UNCLOAKED);
//		assert(eAction != ACT_CAST);
//			break;
	default: //rest of units: peasants, merchants, soldier, gypsy, etc
		assert(eAction != ACT_FIDGET_UNCLOAKED);
		assert(eAction != ACT_CAST);
		assert(eAction != ACT_SPECIAL); 
		break;
	}
#endif

	if (eAction != meAction)
	{
		//if we we're just feeding on someone and our action now changed, put attackee back into fidget mode
		/*
		if (meAction == ACT_FEED && mnAttackee)
		{
			Unit * poAttackee = (Unit*)goObjectDB.poGetObject(mnAttackee);
			if (poAttackee)
			{
				poAttackee->vFidget();
			}
			vSetAttackee(NULL);
		}*/

		meAction = eAction;

//		if (mpoWalker) {
//			mpoBitmap = mpoWalker->poGetBitmap(this, meDirection);
//			vSetActionLoop();
//		}
		
		vSetActionLoop();
		if (mpoCycler) {
			mpoCycler->vReset();
		}

	}

}

void Unit::vCue()
{
	switch (meAction)
	{
	//don't need to do anything for these actions
	case ACT_WALK:
		break;
	case ACT_RUN:
		break;
	case ACT_FIDGET_CLOAKED:
		break;
	case ACT_FIDGET_UNCLOAKED:
		break;
	case ACT_FEED:
		break;

	//return to fidget mode after these animations are complete
	case ACT_CAST:
	case ACT_REACT:
	case ACT_ATTACK:
		if (mbCloaked || !bIsVampire())
			vSetAction(ACT_FIDGET_CLOAKED);
		else
			vSetAction(ACT_FIDGET_UNCLOAKED);
		break;

	//remove cycler after last frame of death animation
	case ACT_DIE:
		if (mpoCycler)
		{
			delete mpoCycler;
			mpoCycler = NULL;
		}
		Missile * poMissile = new Missile(MISSILE_ANIMATESTOPANDFADEAWAY, gpoMap->mpoPlane, this, mpoCurNode, mpoCurNode, 6023, 0, FALSE, 1.18f, NULL, 10.0, 10.0, 2, 1);
		poMissile->SetDrawObj(nGetID());
		poMissile->vSetLayer(DP_FLOOR);
		poMissile->SetCyclesRemaining(200);

		break;
	}

}


void Unit::FireAlertClan(int nPlayerNumber)
{
	// FOR THE RANGE, CALL AlertClan(nPlayerNumber, FALSE)
	// FOR ALL HUMAN UNITS
	// MAX ANXIETY BLAST RANGE
	int nRange = vGetClanTemplate()->lGetValue(TAG_MFAC, 15);
	RECT r, r1,r2;
	r1.left = nGetX() - nRange;
	r1.right = nGetX() + nRange;
	r1.top = nGetY() - nRange;
	r1.bottom = nGetY() + nRange;
	r2.left = 0;
	r2.top = 0;
	r2.right = gpoMap->nGetWidth()-1;
	r2.bottom = gpoMap->nGetHeight()-1;

	if (!IntersectRect(&r, &r1, &r2)) return;
	
	for (int i=r.left; i<=r.right; i++)
	{
		for (int j=r.top; j<=r.bottom; j++)
		{
			Unit * poUnit = gpoMap->poGetNode(i, j)->mpoUnit;
			if (poUnit && poUnit->bIsVampire())
			{
				if (poUnit->poOwner() == mpoOwner)
				{
					AlertClan(nPlayerNumber, FALSE);
				}
			}
		}
	}
}


int Unit::GetPronunciationOfClanForPlayer(int nPlayerNumber)
{
	int nSize = game->poPlayers()->GetSize();
	if (nPlayerNumber >= nSize) return 0;
	int nTemplate = game->poPlayers()->GetItem(nPlayerNumber)->mpoClan->GetTemplate()->resourceID;
	long lTag = TAG_Sc01; 
	switch (nTemplate)
	{
	case 10:
		lTag = TAG_Sc00;
		break;
	case 11:
		lTag = TAG_Sc01;
		break;
	case 12:
		lTag = TAG_Sc02;
		break;
	case 13:
		lTag = TAG_Sc03;
		break;
	}
	return vGetClanTemplate()->lGetValue(lTag, 7116);
}

BOOL Unit::AlertClan(int nPlayerNumber, BOOL bWithSound)
{
	BOOL bReturn = FALSE;
	if (bWithSound)
	{
		BOOL bCheck = (0 == mnAlertCyclesRemaining);
		FireAlertClan(nPlayerNumber);
		if (bCheck)
		{
			//PlaySound!
			if (mpoOwner == game->poLocalPlayer()) 
				pApp->pSoundManager->EmitFromObject(this, GetPronunciationOfClanForPlayer(nPlayerNumber));
			bReturn = TRUE;
		}
	}
	mnAlertCyclesRemaining = ((Clan *)vGetClan())->GetAlertCycleLimit();
	return bReturn;
}

void Unit::DecrementCycleCounts()
{
	if (mnAlertCyclesRemaining) 
		mnAlertCyclesRemaining--;

	// IF mnAdrenalinCycleTicker == 0 ADRENALIN CAN BE USED TO BOOST SPEED
	if (mnAdrenalineCycleTicker != 0)
	{
		vProcessAdrenaline();
	}
}

void Unit::vProcessAdrenaline()
{
}

void Unit::SetSpy(Player * poPlayer)
{
	mwSpyFlags = mwSpyFlags | ((WORD)1<<poPlayer->mnPlayer);
}

void Unit::ClearSpy(Player * poPlayer)
{
	mwSpyFlags = mwSpyFlags & (~((WORD)1<<poPlayer->mnPlayer));
}

void Unit::SetInvisible()
{
	mbInvisible = TRUE;
	if (mpoOwner == game->poLocalPlayer())
	{
		mjAlpha = 64;
	}
	else
	{
		mjAlpha = 0;
	}
}

void Unit::ClearInvisible()
{
	mbInvisible = FALSE;
	mjAlpha = 255;
}


BOOL Unit::bInFog()
{
	return mnFog > 0; // mdwFlags & UNIT_FLAG_INFOG;
}

void Unit::SetInFog()
{
	//mdwFlags = mdwFlags | UNIT_FLAG_INFOG;
	mnFog++;
}

void Unit::SetNoFog()
{
	mnFog--;
	//mdwFlags = mdwFlags & (~UNIT_FLAG_INFOG);
}

//calculates and sets loop based on current action and direction
void Unit::vSetActionLoop()
{
//	if (gbMapEditing)
//		return;

	Boolean bFlip = FALSE;
	int nLoop;
	DIRECTION eDir = meDirection;

	if (eDir > D_NORTH)
		bFlip = TRUE;
	
	if (eDir == D_NORTHEAST) 
		eDir = D_NORTHWEST;
	if (eDir == D_EAST) 
		eDir = D_WEST;
	if (eDir == D_SOUTHEAST) 
		eDir = D_SOUTHWEST;
	
	nLoop = ((int)meAction * 10) + (int)eDir;

//	vDebugOut ("%d %d act:%d\n", nLoop, mnLoop, (int)meAction);

	if (nLoop != mnLoop)
	{
//		vSetCel(0);
		vSetLoop(nLoop);
	}
	vSetFlip(bFlip);

}

void Unit::PlayClassSound(int nSoundName, int nWait, int nWithin, int nMax)
{
	BOOL bPlay = FALSE;
	int nCurCycle = game->nGameCycle();

	if (nCurCycle - mclassstaticLastPlayedClassSound > nWait)
	{

		mclassstaticClassSounds = 1;
		mclassstaticLastPlayedClassSound = nCurCycle;

		bPlay = TRUE;
	}
	else if ((mclassstaticClassSounds < nMax) && (nCurCycle - mclassstaticLastPlayedClassSound < nWithin))
	{
		mclassstaticClassSounds++;
		bPlay = TRUE;
	}
	if (bPlay)
	{
		UINT unX, unY;

		// GET THE POSITION
		GetPlayPosition(&unX, &unY);

		// PLAY THE SOUND
		long *plList;
		long lSize;

		plList = mpoTemplate->plGetList(nSoundName, &lSize);

		if (lSize && plList)
		{
			int nWhich = random((short)lSize);
			int nSound = (int)plList[nWhich];

			
			// PLAY THE SOUND FROM THE POSITION
			pApp->pSoundManager->EmitFromPosition(unX, unY, nSound);
		}
	}
}

void Unit::vSetDefaultAttack(DEFAULT_ATTACK eAtt) 
{
	meDefaultAttack = eAtt;
	gpoInterface->vUpdateButtonState(ATTR_DEFAULT_ATTACK);
}



INTLIST::INTLIST()
{
	mnAllocated = 0;
	mnCount = 0;
	anList = NULL;
}

INTLIST::~INTLIST()
{
	if (anList) 
	{
		free(anList);
		anList = NULL;
		mnCount = 0;
	}
}

void INTLIST::AddToEnd(int nValue)
{
	int nNewSize = (mnCount + 1) * sizeof(int);
	int * pNewList;
	if (mnAllocated < nNewSize)
	{
		int nAllocated = mnAllocated + 20 * sizeof(int);
		pNewList = (int *)malloc(nAllocated);
		memmove(pNewList, anList, mnAllocated);
		mnAllocated = nAllocated;
		free(anList);
		anList = pNewList;
	}
	anList[mnCount] = nValue;
	mnCount++;
}

void INTLIST::RemoveAll()
{
	mnCount = 0;
}

int INTLIST::GetSize()
{
	return mnCount;
}

int INTLIST::GetItem(int nIndex)
{
	assert(nIndex < mnCount);
	return anList[nIndex];
}
