#ifndef _VAMPIRE_H_
#define _VAMPIRE_H_

#ifndef _UNIT_
#include "unit.h"
#endif

#ifndef _MISSILE_H_
#include "missile.h"
#endif

#ifndef _NONPLAYERCLASS_H_
#include "NonPlayerClass.h"
#endif

class Vampire : public Unit
{
	friend class Spell;
	friend class VampireDecoy;
	friend HRESULT NonPlayerClass::vCreateConvert(Vampire ** ppoVampire, Vampire * poParentVampire, Unit * poOriginalUnit, MapNode * poNode, MapPlane * poPlane);

	void vProcessAdrenaline();
	void SetSpeedAdrenalineBoost();
	void FindRetreatLocation(int nDistance, int * pnReturnX, int * pnReturnY);
public:
	Vampire(Player *poOwner, MapPlane *poPlane, MapNode *poNode, int nTemplateID);
	~Vampire();

	Boolean bOnMe(sciEvent &oEvent);		//need to overide for additional 'crown' check
	void CallForHelp();
	int RequestHelp(Vampire * poVampireNeedsHelp);

	void FireAnxiety(int nAmount);
	
	BOOL bAbleToPerformCommand(Command & oCommand, BOOL bWithSound = FALSE);
	BOOL bCanAttack();
	BOOL bIsHuman() { return FALSE; }
	BOOL bCanKillVampireLord() { return bIsVampireLord(); }
	void AdjustSpellCost(FLOAT * pfCost);
	// OVERRIDES
	Template * vGetClanTemplate();
	vClan * vGetClan();
	short vGetRelativeHealth(Player * poPlayer);
	void vUpdateRelativeHealth(Player * poPlayer);

// VAMPIRE SPECIFIC
	void vCommandNone();
	FLOAT vDamage(FLOAT nDamage, Unit *poUnitFrom = NULL);
	void vCommand(Command &oCommand, BOOL bInternal = FALSE);
	void vDraw();

	//map editing mode specific stuff
	void vCopySelf(sciEvent &oEvent);	//makes copy of self and inserts into node nearest event
	void vInsertSelf(sciEvent &oEvent);	//inserts self into node nearest event

	// THESE ARE CALLED BY SOME OF THE UNIT CLASSES TOO.
	static float GetPowerForLevel(int nLevel);
	static float GetHealthForLevel(int nLevel);
	static BOOL bSpellCanCastWhileCloaked(SPELL_TYPE eSpell);

	void SetBitmapCloaked(int nBitmap);
	int vGetVisibleRange();
	
protected:
	int mnTracerReset;
	void vMoveOneCycle();
	// OVERRIDES
	
	void vFeed();
	int vConvert();
	void vCloak();
	void vUncloak();
	//void vScan();
	void vAttack();
	void vPowerToHealth();
	void vRegeneratePower();
	void vSetLevel(int nLevel);
	void vUpdateAttributes();
	void vCast();
	int mnBitmapCloaked;
	
	
public:
	void vGameCycle();
	BOOL IsOnSameTeamAs(Player * poPlayer); 
	void GainPowerFromStone(Item * poItem, FLOAT fPower);
	void GainPowerFromOwnedStone(Item * poItem);
	FLOAT GetOwnedPowerGain();
	FLOAT GetMaxPowerGain();
	FLOAT GetMinPowerGain();
	void vFireMissile(MapNode * poTargetNode, MISSILE_TYPE missile);

	static void GroupUpgradeSight(int nNewValue)		{ 	GroupUpgradeUpgrade(UPGRADE_Sight,		nNewValue); }
	static void GroupUpgradeAttack(int nNewValue)		{ 	GroupUpgradeUpgrade(UPGRADE_Attack,		nNewValue);	}
	static void GroupUpgradeDefend(int nNewValue)		{ 	GroupUpgradeUpgrade(UPGRADE_Defend,		nNewValue);	}
	static void GroupUpgradeSpeed(int nNewValue)		{ 	GroupUpgradeUpgrade(UPGRADE_Speed,		nNewValue);	}
	static void GroupUpgradeStealth(int nNewValue)		{ 	GroupUpgradeUpgrade(UPGRADE_Stealth,	nNewValue);	}
	static void GroupUpgradeDetection(int nNewValue)	{ 	GroupUpgradeUpgrade(UPGRADE_Detection,	nNewValue);	}
	static void GroupUpgradeSpells(int nNewValue)		{ 	GroupUpgradeUpgrade(UPGRADE_Spells,		nNewValue);	}
	static void GroupUpgradeLevel(int nNewValue)		{ 	GroupUpgradeUpgrade(UPGRADE_Level,		nNewValue);	}

	BOOL bCloaked();
	BOOL bIsVampireLord();
	static void SetTenacity(SETTING_LMH setting);
	static void SetAggression(SETTING_LMH setting);
	SETTING_LMH GetTenacity() { return mTenacity; }
	SETTING_LMH GetAggression() { return mAggression; }
	Player *poOwner(void) { return mpoOwner; }

	inline int GetConversionSpeed();
	inline FLOAT GetPowerRegenPercent();
	inline FLOAT GetPowerExceedingEmergencyLevel();
	int GetPowerCostToConvert();
	int GetNewConvertHealth();
	int GetNewConvertPower();
	int GetRangeToDecloak();

		//FOR SETTING,CLEARING, AND CHECKING LEVEL/STATUS DISPLAY
	int manApparentHealth[MAX_PLAYERS];
	int manApparentLevel[MAX_PLAYERS];
	short manRelativeHealth[ALLPLAYERINDICES];
	void vSetMarked (int nPlayerNumber, int nApparentHealth, int nApparentLevel);
	void vGetMarked (int nPlayerNumber, int * pnApparentHealth, int * pnApparentLevel);
	void vClearMarked (int nPlayerNumber);
	Boolean bMarked(int nPlayerNumber);
	StatusIndicator moStatus;


	//  UPGRADES
	//
	//	FUNCTIONS TO SET UPGRADES
	//
	//  THESE ARE NET GAME COMPATIBLE.  BEFORE CALLING, VERIFY THAT YOU SHOULD CALL THEM
	//  WITH A CALL TO [ BOOL bCanUpgrade*() ] -- SEE BELOW.
	//
	void UpgradeSight(int nNewValue)	 { 	UpgradeUpgrade(UPGRADE_Sight,		nNewValue); }
	void UpgradeAttack(int nNewValue)	 { 	UpgradeUpgrade(UPGRADE_Attack,		nNewValue);	}
	void UpgradeDefend(int nNewValue)	 { 	UpgradeUpgrade(UPGRADE_Defend,		nNewValue);	}
	void UpgradeSpeed(int nNewValue)	 { 	UpgradeUpgrade(UPGRADE_Speed,		nNewValue);	}
	void UpgradeStealth(int nNewValue)	 { 	UpgradeUpgrade(UPGRADE_Stealth,		nNewValue);	}
	void UpgradeDetection(int nNewValue) { 	UpgradeUpgrade(UPGRADE_Detection,	nNewValue);	}
	void UpgradeSpells(int nNewValue)	 { 	UpgradeUpgrade(UPGRADE_Spells,		nNewValue);	}
	void UpgradeLevel(int nNewValue)	 { 	UpgradeUpgrade(UPGRADE_Level,		nNewValue);	}

	BOOL bCanUpgradeSight(int nNewValue)	 { 	return bCanUpgradeUpgrade(UPGRADE_Sight,		nNewValue);	}
	BOOL bCanUpgradeAttack(int nNewValue)	 { 	return bCanUpgradeUpgrade(UPGRADE_Attack,		nNewValue);	}
	BOOL bCanUpgradeDefend(int nNewValue)	 { 	return bCanUpgradeUpgrade(UPGRADE_Defend,		nNewValue);	}
	BOOL bCanUpgradeSpeed(int nNewValue)	 { 	return bCanUpgradeUpgrade(UPGRADE_Speed,		nNewValue);	}
	BOOL bCanUpgradeStealth(int nNewValue)	 { 	return bCanUpgradeUpgrade(UPGRADE_Stealth,		nNewValue);	}
	BOOL bCanUpgradeDetection(int nNewValue) { 	return bCanUpgradeUpgrade(UPGRADE_Detection,	nNewValue);	}
	BOOL bCanUpgradeSpells(int nNewValue)	 { 	return bCanUpgradeUpgrade(UPGRADE_Spells,		nNewValue);	}
	BOOL bCanUpgradeLevel(int nNewValue)	 { 	return bCanUpgradeUpgrade(UPGRADE_Level,		nNewValue);	}

	//
	//	UPGRADE IN PROGRESS -- INFORMATION
	//
	BOOL bIsUpgradeInProgress() { return mbUpgrading; }
	void GetUpgradeInProgress(UPGRADE_NAME * pnUpgradeName, int * pnUpgradeProgress);

	//
	//	FUNCTIONS TO GET UPGRADE SETTINGS
	//
	
	virtual int GetUpgradeSight()  { return manUpgrade[UPGRADE_Sight]; }
	virtual int GetUpgradeAttack()  { return manUpgrade[UPGRADE_Attack]; }
	virtual int GetUpgradeDefend()  { return manUpgrade[UPGRADE_Defend]; }
	virtual int GetUpgradeSpeed()  { return manUpgrade[UPGRADE_Speed]; }
	virtual int GetUpgradeStealth()  { return manUpgrade[UPGRADE_Stealth]; }
	virtual int GetUpgradeDetection()  { return manUpgrade[UPGRADE_Detection]; }
	virtual int GetUpgradeSpells()  { return manUpgrade[UPGRADE_Spells]; }
	virtual int GetUpgradeLevel()  { return manUpgrade[UPGRADE_Level]; }
	
	int GetDesiredSight()  { return manDesired[UPGRADE_Sight]; }
	int GetDesiredAttack()  { return manDesired[UPGRADE_Attack]; }
	int GetDesiredDefend()  { return manDesired[UPGRADE_Defend]; }
	int GetDesiredSpeed()  { return manDesired[UPGRADE_Speed]; }
	int GetDesiredStealth()  { return manDesired[UPGRADE_Stealth]; }
	int GetDesiredDetection()  { return manDesired[UPGRADE_Detection]; }
	int GetDesiredSpells()  { return manDesired[UPGRADE_Spells]; }
	int GetDesiredLevel()  { return manDesired[UPGRADE_Level]; }

protected:
	short				manUpgrade[UPGRADE_COUNT];
	short				manDesired[UPGRADE_COUNT];
	float				mfUpgradeProgress;
	float				mfUpgradeProgressPerCycle;
	UPGRADE_NAME		mUpgradeName;
	BOOL				mbUpgrading;
	BOOL				mbDesiredCloakState;
	UPGRADE_NAME		mLastUpgrade;
	int					mnCyclesUntilNextSound;
	int					mnCelCastDelay;
	Clan *				mpoClan;
	Bitmap				*mpoCrown;			//the crown(or head) of the vampire lord
	WORD *				mpwTruePalette;

	DWORD mdwUpgradesPending;

	void UpgradeUpgrade(UPGRADE_NAME UpgradeName, int nNewValue); 
	static void GroupUpgradeUpgrade(UPGRADE_NAME UpgradeName, int nNewValue); 
	BOOL bCanUpgradeUpgrade(UPGRADE_NAME upgrade, int nNewValue);
	inline void NetworkUpgradeReceived(UPGRADE_NAME upgrade); 	

	FLOAT GetRequiredPowerToUpgrade(UPGRADE_NAME upgrade);
	int GetRequiredTimeToUpgrade(UPGRADE_NAME upgrade);

	void ProcessUpgrades();

};

class VampireDecoy : public Vampire
{
public:
	VampireDecoy(Player *poOwner, MapPlane *poPlane, MapNode *poNode, int nTemplateID);
	~VampireDecoy();
	FLOAT vDamage(FLOAT fDamage, Unit * poUnitFrom);
	void vGameCycle();
	void vCommand(Command &oCommand, BOOL bInternal = FALSE);
	void vCast();
	void SetUnitParentID(int nUnitID);

protected:
	inline void	CheckNodeForVisibleUnits(MapNode * poNode, int nDistance);
	void vScan();
	int mnUnitParentID;
};


#endif