#include "os.h"

#include "bitmap.h"
#include "app.h"
#include "window.h"
#include "resource.h"
#include "ztimer.h"

//void rmemcpy(BYTE * hpDest, BYTE * hpSource, WORD wWidth);
//void rtmemcpy(BYTE * hpDest, BYTE * hpSource, WORD wWidth);
//void tmemcpy(BYTE * hpDest, BYTE * hpSource, WORD wWidth);

VBitmap::VBitmap(int newResourceID, ResourceManager * newResourceManager)
:Resource(RES_BITMAP, newResourceID, newResourceManager)
{
	mpPixels = NULL;
	bFlip = FALSE;
	mpoPal16 = NULL;
}

VBitmap::VBitmap(WORD wNewWidth, WORD wNewHeight, int nBits)
:Resource(RES_NONE, 0, NULL)
{
	assert(nBits == 8 || nBits == 16);
	int size;

	mnBits = nBits;

	mpoPal16 = NULL;

	mpPixels = NULL;
	bFlip = FALSE;

	wWidth = wNewWidth;
	wHeight = wNewHeight;
	if (nBits == 8) {
		wWidthInBytes = wWidth + 3 - ((wWidth - 1) % 4);
	} else {
		wWidthInBytes = wWidth * 2;
	}

	// Allocation memory for the pixels.
	size = (DWORD) (wWidthInBytes) * (DWORD) (wHeight);
	mpPixels = new BYTE[size];
	memset(mpPixels, 0, size);
}

VBitmap::VBitmap(RESOURCE_TYPES theType, char mask2, int newResourceID, ResourceManager * newResourceManager)
:Resource(theType, newResourceID, newResourceManager)
{
	mpPixels = NULL;
	wWidth = 0;
	wHeight = 0;
	wWidthInBytes = 0;
	bFlip = FALSE;
	mpoPal16 = NULL;
	mnBits = 8;
}

VBitmap::~VBitmap()
{
	if (mpPixels) {
		delete[] mpPixels;
	}
	if (mpoPal16) {
		delete mpoPal16;
	}
}

/*
WORD VBitmap::Width()
{
	return wWidth;
}

WORD VBitmap::Height()
{
	return wHeight;
}
*/

#pragma optimize( "agt", on )

void VBitmap::ClipRects(RECT & rcDest, RECT & rcDestClippingRect,
						 RECT & rcSource, RECT & rcSourceClippingRect)
{
	RECT rcNewSource,
	 rcNewSourceDelta;
	RECT rcNewDest,
	 rcNewDestDelta;

	// Widths and Heights must match - Stretching not supported.
	if ((rcDest.right - rcDest.left != rcSource.right - rcSource.left) ||
		(rcDest.right - rcDest.left != rcSource.right - rcSource.left) ||
		(rcDest.right - rcDest.left != rcSource.right - rcSource.left) ||
		(rcDest.right - rcDest.left != rcSource.right - rcSource.left)) {
		SetRect(&rcDest, 0, 0, 0, 0);
		SetRect(&rcSource, 0, 0, 0, 0);
		return;
	}
	// Clip the source rect to its valid bounds.
	IntersectRect(&rcNewSource, &rcSourceClippingRect, &rcSource);

	// Get the amounts the source rect changed.
	rcNewSourceDelta.left = rcNewSource.left - rcSource.left;
	rcNewSourceDelta.top = rcNewSource.top - rcSource.top;
	// ??? Are right and bottom negative or positive?
	rcNewSourceDelta.right = rcNewSource.right - rcSource.right;
	rcNewSourceDelta.bottom = rcNewSource.bottom - rcSource.bottom;

	// Change the dest rect by the same amounts.
	rcDest.left += rcNewSourceDelta.left;
	rcDest.top += rcNewSourceDelta.top;
	rcDest.right += rcNewSourceDelta.right;
	rcDest.bottom += rcNewSourceDelta.bottom;
	rcSource.left += rcNewSourceDelta.left;
	rcSource.top += rcNewSourceDelta.top;
	rcSource.right += rcNewSourceDelta.right;
	rcSource.bottom += rcNewSourceDelta.bottom;

	// Make sure the dest rect didn't collapse.
	if ((rcDest.right - rcDest.left < 0) ||
		(rcDest.bottom - rcDest.top < 0)) {
		SetRect(&rcDest, 0, 0, 0, 0);
		SetRect(&rcSource, 0, 0, 0, 0);
		return;
	}
	// Clip the dest rect to its valid bounds.
	IntersectRect(&rcNewDest, &rcDestClippingRect, &rcDest);

	// Get the amounts the dest rect changed.
	rcNewDestDelta.left = rcNewDest.left - rcDest.left;
	rcNewDestDelta.top = rcNewDest.top - rcDest.top;
	// ??? Are right and bottom negative or positive?
	rcNewDestDelta.right = rcNewDest.right - rcDest.right;
	rcNewDestDelta.bottom = rcNewDest.bottom - rcDest.bottom;

	// Change the source rect by the same amounts.
	rcSource.left += rcNewDestDelta.left;
	rcSource.top += rcNewDestDelta.top;
	rcSource.right += rcNewDestDelta.right;
	rcSource.bottom += rcNewDestDelta.bottom;
	rcDest.left += rcNewDestDelta.left;
	rcDest.top += rcNewDestDelta.top;
	rcDest.right += rcNewDestDelta.right;
	rcDest.bottom += rcNewDestDelta.bottom;

	// Make sure the source rect didn't collapse.
	if ((rcSource.right - rcSource.left < 0) ||
		(rcSource.bottom - rcSource.top < 0)) {
		SetRect(&rcDest, 0, 0, 0, 0);
		SetRect(&rcSource, 0, 0, 0, 0);
		return;
	}
}

#pragma optimize( "agt", on )

void VBitmap::CopyRow(BYTE far * hpDest, BYTE far * hpSource, WORD wWidth, BOOL bTransparent)
{
	assert(mnBits == 8);
	if (bFlip) {
		if (bTransparent)
			rtmemcpy(hpDest, hpSource, wWidth);
		else
			rmemcpy(hpDest, hpSource, wWidth);
	} else {
		if (bTransparent)
			tmemcpy(hpDest, hpSource, wWidth);
		else
			memcpy(hpDest, hpSource, wWidth);
	}
}

#pragma optimize( "agt", on )

void VBitmap::vDrawToFrame(FrameBuffer * poFrameBuf, POINT & tDest, RECT & rcSource, BYTE jAlpha, COLOR16 wRemapColor, BYTE jRemapAmount, WORD *pwPalette)
{

	// it's already been clipped
	int nPitch = poFrameBuf->mnPitch;
	int nHeight = rcSource.bottom - rcSource.top;
	int nWidth = rcSource.right - rcSource.left;

	BYTE *pjScreen = (BYTE *) poFrameBuf->pGetBuffer();
	BYTE *pjSource = mpPixels;

	// void *pScreenPixels = poFrameBuf->pGetBuffer();

	switch (poFrameBuf->mnBits) {
	case 8:
		assert(mnBits == 8);
		pjScreen += tDest.y * nPitch + tDest.x;
		pjSource += (rcSource.top * wWidthInBytes) + rcSource.left;
		if (bFlip) {
			while (nHeight--) {
				rtmemcpy(pjScreen, pjSource, nWidth);
				pjScreen += nPitch;
				pjSource += wWidthInBytes;
			}
		} else {
			while (nHeight--) {
				tmemcpy(pjScreen, pjSource, nWidth);
				pjScreen += nPitch;
				pjSource += wWidthInBytes;
			}
		}

		break;

	case 16:
		pjScreen += tDest.y * nPitch + (tDest.x * sizeof(WORD));
		switch (mnBits) {
		case 8:

			//assert(mpoPal16);
			//WORD *pwPalette;

			// if we don't have a palette of our own, use the
			// frame buffer's system palette
			if (!pwPalette) {
				if (mpoPal16) {
					pwPalette = mpoPal16->pwGetPalette();
				} else {
					pwPalette = poFrameBuf->pwGetPalette();
				}
			}
			assert(pwPalette);

			pjSource += (rcSource.top * wWidthInBytes) + rcSource.left;
			if (jAlpha == 255) {
				if (bFlip) {
					while (nHeight--) {
						vCopyPixelsPaletteTransRev16((WORD *) pjScreen, pjSource, nWidth, pwPalette);
						pjScreen += nPitch;
						pjSource += wWidthInBytes;
					}
				} else {
					while (nHeight--) {
						vCopyPixelsPaletteTrans16((WORD *) pjScreen, pjSource, nWidth, pwPalette);
						pjScreen += nPitch;
						pjSource += wWidthInBytes;
					}
				}
			} else {
				if (bFlip) {
					while (nHeight--) {
						vCopyPixelsPaletteTransAlphaRev16((WORD *) pjScreen, pjSource, nWidth, pwPalette, jAlpha);
						pjScreen += nPitch;
						pjSource += wWidthInBytes;
					}
				} else {
					while (nHeight--) {
						vCopyPixelsPaletteTransAlpha16((WORD *) pjScreen, pjSource, nWidth, pwPalette, jAlpha);
						pjScreen += nPitch;
						pjSource += wWidthInBytes;
					}
				}
			}
			break;

		case 16:

			pjSource += ((rcSource.top * wWidth) + rcSource.left) * sizeof(WORD);
			if (jAlpha == 255) {
				if (bFlip) {
					while (nHeight--) {
						vCopyPixelsTransRev16((WORD *) pjScreen, (WORD *) pjSource, nWidth);
						pjScreen += nPitch;
						pjSource += wWidthInBytes;
					}
				} else {
					while (nHeight--) {
						vCopyPixelsTrans16((WORD *) pjScreen, (WORD *) pjSource, nWidth);
						pjScreen += nPitch;
						pjSource += wWidthInBytes;
					}
				}
			} else {
				if (bFlip) {
					while (nHeight--) {
						vCopyPixelsTransAlphaRev16((WORD *) pjScreen, (WORD *) pjSource, nWidth, jAlpha);
						pjScreen += nPitch;
						pjSource += wWidthInBytes;
					}
				} else {
					while (nHeight--) {
						vCopyPixelsTransAlpha16((WORD *) pjScreen, (WORD *) pjSource, nWidth, jAlpha);
						pjScreen += nPitch;
						pjSource += wWidthInBytes;
					}
				}
			}
			break;
		}
		break;
	}

	poFrameBuf->vReleaseBuffer();
}


void VBitmap::CopyPixels(Bitmap * pDestBitmap, RECT rcDest, RECT rcSource, BOOL bTransparent)
{
	RECT rcClipSource,
	 rcClipDest;

	assert(mnBits == 8);
	assert(pDestBitmap);
	assert(pDestBitmap->mnBits == 8);

	if (mpPixels == 0 || pDestBitmap->mpPixels == 0)
		return;
	SetRect(&rcClipSource, 0, 0, Width(), Height());
	SetRect(&rcClipDest, 0, 0, pDestBitmap->Width(), pDestBitmap->Height());
	ClipRects(rcDest, rcClipDest, rcSource, rcClipSource);

#ifdef FLIP_DIB
	RECT rcTemp;
	rcTemp.top = pDestBitmap->Height() - rcDest.bottom;
	rcTemp.bottom = pDestBitmap->Height() - rcDest.top;
	rcTemp.right = rcDest.right;
	rcTemp.left = rcDest.left;
	rcDest = rcTemp;

	rcTemp.top = Height() - rcSource.bottom;
	rcTemp.bottom = Height() - rcSource.top;
	rcTemp.right = rcSource.right;
	rcTemp.left = rcSource.left;
	rcSource = rcTemp;
#endif

	if (!IsRectEmpty(&rcSource))
		CopyRectangle(pDestBitmap->mpPixels, mpPixels,
					  (WORD) rcDest.left, (WORD) rcDest.top, (WORD) rcSource.left, (WORD) rcSource.top,
					  (WORD) (rcSource.right - rcSource.left), (WORD) (rcSource.bottom - rcSource.top),
		 pDestBitmap->wWidthInBytes, wWidthInBytes, bTransparent, bFlip);
}

#pragma optimize( "agt", on )

void VBitmap::vErase8(RECT rDest, BYTE jColor)
{
	RECT rcTemp;
	BYTE *pixels;
	int y;

	assert(mnBits == 8);

	if (mpPixels == 0)
		return;

	// Clip the rect to its valid bounds.
	SetRect(&rcTemp, 0, 0, Width(), Height());
	IntersectRect(&rDest, &rcTemp, &rDest);

	// Don't erase if empty rect.
	if (IsRectEmpty(&rDest))
		return;

	pixels = mpPixels;
	if (pixels) {
		for (y = rDest.top; y < rDest.bottom; ++y) {
			memset(pixels + y * wWidthInBytes + rDest.left, jColor, rDest.right - rDest.left);
		}
	}
}

#pragma optimize( "agt", on )

void VBitmap::vErase16(RECT rDest, BYTE jRed, BYTE jGreen, BYTE jBlue)
{
	RECT rcTemp;
	WORD *pixels;
	int y;

	assert(mnBits == 16);

	if (mpPixels == 0)
		return;

	// Clip the rect to its valid bounds.
	SetRect(&rcTemp, 0, 0, Width(), Height());
	IntersectRect(&rDest, &rcTemp, &rDest);

	// Don't erase if empty rect.
	if (IsRectEmpty(&rDest))
		return;
	
	WORD wColor = (WORD)gpoBlender->dwPack(jRed, jGreen, jBlue);

	pixels = (WORD*)mpPixels;
	if (pixels) {
		for (y = rDest.top; y < rDest.bottom; ++y) {
			//pixels += y * Width() + rDest.left;
			vFillPixels16( (pixels + y * Width() + rDest.left), wColor, rDest.right - rDest.left);
//			memset(pixels + y * wWidthInBytes + rDest.left, color, rDest.right - rDest.left);
		}
	}

}

#define RLE_ESCAPE  0
#define RLE_EOL     0
#define RLE_EOF     1
#define RLE_JMP     2
#define RLE_RUN     3
#pragma optimize( "agt", on )

void VBitmap::CopyRLEPixels(WORD wWidthBytes, BYTE * pb, BYTE * prle)
{
	BYTE cnt;
	BYTE b;
	WORD x;
	WORD dx,
	 dy;

	x = 0;

	for (;;) {
		cnt = *prle++;
		b = *prle++;

		if (cnt == RLE_ESCAPE) {
			switch (b) {
			case RLE_EOF:
				return;
			case RLE_EOL:
				// pb += wWidthBytes - x;	// bottom-up
				pb += -wWidthBytes - x;	// top-down
				x = 0;
				break;
			case RLE_JMP:
				dx = (WORD) * prle++;
				dy = (WORD) * prle++;
				// pb += ((DWORD)wWidthBytes * dy) + dx;	// bottom-up
				pb += -((int) wWidthBytes * dy) + dx;	// top-down
				x += dx;
				break;
			default:
				cnt = b;
				x += (WORD) cnt;
				while (cnt-- > 0)
					*pb++ = (*prle++);
				if (b & 1)
					prle++;
				break;
			}
		} else {
			x += (WORD) cnt;

			while (cnt-- > 0)
				*pb++ = (b);
		}
	}
}

#pragma optimize( "agt", on )

void VBitmap::CopyRLEPixelsWithRemap(WORD wWidthBytes, BYTE * pb, BYTE * prle, BYTE(&remapping)[256])
{
	BYTE cnt;
	BYTE b;
	WORD x;
	WORD dx,
	 dy;
	BYTE rb;

	x = 0;

	for (;;) {
		cnt = *prle++;
		b = *prle++;

		if (cnt == RLE_ESCAPE) {
			switch (b) {
			case RLE_EOF:
				return;
			case RLE_EOL:
				// pb += wWidthBytes - x;	// bottom-up
				pb += -wWidthBytes - x;	// top-down
				x = 0;
				break;
			case RLE_JMP:
				dx = (WORD) * prle++;
				dy = (WORD) * prle++;
				// pb += ((DWORD)wWidthBytes * dy) + dx;	// bottom-up
				pb += -((int) wWidthBytes * dy) + dx;	// top-down
				x += dx;
				break;
			default:
				cnt = b;
				x += (WORD) cnt;
				while (cnt-- > 0)
					*pb++ = remapping[(*prle++)];
				if (b & 1)
					prle++;
				break;
			}
		} else {
			x += (WORD) cnt;
			rb = remapping[(b)];
			while (cnt-- > 0)
				*pb++ = rb;
		}
	}
}

#pragma optimize( "agt", on )

void VBitmap::CopyPixelsWithRemap(WORD wWidthBytes, BYTE * pb, BYTE * prle, BYTE(&remapping)[256])
{
	int i,
	 j;

	j = wWidthInBytes * wHeight;
	for (i = 0; i < wHeight; ++i) {
		for (j = 0; j < wWidthInBytes; ++j) {
			*pb++ = remapping[(*prle++)];
		}
		pb -= wWidthInBytes * 2;// top-down
	}
}

/*
void VBitmap::Flip()
{
	bFlip = !bFlip;
}
*/

#pragma optimize( "agt", on )

WORD VBitmap::ColorOfPoint(POINT tWhere)
{
	WORD wOutput;

#ifdef FLIP_DIB
	tWhere.y = Height() - tWhere.y - 1;
#endif
	if (mnBits == 8) {
		wOutput = mpPixels[tWhere.x + (DWORD) tWhere.y * wWidthInBytes];
	} else {
		wOutput = ((WORD *) mpPixels)[tWhere.x + (DWORD) tWhere.y * wWidth];
	}

	return wOutput;
}

void VBitmap::vSetPixel(POINT tWhere, WORD wColor)
{
	if (mnBits == 8) {
		mpPixels[tWhere.x + (DWORD) tWhere.y * wWidthInBytes] = (BYTE)wColor;
	} else {
		((WORD *) mpPixels)[tWhere.x + (DWORD) tWhere.y * wWidth] = wColor;
	}
}

void VBitmap::vMakeRect (WORD wColor, int nThickness) 
{
	int i;
	int nThCnt;					// thickness loop counter

	RECT rDest;
	SetRect(&rDest,0,0, wWidth, wHeight);

	if (mnBits == 8)
	{
		vErase8(rDest,0);
		for (nThCnt = 0; nThCnt < nThickness; ++nThCnt) 
		{

			memset(mpPixels + wWidthInBytes * nThCnt, (BYTE)wColor, wWidth);

			for (i = 1; i < wHeight; i++) {
				// Left vertical
				mpPixels[(i * wWidthInBytes) + nThCnt] = (BYTE)wColor;
				// Right vertical
				mpPixels[(i * wWidthInBytes) + wWidth - nThCnt - 1] = (BYTE)wColor;
			}

			// Bottom horizontal
			memset(mpPixels + (wHeight - nThCnt - 1) * wWidthInBytes, (BYTE)wColor, wWidth);
		}

	}
	else
	{
		WORD *pixels = (WORD*)mpPixels;
		vErase16(rDest,0,0,0);
				
		for (nThCnt = 0; nThCnt < nThickness; ++nThCnt) 
		{
			vFillPixels16(pixels + wWidth * nThCnt, wColor, wWidth);
			for (i = 1; i < wHeight; i++) {
				// Left vertical
				vFillPixels16(pixels + (i * wWidth) + nThCnt, wColor, 1);
				// Right vertical
				vFillPixels16(pixels + (i * wWidth) + wWidth - nThCnt - 1, wColor, 1);
			}

			// Bottom horizontal
			vFillPixels16(pixels + (wHeight - nThCnt - 1) * wWidth, wColor, wWidth);
			
		}

	}
	

}

void VBitmap::CopyRectangle(BYTE * pDestPixels, BYTE * pSourcePixels,
				WORD wDestX, WORD wDestY, WORD wSourceX, WORD wSourceY,
				WORD wBytesToCopy, WORD wRowsToCopy, WORD wDestWidthInBytes, WORD wSourceWidthInBytes,
				BOOL bTransparent, BOOL bFlip)
{
	if (bFlip) {
		pSourcePixels += (DWORD) wSourceY *wSourceWidthInBytes + Width() - 1 - wSourceX;
	} else {
		pSourcePixels += (DWORD) wSourceY *wSourceWidthInBytes + wSourceX;
	}
	pDestPixels += (DWORD) wDestY *wDestWidthInBytes + wDestX;
	while (wRowsToCopy--) {
		CopyRow(pDestPixels, pSourcePixels, wBytesToCopy, bTransparent);
		pSourcePixels += wSourceWidthInBytes;
		pDestPixels += wDestWidthInBytes;
	}
}

void VBitmap::vSetOriginPoint(short nNewX, short nNewY, int nLoop, int nCel)
{

}


// for when no inline
#ifndef _INLINE_BLITS

MY_INLINE void vFillPixels16(WORD *pwDest, WORD wColor, int nLen)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
#endif

	while (nLen--) {
		*pwDest++ = wColor;
	}
}

MY_INLINE void vFillPixelsRev16(WORD *pwDest, WORD wColor, int nLen)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
#endif

	while (nLen--) {
		*pwDest-- = wColor;
	}
}

MY_INLINE void vFillPixelsAlpha16(WORD *pwDest, WORD wColor, int nLen, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
#endif

#if 0
	while (nLen--) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha(*pwDest, wColor, jAlpha));
		pwDest++;
	}
#else
	jAlpha >>= 3;
	DWORD dwColor = wColor | ((DWORD)wColor << 16);
	int nRemainder = nLen & 1;
	nLen >>= 1;
	while (nLen--) {
		*(DWORD*)pwDest = gpoBlender->dwCalc2Alpha5(*(DWORD*)pwDest, dwColor, jAlpha);
		pwDest += 2;
	}
	if (nRemainder) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, wColor, jAlpha));
	}
#endif
}

MY_INLINE void vFillPixelsAlphaRev16(WORD *pwDest, WORD wColor, int nLen, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
#endif

#if 0
	jAlpha >>= 3;
	while (nLen--) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, wColor, jAlpha));
		pwDest--;
	}
#else
	jAlpha >>= 3;
	DWORD dwColor = wColor | ((DWORD)wColor << 16);
	if (nLen & 1) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, wColor, jAlpha));
		pwDest -= 2;
	} else {
		pwDest--;
	}
	nLen >>= 1;
	while (nLen--) {
		*(DWORD*)pwDest = gpoBlender->dwCalc2Alpha5(*(DWORD*)pwDest, dwColor, jAlpha);
		pwDest -= 2;
	}
#endif
}

#define TEST_ALT_FILL
#undef TIME_PIXEL_COPY
MY_INLINE void vCopyPixelsPalette16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

#ifdef TIME_PIXEL_COPY
	TIMER_BEGIN();
#endif

#ifdef TEST_ALT_FILL
	//int nRemainder = nLen & 1;
	if (nLen & 1) {
		*pwDest++ = pwPalette[*pjSource++];
	}
	nLen >>= 1;
	while (nLen--) {
		*(DWORD*)pwDest = ((DWORD)pwPalette[*pjSource]) | ((DWORD)pwPalette[*(pjSource + 1)] << 16);
		pwDest += 2;
		pjSource += 2;
	}

#else
	while (nLen--) {
		*pwDest++ = pwPalette[*pjSource++];
	}
#endif
#ifdef TIME_PIXEL_COPY
	TIMER_END();
#endif
}

MY_INLINE void vCopyPixelsPaletteRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

#ifdef TEST_ALT_FILL
	if (nLen & 1) {
		*pwDest = pwPalette[*pjSource];
		pwDest -= 2;
		pjSource++;
	} else {
		pwDest--;
	}
	nLen >>= 1;
	while (nLen--) {
		*(DWORD*)pwDest = ((DWORD)pwPalette[*(pjSource + 1)]) | ((DWORD)pwPalette[*pjSource] << 16);
		pwDest -= 2;
		pjSource += 2;
	}

#else
	while (nLen--) {
		*pwDest-- = pwPalette[*pjSource++];
	}
#endif
}

MY_INLINE void vCopyPixelsPaletteAlpha16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif
	jAlpha >>= 3;
	int nRemainder = nLen & 1;
	nLen >>= 1;

	while (nLen--) {
		*(DWORD*)pwDest = gpoBlender->dwCalc2Alpha5(*(DWORD*)pwDest, ((DWORD)pwPalette[*pjSource]) | ((DWORD)pwPalette[*(pjSource + 1)] << 16), jAlpha);
		pwDest += 2;
		pjSource += 2;
	}
	if (nRemainder) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource], jAlpha));
	}
}

MY_INLINE void vCopyPixelsPaletteAlphaRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	jAlpha >>= 3;

#if 0
	int nRemainder = nLen & 1;
	nLen >>= 1;

	while (nLen--) {
		*(DWORD*)pwDest = gpoBlender->dwCalc2Alpha5(*(DWORD*)pwDest, ((DWORD)pwPalette[*(pjSource + 1)]) | ((DWORD)pwPalette[*pjSource] << 16), jAlpha);
		pwDest -= 2;
		pjSource += 2;
	}
	if (nRemainder) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource], jAlpha));
	}
#else
	while (nLen--) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource++], jAlpha));
		pwDest--;
	}
#endif
}

MY_INLINE void vCopyPixelsPalGeneric16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BOOL bRev, BYTE jAlpha, WORD wRemapColor, BYTE jRemap)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	jAlpha >>= 3;
	jRemap >>= 3;
	if (jAlpha == 0) {
		// totally transparent
		return;
	}
	if (jAlpha == 31) {
		// no alpha

		if (jRemap) {
			if (bRev) {
				while (nLen--) {
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(pwPalette[*pjSource++], wRemapColor, jRemap));
					pwDest--;
				}
			} else {
				while (nLen--) {
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(pwPalette[*pjSource++], wRemapColor, jRemap));
					pwDest++;
				}
			}
		} else {
			if (bRev) {
				while (nLen--) {
					*pwDest = pwPalette[*pjSource++];
					pwDest--;
				}
			} else {
				while (nLen--) {
					*pwDest = pwPalette[*pjSource++];
					pwDest++;
				}
			}
		}
	} else {
		// there is alpha

		if (jRemap) {
			if (bRev) {
				while (nLen--) {
					WORD wTemp = (WORD)(gpoBlender->dwCalcAlpha5(pwPalette[*pjSource++], wRemapColor, jRemap));
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, wTemp, jAlpha));
					pwDest--;
				}
			} else {
				while (nLen--) {
					WORD wTemp = (WORD)(gpoBlender->dwCalcAlpha5(pwPalette[*pjSource++], wRemapColor, jRemap));
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, wTemp, jAlpha));
					pwDest++;
				}
			}
		} else {
			if (bRev) {
				while (nLen--) {
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource++], jAlpha));
					pwDest--;
				}
			} else {
				while (nLen--) {
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource++], jAlpha));
					pwDest++;
				}
			}
		}
	}
}

MY_INLINE void vCopyPixelsPaletteTrans16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	while (nLen--) {
		if (*pjSource) {
			*pwDest = pwPalette[*pjSource];
		}
		pwDest++;
		pjSource++;
	}
}

MY_INLINE void vCopyPixelsPaletteTransRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	while (nLen--) {
		if (*pjSource) {
			*pwDest = pwPalette[*pjSource];
		}
		pwDest--;
		pjSource++;
	}
}

MY_INLINE void vCopyPixelsPaletteTransAlpha16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	jAlpha >>= 3;
	while (nLen--) {
		if (*pjSource) {
			*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource], jAlpha));
		}
		pwDest++;
		pjSource++;
	}
}

MY_INLINE void vCopyPixelsPaletteTransAlphaRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	jAlpha >>= 3;
	while (nLen--) {
		if (*pjSource) {
			*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource], jAlpha));
		}
		pwDest--;
		pjSource++;
	}
}

MY_INLINE void vCopyPixels16(WORD *pwDest, WORD *pwSource, int nLen)
{
	memcpy(pwDest, pwSource, nLen << 1);
}

MY_INLINE void vCopyPixelsRev16(WORD *pwDest, WORD *pwSource, int nLen)
{
	while (nLen--) {
		*pwDest = *pwSource;
		pwDest--;
		pwSource++;
	}
}

MY_INLINE void vCopyPixelsTrans16(WORD *pwDest, WORD *pwSource, int nLen)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pwSource);
#endif

	while (nLen--) {
		if (*pwSource) {
			*pwDest = *pwSource;
		}
		pwDest++;
		pwSource++;
	}
}

MY_INLINE void vCopyPixelsTransRev16(WORD *pwDest, WORD *pwSource, int nLen)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pwSource);
#endif

	while (nLen--) {
		if (*pwSource) {
			*pwDest = *pwSource;
		}
		pwDest--;
		pwSource++;
	}
}

MY_INLINE void vCopyPixelsTransAlpha16(WORD *pwDest, WORD *pwSource, int nLen, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pwSource);
#endif

	jAlpha >>= 3;
	while (nLen--) {
		if (*pwSource) {
			*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, *pwSource, jAlpha));
		}
		pwDest++;
		pwSource++;
	}
}

MY_INLINE void vCopyPixelsTransAlphaRev16(WORD *pwDest, WORD *pwSource, int nLen, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pwSource);
#endif

	jAlpha >>= 3;
	while (nLen--) {
		if (*pwSource) {
			*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, *pwSource, jAlpha));
		}
		pwDest--;
		pwSource++;
	}
}

MY_INLINE void rmemcpy(BYTE * hpDest, BYTE * hpSource, int nLen)
{
	while (nLen--) {
		*hpDest++ = *hpSource--;
	}
}

MY_INLINE void rtmemcpy(BYTE * hpDest, BYTE * hpSource, int nLen)
{
	while (nLen--) {
		if (*hpSource) {
			*hpDest = *hpSource;
		}
		++hpDest;
		--hpSource;
	}
}

MY_INLINE void tmemcpy(BYTE * hpDest, BYTE * hpSource, int nLen)
{
	while (nLen--) {
		if (*hpSource) {
			*hpDest = *hpSource;
		}
		++hpDest;
		++hpSource;
	}
}
#endif
