#ifndef _vbitmap
#define _vbitmap

#include "resource.h"
#include "palette.h"
#include "framebuf.h"

#ifdef OS_WINDOWS
//#define FLIP_DIB
#endif // OS_WINDOWS

class Window;
class Bitmap;

class VBitmap : public Resource
	{
	public:
		VBitmap(WORD wNewWidth, WORD wNewHeight, int nBits = 8);
		VBitmap(int newResourceID, ResourceManager *newResourceManager);
		VBitmap(RESOURCE_TYPES theType, char mask2, int newResourceID, ResourceManager *newResourceManager);
		virtual ~VBitmap();
		virtual void vDrawToFrame(FrameBuffer * pFrameBuf, POINT & tDest, RECT & rcSource, BYTE jAlpha = 255, COLOR16 wRemapColor = 0, BYTE jRemapAmount = 0, WORD *pwPalette = NULL);
		virtual void CopyPixels(Bitmap *pDestBitmap, RECT rcDest, RECT rcSource, BOOL bTransparent);
		virtual void CopyPixels(Window *pDestWindow, RECT rcDest, RECT rcSource) {};
		void vErase8(RECT rDest, BYTE jColor);
		void vErase16(RECT rDest, BYTE jRed, BYTE jGreen, BYTE jBlue);
		WORD Width() {return wWidth;}
		WORD Height() {return wHeight;}
		void Flip() {bFlip = !bFlip;}
		void vSetFlip(BOOL bFlipStatus) {bFlip = bFlipStatus;}
		virtual WORD ColorOfPoint(POINT point);
		void vSetPixel(POINT tWhere, WORD wColor);
		virtual short nGetOriginX(int nLoop = 0, int nCel = 0) {return 0;};
		virtual short nGetOriginY(int nLoop = 0, int nCel = 0) {return 0;};
		virtual void vSetOriginPoint(short nNewX, short nNewY, int nLoop = 0, int nCel = 0); 
		virtual void vMoveAllOrigins(int nDeltaX, int nDeltaY) {};
		int nBits() {return mnBits;}
		void vMakeRect (WORD wColor, int nThickness);

		/*
		void vSetAlpha(BYTE jAlpha) {mjAlpha = jAlpha;}
		void vSetAlpha(float fAlpha) {
			assert(fAlpha >= 0.0 && fAlpha <= 1.0);
			mjAlpha = (BYTE)(fAlpha * 255.0 + 0.5);
		}
		BYTE jGetAlpha(void) {return mjAlpha;}
		void vSetRemapColor(DWORD dwColor) {mdwRemapColor = dwColor;}
		void vSetRemap(BYTE jRemap) {mjRemap = jRemap;}
		void vSetRemap(float fRemap) {
			assert(fRemap >= 0.0 && fRemap <= 1.0);
			mjRemap = (BYTE)(fRemap * 255.0 + 0.5);
		}
		*/

	protected:
    	// Functions.
		void CopyPixels(BYTE *hpDestBytes, BYTE *hpSourceBytes);
		void CopyRLEPixels(WORD wWidthBytes, BYTE *hpDestBytes,
			BYTE *hpSourceBytes);
		void CopyRLEPixelsWithRemap(WORD wWidthBytes, BYTE *hpDestBytes,
			BYTE *hpSourceBytes, BYTE (&remapping)[256]);
		void CopyPixelsWithRemap(WORD wWidthBytes, BYTE *hpDestBytes,
			BYTE *hpSourceBytes, BYTE (&remapping)[256]);
		void ClipRects(RECT &rcDest, RECT &rcDestClippingRect,
				RECT &rcSource, RECT &rcSourceClippingRect);
		void CopyRow(BYTE far *hpDest, BYTE far *hpSource, WORD wWidth, BOOL bTransparent);
		virtual void CopyRectangle(BYTE *pDestPixels, BYTE *pSourcePixels,
				WORD wDestX, WORD wDestY, WORD wSourceX, WORD wSourceY,
				WORD wBytesToCopy, WORD wRowsToCopy, WORD wDestWidthInBytes, WORD wSourceWidthInBytes,
				BOOL bTransparent, BOOL bFlip);

		// Data.
		BYTE *mpPixels;
		BOOL bFlip;
		int mnBits;
		Palette16 *mpoPal16;

		/*
		BYTE	mjAlpha;
		DWORD	mdwRemapColor;
		BYTE	mjRemap;
		*/

		WORD wWidth;
		WORD wHeight;
		WORD wWidthInBytes;
	};


// remember to undef this when you are satisfied that it's working
#undef _WAY_DEBUG_DUDE
//#pragma inline_recursion(off)

#undef _INLINE_BLITS

#ifdef _INLINE_BLITS
#define MY_INLINE inline
#else
#define MY_INLINE
#endif

#ifdef _INLINE_BLITS

MY_INLINE void vFillPixels16(WORD *pwDest, WORD wColor, int nLen)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
#endif

	while (nLen--) {
		*pwDest++ = wColor;
	}
}

MY_INLINE void vFillPixelsRev16(WORD *pwDest, WORD wColor, int nLen)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
#endif

	while (nLen--) {
		*pwDest-- = wColor;
	}
}

MY_INLINE void vFillPixelsAlpha16(WORD *pwDest, WORD wColor, int nLen, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
#endif

#if 0
	while (nLen--) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha(*pwDest, wColor, jAlpha));
		pwDest++;
	}
#else
	jAlpha >>= 3;
	DWORD dwColor = wColor | ((DWORD)wColor << 16);
	int nRemainder = nLen & 1;
	nLen >>= 1;
	while (nLen--) {
		*(DWORD*)pwDest = gpoBlender->dwCalc2Alpha5(*(DWORD*)pwDest, dwColor, jAlpha);
		pwDest += 2;
	}
	if (nRemainder) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, wColor, jAlpha));
	}
#endif
}

MY_INLINE void vFillPixelsAlphaRev16(WORD *pwDest, WORD wColor, int nLen, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
#endif

#if 1
	jAlpha >>= 3;
	while (nLen--) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, wColor, jAlpha));
		pwDest--;
	}
#else
	jAlpha >>= 3;
	DWORD dwColor = wColor | ((DWORD)wColor << 16);
	int nRemainder = nLen & 1;
	nLen >>= 1;
	while (nLen--) {
		*(DWORD*)pwDest = gpoBlender->dwCalc2Alpha5(*(DWORD*)pwDest, dwColor, jAlpha);
		pwDest -= 2;
	}
	if (nRemainder) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, wColor, jAlpha));
	}
#endif
}

#define TEST_ALT_FILL
MY_INLINE void vCopyPixelsPalette16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

#ifdef TEST_ALT_FILL
	int nRemainder = nLen & 1;
	nLen >>= 1;
	while (nLen--) {
		*(DWORD*)pwDest = ((DWORD)pwPalette[*pjSource]) | ((DWORD)pwPalette[*(pjSource + 1)] << 16);
		pwDest += 2;
		pjSource += 2;
	}
	if (nRemainder) {
		*pwDest = pwPalette[*pjSource];
	}

#else
	while (nLen--) {
		*pwDest++ = pwPalette[*pjSource++];
	}
#endif
}

MY_INLINE void vCopyPixelsPaletteRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

#if 0
	int nRemainder = nLen & 1;
	nLen >>= 1;
	while (nLen--) {
		*(DWORD*)pwDest = ((DWORD)pwPalette[*(pjSource + 1)]) | ((DWORD)pwPalette[*pjSource] << 16);
		pwDest -= 2;
		pjSource += 2;
	}
	if (nRemainder) {
		*pwDest = pwPalette[*pjSource];
	}

#else
	while (nLen--) {
		*pwDest-- = pwPalette[*pjSource++];
	}
#endif
}

MY_INLINE void vCopyPixelsPaletteAlpha16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif
	jAlpha >>= 3;
	int nRemainder = nLen & 1;
	nLen >>= 1;

	while (nLen--) {
		*(DWORD*)pwDest = gpoBlender->dwCalc2Alpha5(*(DWORD*)pwDest, ((DWORD)pwPalette[*pjSource]) | ((DWORD)pwPalette[*(pjSource + 1)] << 16), jAlpha);
		pwDest += 2;
		pjSource += 2;
	}
	if (nRemainder) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource], jAlpha));
	}
}

MY_INLINE void vCopyPixelsPaletteAlphaRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	jAlpha >>= 3;

#if 0
	int nRemainder = nLen & 1;
	nLen >>= 1;

	while (nLen--) {
		*(DWORD*)pwDest = gpoBlender->dwCalc2Alpha5(*(DWORD*)pwDest, ((DWORD)pwPalette[*(pjSource + 1)]) | ((DWORD)pwPalette[*pjSource] << 16), jAlpha);
		pwDest -= 2;
		pjSource += 2;
	}
	if (nRemainder) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource], jAlpha));
	}
#else
	while (nLen--) {
		*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource++], jAlpha));
		pwDest--;
	}
#endif
}

MY_INLINE void vCopyPixelsPalGeneric16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BOOL bRev, BYTE jAlpha, WORD wRemapColor, BYTE jRemap)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	jAlpha >>= 3;
	jRemap >>= 3;
	if (jAlpha == 0) {
		// totally transparent
		return;
	}
	if (jAlpha == 31) {
		// no alpha

		if (jRemap) {
			if (bRev) {
				while (nLen--) {
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(pwPalette[*pjSource++], wRemapColor, jRemap));
					pwDest--;
				}
			} else {
				while (nLen--) {
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(pwPalette[*pjSource++], wRemapColor, jRemap));
					pwDest++;
				}
			}
		} else {
			if (bRev) {
				while (nLen--) {
					*pwDest = pwPalette[*pjSource++];
					pwDest--;
				}
			} else {
				while (nLen--) {
					*pwDest = pwPalette[*pjSource++];
					pwDest++;
				}
			}
		}
	} else {
		// there is alpha

		if (jRemap) {
			if (bRev) {
				while (nLen--) {
					WORD wTemp = (WORD)(gpoBlender->dwCalcAlpha5(pwPalette[*pjSource++], wRemapColor, jRemap));
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, wTemp, jAlpha));
					pwDest--;
				}
			} else {
				while (nLen--) {
					WORD wTemp = (WORD)(gpoBlender->dwCalcAlpha5(pwPalette[*pjSource++], wRemapColor, jRemap));
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, wTemp, jAlpha));
					pwDest++;
				}
			}
		} else {
			if (bRev) {
				while (nLen--) {
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource++], jAlpha));
					pwDest--;
				}
			} else {
				while (nLen--) {
					*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource++], jAlpha));
					pwDest++;
				}
			}
		}
	}
}

MY_INLINE void vCopyPixelsPaletteTrans16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	while (nLen--) {
		if (*pjSource) {
			*pwDest = pwPalette[*pjSource];
		}
		pwDest++;
		pjSource++;
	}
}

MY_INLINE void vCopyPixelsPaletteTransRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	while (nLen--) {
		if (*pjSource) {
			*pwDest = pwPalette[*pjSource];
		}
		pwDest--;
		pjSource++;
	}
}

MY_INLINE void vCopyPixelsPaletteTransAlpha16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	jAlpha >>= 3;
	while (nLen--) {
		if (*pjSource) {
			*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource], jAlpha));
		}
		pwDest++;
		pjSource++;
	}
}

MY_INLINE void vCopyPixelsPaletteTransAlphaRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pjSource);
	assert(pwPalette);
#endif

	jAlpha >>= 3;
	while (nLen--) {
		if (*pjSource) {
			*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, pwPalette[*pjSource], jAlpha));
		}
		pwDest--;
		pjSource++;
	}
}

MY_INLINE void vCopyPixels16(WORD *pwDest, WORD *pwSource, int nLen)
{
	memcpy(pwDest, pwSource, nLen << 1);
}

MY_INLINE void vCopyPixelsRev16(WORD *pwDest, WORD *pwSource, int nLen)
{
	while (nLen--) {
		*pwDest = *pwSource;
		pwDest--;
		pwSource++;
	}
}

MY_INLINE void vCopyPixelsTrans16(WORD *pwDest, WORD *pwSource, int nLen)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pwSource);
#endif

	while (nLen--) {
		if (*pwSource) {
			*pwDest = *pwSource;
		}
		pwDest++;
		pwSource++;
	}
}

MY_INLINE void vCopyPixelsTransRev16(WORD *pwDest, WORD *pwSource, int nLen)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pwSource);
#endif

	while (nLen--) {
		if (*pwSource) {
			*pwDest = *pwSource;
		}
		pwDest--;
		pwSource++;
	}
}

MY_INLINE void vCopyPixelsTransAlpha16(WORD *pwDest, WORD *pwSource, int nLen, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pwSource);
#endif

	jAlpha >>= 3;
	while (nLen--) {
		if (*pwSource) {
			*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, *pwSource, jAlpha));
		}
		pwDest++;
		pwSource++;
	}
}

MY_INLINE void vCopyPixelsTransAlphaRev16(WORD *pwDest, WORD *pwSource, int nLen, BYTE jAlpha)
{
#ifdef _WAY_DEBUG_DUDE
	assert(pwDest);
	assert(nLen > 0 && nLen < 1000);
	assert(pwSource);
#endif

	jAlpha >>= 3;
	while (nLen--) {
		if (*pwSource) {
			*pwDest = (WORD)(gpoBlender->dwCalcAlpha5(*pwDest, *pwSource, jAlpha));
		}
		pwDest--;
		pwSource++;
	}
}

MY_INLINE void rmemcpy(BYTE * hpDest, BYTE * hpSource, int nLen)
{
	while (nLen--) {
		*hpDest++ = *hpSource--;
	}
}

MY_INLINE void rtmemcpy(BYTE * hpDest, BYTE * hpSource, int nLen)
{
	while (nLen--) {
		if (*hpSource) {
			*hpDest = *hpSource;
		}
		++hpDest;
		--hpSource;
	}
}

MY_INLINE void tmemcpy(BYTE * hpDest, BYTE * hpSource, int nLen)
{
	while (nLen--) {
		if (*hpSource) {
			*hpDest = *hpSource;
		}
		++hpDest;
		++hpSource;
	}
}
#else

MY_INLINE void vFillPixels16(WORD *pwDest, WORD wColor, int nLen);
MY_INLINE void vFillPixelsRev16(WORD *pwDest, WORD wColor, int nLen);
MY_INLINE void vFillPixelsAlpha16(WORD *pwDest, WORD wColor, int nLen, BYTE jAlpha);
MY_INLINE void vFillPixelsAlphaRev16(WORD *pwDest, WORD wColor, int nLen, BYTE jAlpha);
MY_INLINE void vCopyPixelsPalette16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette);
MY_INLINE void vCopyPixelsPaletteRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette);
MY_INLINE void vCopyPixelsPaletteAlpha16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha);
MY_INLINE void vCopyPixelsPaletteAlphaRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha);
MY_INLINE void vCopyPixelsPalGeneric16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BOOL bRev, BYTE jAlpha, WORD wRemapColor, BYTE jRemap);
MY_INLINE void vCopyPixelsPaletteTrans16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette);
MY_INLINE void vCopyPixelsPaletteTransRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette);
MY_INLINE void vCopyPixelsPaletteTransAlpha16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha);
MY_INLINE void vCopyPixelsPaletteTransAlphaRev16(WORD *pwDest, BYTE *pjSource, int nLen, WORD *pwPalette, BYTE jAlpha);
MY_INLINE void vCopyPixels16(WORD *pwDest, WORD *pwSource, int nLen);
MY_INLINE void vCopyPixelsRev16(WORD *pwDest, WORD *pwSource, int nLen);
MY_INLINE void vCopyPixelsTrans16(WORD *pwDest, WORD *pwSource, int nLen);
MY_INLINE void vCopyPixelsTransRev16(WORD *pwDest, WORD *pwSource, int nLen);
MY_INLINE void vCopyPixelsTransAlpha16(WORD *pwDest, WORD *pwSource, int nLen, BYTE jAlpha);
MY_INLINE void vCopyPixelsTransAlphaRev16(WORD *pwDest, WORD *pwSource, int nLen, BYTE jAlpha);
MY_INLINE void rmemcpy(BYTE * hpDest, BYTE * hpSource, int nLen);
MY_INLINE void rtmemcpy(BYTE * hpDest, BYTE * hpSource, int nLen);
MY_INLINE void tmemcpy(BYTE * hpDest, BYTE * hpSource, int nLen);

#endif // inline blits

#endif
