#include "os.h"

#include "viewbmp.h"
#include "palette.h"
#include "app.h"
#include "util.h"

extern Application *pApp;
extern RGBQUAD palettecolors[256];

ViewBitmap::ViewBitmap(int iResourceID, ResourceManager * newResourceManager, BOOL activatePalette)
:Bitmap(RES_VIEWBITMAP, 'm', iResourceID, newResourceManager)
{
	int i;//,j;
	//BYTE remapping[256];
	//BOOL identityPalette;
	RGB *viewColors;

	mnBits = 8;

	activatePalette = FALSE;

	type = RES_BITMAP;
	rows = NULL;
	loop = 0;
	cel = 0;
	wWidth = 0;
	wHeight = 0;
	wWidthInBytes = 0;
	loops = NULL;
	if (data) {
		stuff = data;
		i = *(int *) stuff;		// tag = 'tex '
		stuff += sizeof(int);
		i = *(int *) stuff;		// version
		stuff += sizeof(int);
		viewColors = (RGB *) stuff;	// palette
		stuff += 768;
		numLoops = *(short *) stuff;
		stuff += sizeof(short);

		for (i = 0; i < 256; ++i) {
			palettecolors[i].rgbRed = viewColors[i].red;
			palettecolors[i].rgbGreen = viewColors[i].green;
			palettecolors[i].rgbBlue = viewColors[i].blue;
			palettecolors[i].rgbReserved = 0;
		}


		if (activatePalette) {
			ActivatePalette();
			//identityPalette = TRUE;
		}
		// let's generate a 16-bit palette
		if (pApp->poFrameBuf() && pApp->poFrameBuf()->mnBits == 16) {
			mpoPal16 = new Palette16(palettecolors);
		}


		// Generate a remapping table;
		//remapping[0] = 0;
		//identityPalette = TRUE;
		//for (i = 1; i < 256; ++i) {
			//remapping[i] = i;
			/*
			if (pApp->pCurrentPalette) {
				for (j = 1; j < 256; ++j) {
					if ((viewColors[i].red == pApp->pCurrentPalette->colors[j].rgbRed) &&
						(viewColors[i].green == pApp->pCurrentPalette->colors[j].rgbGreen) &&
						(viewColors[i].blue == pApp->pCurrentPalette->colors[j].rgbBlue)
						) {
						remapping[i] = j;
						identityPalette = FALSE;
						break;
					}
				}
			}
			*/
		//}

		loops = new LOOP[numLoops];
		for (i = 0; i < numLoops; ++i) {
			ReadLoop(i);
		}

		// force an init of loop and cel parameters
		loop = -1;
		SetLoop(0);
		cel = -1;
		SetCel(0);
	}
	if (mpPixels) {
		delete[] mpPixels;
		mpPixels = NULL;
	}
	if (data) {
		// Make the Bitmap class think there's data.
		mpPixels = new BYTE[4];
		memset(mpPixels, 0, 4);
	}
}

ViewBitmap::~ViewBitmap()
{
	int i,
	 j;
	int numCels;

	if (loops) {
		for (i = 0; i < numLoops; ++i) {
			numCels = loops[i].numCels;
			for (j = 0; j < numCels; ++j) {
				delete[] loops[i].cels[j].rows;
			}
			if (loops[i].cels) {
				delete[] loops[i].cels;
			}
		}
		delete[] loops;
	}
}

WORD ViewBitmap::ColorOfPoint(POINT point)
{
	if ((point.y >= wHeight) || (point.x >= wWidth) || (point.x < 0) || (point.y < 0)) {
		return 0;
	}

	assert(GetNumCels() != 0);

	BYTE *pjSource = rows[point.y];	// tw TBD - this may have to change
									// when loops/cels are working
	WORD *pwPalette;
	WORD wRet = 0;
	BYTE jTemp = 0;

	switch (pApp->poFrameBuf()->mnBits) {
	case 8:
		TextureRenderRow(&jTemp, pjSource, point.x, 1);
		wRet = jTemp;
		break;

	case 16:

		// if we don't have a palette of our own, use the
		// frame buffer's system palette
		if (mpoPal16) {
			pwPalette = mpoPal16->pwGetPalette();
		} else {
			pwPalette = pApp->poFrameBuf()->pwGetPalette();
		}
		assert(pwPalette);

		TextureRenderRow16(&wRet, pjSource, point.x, 1, pwPalette);
		break;
	}

	return wRet;
}

//#pragma optimize( "agt", on )

void ViewBitmap::vDrawToFrame(FrameBuffer * poFrameBuf, POINT & tDest, RECT & rcSource, BYTE jAlpha, COLOR16 wRemapColor, BYTE jRemapAmount, WORD *pwPalette)
{

	if (!rows) {
		vDebugOut("Trying to draw nonexistent cel in TEX %d.\n", resourceID);
		return;
	}
	// it's already been clipped
	int nPitch = poFrameBuf->mnPitch;
	int nHeight = rcSource.bottom - rcSource.top;
	int nWidth = rcSource.right - rcSource.left;
	int nSourceX = rcSource.left;
	BYTE **ppRow = rows + rcSource.top;
	BYTE *pjScreen = (BYTE *) poFrameBuf->pGetBuffer();

	switch (poFrameBuf->mnBits) {
	case 8:
		//
		// NOTE: Does not currently handle bFlip properly
		// Will require TextureRenderRowFlip
		//
		assert(mnBits == 8);
		pjScreen += tDest.y * nPitch + tDest.x;
		while (nHeight--) {
			TextureRenderRow(pjScreen, *ppRow, nSourceX, nWidth);
			++ppRow;																											// >colors);
			pjScreen += nPitch;
		}
		break;

	case 16:
		assert(mnBits == 8);
		//assert(mpoPal16);
		pjScreen += tDest.y * nPitch + (tDest.x * sizeof(WORD));
		if (!pwPalette) {
			if (mpoPal16) {
				pwPalette = mpoPal16->pwGetPalette();
			} else {
				pwPalette = poFrameBuf->pwGetPalette();
			}
		}
		assert(pwPalette);

		if (jAlpha == 255) {
			if (bFlip) {
				while (nHeight--) {
					TextureRenderRowFlip16((WORD*)pjScreen, *ppRow, nSourceX, nWidth, pwPalette);
					++ppRow;																											// >colors);
					pjScreen += nPitch;
				}
			} else {
				while (nHeight--) {
					TextureRenderRow16((WORD*)pjScreen, *ppRow, nSourceX, nWidth, pwPalette);
					++ppRow;																											// >colors);
					pjScreen += nPitch;
				}
			}
		} else {
			while (nHeight--) {
				TextureRenderRowAlpha16((WORD*)pjScreen, *ppRow, nSourceX, nWidth, pwPalette, jAlpha, bFlip);
				++ppRow;																											// >colors);
				pjScreen += nPitch;
			}
		}
		break;
	}

	poFrameBuf->vReleaseBuffer();
}


//#pragma optimize( "agt", on )

void ViewBitmap::CopyRectangle(BYTE * pDestPixels, BYTE * pSourcePixels,
				  WORD wDestX, WORD wDestY, WORD wSourceX, WORD wSourceY,
							    WORD wBytesToCopy, WORD wRowsToCopy, WORD wDestWidthInBytes, WORD wSourceWidthInBytes,
							    BOOL bTransparent, BOOL bFlip)
{
	if (!rows) {
		vDebugOut("Trying to draw nonexistent cel in TEX %d.\n", resourceID);
		return;
	}

	pDestPixels += (DWORD) wDestY *wDestWidthInBytes + wDestX;

	// copy some of these locally to avoid a few indirections/size changes
	// in the loop
	BYTE **ppRow = rows + wSourceY;
	int nSourceX = wSourceX;
	int nLength = wBytesToCopy;

	// for now, we are removing alpha, since the implementation is going to
	// change anyway TBD DJM
	while (wRowsToCopy--) {
		TextureRenderRow(pDestPixels, *ppRow, nSourceX, nLength);
		++ppRow;																											// >colors);
		//++wSourceY;
		pDestPixels += wDestWidthInBytes;
	}
}

void ViewBitmap::ReadLoop(int theLoop)
{
	int i;
	int numCels;

	numCels = *(short *) stuff;
	stuff += sizeof(short);
	loops[theLoop].numCels = numCels;
	if (numCels) {
		loops[theLoop].cels = new CEL[numCels];
	} else {
		loops[theLoop].cels = NULL;
	}

	for (i = 0; i < numCels; ++i) {
		ReadCel(theLoop, i);
	}
}

void ViewBitmap::ReadCel(int theLoop, int theCel)
{
	int i,
	 j;
	int height;

	// save a pointer to the actual data
	// to be able to change origins 'n' stuff
	loops[theLoop].cels[theCel].pData = stuff;

	loops[theLoop].cels[theCel].width = *(short *) stuff;
	stuff += sizeof(short);
	height = *(short *) stuff;
	stuff += sizeof(short);
	loops[theLoop].cels[theCel].height = height;
	loops[theLoop].cels[theCel].xOrigin = *(short *) stuff;
	stuff += sizeof(short);
	loops[theLoop].cels[theCel].yOrigin = *(short *) stuff;
	stuff += sizeof(short);
	loops[theLoop].cels[theCel].x = *(short *) stuff;
	stuff += sizeof(short);
	loops[theLoop].cels[theCel].y = *(short *) stuff;
	stuff += sizeof(short);

	nOriginX = loops[theLoop].cels[theCel].xOrigin;
	nOriginY = loops[theLoop].cels[theCel].yOrigin;


	// Remap the bitmap to the current palette.

	loops[theLoop].cels[theCel].rows = new PBYTE[height];

	for (i = 0; i < height; ++i) {
		j = *(short *) stuff;
		stuff += sizeof(short);
		loops[theLoop].cels[theCel].rows[i] = stuff;
		stuff += j;
	}
}

//inlined in .H file
//
/*void ViewBitmap::SetLoop(int theLoop)
{
	//int theCel;
	if (loop == theLoop) {
		return;
	}

  //if theLoop is greater than the number of loops, set it to 0
	if (theLoop >= numLoops)
		theLoop = 0;

	loop = theLoop;
	//theLoop = GetNumLoops() - loop - 1;
	//theCel = GetNumCels() - cel - 1;
	if (loops[loop].numCels) {
		rows = loops[loop].cels[cel].rows;
		wWidth = loops[loop].cels[cel].width;
		wHeight = loops[loop].cels[cel].height;
		//wWidthInBytes = wWidth + 3 - ((wWidth - 1) % 4);

		nOriginX = loops[loop].cels[cel].xOrigin;
		nOriginY = loops[loop].cels[cel].yOrigin;
	} else {
		rows = NULL;
		wWidth = 0;
		wHeight = 0;
		nOriginX = 0;
		nOriginY = 0;
	}

}
*/

//inlined in .H file
//
/*void ViewBitmap::SetCel(int theCel)
{
	//int theLoop;

	if (cel == theCel) {
		return;
	}
	
	//if theCel is greater than the number of cels, set to 0
	if (theCel >= loops[loop].numCels)
		theCel = 0;

	cel = theCel;
	//theLoop = GetNumLoops() - loop - 1;
	//theCel = GetNumCels() - cel - 1;
	rows = loops[loop].cels[cel].rows;
	wWidth = loops[loop].cels[cel].width;
	wHeight = loops[loop].cels[cel].height;
	//wWidthInBytes = wWidth + 3 - ((wWidth - 1) % 4);

	nOriginX = loops[loop].cels[cel].xOrigin;
	nOriginY = loops[loop].cels[cel].yOrigin;
}
*/
int ViewBitmap::GetNumLoops()
{
	return numLoops;
}

int ViewBitmap::GetNumCels()
{
	return loops[loop].numCels;
}


int ViewBitmap::nGetCel()
{
	return cel;
}

int ViewBitmap::nGetLoop()
{
	return loop;
}

void ViewBitmap::vIncrementCel()
{
	if (cel + 1 >= GetNumCels())
		SetCel(0);
	else
		SetCel(cel + 1);
}

void ViewBitmap::vDecrementCel()
{
	if (cel - 1 < 0)
		SetCel(GetNumCels() - 1);
	else
		SetCel(cel - 1);
}

void ViewBitmap::vIncrementLoop()
{
	if (loop + 1 >= GetNumLoops())
		SetLoop(0);
	else
		SetLoop(loop + 1);
}

void ViewBitmap::vDecrementLoop()
{
	if (loop - 1 < 0)
		SetLoop(GetNumLoops() - 1);
	else
		SetLoop(loop - 1);
}

void ViewBitmap::vSetOriginPoint(short nNewX, short nNewY, int nLoop, int nCel)
{
	//BYTE *stuff;
	DISK_CEL *pCel = (DISK_CEL*)(loops[nLoop].cels[nCel].pData);

	// change the actual data in the disk representation
	pCel->xOrigin = nNewX;
	pCel->yOrigin = nNewY;

	// change our handy directory
	loops[nLoop].cels[nCel].xOrigin = nNewX;
	loops[nLoop].cels[nCel].yOrigin = nNewY;

	// reset the current origin point, if necessary
	if (nLoop == loop && nCel == cel) {
		nOriginX = nNewX;
		nOriginY = nNewY;
	}
}

void ViewBitmap::vMoveAllOrigins(int nDeltaX, int nDeltaY)
{
	int nCurLoop = loop;
	int nLoops = GetNumLoops();
	int i, j;

	for (i = 0; i < nLoops; ++i) {
		loop = i;
		int nCels = GetNumCels();
		for (j = 0; j < nCels; ++j) {
			int nX = nGetOriginX(i, j);
			int nY = nGetOriginY(i, j);

			nX += nDeltaX;
			nY += nDeltaY;

			vSetOriginPoint(nX, nY, i, j);
		}
	}
	loop = nCurLoop;	//restore
}
