#include "os.h"

#include "vwindow.h"
#include "bitmap.h"
#include "scigame.h"
#include "sciroom.h"
#include "sciplane.h"
#include "game.h"

VWindow::VWindow()
{
	//pBackgroundBitmap = NULL;
	pWorkBitmap = NULL;
	dirty = FALSE;
	//plane = NULL;
	SetRect(&dirtyRect, 0, 0, 0, 0);
	x = 0;
	y = 0;
}

VWindow::~VWindow()
{
	//if (pBackgroundBitmap)
	//	delete pBackgroundBitmap;
	if (pWorkBitmap) {
		delete pWorkBitmap;
	}
}

Boolean VWindow::HandleEvent(sciEvent & event)
{
	//int i, nSize;
	Boolean bClaimed;
	sciPlane *poPlane;
	sciPlane **ppoPlane;

	event.vGlobalize(); // just to be safe

	// global event handlers get first crack
	//
	// Designed for debug handlers -- not for general use
	Object **ppoTemp = moGlobalHandlers.GetBegin();
	Object *poTemp;
	while (ppoTemp != moGlobalHandlers.GetEnd()) {
		poTemp = *ppoTemp++;
		if (poTemp->HandleEvent(event)) {
			goto exit;
		}
	}

	// if someone's hogging events, they're next
	//
	// used by buttons and draggable things
	ppoTemp = moEventHogs.GetBegin();
	while (ppoTemp != moEventHogs.GetEnd()) {
		poTemp = *ppoTemp++;
		poPlane = poTemp->poGetPlane();
		if (poPlane)
			event.vLocalize(poPlane);
		if (poTemp->HandleEvent(event)) {
			goto exit;
		}
	}

	// special key handlers get first crack at key events
	//
	// Used for anything that just handles keys, like hotkeys or menu
	// keys
	if (event.type == KEY_PRESS) {
		ppoTemp = moKeyHandlers.GetBegin();
		while (ppoTemp != moKeyHandlers.GetEnd()) {
			poTemp = *ppoTemp++;
			if (poTemp->HandleEvent(event)) {
				goto exit;
			}
		}
	}

	// give the game (and hence the current room) a chance first
	// for old times' sake
	if (game) {
		if (game->HandleEvent(event)) {
			goto exit;
		}
	}

	// walk planes front to back (reverse)
	ppoPlane = moPlanes.GetEnd();
	while (ppoPlane != moPlanes.GetBegin()) {
		poPlane = *(--ppoPlane);
		assert(poPlane);

		event.vLocalize(poPlane);
		Boolean bModal = poPlane->IsModal();
		bClaimed = poPlane->HandleEvent(event);

		// if a plane claims the event or is modal,
		// no lower plane gets the event.
		if (bClaimed || bModal) {
			goto exit;
		}
	}

exit:
	return (event.claimed);
}

void VWindow::MakeWindow(int nWidth, int nHeight)
{
}

void VWindow::Show()
{
}

void VWindow::Hide()
{
}

WORD VWindow::Width()
{
	return (WORD) (rcInner.right - rcInner.left);
}

WORD VWindow::Height()
{
	return (WORD) (rcInner.bottom - rcInner.top);
}

void VWindow::Paint(RECT & rcPaint)
{
}

void VWindow::OffscreenUpdate(RECT & updateRect)
{
	int i, nSize;
	//if (plane)
	//	plane->Render(updateRect);

	nSize = moPlanes.GetSize();

	// walk back to front
	for (i = 0; i < nSize; ++i) {
		sciPlane *poPlane = (sciPlane*)moPlanes.GetItem(i);
		if (poPlane) {
			poPlane->Render(updateRect);
		}
	}

}

void VWindow::OffscreenErase(RECT & rect)
{
	//pBackgroundBitmap->CopyPixels(pWorkBitmap, rect, rect, FALSE);
}

void VWindow::Update(RECT & rect)
{
	dirty = TRUE;
	//Paint(rect);
}

void VWindow::vAddPlane(sciPlane *poPlane)
{
	int i, nSize;

	assert(poPlane);

	nSize = moPlanes.GetSize();

	// hunt for the first one with a higher priority,
	// then slide this one in just in front of it
	for (i = 0; i < nSize; ++i) {
		sciPlane *poTemp = (sciPlane*)moPlanes.GetItem(i);
		if (poTemp && poTemp->nGetPri() >= poPlane->nGetPri()) {
			break;
		}
	}

	moPlanes.Insert(poPlane, i);
	//poPlane->window = this;

	// probably need to mark the whole area this plane covers
	// for updating. TBD
}

void VWindow::vDeletePlane(sciPlane *poPlane)
{
	assert(poPlane);

	// if it's not there, nothing will happen
	moPlanes.Remove(poPlane);
	//poPlane->window = NULL;

	// probably need to mark the whole area this plane covers
	// for updating. TBD
}

int VWindow::nGetHighPriority()
{
	if (!moPlanes.GetSize()) {
		return 0;
	} else {
		return (((sciPlane*)moPlanes.GetLast())->nGetPri());
	}
};

//Find the 1st modal plane(if any) and give it all the doit's, else if no modal plane give all
//planes doit's as usual.
//
// NOTE - special modal planes(like DialogPlane) will never get Doits through here because they 
// use the 'special' GetInputEvent procedure which does not go through SCIGAME's doit.
//
void VWindow::Doit()
{
	sciPlane *poPlane;
	sciPlane **ppoPlane;
	Boolean bModal = FALSE;

	// walk planes front to back (reverse)
	ppoPlane = moPlanes.GetEnd();
	while (ppoPlane != moPlanes.GetBegin()) {
		poPlane = *(--ppoPlane);
		bModal = poPlane->IsModal();
		if (bModal) {
			break;
		}
	}

	if (bModal) {
		poPlane->Doit();
	} else {
		int i;
		int nSize = moPlanes.GetSize();
		for (i = 0; i < nSize; ++i) {
			poPlane = (sciPlane *)moPlanes.GetItem(i);
			poPlane->Doit();
		}
	}

}