// WALKER.H -- the object that handles/manages the various bitmaps for walking, running, fidgeting, attacking, etc
//
// TW
//

#ifndef _walker_
#define _walker_

#include "scimain.h"
#include "unit.h"

#define NUM_DIRECTIONS	5

//The offsets for the animations
#define OFF_WALK		0
#define	OFF_RUN			1	//run or mob
#define OFF_ATTACK		2
#define OFF_REACT		3
#define OFF_FEED		4	//feed or fed upon
#define OFF_DIE			5
#define OFF_FIDGET_C	6	//fidget cloaked as you see yourself
#define OFF_FIDGET_U	7	//fidget uncloaked as everyone(including you) sees you
#define OFF_CAST		8
#define OFF_SPECIAL		9	//things like arise for revenants


// The Walker object
class Walker : public Object
{
public:
	Walker(Unit *poUnit, int nTemplateID);
	~Walker();
	void vAnimationCycle();
	void vPause();
	Bitmap *poGetBitmap(Unit *poUnit, DIRECTION eDir);
	Boolean bIsSame(int nID) {return mnTemplateID == nID;}
protected:
	void	vInitArrays(Unit *poUnit);

	int		mnTemplateID;		//id number for WalkerDB list
	
/*
	//Starting resource numbers for the various actions both for vampires and peasants
	//								VAMPIRE							PEASANT
	int		mnWalk;				// see self Walking(as cloaked)	  standard Walk
	int		mnWalkOther;		// others see self walking(as		  N/A				//this case is for cloaked non-local players/vampires
								// a peasant)
	int		mnRun;				// Run(uncloaked)				  Run Mobbing
	int		mnAttack;			//  attack							attack
	int		mnReact;			// taking damage				   taking damage??
	int		mnFeed;				// feed on peasants				 fed upon by vampires
	int		mnCast;				//  cast a spell					  N/A
	int		mnDie;				//	   Die							  Die
	int		mnFidgetCloaked;	//  see self fidgeting(as cloaked)	  N/A
	int		mnFidgetUncloaked;	//  see self fidgeting(uncloaked)	standard fidget for peasants
	int		mnFidgetCloakedOther;//	others see self fidgeting		 N/A				//for cloaked fidgeting as seen by other non-local players/vampires
								//	(as a peasant)
	int		mnSpecial;			//	special
*/
	int		mnNormal;			// normal set of animations
	int		mnCloaked;			// cloaked set of animations(as you appear to others)

	//array of pointers to the different action bitmaps
	Bitmap	*mapoWalk[NUM_DIRECTIONS];
	Bitmap	*mapoWalkOther[NUM_DIRECTIONS];					
	Bitmap	*mapoRun[NUM_DIRECTIONS];
	Bitmap	*mapoAttack[NUM_DIRECTIONS];
	Bitmap	*mapoReact[NUM_DIRECTIONS];
	Bitmap	*mapoFeed[NUM_DIRECTIONS];
	Bitmap	*mapoCast[NUM_DIRECTIONS];
	Bitmap	*mapoDie[NUM_DIRECTIONS];
	Bitmap	*mapoFidgetCloaked[NUM_DIRECTIONS];
	Bitmap	*mapoFidgetUncloaked[NUM_DIRECTIONS];
	Bitmap	*mapoFidgetCloakedOther[NUM_DIRECTIONS];		
	Bitmap	*mapoSpecial[NUM_DIRECTIONS];

};


// Manages the number of Walkers. Only one Walker object per unique templateID number.
class WalkerDB
{
public:
	Walker *poAddObject(Unit *poUnit, int nTemplateID);
	WalkerDB();
	~WalkerDB();
protected:
	List<Walker>	moObjects;
};

extern WalkerDB goWalkerDB;

#endif 
