#include "os.h"
#include "window.h"
#include "app.h"
#include "bitmap.h"
#include "palette.h"
#include "sounder.h"
#include "inifile.h"
#include "rc.h"
#include "rooms.h"
#include "util.h"
#include "version.h"
#include "scievent.h"
#include "scigame.h"
#include "sciroom.h"
#include "sciplane.h"
#include "scicursr.h"
#include "scisound.h"
#include "sigs.h"
#include "game.h"

#include <string.h>
#include <dbt.h>
#include <stdio.h>

extern "C" long FAR PASCAL WndProc(HWND hWnd, unsigned message, unsigned wParam, LONG lParam);
Boolean bCheckInput(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam, sciEvent *poEvent);

#define DEBUG_LOCALMESSAGES 0

extern SIGS *sigs;
typedef int (*GameCloseFunc) ();
extern GameCloseFunc GameClose;

void CheckMenuItems(HWND, WPARAM);
void NotImplementedYet(char);
extern char *szAppTitle;
extern HINSTANCE hInstance;

extern char cdDriveLetter;
extern char *szResourceFilename;
extern double defaultGamma;
extern BOOL win32s;

char szEntry[32];
extern char *helpFilename;
HWND hDlg1;
WNDCLASS sWC;

Window::Window()
{
	hDC = NULL;
	hWaitCursor = NULL;
	hNormalCursor = NULL;
	hHandCursor = NULL;
	hWnd = NULL;

	hWaitCursor = LoadCursor(NULL, IDC_WAIT);
	hNormalCursor = LoadCursor(NULL, IDC_ARROW);
	hHandCursor = LoadCursor(hInstance, MAKEINTRESOURCE(IDC_CURSOR1));
	sWC.style = NULL;
	sWC.lpfnWndProc = WndProc;
	sWC.lpszClassName = szAppTitle;
	sWC.cbClsExtra = 0;
	sWC.cbWndExtra = sizeof(Window far *);
	sWC.hInstance = hInstance;
	sWC.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
	sWC.hCursor = NULL; // hNormalCursor;
	sWC.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);
	sWC.lpszMenuName = NULL;//MAKEINTRESOURCE(ROOM0_MENU);

	RegisterClass(&sWC);

}

Window::~Window()
{
	// Shut down help system.
	if (hWnd) {
		WinHelp(hWnd, helpFilename, HELP_QUIT, 0L);
	}

	if (hDC) {
		SendMessage(hWnd, WM_ERASEBKGND, (WPARAM) hDC, 0);
		ReleaseDC(hWnd, hDC);
		hDC = NULL;
	}
	// Remove the reference to this from the window class.
	// Otherwise, member functions might be called.
	if (hWnd) {
		SetWindowLong(hWnd, 0, 0L);
		DestroyWindow(hWnd);
	}
}

void Window::MakeWindow(int nWidth, int nHeight, BOOL bForceBorderless)
{
	HDC hScreenDC;
	DWORD dwStyle;
	RECT rect;
	int screenWidth,
	 screenHeight;
	int extraWidth,
	 extraHeight;

	if (hDC) {
		SendMessage(hWnd, WM_ERASEBKGND, (WPARAM) hDC, 0);
		ReleaseDC(hWnd, hDC);
		hDC = NULL;
	}
	if (hWnd) {
		SetWindowLong(hWnd, 0, 0L);
		DestroyWindow(hWnd);
		hWnd = NULL;
	}

	pApp->pIniFile->ReadString("Gamma", "1.3", szEntry, 5);
	defaultGamma = atof(szEntry);
	pApp->pIniFile->ReadString("FullScreen", "No", szEntry, 5);
	BOOL bFullScreen = !(stricmp(szEntry, "No") == 0);

	hScreenDC = GetDC(NULL);

	rect.top = 0;
	rect.left = 0;
	rect.bottom = 0;
	rect.right = 0;
	screenWidth = GetSystemMetrics(SM_CXSCREEN);
	screenHeight = GetSystemMetrics(SM_CYSCREEN);
	// Force 640x480 windows to FullScreen mode.
	if (screenWidth <= nWidth) {
		bFullScreen = TRUE;
	}
	if (bFullScreen) {
		screenWidth = GetSystemMetrics(SM_CXSCREEN);
		screenHeight = GetSystemMetrics(SM_CYSCREEN);
		rect.top = 0;
		rect.left = 0;
		rect.bottom = screenHeight;
		rect.right = screenWidth;
	} else {
		if (SystemParametersInfo(SPI_GETWORKAREA, 0, &rect, 0)) {
			screenWidth = rect.right - rect.left;
			screenHeight = rect.bottom - rect.top;
		} else {
			screenWidth = GetSystemMetrics(SM_CXSCREEN);
			screenHeight = GetSystemMetrics(SM_CYSCREEN);
			rect.top = 0;
			rect.left = 0;
			rect.bottom = screenHeight;
			rect.right = screenWidth;
		}
	}

	extraWidth = GetSystemMetrics(SM_CXDLGFRAME) * 2;
	extraHeight = GetSystemMetrics(SM_CYCAPTION) +
		GetSystemMetrics(SM_CYDLGFRAME) * 2;	// +
	// GetSystemMetrics(SM_CYMENU);

	if (bFullScreen) {
		rcOuter.left = 0;
		rcOuter.top = 0;
		rcOuter.right = screenWidth;
		rcOuter.bottom = screenHeight;
		rcInner = rcOuter;
	} else {
		if (bForceBorderless) {
			//rcOuter.left = rect.left + (screenWidth - nWidth) / 2;
			//rcOuter.top = rect.top + (screenHeight - nHeight) / 2;
			rcOuter.left = 0;
			rcOuter.top = 0;
			rcOuter.right = rcOuter.left + nWidth;
			rcOuter.bottom = rcOuter.top + nHeight;
			rcInner = rcOuter;
		} else {
			// Set Width and Height to just what is needed.
			rcOuter.left = rect.left + (screenWidth - (nWidth + extraWidth)) / 2;
			rcOuter.top = rect.top + (screenHeight - (nHeight + extraHeight)) / 2;
			if (rcOuter.left < rect.left)
				rcOuter.left = rect.left;
			if (rcOuter.top < rect.top)
				rcOuter.top = rect.top;
			rcOuter.right = rcOuter.left + nWidth + extraWidth;
			rcOuter.bottom = rcOuter.top + nHeight + extraHeight;
			if ((screenWidth <= nWidth) && (rcOuter.right > nWidth))
				rcOuter.right = nWidth;
			if ((screenHeight <= nHeight) && (rcOuter.bottom > nHeight))
				rcOuter.bottom = nHeight;
			rcInner.left = ((rcOuter.right - rcOuter.left) - nWidth - extraWidth) / 2;
			if (rcInner.left < 0)
				rcInner.left = 0;
			rcInner.top = ((rcOuter.bottom - rcOuter.top) - nHeight - extraHeight) / 2;
			if (rcInner.top < 0)
				rcInner.top = 0;
			rcInner.right = rcInner.left + nWidth;
			rcInner.bottom = rcInner.top + nHeight;
		}
	}

	x = rcOuter.left;
	y = rcOuter.top;

	ReleaseDC(NULL, hScreenDC);

	// pBackgroundBitmap = NULL;
	//pWorkBitmap = new Bitmap((WORD) (rcInner.right - rcInner.left), (WORD) (rcInner.bottom - rcInner.top));

	if (bFullScreen || bForceBorderless) {
		dwStyle = WS_VISIBLE | WS_POPUP;
	} else {
		dwStyle = WS_VISIBLE | WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX;
	}

	hWnd = CreateWindow(sWC.lpszClassName, szAppTitle, dwStyle,
						rcOuter.left, rcOuter.top, rcOuter.right - rcOuter.left, rcOuter.bottom - rcOuter.top,
						NULL, 0, hInstance, this);

	hDC = GetDC(hWnd);
}

void Window::vMoveUL(void)
{
	int nWidth = rcOuter.right - rcOuter.left;
	int nHeight = rcOuter.bottom - rcOuter.top;
	MoveWindow(hWnd, 0, 0, nWidth, nHeight, FALSE);
}

void Window::vRestoreFromUL(void)
{
	int nWidth = rcOuter.right - rcOuter.left;
	int nHeight = rcOuter.bottom - rcOuter.top;
 	MoveWindow(hWnd, rcOuter.left, rcOuter.right, nWidth, nHeight, TRUE);
}

HWND Window::hMakeDDWindow(int nWidth, int nHeight)
{
	hWnd = CreateWindowEx(WS_EX_TOPMOST, sWC.lpszClassName, szAppTitle, WS_VISIBLE | WS_POPUP,
					0, 0, nWidth, nHeight,
					NULL, 0, hInstance, NULL);
	return hWnd;
}

void Window::Show()
{
	ShowWindow(hWnd, SW_SHOW);
}

void Window::Hide()
{
	ShowWindow(hWnd, SW_HIDE);
}

void Window::Erase(RECT & rcPaint)
{
	PatBlt(hDC, rcInner.left, rcInner.top, Width(), Height(), BLACKNESS);
}

void Window::Paint(RECT & rcPaint)
{
	//UnionRect(&dirtyRect, &dirtyRect, &rcPaint);

	// stop relying on Windows for update regions
	//InvalidateRect(hWnd, &rcPaint, FALSE);
}

void Window::TruePaint(RECT & rcPaint)
{

	if (dirty) {
		//dirty = FALSE;
		//OffscreenErase(dirtyRect);
		SetRect(&dirtyRect, 0, 0, 640, 480);

		OffscreenUpdate(dirtyRect);
		dirty = FALSE;	//having dirty as above screws up any ForceUpdate calls
		//SetRect(&dirtyRect, 0, 0, 0, 0);
#if DEBUG_LOCALMESSAGES
	DebugString("Copying pixels to window\n");
#endif

	//pWorkBitmap->CopyPixels((Window*)this, rcPaint, rcPaint);
	//pWorkBitmap->CopyPixels((Window *) this, rcInner, rcInner);
		if (pApp && pApp->poFrameBuf()) {
			pApp->poFrameBuf()->vFrameDone();
		}
	}
}

void Window::PaintIfDirty()
{
	if (dirty) {
		TruePaint(rcInner);
	}
}

void Window::ForceUpdate()
{
	if (hDC) {
		TruePaint(rcInner);
	} else {
		hDC = GetDC(hWnd);
		TruePaint(rcInner);
		hDC = NULL;
	}
}

BOOL bProcessingMouseMove = FALSE;

extern "C" LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
	Window *pWindow;
//	POINT point;
	sciEvent event;
	DEV_BROADCAST_HDR *devBroadcastHdr;
	DEV_BROADCAST_VOLUME *devBroadcastVolume;
	int driveBit;
	char filename[32];
	FILE *file;

	pWindow = (Window *) GetWindowLong(hWnd, 0);

	if (message != WM_CREATE && pWindow == NULL)
		return DefWindowProc(hWnd, message, wParam, lParam);

	if (bCheckInput(hWnd, message, wParam, lParam, &event) == TRUE) {
		pWindow->HandleEvent(event);
	}
	else {

		switch (message) {
		case WM_SETCURSOR:
			return NULL;
			break;
		case WM_CREATE:
			SetWindowLong(hWnd, 0, *(long *) lParam);
			pWindow = (Window *) GetWindowLong(hWnd, 0);
			pWindow->hWnd = hWnd;
			break;
		case WM_PAINT:
			BeginPaint(hWnd, &ps);
			OffsetRect(&(ps.rcPaint), -pWindow->rcInner.left, -pWindow->rcInner.top);
			pWindow->TruePaint(ps.rcPaint);
			EndPaint(hWnd, &ps);
			break;
		case WM_ERASEBKGND:
			break;
		case WM_ACTIVATEAPP:
			if (pApp) {
				if ((BOOL) wParam) {
					// if (game && game->curRoom && game->curRoom->plane)
					// game->curRoom->plane->cursor->Show();
				} else {
					// if (game && game->curRoom && game->curRoom->plane)
					// game->curRoom->plane->cursor->Hide();
				}
			}
			break;
		case WM_ACTIVATE:
			// pWindow->Paint(pWindow->rcInner);
			return DefWindowProc(hWnd, message, wParam, lParam);
		//case WM_COMMAND:
			//CheckMenuItems(hWnd, wParam);
			//break;
		//case WM_ENTERMENULOOP:
			//game->curRoom->plane->cursor->Hide();
			//break;
		//case WM_EXITMENULOOP:
			//game->curRoom->plane->cursor->Show();
			//break;
		//case WM_NCHITTEST:
			//if (game->curRoom->GetModalPlane())
			//	return HTERROR;
			//else
			//	return DefWindowProc(hWnd, message, wParam, lParam);
		case WM_CLOSE:
			PostQuitMessage(0);
			break;
		case WM_QUERYNEWPALETTE:
			if (pApp && (pApp->pCurrentPalette)) {
				pApp->pCurrentPalette->ActivatePalette();
				InvalidateRect(pWindow->hWnd, NULL, TRUE);
			}
			break;
		case WM_PALETTECHANGED:
			if (((HWND) wParam != pWindow->hWnd) && pApp && pApp->pCurrentPalette) {
				pApp->pCurrentPalette->ActivatePalette();
				InvalidateRect(pWindow->hWnd, NULL, TRUE);
			}
			break;
		case WM_DEVICECHANGE:
			switch ((UINT) wParam) {
			case DBT_DEVICEREMOVECOMPLETE:
				devBroadcastHdr = (DEV_BROADCAST_HDR *) lParam;
				if (devBroadcastHdr->dbch_devicetype == DBT_DEVTYP_VOLUME) {
					devBroadcastVolume = (DEV_BROADCAST_VOLUME *) lParam;
					driveBit = devBroadcastVolume->dbcv_unitmask;
					while (cdDriveLetter &&
						   (driveBit == (1 << (cdDriveLetter - 'a')))) {
						if (IDYES == MessageBox(NULL,
												"Do you want to quit?\n\nIf not, re-insert the CD and click <No>", "CD ejected!", MB_YESNO | MB_TASKMODAL)) {
							game->QuitGame(TRUE);
							driveBit = 0;
						} else {
							Sleep(5000);
							wsprintf(filename, "%c:\\%s", cdDriveLetter, szResourceFilename);
							file = fopen(filename, "rb");
							if (file != NULL) {
								fclose(file);
								driveBit = 0;
							}
						}
					}
				}
				break;
			case DBT_DEVICEQUERYREMOVE:
				// This message isn't sent on my machine.
				// I'm not sure how to handle this message.
				// game->QuitGame();
				break;
			case DBT_DEVICEREMOVEPENDING:
				// This message isn't sent on my machine.
				// I'm not sure how to handle this message.
				// game->QuitGame();
				break;

			}
			break;
		case WM_HELP:				// <F1> on a dialog.
			WinHelp(hWnd, helpFilename, HELP_CONTENTS, 0L);
			break;
		case WM_MOVE:
			pWindow->x = (int) LOWORD(lParam);
			pWindow->y = (int) HIWORD(lParam);
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
		}
	
	}
	
	return NULL;
}

//Checks for an input event, based on the current message, and returns TRUE if an SCI 'input'
//event has occured and fills in poEvent with the appropiate data. 
//
//Returns FALSE if not an SCI 'input' event.
//
Boolean bCheckInput(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam, sciEvent *poEvent) {
	Window *pWindow;
	POINT point;

	pWindow = (Window *) GetWindowLong(hWnd, 0);
	
	switch (message) {
	case WM_LBUTTONDOWN:
		SetCapture(hWnd);
		point.x = LOWORD(lParam) - pWindow->rcInner.left;
		point.y = HIWORD(lParam) - pWindow->rcInner.top;

		// Now do an SCI event.
		poEvent->type = MOUSE_DOWN;
		poEvent->modifiers = LEFT_MOUSE;
		if (GetKeyState(VK_CONTROL) & 0xf000)
			poEvent->modifiers |= CTRL_DOWN;
		if (GetKeyState(VK_SHIFT) & 0xf000)
			poEvent->modifiers |= SHIFT_DOWN;
		if (GetKeyState(VK_MENU) & 0xf000)
			poEvent->modifiers |= ALT_DOWN;
		poEvent->claimed = FALSE;
		poEvent->x = point.x;
		poEvent->y = point.y;
		poEvent->plane = NULL;		// pWindow->plane;
		return TRUE;
		break;
	case WM_LBUTTONUP:
		ReleaseCapture();
		point.x = LOWORD(lParam) - pWindow->rcInner.left;
		point.y = HIWORD(lParam) - pWindow->rcInner.top;

		// Now do an SCI event.
		poEvent->type = MOUSE_UP;
		poEvent->modifiers = LEFT_MOUSE;
		if (GetKeyState(VK_CONTROL) & 0xf000)
			poEvent->modifiers |= CTRL_DOWN;
		if (GetKeyState(VK_SHIFT) & 0xf000)
			poEvent->modifiers |= SHIFT_DOWN;
		if (GetKeyState(VK_MENU) & 0xf000)
			poEvent->modifiers |= ALT_DOWN;
		poEvent->claimed = FALSE;
		poEvent->x = point.x;
		poEvent->y = point.y;
		poEvent->plane = NULL;		// pWindow->plane;
		return TRUE;
		break;
	case WM_RBUTTONDOWN:
		point.x = LOWORD(lParam) - pWindow->rcInner.left;
		point.y = HIWORD(lParam) - pWindow->rcInner.top;

		// Now do an SCI event.
		poEvent->type = MOUSE_DOWN;
		poEvent->modifiers = RIGHT_MOUSE;
		if (GetKeyState(VK_CONTROL) & 0xf000)
			poEvent->modifiers |= CTRL_DOWN;
		if (GetKeyState(VK_SHIFT) & 0xf000)
			poEvent->modifiers |= SHIFT_DOWN;
		if (GetKeyState(VK_MENU) & 0xf000)
			poEvent->modifiers |= ALT_DOWN;
		poEvent->claimed = FALSE;
		poEvent->x = point.x;
		poEvent->y = point.y;
		poEvent->plane = NULL;		// pWindow->plane;
		return TRUE;
		break;
	case WM_RBUTTONUP:
		point.x = LOWORD(lParam) - pWindow->rcInner.left;
		point.y = HIWORD(lParam) - pWindow->rcInner.top;

		// Now do an SCI event.
		poEvent->type = MOUSE_UP;
		poEvent->modifiers = RIGHT_MOUSE;
		if (GetKeyState(VK_CONTROL) & 0xf000)
			poEvent->modifiers |= CTRL_DOWN;
		if (GetKeyState(VK_SHIFT) & 0xf000)
			poEvent->modifiers |= SHIFT_DOWN;
		if (GetKeyState(VK_MENU) & 0xf000)
			poEvent->modifiers |= ALT_DOWN;
		poEvent->claimed = FALSE;
		poEvent->x = point.x;
		poEvent->y = point.y;
		poEvent->plane = NULL;		// pWindow->plane;
		return TRUE;
		break;
	case WM_MOUSEMOVE:
		if (bProcessingMouseMove)
			break;
		bProcessingMouseMove = TRUE;
		pApp->nMouseX = LOWORD(lParam) - (WORD)pWindow->rcInner.left;
		pApp->nMouseY = HIWORD(lParam) - (WORD)pWindow->rcInner.top;

		// Now do an SCI event.
		poEvent->type = MOUSE_MOVE;
		poEvent->modifiers = (EVENT_MODIFIER) 0;

		poEvent->claimed = FALSE;
		poEvent->x = pApp->nMouseX;
		poEvent->y = pApp->nMouseY;
		poEvent->plane = NULL;	
		
		bProcessingMouseMove = FALSE;
		return TRUE;
		break;
	case WM_CHAR:
		// Now do an SCI event.
		poEvent->type = KEY_CHAR;
		poEvent->modifiers = (EVENT_MODIFIER) 0;
		poEvent->claimed = FALSE;
		poEvent->x = 0;
		poEvent->y = 0;
		poEvent->c = (char) wParam;
		poEvent->plane = NULL;		// pWindow->plane;
		return TRUE;
		break;
	case WM_SYSKEYDOWN:
		// do an SCI event.
		poEvent->type = KEY_PRESS;
		poEvent->modifiers = (EVENT_MODIFIER) 0;
		poEvent->modifiers |= ALT_DOWN;
		if (GetKeyState(VK_CONTROL) & 0xf000)
			poEvent->modifiers |= CTRL_DOWN;
		if (GetKeyState(VK_SHIFT) & 0xf000)
			poEvent->modifiers |= SHIFT_DOWN;
		poEvent->claimed = FALSE;
		poEvent->x = 0;
		poEvent->y = 0;
//		if (('A' <= wParam) && (wParam <= 'Z') && !(poEvent->modifiers & SHIFT_DOWN))
//			wParam += 32;
		poEvent->c = (char) wParam;
		poEvent->plane = NULL;		
		return TRUE;
		break;
	case WM_KEYDOWN:
		// Now do an SCI event.
		if (wParam == VK_SHIFT)
		{
			return FALSE;
		} 
		poEvent->type = KEY_PRESS;
		poEvent->modifiers = (EVENT_MODIFIER) 0;
		if (GetKeyState(VK_MENU) & 0xf000)
			poEvent->modifiers |= ALT_DOWN;
		if (GetKeyState(VK_CONTROL) & 0xf000)
			poEvent->modifiers |= CTRL_DOWN;
		if (GetKeyState(VK_SHIFT) & 0xf000)
			poEvent->modifiers |= SHIFT_DOWN;
		poEvent->claimed = FALSE;
		poEvent->x = 0;
		poEvent->y = 0;
		poEvent->c = (char) wParam;
		poEvent->plane = NULL;	
		return TRUE;
		break;
	}
	
	//not an SCI 'input' event 
	return FALSE;
}

Boolean bGetInputEvent(sciEvent *poEvent) {
	MSG msg;

	if (PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE)) {
		TranslateMessage(&msg);
		if (bCheckInput(msg.hwnd, msg.message, msg.wParam, msg.lParam, poEvent) == TRUE) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	}
	else {
		return FALSE;
	}
}


void NotImplementedYet(char *szComment)
{
	MessageBox(NULL, szComment, "Work In Progress",
			   MB_OK | MB_ICONINFORMATION | MB_TASKMODAL);
	return;
}

void CheckMenuItems(HWND hWnd, WPARAM wParam)
{
	switch (LOWORD(wParam)) {
		// Menu Items
		// Help Menu
 case ID_HELP_CONTENTS:
		WinHelp(pApp->pWindow->hWnd, helpFilename, HELP_CONTENTS, 0L);
		break;
	case ID_HELP_HELPONHELP:
		WinHelp(pApp->pWindow->hWnd, helpFilename, HELP_HELPONHELP, 0L);
		break;
	case ID_HELP_MYINDEX:
		WinHelp(pApp->pWindow->hWnd, helpFilename, HELP_FINDER, 0L);
		break;

		// Game Menu
	case ID_GAME_QUIT:
		PostQuitMessage(0);
		break;
	case ID_NETWORK_STARTNETWORKGAME:
		switch (game->GetCurrentRoomNum()) {
		case ROOM_0:
			game->SetNextRoom(ROOM_0);
			break;
		}
		break;
	case ID_NETWORK_QUITNETWORKGAME:
	case ID_NETWORK_LEAVENETWORKGAME:
		if (sigs->Status() == IN_GAME) {
			// return to GP.
			GameClose();
			switch (game->GetCurrentRoomNum()) {
			case ROOM_0:
				game->SetNextRoom(ROOM_0);
				break;
			}
		} else {
			delete sigs;
			sigs = NULL;
			switch (game->GetCurrentRoomNum()) {
			case ROOM_0:
#ifdef DEMO
				game->SetNextRoom(ROOM_END);
#else
				game->SetNextRoom(ROOM_0);
#endif // DEMO
				break;
			}
		}
		break;
		// End of Menu Items

		// about Hoyle stuff
	case ID_HELP_ABOUTNEWGAME:
		game->SetNextRoom(ROOM_CREDITS);
		break;
	case ID_GAME_EXITCREDITS:
		game->SetNextRoom(ROOM_0);
		break;
	}
	return;
}
